(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair
imports Term_Order Trs DP_Framework
begin

context redpair
begin  
lemma redPairSound :
  assumes rules: "R \<subseteq> \<succeq>"
  and nonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  from chain have allP: "\<forall> i. (s i, t i) \<in> P" by auto
  from chain have steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*" by auto
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP"
  proof
    fix i
    from substNS wmNS rules trs_inclusion have "rstep R \<subseteq> \<succeq>" by auto
    hence "(rstep R)^* \<subseteq> \<succeq>^*" by (rule rtrancl_mono)
    with steps have stepsNS: "(?ts i, ?ss (Suc i)) \<in> \<succeq>^*" by blast
    show "(?ts i, ?ts (Suc i)) \<in> \<succ> \<or> (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<and> (?next i) \<in> rP" 
    proof (cases "?next i \<in> \<succ> \<inter> (P \<union> rP)")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> \<succ>" using substS by (auto simp: subst_closed_def)
      with stepsNS trCompat show ?thesis by auto 
    next
      case False
      with strict nonStrict allP
      have one: "?next i \<in> \<succeq>" and two: "?next i \<in> rP" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> \<succeq>" using substNS by (auto simp: subst_closed_def)
      with stepsNS have "(?ts i, ?ts (Suc i)) \<in> \<succeq>^*" by auto
      with two show ?thesis by simp
    qed
  qed  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* \<union> \<succ>" by auto
  from SN have "SN_elt \<succ> (?ts 0)" by best
  from infSeq trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" by (rule non_strict_ending)  
  from this obtain j where "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> \<succeq>^* - \<succ>" ..
  with piece1 have "\<forall> i \<ge> j. ?next i \<in> rP" by blast
  hence "\<forall> i. (s (i+(Suc j)), t (i + (Suc j))) \<in> rP" by auto
  thus ?thesis using chain by (auto, blast)
qed
end


abbreviation redPairProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
 where "redPairProc S NS P R rP rR \<equiv> R \<subseteq> NS \<and> rP \<inter> P \<subseteq> NS \<union> S \<and> P - rP \<subseteq> S \<and> rR = R"

context redpair
begin
lemma redPairProc_ci: "ci_proc (redPairProc \<succ> \<succeq>)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P R rP rR
  assume proc: "redPairProc \<succ> \<succeq> P R rP rR"
  show "rR \<subseteq>ci R \<and> (\<forall> s t \<sigma> . min_ichain (P, R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (intro conjI, simp add: proc ci_subset_refl, intro allI impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,R) s t \<sigma>"
    hence "ichain (P,R) s t \<sigma>" by simp
    with proc show "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using redPairSound by auto
  qed
qed


lemma monoRedPairSound: (* SW *)
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and rNonStrict: "rR \<inter> R \<subseteq> \<succeq> \<union> \<succ>"
  and rStrict: "R - rR \<subseteq> \<succ>"
  and pNonStrict: "rP \<inter> P \<subseteq> \<succeq> \<union> \<succ>"
  and pStrict: "P - rP \<subseteq> \<succ>"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
proof -
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?nextr = "\<lambda> i. (?ts i, ?ss (Suc i))"
  let ?nextp = "\<lambda> i. (s i, t i)"
  from chain have allP: "\<forall> i. (s i, t i) \<in> P" by auto
  from chain have steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*" by auto
  from monotonic wmNS have monoNSS:"one_step_ctxt_closed (\<succeq> \<union> \<succ>)" unfolding one_step_ctxt_closed_def by auto
  from substS substNS have substNSS:"subst_closed (\<succeq> \<union> \<succ>)" unfolding subst_closed_def using cs_union_distr[where ?R="\<succeq>" and ?S=\<succ>] by auto
  let ?SS = "\<succ> O \<succ>^* O \<succeq>^*"
  from rStrict rNonStrict have "R \<subseteq> \<succeq> \<union> \<succ>" by auto
  with monoNSS substNSS trs_inclusion have "(rstep R) \<subseteq> (\<succeq> \<union> \<succ>)" by auto
  hence "(rstep R)^* \<subseteq> (\<succeq> \<union> \<succ>)^*" by (rule rtrancl_mono)
  hence rsteps_in_NSS_or_NS:"(rstep R)^* \<subseteq> ?SS \<union> \<succeq>^*" using trans_split_NS_S_union by auto
  have piece1: "\<forall> i. (?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> ((?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP)"
  proof
    fix i
    show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
    proof (cases "?nextp i \<in> \<succ> \<inter> (P \<union> rP)")
      (* case 1: strict decrease with DP step *)
      case True
      hence "(?ss i, ?ts i) \<in> \<succ>" using substS by (auto simp: subst_closed_def)
      from steps rsteps_in_NSS_or_NS have "(?nextr i) \<in> ?SS \<union> \<succeq>^*" by auto
      with `(?ss i, ?ts i) \<in> \<succ>` have "(?ss i, ?ss (Suc i)) \<in> \<succ> O (?SS \<union> \<succeq>^*)" by auto
      hence "(?ss i, ?ss (Suc i)) \<in> ?SS" by simp
      thus ?thesis by auto
    next
      case False
      with pStrict pNonStrict allP
      have "?nextp i \<in> \<succeq>" and "?nextp i \<in> rP" by auto
      hence dp_step_in_NS: "(?ss i, ?ts i) \<in> \<succeq>" using substNS by (auto simp: subst_closed_def)
      show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> \<succeq>^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
      proof (cases "(?nextr i) \<in> ((rstep rR)^* - ?SS)")
        (* case 2: strict decrease in R steps *)
        case False
        hence or:"(?nextr i) \<notin> ((rstep rR)^*) \<or> (?nextr i) \<in> ?SS" by auto
        have "(?nextr i) \<in> ?SS"
        proof (cases "(?nextr i) \<in> ?SS", simp)
          case False
          with or steps have "(?nextr i) \<in> ((rstep R)^* ) - ((rstep rR)^* )" by auto
          hence "(?nextr i) \<in> (rstep R)^* O ((rstep R) - (rstep rR)) O (rstep R)^*"
            using rtrancl_diff_decomp[where ?A="rstep R"] by auto
          hence "\<exists> u v. (?ts i, u) \<in> (rstep R)^* \<and> (u, v) \<in> ((rstep R) - (rstep rR)) \<and> (v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          then obtain u v where
            tu:"(?ts i, u) \<in> (rstep R)^*" and uv:"(u, v) \<in> ((rstep R) - (rstep rR))" and vs:"(v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
          with rsteps_in_NSS_or_NS have tu:"(?ts i, u) \<in> ?SS \<union> \<succeq>^*" and vs:"(v, ?ss (Suc i)) \<in> ?SS \<union> \<succeq>^*" by auto
          from rStrict monotonic substS trs_inclusion have "rstep (R - rR) \<subseteq> \<succ>" by auto
          with uv have "(u, v) \<in> \<succ>" using rstep_diff by auto
          with tu have "(?ts i, v) \<in> (?SS O \<succ>) \<union> (\<succeq>^* O \<succ>)" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O  \<succ>^* O \<succeq>^* O \<succ>) \<union> \<succ>" by auto
          with trCompat have "(?ts i, v) \<in> (\<succ> O \<succ>^* O \<succ>) \<union> \<succ>" by auto
          hence "(?ts i, v) \<in> \<succ> O \<succ>^*" by auto
          with vs have nextrstep:"(?ts i, ?ss (Suc i)) \<in> (\<succ> O \<succ>^* O \<succ> O \<succ>^* O \<succeq>^*) \<union> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
          have "\<succ> O \<succ>^* O \<succ> O \<succ>^* \<subseteq> \<succ> O (\<succ>^* O \<succ>^*)" using O_assoc[where ?R="\<succ>^*" and ?S="\<succ>^*" and ?T="\<succ>"] by auto
          hence "\<succ> O \<succ>^* O \<succ> O \<succ>^* \<subseteq> \<succ> O \<succ>^*" using rtrancl_idemp_self_comp by auto
          hence "(\<succ> O \<succ>^* O \<succ> O \<succ>^*) O \<succeq>^* \<subseteq> (\<succ> O \<succ>^*) O \<succeq>^*" by (rule rel_comp_mono) simp
          hence "(\<succ> O \<succ>^* O \<succ> O \<succ>^* O \<succeq>^*) \<union> (\<succ> O \<succ>^* O \<succeq>^*) \<subseteq> (\<succ> O \<succ>^* O \<succeq>^*)" by (auto simp: O_assoc)
          from this nextrstep show  "(?ts i, ?ss (Suc i)) \<in> \<succ> O \<succ>^* O \<succeq>^*" 
          by (rule set_mp)
        qed
        with dp_step_in_NS trCompat show ?thesis by auto
      next
        (* case 3: only weak decrease with both DP and R steps *)
        case True
        hence "(?nextr i) \<in> ((rstep rR)^*)" by auto
        from steps rsteps_in_NSS_or_NS True have "(?nextr i) \<in> \<succeq>^*" by auto
        from dp_step_in_NS `(?nextr i) \<in> \<succeq>^*` have  "(?ss i, ?ss (Suc i)) \<in> \<succeq>^*" by auto
        with `?nextp i \<in> rP` `(?nextr i) \<in> ((rstep rR)^*)` show ?thesis by simp
      qed
    qed
  qed
  hence infSeq: "\<forall> i. (?ss i, ?ss (Suc i)) \<in>  \<succeq>^* \<union> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
  from trCompat have compat2:"\<succeq>^* O (\<succ> O \<succ>^* O \<succeq>^*) \<subseteq> (\<succ> O \<succ>^* O \<succeq>^*)" by auto
  have "SN (\<succ> O \<succ>^* O \<succeq>^*)" using compatible_SN[of S NS] SN compat by auto
  hence "SN_elt (\<succ> O \<succ>^* O \<succeq>^*) (?ss 0)" by auto
  from infSeq compat2 this have "\<exists> j. \<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succ> O \<succ>^* O \<succeq>^*)" using non_strict_ending by auto
  from this obtain j where "\<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> \<succeq>^* - (\<succ> O \<succ>^* O \<succeq>^*)" ..
  with piece1 have "\<forall> i \<ge> j. (?nextr i \<in> ((rstep rR)^*) \<and> (?nextp i) \<in> rP)" by blast
  hence "\<forall> i. ((?ts (i + j), ?ss (Suc i + j)) \<in> ((rstep rR)^*)  \<and> (s (i + j), t (i + j)) \<in> rP)" by auto
  with chain show ?thesis by auto
qed
end


abbreviation monoRedPairProc :: "('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)trs \<Rightarrow> bool"
 where "monoRedPairProc S NS P R rP rR \<equiv> one_step_ctxt_closed S \<and> rR \<subseteq> R \<and> rR \<inter> R \<subseteq> NS \<union> S \<and> R - rR \<subseteq> S \<and> rP \<inter> P \<subseteq> NS \<union> S \<and> P - rP \<subseteq> S"

context redpair
begin
lemma monoRedPairProc_ci: "ci_proc (monoRedPairProc \<succ> \<succeq>)"
unfolding ci_proc_def
proof (intro allI impI)
  fix P R rP rR
  assume proc: "monoRedPairProc \<succ> \<succeq> P R rP rR"
  show "rR \<subseteq>ci R \<and> (\<forall> s t \<sigma> . min_ichain (P, R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (rP, rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (rule conjI, simp add: proc ci_subsetI[of rR R], (rule allI)+, rule impI)
    fix s t \<sigma>
    assume min: "min_ichain (P,R) s t \<sigma>"
    hence "ichain (P,R) s t \<sigma>" by simp
    with proc show "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using monoRedPairSound by auto
  qed
qed

lemma ruleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "R \<subseteq> \<succeq> \<union> \<succ>"
  and strict: "R - rR \<subseteq> \<succ>"
  and ideriv: "ideriv R t"
  shows "\<exists> i. ideriv rR (shift t i)"
proof -
  from ideriv have alli:"\<forall>i. (t i, t (Suc i)) \<in> rstep R" unfolding ideriv_def by auto
  show ?thesis 
  proof (cases "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<notin> rstep rR")
    case False
    then obtain i where "\<forall> j \<ge> i. (t j, t (Suc j)) \<in> rstep rR" by auto
    hence "\<forall> k. (shift t i k, shift t i (Suc k)) \<in> (rstep rR)" by auto
    thus ?thesis  unfolding ideriv_def by auto
  next
    case True
    from monotonic strict substS trs_inclusion have "(rstep (R - rR)) \<subseteq> \<succ>" by auto
    with rstep_diff have "rstep R - rstep rR \<subseteq> \<succ>" by auto
    with True alli have strict: "\<forall> i. \<exists> j \<ge> i. (t j, t (Suc j)) \<in> \<succ>" by blast 
    from trs_inclusion[OF union_one_step_ctxt_closed[OF wmNS monotonic] subst_closed_union[OF substNS substS] nonStrict]
      alli have "\<forall>i. (t i, t (Suc i)) \<in> (\<succeq> \<union> \<succ>)" by auto
    from non_strict_ending[OF this compat] SN obtain j where "\<forall> i \<ge> j. (t i, t (Suc i)) \<in> \<succeq> - \<succ>" unfolding SN_def by auto
    with strict have False by auto
    thus ?thesis ..
  qed
qed

lemma manna_ness:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and strict: "R \<subseteq> \<succ>"
  shows "SN (rstep R)"
proof -
  from monotonic strict substS trs_inclusion have "(rstep R) \<subseteq> \<succ>" by auto
  from SN_subset[OF SN this] show "SN (rstep R)"  .
qed

end

(* now the implementation starts *)

fun
  check_redpair_proc ::
    "('f :: show,'v :: show)redpair_impl \<Rightarrow>
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where
  "check_redpair_proc rp P rP r = do {
     let (Sp::('f,'v)rules,NSp) = List.partition (\<lambda>foo. isOK(rp_strict rp foo)) P;
     (rp_valid rp);
     check_allm (rp_non_strict rp) r
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSp
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_subseteq NSp rP
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     (rp_desc rp) +@+ shows_nl +@+ s)"

fun
  check_mono_redpair_proc ::
    "('f :: show,'v :: show)redpair_impl \<Rightarrow> 
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f,'v)rules  \<Rightarrow>  ('f,'v)rules  \<Rightarrow> shows check"
where
  "check_mono_redpair_proc rp P rP R rR  = do {
     let (Sp :: ('f,'v) rules , NSp) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) P;
     let (Sr :: ('f,'v) rules , NSr) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) R;
     rp_valid rp;
     rp_mono rp
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSr
       <+? (\<lambda>s. ''problem when orienting TRS'' +#+ shows_nl +@+ s);
     check_allm (rp_non_strict rp) NSp
       <+? (\<lambda>s. ''problem when orienting DPs'' +#+ shows_nl +@+ s);
     check_subseteq rR R
       <+? (\<lambda>s. shows_string ''problem: additional '');
     check_subseteq NSr rR
       <+? (\<lambda>s. case (rp_strict rp s) of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error);
     check_subseteq NSp rP
       <+? (\<lambda>s. case (rp_strict rp s) of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
   } <+? (\<lambda>s. ''could not apply the reduction pair processor with the following'' +#+ shows_nl +@+
     rp_desc rp +@+ shows_nl +@+ s)"

fun
  check_rule_removal_proc ::
    "('f :: show, 'v::show)redpair_impl \<Rightarrow> 
     ('f,'v)rules \<Rightarrow> ('f,'v)rules \<Rightarrow> shows check"
where
  "check_rule_removal_proc rp R rR  = do {
     let (Sr :: ('f,'v) rules , NSr) = List.partition (\<lambda> foo. isOK(rp_strict rp foo)) R;
     rp_valid rp;
     rp_mono rp 
       <+? (\<lambda>s. ''problem with monotonicity of strict order'' +#+ shows_nl +@+ s);
     check_subseteq rR R
       <+? (\<lambda>s. ''not allowed to introduce new rule '' +#+ shows_nl +@+ shows_rule s);
     check_allm (rp_non_strict rp) NSr
       <+? (\<lambda>s. ''problem when orienting rules'' +#+ shows_nl +@+ s);
     check_subseteq NSr rR
       <+? (\<lambda>s. case (rp_strict rp) s of Inl err \<Rightarrow> err | Inr ok \<Rightarrow> internal_error)
    } <+? (\<lambda>s. ''could not apply the rule removal processor with the following'' +#+ shows_nl +@+
         rp_desc rp +@+ shows_nl +@+ s)"

context generic_redpair_impl
begin

lemma check_redpair_proc_sound :
  shows "ci_only_pairs_proc_impl (\<lambda> P R rm rrm rP . isOK(check_redpair_proc (generate_redpair rp) P rP R))"
unfolding ci_only_pairs_proc_impl_def
proof (intro allI, rule impI)
  let ?rp = "generate_redpair rp"
  fix P R rP 
  assume check2: "isOK (check_redpair_proc ?rp P rP R)"
  hence valid: "isOK(rp_valid ?rp)" by auto
  show "\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set R) (shift s i) (shift t i) (shift \<sigma> i))"
  proof (cases "List.partition (\<lambda> dp. isOK(rp_strict ?rp dp)) P")
    case (Pair Sp NSp)
    let ?NSp = "set NSp \<inter> set rP"
    from Pair check2 have cond:
      "isOK(check_allm (rp_strict ?rp) Sp)" and "isOK(check_allm (rp_non_strict ?rp) (NSp @ R))" by (auto simp: Let_def) 
    from generate_redpair[OF valid this]
    obtain S NS where redpair: "ce_af_redpair S NS (rp_af ?rp)" (is "ce_af_redpair _ _ ?pi") and S: "set Sp \<subseteq> S"  and NS: "set (NSp @ R) \<subseteq> NS" by auto
    from Pair check2 S NS have "set P - set NSp \<subseteq> S \<and> ?NSp \<inter> set P \<subseteq> NS \<union> S \<and> set NSp \<subseteq> set rP" by (auto)
    hence id: "set NSp = ?NSp" and stri: "set P - ?NSp \<subseteq> S" by (auto)
    with S NS have proc: "redPairProc S NS (set P) (set R) (set rP) (set R)" by auto
    from redpair interpret ce_af_redpair S NS ?pi .
    with proc show ?thesis using redPairProc_ci unfolding ci_proc_def by auto
  qed
qed

lemma rule_removal_sound: 
  shows "ci_tt_impl (\<lambda> R rR. isOK(check_rule_removal_proc (generate_redpair rp) R rR))"
unfolding ci_tt_impl_def
proof (intro allI, intro impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix R rR
  assume check2: "isOK(check_rule_removal_proc ?rp R rR)"
  show "set rR \<subseteq> set R \<and> (\<forall> t. ideriv (set R) t \<longrightarrow> (\<exists> i. ideriv (set rR) (shift t i)))"
  proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) R")
    case (Pair Sr NSr)
    let ?NSr = "set NSr \<inter> set rR"
    from Pair check2 have 
      one: "isOK(check_allm (rp_strict ?rp) Sr)" and two: "isOK(check_allm (rp_non_strict ?rp) NSr)" 
      and mono: "isOK(rp_mono ?rp)" and valid: "isOK(rp_valid ?rp)" by (auto simp: Let_def) 
    from generate_redpair[OF valid one two] mono obtain S NS where redpair: "mono_ce_af_redpair S NS ?af" and S: "set Sr \<subseteq> S" and NS: "set NSr \<subseteq> NS" and mono: "one_step_ctxt_closed S"
      by auto
    from Pair S NS check2 have "set R - set NSr \<subseteq> S \<and> ?NSr \<inter> set R \<subseteq> NS \<union> S \<and> set NSr \<subseteq> set rR" by (auto)
    with check2 have NS: "set R  \<subseteq> NS \<union> S" and S: "set R - set rR \<subseteq> S" and subset: "set rR \<subseteq> set R" by auto
    from redpair interpret mono_ce_af_redpair S NS ?af .
    from ruleRemoval[OF mono NS S]
    show ?thesis using subset by auto
  qed
qed

lemma check_mono_redpair_proc_sound :
  shows "ci_proc_impl (\<lambda> P R rm rP rR. isOK(check_mono_redpair_proc (generate_redpair rp) P rP R rR))"
unfolding ci_proc_impl_def
proof (intro allI impI)
  let ?rp = "generate_redpair rp"
  let ?af = "rp_af ?rp"
  fix P R rP rR
  assume check2: "isOK (check_mono_redpair_proc ?rp P rP R rR)"
  show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))"
  proof (cases "List.partition (\<lambda> r. isOK(rp_strict ?rp r)) R")
    case (Pair Sr NSr) note rPair = this
    let ?NSr = "set NSr \<inter> set rR"
    show ?thesis
    proof (cases "List.partition (\<lambda> dp. isOK(rp_strict ?rp dp)) P") 
      case (Pair Sp NSp) 
      let ?NSp = "set NSp \<inter> set rP"
      from Pair rPair check2 have 
        one: "isOK(check_allm (rp_strict ?rp) (Sr @ Sp))" and two: "isOK(check_allm (rp_non_strict ?rp) (NSr @ NSp))" 
        and mono: "isOK(rp_mono ?rp)" and valid: "isOK(rp_valid ?rp)" and subset: "set rR \<subseteq> set R" by (auto simp: Let_def) 
      from generate_redpair[OF valid one two] mono obtain S NS where redpair: "mono_ce_af_redpair S NS ?af" and Sr: "set Sr \<subseteq> S" and NSr: "set NSr \<subseteq> NS" and Sp: "set Sp \<subseteq> S" and NSp: "set NSp \<subseteq> NS" and mono: "one_step_ctxt_closed S"
        by auto
      from rPair Sr NSr check2 have "set R - set NSr \<subseteq> S \<and> ?NSr \<inter> set R \<subseteq> NS \<union> S \<and> set NSr \<subseteq> set rR" by (auto)
      with check2 have NSr: "set R \<subseteq> NS \<union> S" and Sr: "set R - set rR \<subseteq> S" by auto
      from Pair Sp NSp check2 have "set P - set NSp \<subseteq> S \<and> ?NSp \<inter> set P \<subseteq> NS \<union> S \<and> set NSp \<subseteq> set rP" by (auto)
      with check2 have NSp: "set P \<subseteq> NS \<union> S" and Sp: "set P - set rP \<subseteq> S"  by auto
      have mono: "monoRedPairProc S NS (set P) (set R) (set rP) (set rR)"
        using NSr NSp by (simp add: Sr Sp mono subset, blast)
      from redpair interpret mono_ce_af_redpair S NS ?af .      
      show "set rR \<subseteq>ci set R \<and> (\<forall> s t \<sigma>. min_ichain (set P, set R) s t \<sigma> \<longrightarrow> (\<exists> i. ichain (set rP, set rR) (shift s i) (shift t i) (shift \<sigma> i)))" (is "?goal (set P) (set R) (set rP) (set rR)")
      proof -
        let ?prop = "\<lambda> P R rP rR. monoRedPairProc S NS P R rP rR \<longrightarrow> ?goal P R rP rR"
        from monoRedPairProc_ci have "\<forall> P R rP rR. ?prop P R rP rR"
          unfolding ci_proc_def  .
        hence "\<forall> R rP rR. ?prop (set P) R rP rR" ..
        hence "\<forall> rP rR. ?prop (set P) (set R) rP rR" ..
        hence "?prop (set P) (set R) (set rP) (set rR)" using ci_subsetI[of "set rR" "set R"] by force
        with mono show ?thesis by blast
      qed
    qed
  qed
qed

end



end

