{-# LANGUAGE ScopedTypeVariables #-}

module Ceta where {

import qualified Natural;

data Nat = Zero_nat | Suc Nat;

class Orda a where {
  less_eq :: a -> a -> Bool;
  less :: a -> a -> Bool;
};

class (Orda a) => Preorder a where {
};

class (Preorder a) => Order a where {
};

class (Order a) => Linorder a where {
};

data Color = R | B;

data Rbta a b = Emptya | Branch Color (Rbta a b) a b (Rbta a b);

newtype (Linorder a) => Rbt a b = Rbt (Rbta a b);

newtype Rat = Frct (Integer, Integer);

nat_aux :: Integer -> Nat -> Nat;
nat_aux i n = (if i <= 0 then n else nat_aux (i - 1) (Suc n));

nat :: Integer -> Nat;
nat i = nat_aux i Zero_nat;

tl :: forall a. [a] -> [a];
tl [] = [];
tl (x : xs) = xs;

collect :: forall a. (a -> Bool) -> a -> Bool;
collect p = p;

is_none :: forall a. Maybe a -> Bool;
is_none (Just x) = False;
is_none Nothing = True;

dom :: forall a b. (a -> Maybe b) -> a -> Bool;
dom m = collect (\ a -> not (is_none (m a)));

data Xml = Xml [Char] [([Char], [Char])] [Xml] (Maybe [Char]);

data Sum a b = Inl a | Inr b;

binda :: forall a b c. Sum a b -> (b -> Sum a c) -> Sum a c;
binda m f = (case m of {
              Inl a -> Inl a;
              Inr a -> f a;
            });

the :: forall a. Maybe a -> a;
the (Just x) = x;

data Nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5
  | Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC
  | NibbleD | NibbleE | NibbleF;

newline :: Char;
newline = '\n';

shows_string :: [Char] -> [Char] -> [Char];
shows_string = (\ a b -> a ++ b);

class Showa a where {
  shows_prec :: Nat -> a -> [Char] -> [Char];
  shows_list :: [a] -> [Char] -> [Char];
};

shows_prec_list :: forall a. (Showa a) => Nat -> [a] -> [Char] -> [Char];
shows_prec_list d l = shows_list l;

shows_prec_char :: Nat -> Char -> [Char] -> [Char];
shows_prec_char d c = (\ a -> c : a);

shows_list_char :: [Char] -> [Char] -> [Char];
shows_list_char cs = shows_string cs;

instance Showa Char where {
  shows_prec = shows_prec_char;
  shows_list = shows_list_char;
};

shows_map :: forall a. (a -> [Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_map s [] = id;
shows_map s (x : xs) = s x . shows_map s xs;

dquote :: [Char];
dquote = ['\"'];

shows_attr :: ([Char], [Char]) -> [Char] -> [Char];
shows_attr av =
  shows_prec_list Zero_nat (fst av) .
    shows_string ['='] . shows_string (dquote ++ snd av ++ dquote);

shows_attrs :: [([Char], [Char])] -> [Char] -> [Char];
shows_attrs asa = shows_map (\ a -> shows_string [' '] . shows_attr a) asa;

replicatea :: forall a. Nat -> a -> [a];
replicatea Zero_nat x = [];
replicatea (Suc n) x = x : replicatea n x;

shows_XML_indent :: [Char] -> Nat -> Xml -> [Char] -> [Char];
shows_XML_indent ind i (Xml n a c t) =
  shows_string ind .
    shows_string ['<'] .
      shows_prec_list Zero_nat n .
        shows_attrs a .
          (if null c && is_none t then shows_string ['/', '>']
            else shows_string ['>'] .
                   (if not (null c)
                     then shows_map
                            (shows_XML_indent (ind ++ replicatea i ' ') i) c .
                            shows_string ind
                     else id) .
                     (if is_none t then id else shows_string (the t)) .
                       shows_string ['<', '/'] .
                         shows_prec_list Zero_nat n . shows_string ['>']);

shows_prec_xml :: Nat -> Xml -> [Char] -> [Char];
shows_prec_xml d xml = shows_XML_indent [newline] (nat 2) xml;

text :: [Char] -> Xml -> Sum [Char] [Char];
text tag (Xml n atts cs t) =
  (if n == tag &&
        (null atts && (null cs && (not (is_none t) && not (null (the t)))))
    then Inr (the t)
    else Inl (concat
               [['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'x',
                  't', 'r', 'a', 'c', 't', ' ', 't', 'e', 'x', 't', ' ', 'f',
                  'o', 'r', ' '],
                 tag, [' ', 'f', 'r', 'o', 'm', ' '], [newline],
                 shows_prec_xml Zero_nat (Xml n atts cs t) []]));

takea :: forall a. Nat -> [a] -> [a];
takea n [] = [];
takea n (x : xs) = (case n of {
                     Zero_nat -> [];
                     Suc m -> x : takea m xs;
                   });

one_nat :: Nat;
one_nat = Suc Zero_nat;

digit2int :: Char -> Sum [Char] Integer;
digit2int x =
  (if x == '0' then Inr 0
    else (if x == '1' then Inr 1
           else (if x == '2' then Inr 2
                  else (if x == '3' then Inr 3
                         else (if x == '4' then Inr 4
                                else (if x == '5' then Inr 5
                                       else (if x == '6' then Inr 6
      else (if x == '7' then Inr 7
             else (if x == '8' then Inr 8
                    else (if x == '9' then Inr 9
                           else Inl ([x] ++
                                      [' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
'a', ' ', 'd', 'i', 'g', 'i', 't'])))))))))));

string2int_aux :: Integer -> [Char] -> Sum [Char] Integer;
string2int_aux n [] = Inr n;
string2int_aux n (d : s) =
  binda (digit2int d) (\ m -> string2int_aux (10 * n + m) s);

string2int :: [Char] -> Sum [Char] Integer;
string2int s =
  (if null s
    then Inl ['c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v', 'e', 'r',
               't', ' ', 'e', 'm', 'p', 't', 'y', ' ', 's', 't', 'r', 'i', 'n',
               'g', ' ', 'i', 'n', 't', 'o', ' ', 'n', 'u', 'm', 'b', 'e', 'r']
    else (if takea one_nat s == ['-']
           then binda (string2int_aux 0 (tl s)) (\ i -> Inr (0 - i))
           else string2int_aux 0 s));

int :: [Char] -> Xml -> Sum [Char] Integer;
int tag x = binda (text tag x) string2int;

sequencea :: forall a b. [Sum a b] -> Sum a [b];
sequencea [] = Inr [];
sequencea (m : ms) =
  binda m (\ x -> binda (sequencea ms) (\ xs -> Inr (x : xs)));

mapm :: forall a b c. (a -> Sum b c) -> [a] -> Sum b [c];
mapm f ms = sequencea (map f ms);

mapa :: forall a. (Xml -> Sum [Char] a) -> [Xml] -> Sum [Char] [a];
mapa = mapm;

nata :: [Char] -> Xml -> Sum [Char] Nat;
nata tag x =
  binda (text tag x) (\ txt -> binda (string2int txt) (\ i -> Inr (nat i)));

lookupb :: forall a b. (Eq a) => a -> [(a, b)] -> Maybe b;
lookupb uu [] = Nothing;
lookupb aa ((a, b) : xs) = (if aa == a then Just b else lookupb aa xs);

get_name :: Xml -> [Char];
get_name (Xml name uu uv uw) = name;

xml_options :: forall a. [([Char], Xml -> Sum [Char] a)] -> Xml -> Sum [Char] a;
xml_options ps x =
  (case lookupb (get_name x) ps of {
    Nothing ->
      Inl (concat
            [['e', 'x', 'p', 'e', 'c', 't', 'e', 'd', ' ', 'o', 'n', 'e', ' ',
               'o', 'f', ':', ' '],
              concatMap (\ p -> fst p ++ [' ']) ps, [newline],
              ['b', 'u', 't', ' ', 'f', 'o', 'u', 'n', 'd'], [newline],
              shows_prec_xml Zero_nat x []]);
    Just p -> p x;
  });

xml_change ::
  forall a b. (Xml -> Sum [Char] a) -> (a -> b) -> Xml -> Sum [Char] b;
xml_change p f x = binda (p x) (\ a -> Inr (f a));

abs_int :: Integer -> Integer;
abs_int i = (if i < 0 then negate i else i);

apsnd :: forall a b c. (a -> b) -> (c, a) -> (c, b);
apsnd f (x, y) = (x, f y);

sgn_int :: Integer -> Integer;
sgn_int i = (if i == 0 then 0 else (if 0 < i then 1 else negate 1));

divmod_int :: Integer -> Integer -> (Integer, Integer);
divmod_int k l =
  (if k == 0 then (0, 0)
    else (if l == 0 then (0, k)
           else apsnd (\ a -> sgn_int l * a)
                  (if sgn_int k == sgn_int l then divMod (abs k) (abs l)
                    else let {
                           (r, s) = divMod (abs k) (abs l);
                         } in (if s == 0 then (negate r, 0)
                                else (negate r - 1, abs_int l - s)))));

mod_int :: Integer -> Integer -> Integer;
mod_int a b = snd (divmod_int a b);

gcd_int :: Integer -> Integer -> Integer;
gcd_int k l =
  abs_int (if l == 0 then k else gcd_int l (mod_int (abs_int k) (abs_int l)));

div_int :: Integer -> Integer -> Integer;
div_int a b = fst (divmod_int a b);

normalize :: (Integer, Integer) -> (Integer, Integer);
normalize p =
  (if 0 < snd p
    then let {
           a = gcd_int (fst p) (snd p);
         } in (div_int (fst p) a, div_int (snd p) a)
    else (if snd p == 0 then (0, 1)
           else let {
                  a = negate (gcd_int (fst p) (snd p));
                } in (div_int (fst p) a, div_int (snd p) a)));

fract :: Integer -> Integer -> Rat;
fract a b = Frct (normalize (a, b));

equal_nat :: Nat -> Nat -> Bool;
equal_nat (Suc nat) Zero_nat = False;
equal_nat Zero_nat (Suc nat) = False;
equal_nat (Suc nata) (Suc nat) = equal_nat nata nat;
equal_nat Zero_nat Zero_nat = True;

plus_nat :: Nat -> Nat -> Nat;
plus_nat (Suc m) n = plus_nat m (Suc n);
plus_nat Zero_nat n = n;

size_list :: forall a. [a] -> Nat;
size_list [] = Zero_nat;
size_list (a : list) = plus_nat (size_list list) (Suc Zero_nat);

nth :: forall a. [a] -> Nat -> a;
nth (x : xs) (Suc n) = nth xs n;
nth (x : xs) Zero_nat = x;

no_transform :: forall a. [Char] -> Xml -> Sum [Char] a;
no_transform tag xml =
  Inl (concat
        [['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 't', 'r', 'a', 'n',
           's', 'f', 'o', 'r', 'm', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
           'o', 'w', 'i', 'n', 'g', '\n', ' ', ' ', ' ', ' ', ' ', 'x', 'm',
           'l', ' ', 'e', 'l', 'e', 'm', 'e', 'n', 't', ' ', '(', 'e', 'x', 'p',
           'e', 'c', 't', 'e', 'd', ' '],
          tag, [')'], [newline], shows_prec_xml Zero_nat xml []]);

xml2elements ::
  forall a b c.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) -> (a -> b -> c) -> Xml -> Sum [Char] c;
xml2elements tag p1 p2 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (equal_nat (size_list cs) (nat 2) && is_none text))
    then binda (p1 (nth cs Zero_nat))
           (\ x1 -> binda (p2 (nth cs one_nat)) (\ x2 -> Inr (f x1 x2)))
    else no_transform tag (Xml name atts cs text));

rat :: Xml -> Sum [Char] Rat;
rat = xml_options
        [(['i', 'n', 't', 'e', 'g', 'e', 'r'],
           xml_change (int ['i', 'n', 't', 'e', 'g', 'e', 'r'])
             (\ x -> fract x 1)),
          (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l'],
            xml2elements ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l']
              (int ['n', 'u', 'm', 'e', 'r', 'a', 't', 'o', 'r'])
              (int ['d', 'e', 'n', 'o', 'm', 'i', 'n', 'a', 't', 'o', 'r'])
              fract)];

data Term a b = Var b | Fun a [Term a b];

data Hterm a b = HHole | HFun a [Hterm a b];

mergeHTerms ::
  forall a b. (Eq a) => [Hterm a b] -> [Hterm a b] -> Maybe [Hterm a b];
mergeHTerms [] [] = Just [];
mergeHTerms [] (uu : uv) = Nothing;
mergeHTerms (uw : ux) [] = Nothing;
mergeHTerms (HHole : hsl) (h : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (h : res);
  });
mergeHTerms (HFun f hs : hsl) (HHole : hsr) =
  (case mergeHTerms hsl hsr of {
    Nothing -> Nothing;
    Just res -> Just (HFun f hs : res);
  });
mergeHTerms (HFun f fs : hsl) (HFun g gs : hsr) =
  (if f == g
    then (case mergeHTerms fs gs of {
           Nothing -> Nothing;
           Just mhs ->
             (case mergeHTerms hsl hsr of {
               Nothing -> Nothing;
               Just res -> Just (HFun f mhs : res);
             });
         })
    else Nothing);

merge :: forall a b. (Eq a) => Hterm a b -> Hterm a b -> Maybe (Hterm a b);
merge hl hr =
  (case mergeHTerms [hl] [hr] of {
    Nothing -> Nothing;
    Just res -> Just (nth res Zero_nat);
  });

mergeX ::
  forall a b.
    (Eq a,
      Eq b) => a -> Hterm b a -> [(Hterm b a, a)] -> Maybe [(Hterm b a, a)];
mergeX x h [] = Just [];
mergeX x h ((ht, y) : ps) =
  (if x == y
    then (case merge h ht of {
           Nothing -> Nothing;
           Just m -> mergeX x m ps;
         })
    else (case mergeX x h ps of {
           Nothing -> Nothing;
           Just pairs -> Just ((ht, y) : pairs);
         }));

mergeAll :: forall a b. (Eq a, Eq b) => [(Hterm a b, b)] -> Bool;
mergeAll [] = True;
mergeAll ((h, x) : ps) =
  (case mergeX x h ps of {
    Nothing -> False;
    Just a -> mergeAll a;
  });

decomp :: forall a b. [a] -> [b] -> Maybe [(a, b)];
decomp [] [] = Just [];
decomp (uu : uv) [] = Nothing;
decomp [] (uw : ux) = Nothing;
decomp (s : ss) (t : ts) =
  (case decomp ss ts of {
    Nothing -> Nothing;
    Just pairs -> Just ((s, t) : pairs);
  });

gcMatchList ::
  forall a b. (Eq a) => [(Hterm a b, Term a b)] -> Maybe [(Hterm a b, b)];
gcMatchList [] = Just [];
gcMatchList ((HHole, uu) : ps) = gcMatchList ps;
gcMatchList ((HFun f hs, Fun g ss) : ps) =
  (if f == g
    then (case decomp hs ss of {
           Nothing -> Nothing;
           Just pairs -> gcMatchList (pairs ++ ps);
         })
    else Nothing);
gcMatchList ((HFun f hs, Var x) : ps) =
  (case gcMatchList ps of {
    Nothing -> Nothing;
    Just pairs -> Just ((HFun f hs, x) : pairs);
  });

gcMatchI :: forall a b. (Eq a, Eq b) => Hterm a b -> Term a b -> Bool;
gcMatchI h t =
  (case gcMatchList [(h, t)] of {
    Nothing -> False;
    Just a -> mergeAll a;
  });

capI ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Term a b -> Hterm a b;
capI uu (Var uv) = HHole;
capI r (Fun f ts) =
  let {
    h = HFun f (map (capI r) ts);
  } in (if any (\ ra -> gcMatchI h (fst ra)) r then HHole else h);

class (Preorder a) => Bot a where {
  bot :: a;
};

bot_fun :: forall a b. (Bot b) => a -> b;
bot_fun = (\ _ -> bot);

less_eq_bool :: Bool -> Bool -> Bool;
less_eq_bool True b = b;
less_eq_bool False b = True;

less_bool :: Bool -> Bool -> Bool;
less_bool True b = False;
less_bool False b = b;

instance Orda Bool where {
  less_eq = less_eq_bool;
  less = less_bool;
};

instance Preorder Bool where {
};

bot_bool :: Bool;
bot_bool = False;

instance Bot Bool where {
  bot = bot_bool;
};

inserta :: forall a. (Eq a) => a -> (a -> Bool) -> a -> Bool;
inserta y a x = y == x || a x;

set :: forall a. (Eq a) => [a] -> a -> Bool;
set [] = bot_fun;
set (x : xs) = inserta x (set xs);

less_nat :: Nat -> Nat -> Bool;
less_nat m (Suc n) = less_eq_nat m n;
less_nat n Zero_nat = False;

less_eq_nat :: Nat -> Nat -> Bool;
less_eq_nat (Suc m) n = less_nat m n;
less_eq_nat Zero_nat n = True;

upt :: Nat -> Nat -> [Nat];
upt i j = (if less_nat i j then i : upt (Suc i) j else []);

data Ctxt a b = Hole | More a [Term a b] (Ctxt a b) [Term a b];

string2bool :: [Char] -> Sum [Char] Bool;
string2bool s =
  (if s == ['t', 'r', 'u', 'e'] then Inr True
    else (if s == ['f', 'a', 'l', 's', 'e'] then Inr False
           else Inl (['c', 'a', 'n', 'n', 'o', 't', ' ', 'c', 'o', 'n', 'v',
                       'e', 'r', 't', ' '] ++
                      s ++ [' ', 'i', 'n', 't', 'o', ' ', 'B', 'o', 'o', 'l',
                             'e', 'a', 'n'])));

bool :: [Char] -> Xml -> Sum [Char] Bool;
bool tag node = binda (text tag node) string2bool;

capRM2 ::
  forall a b.
    (Eq a,
      Eq b) => ((a, Nat) -> [(Term a b, Term a b)]) -> Term a b -> Hterm a b;
capRM2 uu (Var uv) = HHole;
capRM2 rm (Fun f ts) =
  let {
    h = HFun f (map (capRM2 rm) ts);
    n = size_list ts;
  } in (if any (\ r -> gcMatchI h (fst r)) (rm (f, n)) then HHole else h);

capRM ::
  forall a b.
    (Eq a,
      Eq b) => Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                 Term a b -> Hterm a b;
capRM r = (case r of {
            Nothing -> (\ _ -> HHole);
            Just a -> capRM2 a;
          });

dropa :: forall a. Nat -> [a] -> [a];
dropa n [] = [];
dropa n (x : xs) = (case n of {
                     Zero_nat -> x : xs;
                     Suc m -> dropa m xs;
                   });

empty :: forall a b. (Linorder a) => Rbt a b;
empty = Rbt Emptya;

class Times a where {
  times :: a -> a -> a;
};

class (Times a) => Dvd a where {
};

args :: forall a b. Term a b -> [Term a b];
args (Var x) = [];
args (Fun f ts) = ts;

root :: forall a b. Term a b -> Maybe a;
root (Var uu) = Nothing;
root (Fun f uv) = Just f;

membera :: forall a. (Eq a) => [a] -> a -> Bool;
membera [] y = False;
membera (x : xs) y = x == y || membera xs y;

instance (Eq a, Eq b) => Eq (Term a b) where {
  a == b = equal_term a b;
};

equal_term :: forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Bool;
equal_term (Fun f list) (Var v) = False;
equal_term (Var v) (Fun f list) = False;
equal_term (Fun fa lista) (Fun f list) = fa == f && lista == list;
equal_term (Var va) (Var v) = va == v;

supt_impl :: forall a b. (Eq a, Eq b) => Term a b -> Term a b -> Bool;
supt_impl (Var x) t = False;
supt_impl (Fun f ss) t = membera ss t || any (\ s -> supt_impl s t) ss;

supt :: forall a b. (Eq a, Eq b) => (Term a b, Term a b) -> Bool;
supt = (\ (a, b) -> supt_impl a b);

find :: forall a. (a -> Bool) -> [a] -> Maybe a;
find uu [] = Nothing;
find p (x : xs) = (if p x then Just x else find p xs);

class One a where {
  one :: a;
};

instance Times Integer where {
  times = (\ a b -> a * b);
};

instance Dvd Integer where {
};

one_int :: Integer;
one_int = 1;

instance One Integer where {
  one = one_int;
};

instance Orda Integer where {
  less_eq = (\ a b -> a <= b);
  less = (\ a b -> a < b);
};

foldla :: forall a b. (a -> b -> a) -> a -> [b] -> a;
foldla f a [] = a;
foldla f a (x : xs) = foldla f (f a x) xs;

foldra :: forall a b. (a -> b -> b) -> [a] -> b -> b;
foldra f [] a = a;
foldra f (x : xs) a = f x (foldra f xs a);

vec :: forall a. Nat -> [a] -> Bool;
vec n x = equal_nat (size_list x) n;

mat :: forall a. Nat -> Nat -> [[a]] -> Bool;
mat nr nc m = equal_nat (size_list m) nc && all (vec nr) m;

funpow :: forall a. Nat -> (a -> a) -> a -> a;
funpow Zero_nat f = id;
funpow (Suc n) f = f . funpow n f;

instance Orda Nat where {
  less_eq = less_eq_nat;
  less = less_nat;
};

paint :: forall a b. Color -> Rbta a b -> Rbta a b;
paint c Emptya = Emptya;
paint c (Branch uu l k v r) = Branch c l k v r;

balance :: forall a b. Rbta a b -> a -> b -> Rbta a b -> Rbta a b;
balance (Branch R a w x b) s t (Branch R c y z d) =
  Branch R (Branch B a w x b) s t (Branch B c y z d);
balance (Branch R (Branch R a w x b) s t c) y z Emptya =
  Branch R (Branch B a w x b) s t (Branch B c y z Emptya);
balance (Branch R (Branch R a w x b) s t c) y z (Branch B va vb vc vd) =
  Branch R (Branch B a w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch R Emptya w x (Branch R b s t c)) y z Emptya =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z Emptya);
balance (Branch R (Branch B va vb vc vd) w x (Branch R b s t c)) y z Emptya =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z Emptya);
balance (Branch R Emptya w x (Branch R b s t c)) y z (Branch B va vb vc vd) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch R (Branch B ve vf vg vh) w x (Branch R b s t c)) y z
  (Branch B va vb vc vd) =
  Branch R (Branch B (Branch B ve vf vg vh) w x b) s t
    (Branch B c y z (Branch B va vb vc vd));
balance Emptya w x (Branch R b s t (Branch R c y z d)) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z d);
balance (Branch B va vb vc vd) w x (Branch R b s t (Branch R c y z d)) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z d);
balance Emptya w x (Branch R (Branch R b s t c) y z Emptya) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z Emptya);
balance Emptya w x (Branch R (Branch R b s t c) y z (Branch B va vb vc vd)) =
  Branch R (Branch B Emptya w x b) s t (Branch B c y z (Branch B va vb vc vd));
balance (Branch B va vb vc vd) w x (Branch R (Branch R b s t c) y z Emptya) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t (Branch B c y z Emptya);
balance (Branch B va vb vc vd) w x
  (Branch R (Branch R b s t c) y z (Branch B ve vf vg vh)) =
  Branch R (Branch B (Branch B va vb vc vd) w x b) s t
    (Branch B c y z (Branch B ve vf vg vh));
balance Emptya s t Emptya = Branch B Emptya s t Emptya;
balance Emptya s t (Branch B va vb vc vd) =
  Branch B Emptya s t (Branch B va vb vc vd);
balance Emptya s t (Branch v Emptya vb vc Emptya) =
  Branch B Emptya s t (Branch v Emptya vb vc Emptya);
balance Emptya s t (Branch v (Branch B ve vf vg vh) vb vc Emptya) =
  Branch B Emptya s t (Branch v (Branch B ve vf vg vh) vb vc Emptya);
balance Emptya s t (Branch v Emptya vb vc (Branch B vf vg vh vi)) =
  Branch B Emptya s t (Branch v Emptya vb vc (Branch B vf vg vh vi));
balance Emptya s t
  (Branch v (Branch B ve vj vk vl) vb vc (Branch B vf vg vh vi)) =
  Branch B Emptya s t
    (Branch v (Branch B ve vj vk vl) vb vc (Branch B vf vg vh vi));
balance (Branch B va vb vc vd) s t Emptya =
  Branch B (Branch B va vb vc vd) s t Emptya;
balance (Branch B va vb vc vd) s t (Branch B ve vf vg vh) =
  Branch B (Branch B va vb vc vd) s t (Branch B ve vf vg vh);
balance (Branch B va vb vc vd) s t (Branch v Emptya vf vg Emptya) =
  Branch B (Branch B va vb vc vd) s t (Branch v Emptya vf vg Emptya);
balance (Branch B va vb vc vd) s t
  (Branch v (Branch B vi vj vk vl) vf vg Emptya) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v (Branch B vi vj vk vl) vf vg Emptya);
balance (Branch B va vb vc vd) s t
  (Branch v Emptya vf vg (Branch B vj vk vl vm)) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v Emptya vf vg (Branch B vj vk vl vm));
balance (Branch B va vb vc vd) s t
  (Branch v (Branch B vi vn vo vp) vf vg (Branch B vj vk vl vm)) =
  Branch B (Branch B va vb vc vd) s t
    (Branch v (Branch B vi vn vo vp) vf vg (Branch B vj vk vl vm));
balance (Branch v Emptya vb vc Emptya) s t Emptya =
  Branch B (Branch v Emptya vb vc Emptya) s t Emptya;
balance (Branch v Emptya vb vc (Branch B ve vf vg vh)) s t Emptya =
  Branch B (Branch v Emptya vb vc (Branch B ve vf vg vh)) s t Emptya;
balance (Branch v (Branch B vf vg vh vi) vb vc Emptya) s t Emptya =
  Branch B (Branch v (Branch B vf vg vh vi) vb vc Emptya) s t Emptya;
balance (Branch v (Branch B vf vg vh vi) vb vc (Branch B ve vj vk vl)) s t
  Emptya =
  Branch B (Branch v (Branch B vf vg vh vi) vb vc (Branch B ve vj vk vl)) s t
    Emptya;
balance (Branch v Emptya vf vg Emptya) s t (Branch B va vb vc vd) =
  Branch B (Branch v Emptya vf vg Emptya) s t (Branch B va vb vc vd);
balance (Branch v Emptya vf vg (Branch B vi vj vk vl)) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v Emptya vf vg (Branch B vi vj vk vl)) s t
    (Branch B va vb vc vd);
balance (Branch v (Branch B vj vk vl vm) vf vg Emptya) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v (Branch B vj vk vl vm) vf vg Emptya) s t
    (Branch B va vb vc vd);
balance (Branch v (Branch B vj vk vl vm) vf vg (Branch B vi vn vo vp)) s t
  (Branch B va vb vc vd) =
  Branch B (Branch v (Branch B vj vk vl vm) vf vg (Branch B vi vn vo vp)) s t
    (Branch B va vb vc vd);

ins ::
  forall a b.
    (Linorder a) => (a -> b -> b -> b) -> a -> b -> Rbta a b -> Rbta a b;
ins f k v Emptya = Branch R Emptya k v Emptya;
ins f k v (Branch B l x y r) =
  (if less k x then balance (ins f k v l) x y r
    else (if less x k then balance l x y (ins f k v r)
           else Branch B l x (f k y v) r));
ins f k v (Branch R l x y r) =
  (if less k x then Branch R (ins f k v l) x y r
    else (if less x k then Branch R l x y (ins f k v r)
           else Branch R l x (f k y v) r));

insert_with_key ::
  forall a b.
    (Linorder a) => (a -> b -> b -> b) -> a -> b -> Rbta a b -> Rbta a b;
insert_with_key f k v t = paint B (ins f k v t);

insertb :: forall a b. (Linorder a) => a -> b -> Rbta a b -> Rbta a b;
insertb = insert_with_key (\ _ _ nv -> nv);

impl_of :: forall a b. (Linorder a) => Rbt a b -> Rbta a b;
impl_of (Rbt x) = x;

insert :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
insert k v t = Rbt (insertb k v (impl_of t));

lookupc :: forall a b. (Linorder a) => Rbta a b -> a -> Maybe b;
lookupc Emptya k = Nothing;
lookupc (Branch uu l x y r) k =
  (if less k x then lookupc l k
    else (if less x k then lookupc r k else Just y));

lookupa :: forall a b. (Linorder a) => Rbt a b -> a -> Maybe b;
lookupa t = lookupc (impl_of t);

quotient_of :: Rat -> (Integer, Integer);
quotient_of (Frct x) = x;

times_rat :: Rat -> Rat -> Rat;
times_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * b, c * d));

instance Times Rat where {
  times = times_rat;
};

instance Dvd Rat where {
};

one_rat :: Rat;
one_rat = Frct (one_int, one_int);

instance One Rat where {
  one = one_rat;
};

less_eq_rat :: Rat -> Rat -> Bool;
less_eq_rat p q =
  let {
    (a, c) = quotient_of p;
    (b, d) = quotient_of q;
  } in a * d <= c * b;

less_rat :: Rat -> Rat -> Bool;
less_rat p q =
  let {
    (a, c) = quotient_of p;
    (b, d) = quotient_of q;
  } in a * d < c * b;

instance Orda Rat where {
  less_eq = less_eq_rat;
  less = less_rat;
};

member :: forall a. a -> (a -> Bool) -> Bool;
member x a = a x;

debug :: forall a. [Char] -> [Char] -> a -> a;
debug i t x = x;

fun_upd :: forall a b. (Eq a) => (a -> b) -> a -> b -> a -> b;
fun_upd f a b = (\ x -> (if x == a then b else f x));

class Plus a where {
  plus :: a -> a -> a;
};

class Zero a where {
  zero :: a;
};

implies :: Bool -> Bool -> Bool;
implies p True = True;
implies p False = not p;
implies True p = p;
implies False p = True;

instance Plus Integer where {
  plus = (\ a b -> a + b);
};

zero_int :: Integer;
zero_int = 0;

instance Zero Integer where {
  zero = zero_int;
};

data Pos = Empty | PCons Nat Pos;

class (One a, Times a) => Power a where {
};

powera :: forall a. a -> (a -> a -> a) -> a -> Nat -> a;
powera one times a (Suc n) = times a (powera one times a n);
powera one times a Zero_nat = one;

power :: forall a. (Power a) => a -> Nat -> a;
power = powera one times;

plus_rat :: Rat -> Rat -> Rat;
plus_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * d + b * c, c * d));

instance Plus Rat where {
  plus = plus_rat;
};

zero_rat :: Rat;
zero_rat = Frct (zero_int, one_int);

instance Zero Rat where {
  zero = zero_rat;
};

is_var :: forall a b. Term a b -> Bool;
is_var (Var uu) = True;
is_var (Fun uv uw) = False;

supteq :: forall a b. (Eq a, Eq b) => (Term a b, Term a b) -> Bool;
supteq = (\ (x, y) -> equal_term x y || supt_impl x y);

shows_nl :: [Char] -> [Char];
shows_nl = shows_prec_char Zero_nat newline;

shows_between ::
  ([Char] -> [Char]) ->
    ([Char] -> [Char]) -> ([Char] -> [Char]) -> [Char] -> [Char];
shows_between l p r = l . p . r;

shows_sep ::
  forall a.
    (a -> [Char] -> [Char]) -> ([Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_sep s sep [] = shows_string [];
shows_sep s sep [x] = s x;
shows_sep s sep (x : v : va) = s x . sep . shows_sep s sep (v : va);

shows_list_gen ::
  forall a.
    (a -> [Char] -> [Char]) ->
      [Char] -> [Char] -> [Char] -> [Char] -> [a] -> [Char] -> [Char];
shows_list_gen elt e l s r xs =
  (if null xs then shows_string e
    else shows_between (shows_string l) (shows_sep elt (shows_string s) xs)
           (shows_string r));

shows_term :: forall a b. (Showa a, Showa b) => Term a b -> [Char] -> [Char];
shows_term (Var x) = shows_prec Zero_nat x;
shows_term (Fun f ts) =
  shows_prec Zero_nat f .
    shows_list_gen id [] ['('] [',', ' '] [')'] (map shows_term ts);

shows_prec_term ::
  forall a b. (Showa a, Showa b) => Nat -> Term a b -> [Char] -> [Char];
shows_prec_term d t = shows_term t;

shows_rule ::
  forall a b. (Showa a, Showa b) => (Term a b, Term a b) -> [Char] -> [Char];
shows_rule (l, r) =
  shows_prec_term Zero_nat l .
    shows_string [' ', '-', '>', ' '] . shows_prec_term Zero_nat r;

shows_rules ::
  forall a b. (Showa a, Showa b) => [(Term a b, Term a b)] -> [Char] -> [Char];
shows_rules trs = shows_list_gen shows_rule [] [] [newline] [] trs . shows_nl;

shows_p ::
  forall a b. (Showa a, Showa b) => [(Term a b, Term a b)] -> [Char] -> [Char];
shows_p p =
  shows_string
    ['d', 'e', 'p', 'e', 'n', 'd', 'e', 'n', 'c', 'y', ' ', 'p', 'a', 'i', 'r',
      's', ':'] .
    shows_nl . shows_rules p;

instance Power Integer where {
};

remdups :: forall a. (Eq a) => [a] -> [a];
remdups [] = [];
remdups (x : xs) = (if membera xs x then remdups xs else x : remdups xs);

it_add ::
  forall a b c.
    (a -> b -> c -> c) -> ((a -> b -> c -> c) -> c -> c -> c) -> c -> c -> c;
it_add update iteratea m1 m2 = iteratea update m2 m1;

vec0I :: forall a. a -> Nat -> [a];
vec0I ze n = replicatea n ze;

mat0I :: forall a. a -> Nat -> Nat -> [[a]];
mat0I ze nr nc = replicatea nc (vec0I ze nr);

minus_nat :: Nat -> Nat -> Nat;
minus_nat (Suc m) (Suc n) = minus_nat m n;
minus_nat Zero_nat n = Zero_nat;
minus_nat m Zero_nat = m;

vec1I :: forall a. a -> a -> Nat -> Nat -> [a];
vec1I ze on n i =
  replicatea i ze ++ on : replicatea (minus_nat (minus_nat n one_nat) i) ze;

mat1I :: forall a. a -> a -> Nat -> [[a]];
mat1I ze on n = map (vec1I ze on n) (upt Zero_nat n);

less_eq_prod_aux ::
  forall a b. (Eq a, Orda a, Orda b) => (a, b) -> (a, b) -> Bool;
less_eq_prod_aux (a1, a2) (b1, b2) = less a1 b1 || a1 == b1 && less_eq a2 b2;

less_eq_prod :: forall a b. (Eq a, Orda a, Orda b) => (a, b) -> (a, b) -> Bool;
less_eq_prod a b = less_eq_prod_aux a b;

less_prod ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => (a, b) -> (a, b) -> Bool;
less_prod a b = not (a == b) && less_eq_prod_aux a b;

instance (Eq a, Orda a, Eq b, Orda b) => Orda (a, b) where {
  less_eq = less_eq_prod;
  less = less_prod;
};

instance Eq Nat where {
  a == b = equal_nat a b;
};

instance Preorder Nat where {
};

instance Order Nat where {
};

equal_rat :: Rat -> Rat -> Bool;
equal_rat a b = quotient_of a == quotient_of b;

instance Eq Rat where {
  a == b = equal_rat a b;
};

instance Power Rat where {
};

class Non_strict_order a where {
  ge :: a -> a -> Bool;
};

digit2string :: Nat -> [Char];
digit2string n =
  (if equal_nat n Zero_nat then ['0']
    else (if equal_nat n one_nat then ['1']
           else (if equal_nat n (nat 2) then ['2']
                  else (if equal_nat n (nat 3) then ['3']
                         else (if equal_nat n (nat 4) then ['4']
                                else (if equal_nat n (nat 5) then ['5']
                                       else (if equal_nat n (nat 6) then ['6']
      else (if equal_nat n (nat 7) then ['7']
             else (if equal_nat n (nat 8) then ['8']
                    else (if equal_nat n (nat 9) then ['9']
                           else ['n', 'o', 't', ' ', 'a', ' ', 'd', 'i', 'g',
                                  'i', 't']))))))))));

divmod_nat :: Nat -> Nat -> (Nat, Nat);
divmod_nat m n =
  (if equal_nat n Zero_nat || less_nat m n then (Zero_nat, m)
    else let {
           (q, a) = divmod_nat (minus_nat m n) n;
         } in (Suc q, a));

div_nat :: Nat -> Nat -> Nat;
div_nat m n = fst (divmod_nat m n);

mod_nat :: Nat -> Nat -> Nat;
mod_nat m n = snd (divmod_nat m n);

shows_nat :: Nat -> [Char] -> [Char];
shows_nat n =
  (if less_nat n (nat 10) then shows_string (digit2string n)
    else shows_nat (div_nat n (nat 10)) .
           shows_string (digit2string (mod_nat n (nat 10))));

shows_prec_nat :: Nat -> Nat -> [Char] -> [Char];
shows_prec_nat d n = shows_nat n;

shows_prec_int :: Nat -> Integer -> [Char] -> [Char];
shows_prec_int d i =
  (if i < 0
    then shows_prec_char Zero_nat '-' . shows_prec_nat Zero_nat (nat (0 - i))
    else shows_prec_nat Zero_nat (nat i));

shows_list_aux :: forall a. (a -> [Char] -> [Char]) -> [a] -> [Char] -> [Char];
shows_list_aux s xs = shows_list_gen s ['[', ']'] ['['] [','] [']'] xs;

shows_list_int :: [Integer] -> [Char] -> [Char];
shows_list_int is = shows_list_aux (shows_prec_int Zero_nat) is;

instance Showa Integer where {
  shows_prec = shows_prec_int;
  shows_list = shows_list_int;
};

shows_list_nat :: [Nat] -> [Char] -> [Char];
shows_list_nat ns = shows_list_aux (shows_prec_nat Zero_nat) ns;

instance Showa Nat where {
  shows_prec = shows_prec_nat;
  shows_list = shows_list_nat;
};

show_rat :: Rat -> [Char];
show_rat x =
  let {
    (den, num) = quotient_of x;
  } in (if num == 1 then shows_prec_int Zero_nat den []
         else ((shows_prec_int Zero_nat den . shows_prec_char Zero_nat '/') .
                shows_prec_int Zero_nat num)
                []);

shows_prec_rat :: Nat -> Rat -> [Char] -> [Char];
shows_prec_rat d x = shows_string (show_rat x);

shows_list_rat :: [Rat] -> [Char] -> [Char];
shows_list_rat rs = shows_list_aux (shows_prec_rat Zero_nat) rs;

instance Showa Rat where {
  shows_prec = shows_prec_rat;
  shows_list = shows_list_rat;
};

subt_at :: forall a b. Term a b -> Pos -> Term a b;
subt_at s Empty = s;
subt_at (Fun f ss) (PCons i p) = subt_at (nth ss i) p;

the_var :: forall a b. Term a b -> b;
the_var (Var x) = x;

map_term :: forall a b c. (a -> b) -> Term a c -> Term b c;
map_term fun (Var x) = Var x;
map_term fun (Fun f ts) = Fun (fun f) (map (map_term fun) ts);

map_rule ::
  forall a b c. (a -> b) -> (Term a c, Term a c) -> (Term b c, Term b c);
map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr));

data Lab a b = Lab (Lab a b) b | FunLab (Lab a b) [Lab a b] | UnLab a
  | Sharp (Lab a b);

maxaa :: forall a. (a -> a -> Bool) -> a -> a -> a;
maxaa less_eq a b = (if less_eq a b then b else a);

maxa :: forall a. (Orda a) => a -> a -> a;
maxa = maxaa less_eq;

max_list :: [Nat] -> Nat;
max_list [] = Zero_nat;
max_list (x : xs) = maxa x (max_list xs);

label_depth :: forall a b. Lab a b -> Nat;
label_depth (UnLab uu) = Zero_nat;
label_depth (Lab f uv) = Suc (label_depth f);
label_depth (FunLab f uw) = Suc (label_depth f);
label_depth (Sharp f) = Suc (label_depth f);

instance (Eq a, Eq b) => Eq (Lab a b) where {
  a == b = equal_lab a b;
};

equal_lab :: forall a b. (Eq a, Eq b) => Lab a b -> Lab a b -> Bool;
equal_lab (Sharp lab) (UnLab f) = False;
equal_lab (UnLab f) (Sharp lab) = False;
equal_lab (Sharp laba) (FunLab lab list) = False;
equal_lab (FunLab laba list) (Sharp lab) = False;
equal_lab (UnLab f) (FunLab lab list) = False;
equal_lab (FunLab lab list) (UnLab f) = False;
equal_lab (Sharp laba) (Lab lab l) = False;
equal_lab (Lab laba l) (Sharp lab) = False;
equal_lab (UnLab f) (Lab lab l) = False;
equal_lab (Lab lab l) (UnLab f) = False;
equal_lab (FunLab laba list) (Lab lab l) = False;
equal_lab (Lab laba l) (FunLab lab list) = False;
equal_lab (Sharp laba) (Sharp lab) = equal_lab laba lab;
equal_lab (UnLab fa) (UnLab f) = fa == f;
equal_lab (FunLab laba lista) (FunLab lab list) =
  equal_lab laba lab && lista == list;
equal_lab (Lab laba la) (Lab lab l) = equal_lab laba lab && la == l;

gen_label :: forall a b. Lab a b -> Nat -> Lab a b;
gen_label f Zero_nat = f;
gen_label f (Suc n) = FunLab (gen_label f n) [];

fmapa ::
  forall a b.
    (Eq a,
      Eq b) => Lab a b ->
                 Nat ->
                   [((Lab a b, Nat), [Lab a b])] -> Lab a b -> Nat -> Lab a b;
fmapa a nn sml =
  let {
    m = Suc (max_list (map label_depth (a : concatMap snd sml)));
  } in (\ f n -> (if (f, n) == (a, nn) then a else gen_label f m));

extract :: forall a. (a -> Bool) -> [a] -> Maybe ([a], (a, [a]));
extract p [] = Nothing;
extract p (x : xs) =
  (if p x then Just ([], (x, xs))
    else (case extract p xs of {
           Nothing -> Nothing;
           Just (ys, (y, zs)) -> Just (x : ys, (y, zs));
         }));

distinct :: forall a. (Eq a) => [a] -> Bool;
distinct [] = True;
distinct (x : xs) = not (membera xs x) && distinct xs;

vec_comp_all :: forall a. (a -> a -> Bool) -> [a] -> [a] -> Bool;
vec_comp_all rel v w = all (\ (a, b) -> rel a b) (zip v w);

mat_comp_all :: forall a. (a -> a -> Bool) -> [[a]] -> [[a]] -> Bool;
mat_comp_all rel m1 m2 = all (\ (a, b) -> vec_comp_all rel a b) (zip m1 m2);

mat_ge :: forall a. (Non_strict_order a) => [[a]] -> [[a]] -> Bool;
mat_ge = mat_comp_all ge;

times_nat :: Nat -> Nat -> Nat;
times_nat Zero_nat n = Zero_nat;
times_nat (Suc m) n = plus_nat n (times_nat m n);

minaa :: forall a. (a -> a -> Bool) -> a -> a -> a;
minaa less_eq a b = (if less_eq a b then a else b);

mina :: forall a. (Orda a) => a -> a -> a;
mina = minaa less_eq;

class (Preorder a) => Top a where {
  top :: a;
};

minus_rat :: Rat -> Rat -> Rat;
minus_rat p q =
  Frct (let {
          (a, c) = quotient_of p;
          (b, d) = quotient_of q;
        } in normalize (a * d - b * c, c * d));

shows_list_list :: forall a. (Showa a) => [[a]] -> [Char] -> [Char];
shows_list_list ls = shows_list_aux (shows_prec_list Zero_nat) ls;

instance (Showa a) => Showa [a] where {
  shows_prec = shows_prec_list;
  shows_list = shows_list_list;
};

shows_paren :: ([Char] -> [Char]) -> [Char] -> [Char];
shows_paren p = shows_prec_char Zero_nat '(' . p . shows_prec_char Zero_nat ')';

shows_prec_prod ::
  forall a b. (Showa a, Showa b) => Nat -> (a, b) -> [Char] -> [Char];
shows_prec_prod d p =
  shows_paren
    (shows_prec Zero_nat (fst p :: a) .
      shows_prec_char Zero_nat ',' . shows_prec Zero_nat (snd p :: b));

shows_list_prod ::
  forall a b. (Showa a, Showa b) => [(a, b)] -> [Char] -> [Char];
shows_list_prod ps = shows_list_aux (shows_prec_prod Zero_nat) ps;

instance (Showa a, Showa b) => Showa (a, b) where {
  shows_prec = shows_prec_prod;
  shows_list = shows_list_prod;
};

map_vars :: forall a b c. (a -> b) -> Term c a -> Term c b;
map_vars m (Var x) = Var (m x);
map_vars m (Fun f ts) = Fun f (map (map_vars m) ts);

num_args :: forall a b. Term a b -> Nat;
num_args (Var x) = Zero_nat;
num_args (Fun f ts) = size_list ts;

shows_list_term ::
  forall a b. (Showa a, Showa b) => [Term a b] -> [Char] -> [Char];
shows_list_term is = shows_list_aux (shows_prec_term Zero_nat) is;

instance (Showa a, Showa b) => Showa (Term a b) where {
  shows_prec = shows_prec_term;
  shows_list = shows_list_term;
};

shows_trs ::
  forall a b. (Showa a, Showa b) => [(Term a b, Term a b)] -> [Char] -> [Char];
shows_trs r =
  shows_string
    ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'y', 's', 't', 'e', 'm',
      ':'] .
    shows_nl . shows_rules r;

shows_dpp ::
  forall a b.
    (Showa a,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> [Char] -> [Char];
shows_dpp p r = shows_p p . shows_trs r;

unapp :: forall a b. (Eq a) => a -> Term a b -> (Term a b, [Term a b]);
unapp a (Var x) = (Var x, []);
unapp a (Fun f ss) =
  (if f == a && equal_nat (size_list ss) (nat 2)
    then let {
           (r, ts) = unapp a (nth ss Zero_nat);
         } in (r, ts ++ [nth ss one_nat])
    else (Fun f ss, []));

data Af_entry = Collapse Nat | AFList [Nat];

xml_many_elements ::
  forall a b.
    [Char] -> (Xml -> Sum [Char] a) -> ([a] -> b) -> Xml -> Sum [Char] b;
xml_many_elements tag p f (Xml name atts cs text) =
  (if name == tag && (null atts && is_none text)
    then binda (mapa p cs) (Inr . f)
    else no_transform tag (Xml name atts cs text));

xml3elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) -> (a -> b -> c -> d) -> Xml -> Sum [Char] d;
xml3elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (equal_nat (size_list cs) (nat 3) && is_none text))
    then binda (p1 (nth cs Zero_nat))
           (\ x1 ->
             binda (p2 (nth cs one_nat))
               (\ x2 -> binda (p3 (nth cs (nat 2))) (\ x3 -> Inr (f x1 x2 x3))))
    else no_transform tag (Xml name atts cs text));

position :: Xml -> Sum [Char] Nat;
position =
  xml_change (nata ['p', 'o', 's', 'i', 't', 'i', 'o', 'n'])
    (\ n -> minus_nat n one_nat);

afs ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [((a, Nat), Af_entry)];
afs xml2name =
  xml_many_elements
    ['a', 'r', 'g', 'u', 'm', 'e', 'n', 't', 'F', 'i', 'l', 't', 'e', 'r']
    (xml3elements
      ['a', 'r', 'g', 'u', 'm', 'e', 'n', 't', 'F', 'i', 'l', 't', 'e', 'r',
        'E', 'n', 't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (xml_options
        [(['c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'],
           xml_change (nata ['c', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'])
             (\ i -> Collapse (minus_nat i one_nat))),
          (['n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g'],
            xml_many_elements
              ['n', 'o', 'n', 'C', 'o', 'l', 'l', 'a', 'p', 's', 'i', 'n', 'g']
              position AFList)])
      (\ f a -> (\ b -> ((f, a), b))))
    id;

list2position :: [Nat] -> Pos;
list2position [] = Empty;
list2position (n : ns) = PCons n (list2position ns);

pos :: Xml -> Sum [Char] Pos;
pos = xml_many_elements
        ['p', 'o', 's', 'i', 't', 'i', 'o', 'n', 'I', 'n', 'T', 'e', 'r', 'm']
        position list2position;

xml1many_elements_gen ::
  forall a b c.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (a -> Xml -> Sum [Char] b) -> (a -> [b] -> c) -> Xml -> Sum [Char] c;
xml1many_elements_gen tag p1 p2 f (Xml name atts cs text) =
  (if name == tag && (null atts && (is_none text && not (null cs)))
    then binda (p1 (head cs))
           (\ x -> binda (mapa (p2 x) (tl cs)) (\ xs -> Inr (f x xs)))
    else no_transform tag (Xml name atts cs text));

xml1many_elements ::
  forall a b c.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) -> (a -> [b] -> c) -> Xml -> Sum [Char] c;
xml1many_elements tag p1 p2 = xml1many_elements_gen tag p1 (\ _ -> p2);

xml1element ::
  forall a b.
    [Char] -> (Xml -> Sum [Char] a) -> (a -> b) -> Xml -> Sum [Char] b;
xml1element tag p1 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (equal_nat (size_list cs) one_nat && is_none text))
    then binda (p1 (nth cs Zero_nat)) (\ x1 -> Inr (f x1))
    else no_transform tag (Xml name atts cs text));

var :: forall a. Xml -> Sum [Char] (Term a [Char]);
var x = binda (text ['v', 'a', 'r'] x) (Inr . Var);

term :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Term a [Char]);
term xml2name x =
  xml_options
    [(['v', 'a', 'r'], var),
      (['f', 'u', 'n', 'a', 'p', 'p'],
        xml1many_elements ['f', 'u', 'n', 'a', 'p', 'p'] xml2name
          (xml1element ['a', 'r', 'g'] (term xml2name) id) Fun)]
    x;

rule ::
  forall a.
    (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Term a [Char], Term a [Char]);
rule xml2name =
  xml2elements ['r', 'u', 'l', 'e']
    (xml1element ['l', 'h', 's'] (term xml2name) id)
    (xml1element ['r', 'h', 's'] (term xml2name) id) (\ a b -> (a, b));

scg_position :: Xml -> Sum [Char] Nat;
scg_position = nata ['p', 'o', 's', 'i', 't', 'i', 'o', 'n'];

map_filter :: forall a b. (a -> Maybe b) -> [a] -> [b];
map_filter f [] = [];
map_filter f (x : xs) =
  (case f x of {
    Nothing -> map_filter f xs;
    Just y -> y : map_filter f xs;
  });

scg ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          ((Term a [Char], Term a [Char]), ([(Nat, Nat)], [(Nat, Nat)]));
scg xml2name =
  xml1many_elements
    ['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'G', 'r', 'a', 'p', 'h']
    (rule xml2name)
    (xml3elements ['e', 'd', 'g', 'e'] scg_position
      (bool ['s', 't', 'r', 'i', 'c', 't']) scg_position
      (\ p s q -> (s, (p, q))))
    (\ lr edges ->
      (lr, (map_filter (\ x -> (if fst x then Just (snd x) else Nothing)) edges,
             map_filter (\ x -> (if not (fst x) then Just (snd x) else Nothing))
               edges)));

instance (Eq a) => Eq (Hterm a b) where {
  a == b = equal_hterm a b;
};

equal_hterm :: forall a b. (Eq a) => Hterm a b -> Hterm a b -> Bool;
equal_hterm (HFun f list) HHole = False;
equal_hterm HHole (HFun f list) = False;
equal_hterm (HFun fa lista) (HFun f list) = fa == f && lista == list;
equal_hterm HHole HHole = True;

partition :: forall a. (a -> Bool) -> [a] -> ([a], [a]);
partition p [] = ([], []);
partition p (x : xs) =
  let {
    (yes, no) = partition p xs;
  } in (if p x then (x : yes, no) else (yes, x : no));

vec_pre_gtI :: forall a. (a -> a -> Bool) -> [a] -> [a] -> Bool;
vec_pre_gtI gt v w = any (\ (a, b) -> gt a b) (zip v w);

mat_pre_gtI :: forall a. (a -> a -> Bool) -> [[a]] -> [[a]] -> Bool;
mat_pre_gtI gt m1 m2 = any (\ (a, b) -> vec_pre_gtI gt a b) (zip m1 m2);

sub_vec :: forall a. Nat -> [a] -> [a];
sub_vec = takea;

sub_mat :: forall a. Nat -> Nat -> [[a]] -> [[a]];
sub_mat nr nc m = map (sub_vec nr) (takea nc m);

mat_gtI ::
  forall a.
    (Non_strict_order a) => (a -> a -> Bool) -> Nat -> [[a]] -> [[a]] -> Bool;
mat_gtI gt sd m1 m2 =
  mat_ge m1 m2 && mat_pre_gtI gt (sub_mat sd sd m1) (sub_mat sd sd m2);

vec_map :: forall a. (a -> a) -> [a] -> [a];
vec_map = map;

mat_map :: forall a. (a -> a) -> [[a]] -> [[a]];
mat_map f = map (vec_map f);

instance (Eq a, Order a, Eq b, Order b) => Preorder (a, b) where {
};

instance (Eq a, Order a, Eq b, Order b) => Order (a, b) where {
};

class (Plus a) => Semigroup_add a where {
};

class (Semigroup_add a) => Ab_semigroup_add a where {
};

class (Times a) => Semigroup_mult a where {
};

class (Ab_semigroup_add a, Semigroup_mult a) => Semiring a where {
};

class (Semigroup_add a, Zero a) => Monoid_add a where {
};

class (Ab_semigroup_add a, Non_strict_order a) => Ordered_ab_semigroup a where {
};

class (Monoid_add a, Ordered_ab_semigroup a) => Max_ordered_monoid_add a where {
  max0 :: a -> a;
};

s_ins :: forall a b c. (a -> () -> b -> c) -> a -> b -> c;
s_ins update x s = update x () s;

poss_list :: forall a b. Term a b -> [Pos];
poss_list (Var x) = [Empty];
poss_list (Fun f ss) =
  Empty :
    concatMap (\ (i, a) -> map (PCons i) a)
      (zip (upt Zero_nat (size_list ss)) (map poss_list ss));

sharp_term :: forall a b. (a -> a) -> Term a b -> Term a b;
sharp_term shp (Var x) = Var x;
sharp_term shp (Fun f ss) = Fun (shp f) ss;

aarity :: forall a. (a -> Nat -> [a]) -> a -> Nat -> Nat;
aarity sm f n = minus_nat (size_list (sm f n)) one_nat;

list_diff :: forall a. (Eq a) => [a] -> [a] -> [a];
list_diff [] ys = [];
list_diff (x : xs) ys =
  let {
    zs = list_diff xs ys;
  } in (if membera ys x then zs else x : zs);

min_list :: forall a. (Linorder a) => [a] -> a;
min_list [x] = x;
min_list (x : v : va) = mina x (min_list (v : va));

quote :: ([Char] -> [Char]) -> [Char] -> [Char];
quote s =
  shows_between (shows_prec_char Zero_nat '\'') s
    (shows_prec_char Zero_nat '\'');

expecting :: forall a b. (Showa a) => [Char] -> [a] -> Sum [Char] (b, [a]);
expecting msg ts =
  Inl (['e', 'x', 'p', 'e', 'c', 't', 'i', 'n', 'g', ':', ' '] ++
        msg ++
          [',', ' ', 'b', 'u', 't', ' ', 'f', 'o', 'u', 'n', 'd', ':', ' '] ++
            quote (shows_prec_list Zero_nat (takea (nat 30) ts)) []);

eoi :: forall a. (Showa a) => [a] -> Sum [Char] ((), [a]);
eoi [] = Inr ((), []);
eoi (v : va) =
  expecting ['e', 'n', 'd', ' ', 'o', 'f', ' ', 'i', 'n', 'p', 'u', 't']
    (v : va);

xml_choice :: forall a. [Char] -> [Xml -> Sum [Char] a] -> Xml -> Sum [Char] a;
xml_choice e [] x =
  Inl (concat
        [['e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'p', 'a', 'r', 's', 'i',
           'n', 'g', ' ', 'c', 'h', 'o', 'i', 'c', 'e', ' ', 'f', 'o', 'r',
           ' '],
          e, [newline], shows_prec_xml Zero_nat x []]);
xml_choice e (p : ps) x =
  (case p x of {
    Inl _ -> xml_choice e ps x;
    Inr a -> Inr a;
  });

data Alter a b c d e f g h i j = Alter1 a | Alter2 b | Alter3 c | Alter4 d
  | Alter5 e | Alter6 f | Alter7 g | Alter8 h | Alter9 i | Alter10 j;

xml0elements :: forall a. [Char] -> a -> Xml -> Sum [Char] a;
xml0elements tag x (Xml name atts cs text) =
  (if name == tag && (null atts && (null cs && is_none text)) then Inr x
    else no_transform tag (Xml name atts cs text));

xml4elements ::
  forall a b c d e.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) ->
            (Xml -> Sum [Char] d) ->
              (a -> b -> c -> d -> e) -> Xml -> Sum [Char] e;
xml4elements tag p1 p2 p3 p4 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (equal_nat (size_list cs) (nat 4) && is_none text))
    then binda (p1 (nth cs Zero_nat))
           (\ x1 ->
             binda (p2 (nth cs one_nat))
               (\ x2 ->
                 binda (p3 (nth cs (nat 2)))
                   (\ x3 ->
                     binda (p4 (nth cs (nat 3)))
                       (\ x4 -> Inr (f x1 x2 x3 x4)))))
    else no_transform tag (Xml name atts cs text));

ctxt :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Ctxt a [Char]);
ctxt xml2name x =
  xml_options
    [(['b', 'o', 'x'], xml0elements ['b', 'o', 'x'] Hole),
      (['f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'],
        xml4elements ['f', 'u', 'n', 'C', 'o', 'n', 't', 'e', 'x', 't'] xml2name
          (xml_many_elements ['b', 'e', 'f', 'o', 'r', 'e'] (term xml2name) id)
          (ctxt xml2name)
          (xml_many_elements ['a', 'f', 't', 'e', 'r'] (term xml2name) id)
          More)]
    x;

xml3to4elements ::
  forall a b c d e.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) ->
            (Xml -> Sum [Char] d) ->
              (a -> b -> Maybe c -> d -> e) -> Xml -> Sum [Char] e;
xml3to4elements tag p1 p2 p3 p4 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             (null atts &&
               (less_eq_nat (nat 3) l &&
                 (less_eq_nat l (nat 4) && is_none text)))
         then binda (p1 (nth cs Zero_nat))
                (\ x1 ->
                  binda (p2 (nth cs one_nat))
                    (\ x2 ->
                      (if equal_nat l (nat 4)
                        then binda (p3 (nth cs (nat 2)))
                               (\ x3 ->
                                 binda (p4 (nth cs (nat 3)))
                                   (\ x4 -> Inr (f x1 x2 (Just x3) x4)))
                        else binda (p4 (nth cs (nat 2)))
                               (\ x4 -> Inr (f x1 x2 Nothing x4)))))
         else no_transform tag (Xml name atts cs text));

rstep ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          (Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])));
rstep xml2name =
  xml3to4elements ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 't', 'e', 'p'] pos
    (rule xml2name) (xml0elements ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e'] ())
    (term xml2name) (\ p r rel t -> (p, (r, (is_none rel, t))));

rsteps ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          (Term a [Char],
            [(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))]);
rsteps xml2name =
  xml1many_elements
    ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e']
    (xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm'] (term xml2name)
      id)
    (rstep xml2name) (\ a b -> (a, b));

subst ::
  forall a.
    (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [([Char], Term a [Char])];
subst xml2name =
  xml_many_elements ['s', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i', 'o', 'n']
    (xml2elements ['s', 'u', 'b', 's', 't', 'E', 'n', 't', 'r', 'y']
      (text ['v', 'a', 'r']) (term xml2name) (\ a b -> (a, b)))
    id;

loopa ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          (Term a [Char],
            ([(Pos, ((Term a [Char], Term a [Char]), (Bool, Term a [Char])))],
              ([([Char], Term a [Char])], Ctxt a [Char])));
loopa xml2name =
  xml3elements ['l', 'o', 'o', 'p'] (rsteps xml2name) (subst xml2name)
    (ctxt xml2name) (\ (s, rseq) sigma c -> (s, (rseq, (sigma, c))));

newtype ProjL a = Projection [(a, Nat)];

proj :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (ProjL a);
proj xml2name =
  xml_change (afs xml2name)
    (\ afl ->
      Projection
        (map (\ (a, b) ->
               let {
                 (f, aa) = a;
               } in (\ ba ->
                      (case ba of {
                        Collapse ab -> (f, ab);
                        AFList _ -> (f, aa);
                      }))
                 b)
          afl));

start_term ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] a) ->
                Term a [Char] -> Xml -> Sum [Char] (Term a [Char]);
start_term xml2name t =
  xml1element ['s', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm']
    (\ xml ->
      binda (term xml2name xml)
        (\ s ->
          (if equal_term s t then Inr t
            else Inl ['<', 's', 't', 'a', 'r', 't', 'T', 'e', 'r', 'm', '>',
                       ' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'm',
                       'a', 't', 'c', 'h', ' ', 'l', 'h', 's'])))
    id;

proj_term :: forall a b. (a -> Nat) -> Term a b -> Term a b;
proj_term p (Var x) = Var x;
proj_term p (Fun f ts) =
  (if less_nat (p f) (size_list ts) then nth ts (p f) else Fun f ts);

rseq ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] a) ->
                (a -> Nat) ->
                  (Term a [Char], Term a [Char]) ->
                    Xml ->
                      Sum [Char]
                        ((Term a [Char], Term a [Char]),
                          [(Pos, ((Term a [Char], Term a [Char]),
                                   Term a [Char]))]);
rseq xml2name pia r =
  xml1many_elements
    ['r', 'e', 'w', 'r', 'i', 't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e']
    (start_term xml2name (proj_term pia (fst r))) (rstep xml2name)
    (\ _ rseq -> (r, map (\ (x, (y, (_, z))) -> (x, (y, z))) rseq));

instance Semigroup_add Integer where {
};

instance Ab_semigroup_add Integer where {
};

instance Semigroup_mult Integer where {
};

instance Semiring Integer where {
};

mat_max0 :: forall a. (Max_ordered_monoid_add a) => [[a]] -> [[a]];
mat_max0 = mat_map max0;

instance Linorder Nat where {
};

data Tpoly a b = PVar a | PNum b | PSum [Tpoly a b] | PMult [Tpoly a b];

instance Semigroup_add Rat where {
};

instance Ab_semigroup_add Rat where {
};

instance Semigroup_mult Rat where {
};

instance Semiring Rat where {
};

class (Times a, Zero a) => Mult_zero a where {
};

s_memb :: forall a b c. (a -> b -> Maybe c) -> a -> b -> Bool;
s_memb lookupa x s = not (is_none (lookupa x s));

ctxt_apply :: forall a b. Ctxt a b -> Term a b -> Term a b;
ctxt_apply Hole s = s;
ctxt_apply (More f ss1 c ss2) s = Fun f (ss1 ++ ctxt_apply c s : ss2);

equal_ctxt :: forall a b. (Eq a, Eq b) => Ctxt a b -> Ctxt a b -> Bool;
equal_ctxt (More f list1 ctxt list2) Hole = False;
equal_ctxt Hole (More f list1 ctxt list2) = False;
equal_ctxt (More fa list1a ctxta list2a) (More f list1 ctxt list2) =
  fa == f && (list1a == list1 && (equal_ctxt ctxta ctxt && list2a == list2));
equal_ctxt Hole Hole = True;

shows_ctxt :: forall a b. (Showa a, Showa b) => Ctxt a b -> [Char] -> [Char];
shows_ctxt Hole = shows_string ['[', ']'];
shows_ctxt (More f ss1 d ss2) =
  shows_prec Zero_nat f .
    shows_string ['('] .
      shows_list_gen (shows_prec_term Zero_nat) [] [] [',', ' '] [',', ' ']
        ss1 .
        shows_ctxt d .
          shows_list_gen (shows_prec_term Zero_nat) [')'] [',', ' '] [',', ' ']
            [')'] ss2;

hvf_top :: forall a b. (Eq a) => a -> Nat -> Term a b -> Bool;
hvf_top a n (Fun f ts) =
  implies (f == a && equal_nat (size_list ts) n) (not (is_var (head ts)));
hvf_top a n (Var uu) = False;

list_inter :: forall a. (Eq a) => [a] -> [a] -> [a];
list_inter [] bs = [];
list_inter (a : asa) bs =
  (if membera bs a then a : list_inter asa bs else list_inter asa bs);

list_union :: forall a. (Eq a) => [a] -> [a] -> [a];
list_union [] ys = ys;
list_union (x : xs) ys =
  let {
    zs = list_union xs ys;
  } in (if membera zs x then zs else x : zs);

bind ::
  forall a b c.
    ([a] -> Sum [Char] (b, [a])) ->
      (b -> [a] -> Sum [Char] (c, [a])) -> [a] -> Sum [Char] (c, [a]);
bind m f ts = binda (m ts) (\ (a, b) -> f a b);

many :: (Char -> Bool) -> [Char] -> Sum [Char] ([Char], [Char]);
many p [] = Inr ([], []);
many p (t : ts) =
  (if p t then binda (many p ts) (\ (rs, tsa) -> Inr (t : rs, tsa))
    else Inr ([], t : ts));

wspace :: [Char];
wspace = [newline, '\t', '\r', ' '];

trim :: [Char] -> [Char];
trim [] = [];
trim (c : s) = (if membera wspace c then trim s else c : s);

data Domain = Naturala Nat | Arctic | Arctic_rat Rat | Int_mat Nat Nat
  | Arc_mat Nat | Arc_rat_mat Rat Nat | Rationala Rat Nat | Rat_mat Rat Nat Nat;

label :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Sum [Nat] [a]);
label xml2name =
  xml_options
    [(['n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l'],
       xml_many_elements ['n', 'u', 'm', 'b', 'e', 'r', 'L', 'a', 'b', 'e', 'l']
         (nata ['n', 'u', 'm', 'b', 'e', 'r']) Inl),
      (['s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l'],
        xml_many_elements
          ['s', 'y', 'm', 'b', 'o', 'l', 'L', 'a', 'b', 'e', 'l'] xml2name
          Inr)];

rules ::
  forall a.
    (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [(Term a [Char], Term a [Char])];
rules xml2name = xml_many_elements ['r', 'u', 'l', 'e', 's'] (rule xml2name) id;

state :: Xml -> Sum [Char] [Char];
state = text ['s', 't', 'a', 't', 'e'];

isOK :: forall a b. Sum a b -> Bool;
isOK m = (case m of {
           Inl _ -> False;
           Inr _ -> True;
         });

instance Mult_zero Integer where {
};

list_update :: forall a. [a] -> Nat -> a -> [a];
list_update (x : xs) (Suc i) y = x : list_update xs i y;
list_update (x : xs) Zero_nat y = y : xs;
list_update [] i y = [];

base :: forall a. (a, Nat) -> a;
base (f, h) = f;

lift :: forall a. Nat -> a -> (a, Nat);
lift h f = (f, h);

vars_term_lista :: forall a b. Term a b -> [b];
vars_term_lista (Var x) = [x];
vars_term_lista (Fun uu ts) = concatMap vars_term_lista ts;

vars_term_list :: forall a b. (Eq b) => Term a b -> [b];
vars_term_list = remdups . vars_term_lista;

roof :: forall a b. (Eq b) => (Term a b, Term a b) -> Term a b -> Bool;
roof (l, r) =
  let {
    xs = vars_term_list r;
  } in (\ t -> let {
                 xt = vars_term_list t;
               } in all (membera xt) xs);

mat_monoI :: forall a. (a -> Bool) -> Nat -> [[a]] -> Bool;
mat_monoI geq1 sd m = all (any geq1) (sub_mat sd sd m);

scalar_prodI ::
  forall a. a -> (a -> a -> a) -> (a -> a -> a) -> [a] -> [a] -> a;
scalar_prodI ze pl ti v w = foldra (\ (x, y) -> pl (ti x y)) (zip v w) ze;

matT_vec_multI ::
  forall a. a -> (a -> a -> a) -> (a -> a -> a) -> [[a]] -> [a] -> [a];
matT_vec_multI ze pl ti m v = map (\ w -> scalar_prodI ze pl ti w v) m;

transpose :: forall a. Nat -> [[a]] -> [[a]];
transpose nr [] = replicatea nr [];
transpose nr (v : m) = map (\ (a, b) -> a : b) (zip v (transpose nr m));

mat_multI ::
  forall a.
    a -> (a -> a -> a) -> (a -> a -> a) -> Nat -> [[a]] -> [[a]] -> [[a]];
mat_multI ze pl ti nr m1 m2 =
  map (matT_vec_multI ze pl ti (transpose nr m1)) m2;

vec_plusI :: forall a. (a -> a -> a) -> [a] -> [a] -> [a];
vec_plusI pl v w = map (\ xy -> pl (fst xy) (snd xy)) (zip v w);

mat_plusI :: forall a. (a -> a -> a) -> [[a]] -> [[a]] -> [[a]];
mat_plusI pl m1 m2 = map (\ uv -> vec_plusI pl (fst uv) (snd uv)) (zip m1 m2);

equal_color :: Color -> Color -> Bool;
equal_color B R = False;
equal_color R B = False;
equal_color B B = True;
equal_color R R = True;

equal_rbta :: forall a b. (Eq a, Eq b) => Rbta a b -> Rbta a b -> Bool;
equal_rbta (Branch color rbt1 a b rbt2) Emptya = False;
equal_rbta Emptya (Branch color rbt1 a b rbt2) = False;
equal_rbta (Branch colora rbt1a aa ba rbt2a) (Branch color rbt1 a b rbt2) =
  equal_color colora color &&
    (equal_rbta rbt1a rbt1 && (aa == a && (ba == b && equal_rbta rbt2a rbt2)));
equal_rbta Emptya Emptya = True;

equal_rbt :: forall a b. (Eq a, Linorder a, Eq b) => Rbt a b -> Rbt a b -> Bool;
equal_rbt ra r = equal_rbta (impl_of ra) (impl_of r);

instance (Eq a, Linorder a, Eq b) => Eq (Rbt a b) where {
  a == b = equal_rbt a b;
};

instance Mult_zero Rat where {
};

class (Ab_semigroup_add a, Monoid_add a) => Comm_monoid_add a where {
};

class (Comm_monoid_add a, Mult_zero a, Semiring a) => Semiring_0 a where {
};

class (Semigroup_mult a, Power a) => Monoid_mult a where {
};

class (One a, Zero a) => Zero_neq_one a where {
};

class (Monoid_mult a, Semiring_0 a, Zero_neq_one a) => Semiring_1 a where {
};

s_alpha :: forall a b. (a -> b -> Maybe ()) -> a -> b -> Bool;
s_alpha alpha m = dom (alpha m);

s_empty :: forall a. a -> a;
s_empty empt = empt;

s_union :: forall a. a -> a;
s_union add = add;

linear_term_impl :: forall a b. (Eq a) => [a] -> Term b a -> Maybe [a];
linear_term_impl xs (Var x) = (if membera xs x then Nothing else Just (x : xs));
linear_term_impl xs (Fun uu []) = Just xs;
linear_term_impl xs (Fun f (t : ts)) =
  (case linear_term_impl xs t of {
    Nothing -> Nothing;
    Just ys -> linear_term_impl ys (Fun f ts);
  });

linear_term :: forall a b. (Eq b) => Term a b -> Bool;
linear_term t =
  (case linear_term_impl [] t of {
    Nothing -> False;
    Just _ -> True;
  });

map_term_wa :: forall a b c. (a -> Nat -> b) -> Term a c -> Term b c;
map_term_wa fg (Var x) = Var x;
map_term_wa fg (Fun f ts) = Fun (fg f (size_list ts)) (map (map_term_wa fg) ts);

rp_af ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      a -> Nat -> Nat -> Bool;
rp_af (v, (s, (n, (a, (m, d))))) = a;

data Ta_rule a b = TA_rule b [a] a;

rq :: forall a b. Ta_rule a b -> a;
rq (TA_rule f qs q) = q;

map_rules_wa ::
  forall a b c.
    (a -> Nat -> b) -> [(Term a c, Term a c)] -> [(Term b c, Term b c)];
map_rules_wa fg r = map (\ (l, ra) -> (map_term_wa fg l, map_term_wa fg ra)) r;

map_trs_list ::
  forall a b c. (a -> b) -> [(Term a c, Term a c)] -> [(Term b c, Term b c)];
map_trs_list fg r = map (map_rule fg) r;

supteq_list :: forall a b. Term a b -> [Term a b];
supteq_list (Var x) = [Var x];
supteq_list (Fun f ts) = Fun f ts : concatMap supteq_list ts;

dP_list ::
  forall a b.
    (Eq a,
      Eq b) => (a -> a) ->
                 [(Term a b, Term a b)] ->
                   Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                     [(Term a b, Term a b)];
dP_list shp r rm =
  let {
    rma = the rm;
  } in concatMap
         (\ lr ->
           let {
             l = fst lr;
             s = sharp_term shp l;
           } in map_filter
                  (\ x ->
                    (if not (member (l, x) supt) &&
                          (not (is_var x) &&
                            not (null (rma (the (root x), num_args x))))
                      then Just (s, sharp_term shp x) else Nothing))
                  (supteq_list (snd lr)))
         r;

hvf_term :: forall a b. (Eq a) => a -> Term a b -> Bool;
hvf_term a t =
  (case unapp a t of {
    (Var _, ts) -> null ts;
    (Fun _ us, ts) -> all (hvf_term a) (us ++ ts);
  });

oneof_aux :: [[Char]] -> [[Char]] -> [Char] -> Sum [Char] ([Char], [Char]);
oneof_aux allowed [] ts =
  expecting (['o', 'n', 'e', ' ', 'o', 'f', ' '] ++ shows_list_list allowed [])
    ts;
oneof_aux allowed (x : xs) ts =
  (if map snd (zip x ts) == x then Inr (x, trim (dropa (size_list x) ts))
    else oneof_aux allowed xs ts);

oneof :: [[Char]] -> [Char] -> Sum [Char] ([Char], [Char]);
oneof xs = oneof_aux xs xs;

data Xmldoc = Xmldoc [Char] Xml;

xml5elements ::
  forall a b c d e f.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) ->
            (Xml -> Sum [Char] d) ->
              (Xml -> Sum [Char] e) ->
                (a -> b -> c -> d -> e -> f) -> Xml -> Sum [Char] f;
xml5elements tag p1 p2 p3 p4 p5 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (equal_nat (size_list cs) (nat 5) && is_none text))
    then binda (p1 (nth cs Zero_nat))
           (\ x1 ->
             binda (p2 (nth cs one_nat))
               (\ x2 ->
                 binda (p3 (nth cs (nat 2)))
                   (\ x3 ->
                     binda (p4 (nth cs (nat 3)))
                       (\ x4 ->
                         binda (p5 (nth cs (nat 4)))
                           (\ x5 -> Inr (f x1 x2 x3 x4 x5))))))
    else no_transform tag (Xml name atts cs text));

catcha :: forall a b c. Sum a b -> (a -> Sum c b) -> Sum c b;
catcha m f = (case m of {
               Inl a -> f a;
               Inr a -> Inr a;
             });

forallm :: forall a b. (a -> Sum b ()) -> [a] -> Sum (a, b) ();
forallm f [] = Inr ();
forallm f (x : xs) =
  binda (catcha (f x) (\ xa -> Inl (x, xa))) (\ _ -> forallm f xs);

check :: forall a. Bool -> a -> Sum a ();
check b e = (if b then Inr () else Inl e);

check_edges ::
  forall a.
    (a -> [Char] -> [Char]) ->
      ((a, a) -> Bool) -> [a] -> [a] -> Sum ([Char] -> [Char]) ();
check_edges ss g c d =
  catcha
    (forallm
      (\ n ->
        catcha
          (forallm
            (\ m ->
              check (not (g (m, n)))
                (shows_string
                   ['e', 'd', 'g', 'e', ' ', 'f', 'r', 'o', 'm', ' '] .
                  ss m . shows_string [' ', 't', 'o', ' '] . ss n))
            d)
          (\ x -> Inl (snd x)))
      c)
    (\ x -> Inl (snd x));

instance Monoid_add Integer where {
};

instance Comm_monoid_add Integer where {
};

instance Semiring_0 Integer where {
};

instance Monoid_mult Integer where {
};

instance Zero_neq_one Integer where {
};

instance Semiring_1 Integer where {
};

match :: forall a b. (Term a b, Term a b) -> Term a b -> Bool;
match = (\ _ _ -> True);

instance (Eq a, Linorder a, Eq b, Linorder b) => Linorder (a, b) where {
};

top_bool :: Bool;
top_bool = True;

instance Top Bool where {
  top = top_bool;
};

top_fun :: forall a b. (Top b) => a -> b;
top_fun = (\ _ -> top);

equal_pos :: Pos -> Pos -> Bool;
equal_pos (PCons nat pos) Empty = False;
equal_pos Empty (PCons nat pos) = False;
equal_pos (PCons nata posa) (PCons nat pos) =
  equal_nat nata nat && equal_pos posa pos;
equal_pos Empty Empty = True;

instance Eq Pos where {
  a == b = equal_pos a b;
};

data Arctic = MinInfty | Num_arc Integer;

data Arctic_rat = MinInfty_rat | Num_arc_rat Rat;

data Redpair a = Int_carrier [((a, Nat), (Integer, [Integer]))]
  | Int_nl_carrier [((a, Nat), [([(Nat, Nat)], Integer)])]
  | Rat_carrier Rat [((a, Nat), (Rat, [Rat]))]
  | Rat_nl_carrier Rat [((a, Nat), [([(Nat, Nat)], Rat)])]
  | Arctic_carrier [((a, Nat), (Arctic, [Arctic]))]
  | Arctic_rat_carrier Rat [((a, Nat), (Arctic_rat, [Arctic_rat]))]
  | Int_mat_carrier Nat Nat [((a, Nat), ([[Integer]], [[[Integer]]]))]
  | Rat_mat_carrier Rat Nat Nat [((a, Nat), ([[Rat]], [[[Rat]]]))]
  | Arc_mat_carrier Nat [((a, Nat), ([[Arctic]], [[[Arctic]]]))]
  | Arc_rat_mat_carrier Rat Nat [((a, Nat), ([[Arctic_rat]], [[[Arctic_rat]]]))]
  | Lpo [((a, Nat), Nat)] [((a, Nat), Af_entry)];

data ArithFun = Arg Nat | Const Nat | Sum [ArithFun] | Max [ArithFun]
  | Min [ArithFun] | Prod [ArithFun];

data Sl_inter a b = SL_Inter Nat [((Lab a b, Nat), ArithFun)];

data DPProof a b c = PisEmpty
  | RedPairProc (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      (DPProof a b c)
  | RedPairUrProc (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | DepGraphProc
      [(Maybe (DPProof a b c), [(Term (Lab a b) c, Term (Lab a b) c)])]
  | MonoRedPairProc (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | MonoRedPairUrProc (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | UnlabProc [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | SemlabProc (Sl_inter a b)
      (Sum (Lab a b -> [Nat] -> Sum b [Lab a b]) (c, Bool))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | RootlabProc (Maybe (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | SubtermCriterionProc (ProjL (Lab a b))
      [((Term (Lab a b) c, Term (Lab a b) c),
         [(Pos, ((Term (Lab a b) c, Term (Lab a b) c), Term (Lab a b) c))])]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | SizeChangeSubTerm
      [((Term (Lab a b) c, Term (Lab a b) c), ([(Nat, Nat)], [(Nat, Nat)]))]
  | SizeChangeRedPair (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      [((Term (Lab a b) c, Term (Lab a b) c), ([(Nat, Nat)], [(Nat, Nat)]))]
  | UncurryProc (Maybe Nat)
      (Lab a b,
        ([((Lab a b, Nat), [Lab a b])],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | FccProc (Lab a b) [Ctxt (Lab a b) c] [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c);

data Boundstype = Roof | Match;

data Tree_automaton a b = Tree_Automaton [a] [Ta_rule a b] [(a, a)];

data Bounds_info a b =
  Bounds_Info Boundstype Nat [b] (Tree_automaton b (a, Nat));

data TRSProof a b c =
  DPTrans [(Term (Lab a b) c, Term (Lab a b) c)] (DPProof a b c)
  | RuleRemoval (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
      (TRSProof a b c)
  | Unlab [(Term (Lab a b) c, Term (Lab a b) c)] (TRSProof a b c)
  | Semlab (Sl_inter a b) (Sum (Lab a b -> [Nat] -> Sum b [Lab a b]) c)
      [(Term (Lab a b) c, Term (Lab a b) c)] (TRSProof a b c)
  | Rootlab [(Term (Lab a b) c, Term (Lab a b) c)] (TRSProof a b c)
  | Fcc [Ctxt (Lab a b) c] [(Term (Lab a b) c, Term (Lab a b) c)]
      (TRSProof a b c)
  | StringReversal [(Term (Lab a b) c, Term (Lab a b) c)] (TRSProof a b c)
  | Bounds (Bounds_info (Lab a b) c)
  | Uncurry
      (Lab a b,
        ([((Lab a b, Nat), [Lab a b])],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])))
      [(Term (Lab a b) c, Term (Lab a b) c)] (TRSProof a b c)
  | RisEmpty;

rm_update :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
rm_update = insert;

itoi_iterateo ::
  forall a b c d.
    ((a -> Bool) -> (b -> c -> a -> a) -> d -> a -> a) ->
      (b -> c -> a -> a) -> d -> a -> a;
itoi_iterateo it = it (\ _ -> True);

rm_iterateoi ::
  forall a b c. (a -> Bool) -> (b -> c -> a -> a) -> Rbta b c -> a -> a;
rm_iterateoi c f Emptya sigma = sigma;
rm_iterateoi c f (Branch col l k v r) sigma =
  (if c sigma
    then let {
           sigmaa = rm_iterateoi c f l sigma;
         } in (if c sigmaa then rm_iterateoi c f r (f k v sigmaa) else sigmaa)
    else sigma);

rm_iterateoia ::
  forall a b c.
    (Linorder b) => (a -> Bool) -> (b -> c -> a -> a) -> Rbt b c -> a -> a;
rm_iterateoia c f r = rm_iterateoi c f (impl_of r);

rm_iterateo ::
  forall a b c. (Linorder a) => (a -> b -> c -> c) -> Rbt a b -> c -> c;
rm_iterateo = itoi_iterateo rm_iterateoia;

rm_iterate ::
  forall a b c. (Linorder a) => (a -> b -> c -> c) -> Rbt a b -> c -> c;
rm_iterate = rm_iterateo;

rm_add :: forall a b. (Linorder a) => Rbt a b -> Rbt a b -> Rbt a b;
rm_add = it_add rm_update rm_iterate;

rs_ins :: forall a. (Linorder a) => a -> Rbt a () -> Rbt a ();
rs_ins = s_ins rm_update;

anyNstriE :: forall a b. (a -> b -> ((Bool, Bool), a)) -> a -> [b] -> (Bool, a);
anyNstriE f m [] = (False, m);
anyNstriE f m (s : ss) =
  let {
    (a, b) = f m s;
  } in let {
         (_, nstri) = a;
       } in (\ ma -> (if nstri then (nstri, ma) else anyNstriE f ma ss))
         b;

allStriE :: forall a b. (a -> b -> ((Bool, Bool), a)) -> a -> [b] -> (Bool, a);
allStriE f m [] = (True, m);
allStriE f m (t : ts) =
  let {
    (a, b) = f m t;
  } in let {
         (stri, _) = a;
       } in (\ ma -> (if stri then allStriE f ma ts else (False, ma)))
         b;

lex_extE ::
  forall a b.
    (a -> b -> b -> ((Bool, Bool), a)) -> a -> [b] -> [b] -> ((Bool, Bool), a);
lex_extE f m [] [] = ((False, True), m);
lex_extE f m (uu : uv) [] = ((True, True), m);
lex_extE f m [] (uw : ux) = ((False, False), m);
lex_extE f m (a : asa) (b : bs) =
  let {
    (aa, ba) = f m a b;
  } in let {
         (stri, nstri) = aa;
       } in (\ ma ->
              (if stri then ((True, True), ma)
                else (if nstri then lex_extE f ma asa bs
                       else ((False, False), ma))))
         ba;

lpoE_m ::
  forall a b c.
    (Eq c) => (a -> Term b c -> Term b c -> Maybe (Bool, Bool),
                a -> Term b c -> Term b c -> (Bool, Bool) -> a) ->
                a -> (b -> Nat -> Nat) ->
                       Term b c -> Term b c -> ((Bool, Bool), a);
lpoE_m m_ops m pr s t =
  (case let {
          (x, _) = m_ops;
        } in x
          m
          s
          t
    of {
    Nothing ->
      let {
        (v, new_m) =
          (case (s, t) of {
            (Var x, Var y) -> ((False, x == y), m);
            (Var x, Fun g ts) ->
              ((False, null ts && equal_nat (pr g Zero_nat) Zero_nat), m);
            (Fun f ss, Var _) ->
              let {
                (con, a) = anyNstriE (\ lm sa -> lpoE_m m_ops lm pr sa t) m ss;
              } in ((con, con), a);
            (Fun f ss, Fun g ts) ->
              (case anyNstriE (\ lm sa -> lpoE_m m_ops lm pr sa t) m ss of {
                (True, ma) -> ((True, True), ma);
                (False, ma) ->
                  let {
                    prl = pr f (size_list ss);
                    prr = pr g (size_list ts);
                    (sAllStri, mb) =
                      allStriE (\ lm -> lpoE_m m_ops lm pr s) ma ts;
                  } in (if less_eq_nat prr prl && sAllStri
                         then (if less_nat prr prl then ((True, True), mb)
                                else lex_extE (\ lm -> lpoE_m m_ops lm pr) mb ss
                                       ts)
                         else ((False, False), mb));
              });
          });
      } in (v, let {
                 (_, y) = m_ops;
               } in y
                 new_m
                 s
                 t
                 v);
    Just v -> (v, m);
  });

lookup_2 :: forall a b. (Eq a) => [(a, [(a, b)])] -> a -> a -> Maybe b;
lookup_2 [] s t = Nothing;
lookup_2 (heada : taila) s t =
  (if fst heada == s then lookupb t (snd heada) else lookup_2 taila s t);

store_2 ::
  forall a b. (Eq a) => [(a, [(a, b)])] -> a -> a -> b -> [(a, [(a, b)])];
store_2 [] s t v = [(s, [(t, v)])];
store_2 (heada : taila) s t v =
  (if fst heada == s then (s, (t, v) : snd heada) : taila
    else heada : store_2 taila s t v);

lpo_E_2 ::
  forall a b.
    (Eq a, Eq b) => (a -> Nat -> Nat) -> Term a b -> Term a b -> (Bool, Bool);
lpo_E_2 pr s t = fst (lpoE_m (lookup_2, store_2) [] pr s t);

lpo_unbounded ::
  forall a b.
    (Eq a, Eq b) => (a -> Nat -> Nat) -> Term a b -> Term a b -> (Bool, Bool);
lpo_unbounded = lpo_E_2;

instance Monoid_add Rat where {
};

instance Comm_monoid_add Rat where {
};

instance Semiring_0 Rat where {
};

instance Monoid_mult Rat where {
};

instance Zero_neq_one Rat where {
};

instance Semiring_1 Rat where {
};

equal_sum :: forall a b. (Eq a, Eq b) => Sum a b -> Sum a b -> Bool;
equal_sum (Inr b) (Inl a) = False;
equal_sum (Inl a) (Inr b) = False;
equal_sum (Inr ba) (Inr b) = ba == b;
equal_sum (Inl aa) (Inl a) = aa == a;

instance (Eq a, Eq b) => Eq (Sum a b) where {
  a == b = equal_sum a b;
};

match_vars ::
  forall a b. (Eq a, Eq b) => [(a, Term b a)] -> Maybe (a -> Term b a);
match_vars [] = Just Var;
match_vars ((x, t) : ps) =
  (case lookupb x ps of {
    Nothing ->
      (case match_vars ps of {
        Nothing -> Nothing;
        Just sigma -> Just (fun_upd sigma x t);
      });
    Just s -> (if equal_term t s then match_vars ps else Nothing);
  });

match_term_list ::
  forall a b. (Eq a) => [(Term a b, Term a b)] -> Maybe [(b, Term a b)];
match_term_list [] = Just [];
match_term_list ((Var uu, Fun uv uw) : ps) = Nothing;
match_term_list ((Fun f ts, Fun g ls) : ps) =
  (if f == g
    then (case decomp ts ls of {
           Nothing -> Nothing;
           Just eqs -> match_term_list (eqs ++ ps);
         })
    else Nothing);
match_term_list ((t, Var x) : ps) =
  (case match_term_list ps of {
    Nothing -> Nothing;
    Just eqs -> Just ((x, t) : eqs);
  });

match_list ::
  forall a b. (Eq a, Eq b) => [(Term a b, Term a b)] -> Maybe (b -> Term a b);
match_list tls =
  (case match_term_list tls of {
    Nothing -> Nothing;
    Just a -> match_vars a;
  });

instance_rule ::
  forall a b.
    (Eq a, Eq b) => (Term a b, Term a b) -> (Term a b, Term a b) -> Bool;
instance_rule lr st =
  not (is_none (match_list [(fst lr, fst st), (snd lr, snd st)]));

check_pairwise :: forall a b. (a -> a -> Sum b ()) -> [a] -> Sum b ();
check_pairwise c [] = Inr ();
check_pairwise c (x : xs) =
  binda (catcha (forallm (c x) xs) (\ xa -> Inl (snd xa)))
    (\ _ -> check_pairwise c xs);

check_disjoint :: forall a. (Eq a) => [a] -> [a] -> Sum a ();
check_disjoint xs ys =
  catcha (forallm (\ x -> (if not (membera ys x) then Inr () else Inl x)) xs)
    (\ x -> Inl (snd x));

check_partition :: forall a. (Eq a) => [[a]] -> Sum a ();
check_partition xss = check_pairwise check_disjoint xss;

shows_lab :: forall a b. (Showa a, Showa b) => Lab a b -> [Char] -> [Char];
shows_lab (UnLab f) = shows_prec Zero_nat f;
shows_lab (Lab f l) =
  shows_lab f . shows_string ['['] . shows_prec Zero_nat l . shows_string [']'];
shows_lab (Sharp f) = shows_lab f . shows_string ['#'];
shows_lab (FunLab f l) =
  shows_lab f . shows_string ['['] . shows_map shows_lab l . shows_string [']'];

shows_prec_lab ::
  forall a b. (Showa a, Showa b) => Nat -> Lab a b -> [Char] -> [Char];
shows_prec_lab d l = shows_lab l;

shows_list_lab ::
  forall a b. (Showa a, Showa b) => [Lab a b] -> [Char] -> [Char];
shows_list_lab fs = shows_list_aux (shows_prec_lab Zero_nat) fs;

instance (Showa a, Showa b) => Showa (Lab a b) where {
  shows_prec = shows_prec_lab;
  shows_list = shows_list_lab;
};

check_inj ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Lab a b ->
                    Nat ->
                      [((Lab a b, Nat), [Lab a b])] ->
                        Sum ([Char] -> [Char]) ();
check_inj a nn sml =
  let {
    symbols =
      map (\ (aa, b) ->
            let {
              (_, n) = aa;
            } in (\ fs ->
                   map (\ (g, i) ->
                         (g, plus_nat n (times_nat i (minus_nat nn one_nat))))
                     (zip fs (upt Zero_nat (size_list fs))))
              b)
        sml;
    fsymbols = concat symbols;
  } in binda (catcha (check_partition symbols)
               (\ x ->
                 Inl (shows_prec_list Zero_nat
                        ['s', 'y', 'm', 'b', 'o', 'l', ' '] .
                       shows_prec_prod Zero_nat x .
                         shows_prec_list Zero_nat
                           [' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w',
                             'i', 'c', 'e', ')'])))
         (\ _ ->
           binda (check (not (membera fsymbols (a, nn)))
                   (shows_prec_list Zero_nat
                      ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o', 'n',
                        ' ', 's', 'y', 'm', 'b', 'o', 'l'] .
                     shows_prec_lab Zero_nat a .
                       shows_prec_list Zero_nat
                         [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'o',
                           'c', 'c', 'u', 'r', ' ', 'a', 's', ' ', 'n', 'e',
                           'w', ' ', 's', 'y', 'm', 'b', 'o', 'l']))
             (\ _ ->
               binda (check (not (membera (map fst sml) (a, nn)))
                       (shows_prec_list Zero_nat
                          ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o',
                            'n', ' ', 's', 'y', 'm', 'b', 'o', 'l'] .
                         shows_prec_lab Zero_nat a .
                           shows_prec_list Zero_nat
                             [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ',
                               'b', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'i',
                               'e', 'd']))
                 (\ _ ->
                   (if less_eq_nat nn one_nat
                     then catcha
                            (forallm
                              (check_pairwise
                                (\ gn1 gn2 ->
                                  check (not (gn1 == gn2))
                                    (shows_prec_list Zero_nat
                                       ['s', 'y', 'm', 'b', 'o', 'l', ' '] .
                                      shows_prec_prod Zero_nat gn1 .
shows_prec_list Zero_nat
  [' ', 'o', 'c', 'c', 'u', 'r', 's', ' ', 't', 'w', 'i', 'c', 'e'])))
                              symbols)
                            (\ x -> Inl (snd x))
                     else Inr ()))));

manyof :: [Char] -> [Char] -> Sum [Char] ([Char], [Char]);
manyof cs = many (membera cs);

spaces :: [Char] -> Sum [Char] ((), [Char]);
spaces = bind (manyof wspace) (\ _ ts -> Inr ((), ts));

xml2nd_choice ::
  forall a b c.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        [Char] ->
          (Xml -> Sum [Char] b) -> (a -> Maybe b -> c) -> Xml -> Sum [Char] c;
xml2nd_choice tag p1 cn p2 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag && (null atts && (less_eq_nat (nat 2) l && is_none text))
         then binda (p1 (nth cs Zero_nat))
                (\ x1 ->
                  binda (bool cn (nth cs one_nat))
                    (\ b ->
                      (if b then binda (p2 (nth cs (minus_nat l one_nat)))
                                   (\ x2 -> Inr (f x1 (Just x2)))
                        else Inr (f x1 Nothing))))
         else no_transform tag (Xml name atts cs text));

alter1a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> a;
alter1a (Alter1 x) = x;

alter2a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> b;
alter2a (Alter2 x) = x;

alter3a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> c;
alter3a (Alter3 x) = x;

alter4a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> d;
alter4a (Alter4 x) = x;

alter5a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> e;
alter5a (Alter5 x) = x;

alter6a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> f;
alter6a (Alter6 x) = x;

alter7a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> g;
alter7a (Alter7 x) = x;

alter8a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> h;
alter8a (Alter8 x) = x;

alter9a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> i;
alter9a (Alter9 x) = x;

rat_domain :: Xml -> Sum [Char] Rat;
rat_domain =
  xml1element ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's']
    (xml1element ['d', 'e', 'l', 't', 'a'] rat id) id;

basic_domain :: Xml -> Sum [Char] (Nat -> Domain);
basic_domain =
  xml_options
    [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'],
       xml0elements ['n', 'a', 't', 'u', 'r', 'a', 'l', 's'] Naturala),
      (['a', 'r', 'c', 't', 'i', 'c'],
        xml1element ['a', 'r', 'c', 't', 'i', 'c']
          (xml1element ['d', 'o', 'm', 'a', 'i', 'n']
            (xml_options
              [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'],
                 xml0elements ['n', 'a', 't', 'u', 'r', 'a', 'l', 's']
                   (\ _ -> Arctic)),
                (['i', 'n', 't', 'e', 'g', 'e', 'r', 's'],
                  xml0elements ['i', 'n', 't', 'e', 'g', 'e', 'r', 's']
                    (\ _ -> Arctic)),
                (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'],
                  xml_change rat_domain (\ r _ -> Arctic_rat r))])
            id)
          id),
      (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'],
        xml_change rat_domain Rationala)];

interpretation_type :: Xml -> Sum [Char] Domain;
interpretation_type =
  xml1element ['t', 'y', 'p', 'e']
    (xml_options
      [(['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l'],
         xml2elements ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']
           (xml1element ['d', 'o', 'm', 'a', 'i', 'n']
             (xml_options
               [(['n', 'a', 't', 'u', 'r', 'a', 'l', 's'], basic_domain),
                 (['a', 'r', 'c', 't', 'i', 'c'], basic_domain),
                 (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's'], basic_domain),
                 (['m', 'a', 't', 'r', 'i', 'c', 'e', 's'],
                   xml3elements ['m', 'a', 't', 'r', 'i', 'c', 'e', 's']
                     (nata ['d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'])
                     (nata ['s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e',
                             'n', 's', 'i', 'o', 'n'])
                     (xml1element ['d', 'o', 'm', 'a', 'i', 'n'] basic_domain
                       id)
                     (\ di sd domain d ->
                       (case domain d of {
                         Naturala _ -> Int_mat di sd;
                         Arctic -> Arc_mat di;
                         Arctic_rat delta -> Arc_rat_mat delta di;
                         Rationala delta _ -> Rat_mat delta di sd;
                       })))])
             id)
           (nata ['d', 'e', 'g', 'r', 'e', 'e']) (\ typea -> typea)),
        (['m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p', 'r', 'e',
           't', 'a', 't', 'i', 'o', 'n'],
          xml3elements
            ['m', 'a', 't', 'r', 'i', 'x', 'I', 'n', 't', 'e', 'r', 'p', 'r',
              'e', 't', 'a', 't', 'i', 'o', 'n']
            (xml1element ['d', 'o', 'm', 'a', 'i', 'n'] basic_domain id)
            (nata ['d', 'i', 'm', 'e', 'n', 's', 'i', 'o', 'n'])
            (nata ['s', 't', 'r', 'i', 'c', 't', 'D', 'i', 'm', 'e', 'n', 's',
                    'i', 'o', 'n'])
            (\ domain di sd ->
              (case domain Zero_nat of {
                Naturala _ -> Int_mat di sd;
                Arctic -> Arc_mat di;
                Arctic_rat delta -> Arc_rat_mat delta di;
                Rationala delta _ -> Rat_mat delta di sd;
              })))])
    id;

polynomial ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Tpoly Nat a);
polynomial xml2coeff x =
  xml1element ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']
    (xml_options
      [(['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't'],
         xml1element ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't']
           xml2coeff PNum),
        (['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'],
          xml_change (nata ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'])
            (\ n -> PVar (minus_nat n one_nat))),
        (['s', 'u', 'm'],
          xml_many_elements ['s', 'u', 'm'] (polynomial xml2coeff) PSum),
        (['p', 'r', 'o', 'd', 'u', 'c', 't'],
          xml_many_elements ['p', 'r', 'o', 'd', 'u', 'c', 't']
            (polynomial xml2coeff) PMult)])
    id x;

int_coeff :: Xml -> Sum [Char] Integer;
int_coeff = int ['i', 'n', 't', 'e', 'g', 'e', 'r'];

sum_lpoly :: forall a. (a -> a -> a) -> [a] -> [a] -> [a];
sum_lpoly uu [] ys = ys;
sum_lpoly uv (v : va) [] = v : va;
sum_lpoly pl (x : xs) (y : ys) = pl x y : sum_lpoly pl xs ys;

lpoly_of ::
  forall a.
    (Eq a) => a -> a -> (a -> a -> a) ->
                          (a -> a -> a) -> Tpoly Nat a -> Sum [Char] (a, [a]);
lpoly_of uu uv uw ux (PNum i) = Inr (i, []);
lpoly_of ze on pl ti (PVar x) = Inr (ze, replicatea x ze ++ [on]);
lpoly_of ze on pl ti (PSum []) = Inr (ze, []);
lpoly_of ze on pl ti (PSum (p : ps)) =
  binda (lpoly_of ze on pl ti p)
    (\ (cp, ncp) ->
      binda (lpoly_of ze on pl ti (PSum ps))
        (\ (cq, ncq) -> Inr (pl cp cq, sum_lpoly pl ncp ncq)));
lpoly_of ze on pl ti (PMult []) = Inr (on, []);
lpoly_of ze on pl ti (PMult (p : ps)) =
  binda (lpoly_of ze on pl ti p)
    (\ (cp, ncp) ->
      binda (lpoly_of ze on pl ti (PMult ps))
        (\ (cq, ncq) ->
          (if all (\ a -> ze == a) ncp then Inr (ti cp cq, map (ti cp) ncq)
            else (if all (\ a -> ze == a) ncq
                   then Inr (ti cp cq, map (\ x -> ti x cq) ncp)
                   else Inl ['c', 'a', 'n', 'n', 'o', 't', ' ', 't', 'r', 'a',
                              'n', 's', 'f', 'o', 'r', 'm', ' ', 'n', 'o', 'n',
                              '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p', 'o',
                              'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 't',
                              'o', ' ', 'l', 'i', 'n', 'e', 'a', 'r', ' ', 'p',
                              'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l']))));

get_comp_0 ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a;
get_comp_0
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  ze;

max0_int :: Integer -> Integer;
max0_int x = maxa 0 x;

ge_int :: Integer -> Integer -> Bool;
ge_int x y = y <= x;

int_name :: [Char];
int_name =
  ['n', 'a', 't', 'u', 'r', 'a', 'l', 's', ' ', 'w', 'i', 't', 'h', ' ', 'n',
    'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't', 'a', 'n',
    't', 's'];

int_mono :: Integer -> Bool;
int_mono x = 1 <= x;

get_comp_1 ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a;
get_comp_1
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  on;

get_comp_plus ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> a -> a;
get_comp_plus
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  pl;

get_comp_times ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> a -> a;
get_comp_times
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  ti;

fit_length :: forall a. a -> Nat -> [a] -> [a];
fit_length uu Zero_nat uv = [];
fit_length a (Suc v) [] = replicatea (Suc v) a;
fit_length a (Suc n) (b : asa) = b : fit_length a n asa;

class (Semigroup_mult a) => Ab_semigroup_mult a where {
};

class (Ab_semigroup_mult a, Monoid_mult a) => Comm_monoid_mult a where {
};

class (Ab_semigroup_mult a, Semiring a) => Comm_semiring a where {
};

class (Comm_semiring a, Semiring_0 a) => Comm_semiring_0 a where {
};

class (Comm_monoid_mult a, Comm_semiring_0 a, Dvd a,
        Semiring_1 a) => Comm_semiring_1 a where {
};

zero_poly :: forall a b. [([(a, Nat)], b)];
zero_poly = [];

eq_monom :: forall a. (Eq a) => [(a, Nat)] -> [(a, Nat)] -> Bool;
eq_monom [] n = null n;
eq_monom ((x, p) : m) n =
  (case extract (\ yq -> fst yq == x) n of {
    Nothing -> False;
    Just (n1, ((_, q), n2)) -> equal_nat p q && eq_monom m (n1 ++ n2);
  });

poly_add ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => [([(a, Nat)], b)] ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
poly_add [] q = q;
poly_add ((m, c) : p) q =
  (case extract (\ mc -> eq_monom (fst mc) m) q of {
    Nothing -> (m, c) : poly_add p q;
    Just (q1, ((_, d), q2)) ->
      (if plus c d == zero then poly_add p (q1 ++ q2)
        else (m, plus c d) : poly_add p (q1 ++ q2));
  });

one_poly :: forall a b. (Semiring_1 b) => [([(a, Nat)], b)];
one_poly = [([], one)];

monom_mult :: forall a. (Eq a) => [(a, Nat)] -> [(a, Nat)] -> [(a, Nat)];
monom_mult [] n = n;
monom_mult ((x, p) : m) n =
  (case extract (\ yq -> fst yq == x) n of {
    Nothing -> (x, p) : monom_mult m n;
    Just (n1, ((_, q), n2)) -> (x, plus_nat p q) : monom_mult m (n1 ++ n2);
  });

monom_mult_poly ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => ([(a, Nat)], b) ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
monom_mult_poly uu [] = [];
monom_mult_poly (ma, c) ((m, d) : p) =
  (if times c d == zero then monom_mult_poly (ma, c) p
    else (monom_mult ma m, times c d) : monom_mult_poly (ma, c) p);

poly_mult ::
  forall a b.
    (Eq a, Eq b,
      Semiring_0 b) => [([(a, Nat)], b)] ->
                         [([(a, Nat)], b)] -> [([(a, Nat)], b)];
poly_mult [] q = [];
poly_mult (mc : p) q = poly_add (monom_mult_poly mc q) (poly_mult p q);

poly_of ::
  forall a b. (Eq a, Eq b, Comm_semiring_1 b) => Tpoly a b -> [([(a, Nat)], b)];
poly_of (PNum i) = (if i == zero then [] else [([], i)]);
poly_of (PVar x) = [([(x, one_nat)], one)];
poly_of (PSum []) = zero_poly;
poly_of (PSum (p : ps)) = poly_add (poly_of p) (poly_of (PSum ps));
poly_of (PMult []) = one_poly;
poly_of (PMult (p : ps)) = poly_mult (poly_of p) (poly_of (PMult ps));

instance Ab_semigroup_mult Integer where {
};

instance Comm_monoid_mult Integer where {
};

instance Comm_semiring Integer where {
};

instance Comm_semiring_0 Integer where {
};

instance Comm_semiring_1 Integer where {
};

arctic_coeff :: Xml -> Sum [Char] Arctic;
arctic_coeff =
  xml_options
    [(['i', 'n', 't', 'e', 'g', 'e', 'r'],
       xml_change (int ['i', 'n', 't', 'e', 'g', 'e', 'r']) Num_arc),
      (['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'],
        xml0elements
          ['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y']
          MinInfty)];

equal_arctic :: Arctic -> Arctic -> Bool;
equal_arctic (Num_arc int) MinInfty = False;
equal_arctic MinInfty (Num_arc int) = False;
equal_arctic (Num_arc inta) (Num_arc int) = inta == int;
equal_arctic MinInfty MinInfty = True;

instance Eq Arctic where {
  a == b = equal_arctic a b;
};

zero_arctic :: Arctic;
zero_arctic = MinInfty;

one_arctic :: Arctic;
one_arctic = Num_arc 0;

plus_arctic :: Arctic -> Arctic -> Arctic;
plus_arctic MinInfty y = y;
plus_arctic (Num_arc v) MinInfty = Num_arc v;
plus_arctic (Num_arc x) (Num_arc y) = Num_arc (maxa x y);

times_arctic :: Arctic -> Arctic -> Arctic;
times_arctic MinInfty y = MinInfty;
times_arctic (Num_arc v) MinInfty = MinInfty;
times_arctic (Num_arc x) (Num_arc y) = Num_arc (x + y);

ge_arctic :: Arctic -> Arctic -> Bool;
ge_arctic x MinInfty = True;
ge_arctic MinInfty (Num_arc uu) = False;
ge_arctic (Num_arc x) (Num_arc y) = y <= x;

gt_arctic :: Arctic -> Arctic -> Bool;
gt_arctic uu MinInfty = True;
gt_arctic MinInfty (Num_arc uv) = False;
gt_arctic (Num_arc x) (Num_arc y) = y < x;

arctic_name :: [Char];
arctic_name =
  ['a', 'r', 'c', 't', 'i', 'c', ' ', 's', 'e', 'm', 'i', 'r', 'i', 'n', 'g'];

pos_arctic :: Arctic -> Bool;
pos_arctic MinInfty = False;
pos_arctic (Num_arc n) = 0 <= n;

arctic_rat_coeff :: Xml -> Sum [Char] Arctic_rat;
arctic_rat_coeff =
  xml_options
    [(['i', 'n', 't', 'e', 'g', 'e', 'r'], xml_change rat Num_arc_rat),
      (['r', 'a', 't', 'i', 'o', 'n', 'a', 'l'], xml_change rat Num_arc_rat),
      (['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y'],
        xml0elements
          ['m', 'i', 'n', 'u', 's', 'I', 'n', 'f', 'i', 'n', 'i', 't', 'y']
          MinInfty_rat)];

equal_arctic_rat :: Arctic_rat -> Arctic_rat -> Bool;
equal_arctic_rat (Num_arc_rat rat) MinInfty_rat = False;
equal_arctic_rat MinInfty_rat (Num_arc_rat rat) = False;
equal_arctic_rat (Num_arc_rat rata) (Num_arc_rat rat) = equal_rat rata rat;
equal_arctic_rat MinInfty_rat MinInfty_rat = True;

instance Eq Arctic_rat where {
  a == b = equal_arctic_rat a b;
};

zero_arctic_rat :: Arctic_rat;
zero_arctic_rat = MinInfty_rat;

one_arctic_rat :: Arctic_rat;
one_arctic_rat = Num_arc_rat (fract 0 1);

plus_arctic_rat :: Arctic_rat -> Arctic_rat -> Arctic_rat;
plus_arctic_rat MinInfty_rat y = y;
plus_arctic_rat (Num_arc_rat v) MinInfty_rat = Num_arc_rat v;
plus_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = Num_arc_rat (maxa x y);

times_arctic_rat :: Arctic_rat -> Arctic_rat -> Arctic_rat;
times_arctic_rat MinInfty_rat y = MinInfty_rat;
times_arctic_rat (Num_arc_rat v) MinInfty_rat = MinInfty_rat;
times_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = Num_arc_rat (plus_rat x y);

ge_arctic_rat :: Arctic_rat -> Arctic_rat -> Bool;
ge_arctic_rat x MinInfty_rat = True;
ge_arctic_rat MinInfty_rat (Num_arc_rat uu) = False;
ge_arctic_rat (Num_arc_rat x) (Num_arc_rat y) = less_eq_rat y x;

rat_gt :: Rat -> Rat -> Rat -> Bool;
rat_gt delta = (\ x y -> less_eq_rat delta (minus_rat x y));

gt_arctic_rat :: Rat -> Arctic_rat -> Arctic_rat -> Bool;
gt_arctic_rat delta uu MinInfty_rat = True;
gt_arctic_rat delta MinInfty_rat (Num_arc_rat uv) = False;
gt_arctic_rat delta (Num_arc_rat x) (Num_arc_rat y) = rat_gt delta x y;

arctic_rat_name :: [Char];
arctic_rat_name =
  ['a', 'r', 'c', 't', 'i', 'c', ' ', 'r', 'a', 't', 'i', 'o', 'n', 'a', 'l',
    's'];

pos_arctic_rat :: Arctic_rat -> Bool;
pos_arctic_rat MinInfty_rat = False;
pos_arctic_rat (Num_arc_rat n) = less_eq_rat (fract 0 1) n;

vec_coeff :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [a];
vec_coeff xml2coeff =
  xml_many_elements ['v', 'e', 'c', 't', 'o', 'r']
    (xml1element ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't']
      xml2coeff id)
    id;

mat_coeff :: forall a. a -> (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [[a]];
mat_coeff ze xml2coeff =
  xml_options
    [(['m', 'a', 't', 'r', 'i', 'x'],
       xml_many_elements ['m', 'a', 't', 'r', 'i', 'x'] (vec_coeff xml2coeff)
         id),
      (['v', 'e', 'c', 't', 'o', 'r'],
        xml_change (vec_coeff xml2coeff)
          (\ v ->
            v : replicatea (minus_nat (size_list v) one_nat)
                  (replicatea (size_list v) ze)))];

get_comp_def ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a;
get_comp_def
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  def;

instance Non_strict_order Integer where {
  ge = ge_int;
};

instance Ordered_ab_semigroup Integer where {
};

instance Max_ordered_monoid_add Integer where {
  max0 = max0_int;
};

get_comp_gt ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> a -> Bool;
get_comp_gt
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  gt;

get_comp_mono ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      Bool;
get_comp_mono
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  mo;

get_comp_description ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      [Char];
get_comp_description
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  de;

mat_mono :: forall a. Nat -> (a -> Bool) -> [[a]] -> Bool;
mat_mono sd m = mat_monoI m sd;

get_comp_check_mono ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> Bool;
get_comp_check_mono
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  cm;

instance Non_strict_order Arctic where {
  ge = ge_arctic;
};

mat_arc_posI :: forall a. (a -> Bool) -> [[a]] -> Bool;
mat_arc_posI ap m = ap (nth (nth m Zero_nat) Zero_nat);

get_comp_apos ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> Bool;
get_comp_apos
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  ap;

instance Non_strict_order Arctic_rat where {
  ge = ge_arctic_rat;
};

max0_rat :: Rat -> Rat;
max0_rat x = maxa (fract 0 1) x;

ge_rat :: Rat -> Rat -> Bool;
ge_rat x y = less_eq_rat y x;

rat_name :: [Char];
rat_name =
  ['r', 'a', 't', 'i', 'o', 'n', 'a', 'l', 's', ' ', '(', 'w', 'i', 't', 'h',
    ' ', 'n', 'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't',
    'a', 'n', 't', 's', ')'];

rat_mono :: Rat -> Bool;
rat_mono x = less_eq_rat (fract 1 1) x;

instance Ab_semigroup_mult Rat where {
};

instance Comm_monoid_mult Rat where {
};

instance Comm_semiring Rat where {
};

instance Comm_semiring_0 Rat where {
};

instance Comm_semiring_1 Rat where {
};

instance Non_strict_order Rat where {
  ge = ge_rat;
};

instance Ordered_ab_semigroup Rat where {
};

instance Max_ordered_monoid_add Rat where {
  max0 = max0_rat;
};

alter10a :: forall a b c d e f g h i j. Alter a b c d e f g h i j -> j;
alter10a (Alter10 x) = x;

interpretation ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Redpair a);
interpretation xml2name =
  xml1many_elements_gen
    ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n']
    interpretation_type
    (\ a ->
      (case a of {
        Naturala deg ->
          (if less_eq_nat deg one_nat
            then xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     binda (polynomial int_coeff x)
                       (lpoly_of
                         (get_comp_0
                           ((\ _ -> True),
                             (0, (1, (1, ((\ aa b -> aa + b),
   ((\ aa b -> aa * b),
     (max0_int,
       (ge_int,
         ((\ xa y -> y < xa),
           (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                         (get_comp_1
                           ((\ _ -> True),
                             (0, (1, (1, ((\ aa b -> aa + b),
   ((\ aa b -> aa * b),
     (max0_int,
       (ge_int,
         ((\ xa y -> y < xa),
           (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                         (get_comp_plus
                           ((\ _ -> True),
                             (0, (1, (1, ((\ aa b -> aa + b),
   ((\ aa b -> aa * b),
     (max0_int,
       (ge_int,
         ((\ xa y -> y < xa),
           (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                         (get_comp_times
                           ((\ _ -> True),
                             (0, (1, (1, ((\ aa b -> aa + b),
   ((\ aa b -> aa * b),
     (max0_int,
       (ge_int,
         ((\ xa y -> y < xa),
           (True, (int_name, ((\ _ -> True), int_mono)))))))))))))))
                   (\ f aa poly ->
                     Alter1
                       ((f, aa),
                         (fst poly,
                           fit_length
                             (get_comp_0
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ ab b -> ab + b),
     ((\ ab b -> ab * b),
       (max0_int,
         (ge_int,
           ((\ x y -> y < x),
             (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                             aa (snd poly))))
            else xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     binda (polynomial int_coeff x) (\ p -> Inr (poly_of p)))
                   (\ f aa poly -> Alter9 ((f, aa), poly)));
        Arctic ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial arctic_coeff x)
                (lpoly_of
                  (get_comp_0
                    ((\ _ -> True),
                      (zero_arctic,
                        (one_arctic,
                          (one_arctic,
                            (plus_arctic,
                              (times_arctic,
                                (id, (ge_arctic,
                                       (gt_arctic,
 (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                  (get_comp_1
                    ((\ _ -> True),
                      (zero_arctic,
                        (one_arctic,
                          (one_arctic,
                            (plus_arctic,
                              (times_arctic,
                                (id, (ge_arctic,
                                       (gt_arctic,
 (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                  (get_comp_plus
                    ((\ _ -> True),
                      (zero_arctic,
                        (one_arctic,
                          (one_arctic,
                            (plus_arctic,
                              (times_arctic,
                                (id, (ge_arctic,
                                       (gt_arctic,
 (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                  (get_comp_times
                    ((\ _ -> True),
                      (zero_arctic,
                        (one_arctic,
                          (one_arctic,
                            (plus_arctic,
                              (times_arctic,
                                (id, (ge_arctic,
                                       (gt_arctic,
 (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))))
            (\ f aa poly ->
              Alter3
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        ((\ _ -> True),
                          (zero_arctic,
                            (one_arctic,
                              (one_arctic,
                                (plus_arctic,
                                  (times_arctic,
                                    (id, (ge_arctic,
   (gt_arctic, (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                      aa (snd poly))));
        Arctic_rat d ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial arctic_rat_coeff x)
                (lpoly_of
                  (get_comp_0
                    ((\ _ -> True),
                      (zero_arctic_rat,
                        (one_arctic_rat,
                          (one_arctic_rat,
                            (plus_arctic_rat,
                              (times_arctic_rat,
                                (id, (ge_arctic_rat,
                                       (gt_arctic_rat d,
 (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                  (get_comp_1
                    ((\ _ -> True),
                      (zero_arctic_rat,
                        (one_arctic_rat,
                          (one_arctic_rat,
                            (plus_arctic_rat,
                              (times_arctic_rat,
                                (id, (ge_arctic_rat,
                                       (gt_arctic_rat d,
 (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                  (get_comp_plus
                    ((\ _ -> True),
                      (zero_arctic_rat,
                        (one_arctic_rat,
                          (one_arctic_rat,
                            (plus_arctic_rat,
                              (times_arctic_rat,
                                (id, (ge_arctic_rat,
                                       (gt_arctic_rat d,
 (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                  (get_comp_times
                    ((\ _ -> True),
                      (zero_arctic_rat,
                        (one_arctic_rat,
                          (one_arctic_rat,
                            (plus_arctic_rat,
                              (times_arctic_rat,
                                (id, (ge_arctic_rat,
                                       (gt_arctic_rat d,
 (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))))
            (\ f aa poly ->
              Alter7
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        ((\ _ -> True),
                          (zero_arctic_rat,
                            (one_arctic_rat,
                              (one_arctic_rat,
                                (plus_arctic_rat,
                                  (times_arctic_rat,
                                    (id, (ge_arctic_rat,
   (gt_arctic_rat d,
     (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                      aa (snd poly))));
        Int_mat n sd ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial (mat_coeff 0 int_coeff) x)
                (lpoly_of
                  (get_comp_0
                    (mat n n,
                      (mat0I 0 n n,
                        (mat1I 0 1 n,
                          (mat1I 0
                             (get_comp_def
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ aa b -> aa + b),
     ((\ aa b -> aa * b),
       (max0_int,
         (ge_int,
           ((\ xa y -> y < xa),
             (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                             n,
                            (mat_plusI (\ aa b -> aa + b),
                              (mat_multI 0 (\ aa b -> aa + b) (\ aa b -> aa * b)
                                 n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (0, (1, (1, ((\ aa b -> aa + b),
                       ((\ aa b -> aa * b),
                         (max0_int,
                           (ge_int,
                             ((\ xa y -> y < xa),
                               (True,
                                 (int_name,
                                   ((\ _ -> True), int_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (0, (1, (1, ((\ aa b -> aa + b),
                        ((\ aa b -> aa * b),
                          (max0_int,
                            (ge_int,
                              ((\ xa y -> y < xa),
                                (True,
                                  (int_name,
                                    ((\ _ -> True),
                                      int_mono))))))))))))))))))))))))))
                  (get_comp_1
                    (mat n n,
                      (mat0I 0 n n,
                        (mat1I 0 1 n,
                          (mat1I 0
                             (get_comp_def
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ aa b -> aa + b),
     ((\ aa b -> aa * b),
       (max0_int,
         (ge_int,
           ((\ xa y -> y < xa),
             (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                             n,
                            (mat_plusI (\ aa b -> aa + b),
                              (mat_multI 0 (\ aa b -> aa + b) (\ aa b -> aa * b)
                                 n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (0, (1, (1, ((\ aa b -> aa + b),
                       ((\ aa b -> aa * b),
                         (max0_int,
                           (ge_int,
                             ((\ xa y -> y < xa),
                               (True,
                                 (int_name,
                                   ((\ _ -> True), int_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (0, (1, (1, ((\ aa b -> aa + b),
                        ((\ aa b -> aa * b),
                          (max0_int,
                            (ge_int,
                              ((\ xa y -> y < xa),
                                (True,
                                  (int_name,
                                    ((\ _ -> True),
                                      int_mono))))))))))))))))))))))))))
                  (get_comp_plus
                    (mat n n,
                      (mat0I 0 n n,
                        (mat1I 0 1 n,
                          (mat1I 0
                             (get_comp_def
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ aa b -> aa + b),
     ((\ aa b -> aa * b),
       (max0_int,
         (ge_int,
           ((\ xa y -> y < xa),
             (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                             n,
                            (mat_plusI (\ aa b -> aa + b),
                              (mat_multI 0 (\ aa b -> aa + b) (\ aa b -> aa * b)
                                 n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (0, (1, (1, ((\ aa b -> aa + b),
                       ((\ aa b -> aa * b),
                         (max0_int,
                           (ge_int,
                             ((\ xa y -> y < xa),
                               (True,
                                 (int_name,
                                   ((\ _ -> True), int_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (0, (1, (1, ((\ aa b -> aa + b),
                        ((\ aa b -> aa * b),
                          (max0_int,
                            (ge_int,
                              ((\ xa y -> y < xa),
                                (True,
                                  (int_name,
                                    ((\ _ -> True),
                                      int_mono))))))))))))))))))))))))))
                  (get_comp_times
                    (mat n n,
                      (mat0I 0 n n,
                        (mat1I 0 1 n,
                          (mat1I 0
                             (get_comp_def
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ aa b -> aa + b),
     ((\ aa b -> aa * b),
       (max0_int,
         (ge_int,
           ((\ xa y -> y < xa),
             (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                             n,
                            (mat_plusI (\ aa b -> aa + b),
                              (mat_multI 0 (\ aa b -> aa + b) (\ aa b -> aa * b)
                                 n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (0, (1, (1, ((\ aa b -> aa + b),
                 ((\ aa b -> aa * b),
                   (max0_int,
                     (ge_int,
                       ((\ xa y -> y < xa),
                         (True, (int_name, ((\ _ -> True), int_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (0, (1, (1, ((\ aa b -> aa + b),
                       ((\ aa b -> aa * b),
                         (max0_int,
                           (ge_int,
                             ((\ xa y -> y < xa),
                               (True,
                                 (int_name,
                                   ((\ _ -> True), int_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (0, (1, (1, ((\ aa b -> aa + b),
                        ((\ aa b -> aa * b),
                          (max0_int,
                            (ge_int,
                              ((\ xa y -> y < xa),
                                (True,
                                  (int_name,
                                    ((\ _ -> True),
                                      int_mono))))))))))))))))))))))))))))
            (\ f aa poly ->
              Alter4
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        (mat n n,
                          (mat0I 0 n n,
                            (mat1I 0 1 n,
                              (mat1I 0
                                 (get_comp_def
                                   ((\ _ -> True),
                                     (0, (1,
   (1, ((\ ab b -> ab + b),
         ((\ ab b -> ab * b),
           (max0_int,
             (ge_int,
               ((\ x y -> y < x),
                 (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                                 n,
                                (mat_plusI (\ ab b -> ab + b),
                                  (mat_multI 0 (\ ab b -> ab + b)
                                     (\ ab b -> ab * b) n,
                                    (mat_max0,
                                      (mat_ge,
(mat_gtI
   (get_comp_gt
     ((\ _ -> True),
       (0, (1, (1, ((\ ab b -> ab + b),
                     ((\ ab b -> ab * b),
                       (max0_int,
                         (ge_int,
                           ((\ x y -> y < x),
                             (True,
                               (int_name, ((\ _ -> True), int_mono)))))))))))))
   sd,
  (get_comp_mono
     ((\ _ -> True),
       (0, (1, (1, ((\ ab b -> ab + b),
                     ((\ ab b -> ab * b),
                       (max0_int,
                         (ge_int,
                           ((\ x y -> y < x),
                             (True,
                               (int_name, ((\ _ -> True), int_mono)))))))))))),
    (concat
       [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
         [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h',
           ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's',
           'i', 'o', 'n', ' '],
         shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
         get_comp_description
           ((\ _ -> True),
             (0, (1, (1, ((\ ab b -> ab + b),
                           ((\ ab b -> ab * b),
                             (max0_int,
                               (ge_int,
                                 ((\ x y -> y < x),
                                   (True,
                                     (int_name,
                                       ((\ _ -> True), int_mono))))))))))))],
      ((\ _ -> True),
        mat_mono sd
          (get_comp_check_mono
            ((\ _ -> True),
              (0, (1, (1, ((\ ab b -> ab + b),
                            ((\ ab b -> ab * b),
                              (max0_int,
                                (ge_int,
                                  ((\ x y -> y < x),
                                    (True,
                                      (int_name,
((\ _ -> True), int_mono))))))))))))))))))))))))))
                      aa (snd poly))));
        Arc_mat n ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial (mat_coeff MinInfty arctic_coeff) x)
                (lpoly_of
                  (get_comp_0
                    (mat n n,
                      (mat0I zero_arctic n n,
                        (mat1I zero_arctic one_arctic n,
                          (mat1I zero_arctic
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic,
                                   (one_arctic,
                                     (one_arctic,
                                       (plus_arctic,
 (times_arctic,
   (id, (ge_arctic,
          (gt_arctic,
            (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic,
                              (mat_multI zero_arctic plus_arctic times_arctic n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic,
                         (False,
                           (arctic_name,
                             (pos_arctic, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_1
                    (mat n n,
                      (mat0I zero_arctic n n,
                        (mat1I zero_arctic one_arctic n,
                          (mat1I zero_arctic
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic,
                                   (one_arctic,
                                     (one_arctic,
                                       (plus_arctic,
 (times_arctic,
   (id, (ge_arctic,
          (gt_arctic,
            (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic,
                              (mat_multI zero_arctic plus_arctic times_arctic n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic,
                         (False,
                           (arctic_name,
                             (pos_arctic, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_plus
                    (mat n n,
                      (mat0I zero_arctic n n,
                        (mat1I zero_arctic one_arctic n,
                          (mat1I zero_arctic
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic,
                                   (one_arctic,
                                     (one_arctic,
                                       (plus_arctic,
 (times_arctic,
   (id, (ge_arctic,
          (gt_arctic,
            (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic,
                              (mat_multI zero_arctic plus_arctic times_arctic n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic,
                         (False,
                           (arctic_name,
                             (pos_arctic, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_times
                    (mat n n,
                      (mat0I zero_arctic n n,
                        (mat1I zero_arctic one_arctic n,
                          (mat1I zero_arctic
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic,
                                   (one_arctic,
                                     (one_arctic,
                                       (plus_arctic,
 (times_arctic,
   (id, (ge_arctic,
          (gt_arctic,
            (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic,
                              (mat_multI zero_arctic plus_arctic times_arctic n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic,
                         (False,
                           (arctic_name,
                             (pos_arctic, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic,
              (one_arctic,
                (one_arctic,
                  (plus_arctic,
                    (times_arctic,
                      (id, (ge_arctic,
                             (gt_arctic,
                               (False,
                                 (arctic_name,
                                   (pos_arctic, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))))
            (\ f aa poly ->
              Alter6
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        (mat n n,
                          (mat0I zero_arctic n n,
                            (mat1I zero_arctic one_arctic n,
                              (mat1I zero_arctic
                                 (get_comp_def
                                   ((\ _ -> True),
                                     (zero_arctic,
                                       (one_arctic,
 (one_arctic,
   (plus_arctic,
     (times_arctic,
       (id, (ge_arctic,
              (gt_arctic,
                (False, (arctic_name, (pos_arctic, (\ _ -> False))))))))))))))
                                 n,
                                (mat_plusI plus_arctic,
                                  (mat_multI zero_arctic plus_arctic
                                     times_arctic n,
                                    (id, (mat_ge,
   (mat_comp_all
      (get_comp_gt
        ((\ _ -> True),
          (zero_arctic,
            (one_arctic,
              (one_arctic,
                (plus_arctic,
                  (times_arctic,
                    (id, (ge_arctic,
                           (gt_arctic,
                             (False,
                               (arctic_name,
                                 (pos_arctic, (\ _ -> False)))))))))))))),
     (False,
       (concat
          [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
            [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e',
              'r', ' '],
            get_comp_description
              ((\ _ -> True),
                (zero_arctic,
                  (one_arctic,
                    (one_arctic,
                      (plus_arctic,
                        (times_arctic,
                          (id, (ge_arctic,
                                 (gt_arctic,
                                   (False,
                                     (arctic_name,
                                       (pos_arctic, (\ _ -> False)))))))))))))],
         (mat_arc_posI
            (get_comp_apos
              ((\ _ -> True),
                (zero_arctic,
                  (one_arctic,
                    (one_arctic,
                      (plus_arctic,
                        (times_arctic,
                          (id, (ge_arctic,
                                 (gt_arctic,
                                   (False,
                                     (arctic_name,
                                       (pos_arctic, (\ _ -> False)))))))))))))),
           (\ _ -> False))))))))))))))
                      aa (snd poly))));
        Arc_rat_mat d n ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial (mat_coeff MinInfty_rat arctic_rat_coeff) x)
                (lpoly_of
                  (get_comp_0
                    (mat n n,
                      (mat0I zero_arctic_rat n n,
                        (mat1I zero_arctic_rat one_arctic_rat n,
                          (mat1I zero_arctic_rat
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic_rat,
                                   (one_arctic_rat,
                                     (one_arctic_rat,
                                       (plus_arctic_rat,
 (times_arctic_rat,
   (id, (ge_arctic_rat,
          (gt_arctic_rat d,
            (False,
              (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic_rat,
                              (mat_multI zero_arctic_rat plus_arctic_rat
                                 times_arctic_rat n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat d,
                         (False,
                           (arctic_rat_name,
                             (pos_arctic_rat, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_1
                    (mat n n,
                      (mat0I zero_arctic_rat n n,
                        (mat1I zero_arctic_rat one_arctic_rat n,
                          (mat1I zero_arctic_rat
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic_rat,
                                   (one_arctic_rat,
                                     (one_arctic_rat,
                                       (plus_arctic_rat,
 (times_arctic_rat,
   (id, (ge_arctic_rat,
          (gt_arctic_rat d,
            (False,
              (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic_rat,
                              (mat_multI zero_arctic_rat plus_arctic_rat
                                 times_arctic_rat n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat d,
                         (False,
                           (arctic_rat_name,
                             (pos_arctic_rat, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_plus
                    (mat n n,
                      (mat0I zero_arctic_rat n n,
                        (mat1I zero_arctic_rat one_arctic_rat n,
                          (mat1I zero_arctic_rat
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic_rat,
                                   (one_arctic_rat,
                                     (one_arctic_rat,
                                       (plus_arctic_rat,
 (times_arctic_rat,
   (id, (ge_arctic_rat,
          (gt_arctic_rat d,
            (False,
              (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic_rat,
                              (mat_multI zero_arctic_rat plus_arctic_rat
                                 times_arctic_rat n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat d,
                         (False,
                           (arctic_rat_name,
                             (pos_arctic_rat, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))
                  (get_comp_times
                    (mat n n,
                      (mat0I zero_arctic_rat n n,
                        (mat1I zero_arctic_rat one_arctic_rat n,
                          (mat1I zero_arctic_rat
                             (get_comp_def
                               ((\ _ -> True),
                                 (zero_arctic_rat,
                                   (one_arctic_rat,
                                     (one_arctic_rat,
                                       (plus_arctic_rat,
 (times_arctic_rat,
   (id, (ge_arctic_rat,
          (gt_arctic_rat d,
            (False,
              (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                             n,
                            (mat_plusI plus_arctic_rat,
                              (mat_multI zero_arctic_rat plus_arctic_rat
                                 times_arctic_rat n,
                                (id, (mat_ge,
                                       (mat_comp_all
  (get_comp_gt
    ((\ _ -> True),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat d,
                         (False,
                           (arctic_rat_name,
                             (pos_arctic_rat, (\ _ -> False)))))))))))))),
 (False,
   (concat
      [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
        [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e', 'r',
          ' '],
        get_comp_description
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))],
     (mat_arc_posI
        (get_comp_apos
          ((\ _ -> True),
            (zero_arctic_rat,
              (one_arctic_rat,
                (one_arctic_rat,
                  (plus_arctic_rat,
                    (times_arctic_rat,
                      (id, (ge_arctic_rat,
                             (gt_arctic_rat d,
                               (False,
                                 (arctic_rat_name,
                                   (pos_arctic_rat, (\ _ -> False)))))))))))))),
       (\ _ -> False))))))))))))))))
            (\ f aa poly ->
              Alter8
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        (mat n n,
                          (mat0I zero_arctic_rat n n,
                            (mat1I zero_arctic_rat one_arctic_rat n,
                              (mat1I zero_arctic_rat
                                 (get_comp_def
                                   ((\ _ -> True),
                                     (zero_arctic_rat,
                                       (one_arctic_rat,
 (one_arctic_rat,
   (plus_arctic_rat,
     (times_arctic_rat,
       (id, (ge_arctic_rat,
              (gt_arctic_rat d,
                (False,
                  (arctic_rat_name, (pos_arctic_rat, (\ _ -> False))))))))))))))
                                 n,
                                (mat_plusI plus_arctic_rat,
                                  (mat_multI zero_arctic_rat plus_arctic_rat
                                     times_arctic_rat n,
                                    (id, (mat_ge,
   (mat_comp_all
      (get_comp_gt
        ((\ _ -> True),
          (zero_arctic_rat,
            (one_arctic_rat,
              (one_arctic_rat,
                (plus_arctic_rat,
                  (times_arctic_rat,
                    (id, (ge_arctic_rat,
                           (gt_arctic_rat d,
                             (False,
                               (arctic_rat_name,
                                 (pos_arctic_rat, (\ _ -> False)))))))))))))),
     (False,
       (concat
          [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
            [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'o', 'v', 'e',
              'r', ' '],
            get_comp_description
              ((\ _ -> True),
                (zero_arctic_rat,
                  (one_arctic_rat,
                    (one_arctic_rat,
                      (plus_arctic_rat,
                        (times_arctic_rat,
                          (id, (ge_arctic_rat,
                                 (gt_arctic_rat d,
                                   (False,
                                     (arctic_rat_name,
                                       (pos_arctic_rat,
 (\ _ -> False)))))))))))))],
         (mat_arc_posI
            (get_comp_apos
              ((\ _ -> True),
                (zero_arctic_rat,
                  (one_arctic_rat,
                    (one_arctic_rat,
                      (plus_arctic_rat,
                        (times_arctic_rat,
                          (id, (ge_arctic_rat,
                                 (gt_arctic_rat d,
                                   (False,
                                     (arctic_rat_name,
                                       (pos_arctic_rat,
 (\ _ -> False)))))))))))))),
           (\ _ -> False))))))))))))))
                      aa (snd poly))));
        Rationala d deg ->
          (if less_eq_nat deg one_nat
            then xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x ->
                     binda (polynomial rat x)
                       (lpoly_of
                         (get_comp_0
                           ((\ _ -> True),
                             (fract 0 1,
                               (fract 1 1,
                                 (d, (plus_rat,
                                       (times_rat,
 (max0_rat,
   (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                         (get_comp_1
                           ((\ _ -> True),
                             (fract 0 1,
                               (fract 1 1,
                                 (d, (plus_rat,
                                       (times_rat,
 (max0_rat,
   (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                         (get_comp_plus
                           ((\ _ -> True),
                             (fract 0 1,
                               (fract 1 1,
                                 (d, (plus_rat,
                                       (times_rat,
 (max0_rat,
   (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                         (get_comp_times
                           ((\ _ -> True),
                             (fract 0 1,
                               (fract 1 1,
                                 (d, (plus_rat,
                                       (times_rat,
 (max0_rat,
   (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))))
                   (\ f aa poly ->
                     Alter2
                       ((f, aa),
                         (fst poly,
                           fit_length
                             (get_comp_0
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                             aa (snd poly))))
            else xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't']
                   xml2name (nata ['a', 'r', 'i', 't', 'y'])
                   (\ x -> binda (polynomial rat x) (\ p -> Inr (poly_of p)))
                   (\ f aa poly -> Alter10 ((f, aa), poly)));
        Rat_mat d n sd ->
          xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
            (nata ['a', 'r', 'i', 't', 'y'])
            (\ x ->
              binda (polynomial (mat_coeff (fract 0 1) rat) x)
                (lpoly_of
                  (get_comp_0
                    (mat n n,
                      (mat0I (fract 0 1) n n,
                        (mat1I (fract 0 1) (fract 1 1) n,
                          (mat1I (fract 0 1)
                             (get_comp_def
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                             n,
                            (mat_plusI plus_rat,
                              (mat_multI (fract 0 1) plus_rat times_rat n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (fract 0 1,
           (fract 1 1,
             (d, (plus_rat,
                   (times_rat,
                     (max0_rat,
                       (ge_rat,
                         (rat_gt d,
                           (True,
                             (rat_name, ((\ _ -> True), rat_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (fract 0 1,
            (fract 1 1,
              (d, (plus_rat,
                    (times_rat,
                      (max0_rat,
                        (ge_rat,
                          (rat_gt d,
                            (True,
                              (rat_name,
                                ((\ _ -> True),
                                  rat_mono))))))))))))))))))))))))))
                  (get_comp_1
                    (mat n n,
                      (mat0I (fract 0 1) n n,
                        (mat1I (fract 0 1) (fract 1 1) n,
                          (mat1I (fract 0 1)
                             (get_comp_def
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                             n,
                            (mat_plusI plus_rat,
                              (mat_multI (fract 0 1) plus_rat times_rat n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (fract 0 1,
           (fract 1 1,
             (d, (plus_rat,
                   (times_rat,
                     (max0_rat,
                       (ge_rat,
                         (rat_gt d,
                           (True,
                             (rat_name, ((\ _ -> True), rat_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (fract 0 1,
            (fract 1 1,
              (d, (plus_rat,
                    (times_rat,
                      (max0_rat,
                        (ge_rat,
                          (rat_gt d,
                            (True,
                              (rat_name,
                                ((\ _ -> True),
                                  rat_mono))))))))))))))))))))))))))
                  (get_comp_plus
                    (mat n n,
                      (mat0I (fract 0 1) n n,
                        (mat1I (fract 0 1) (fract 1 1) n,
                          (mat1I (fract 0 1)
                             (get_comp_def
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                             n,
                            (mat_plusI plus_rat,
                              (mat_multI (fract 0 1) plus_rat times_rat n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (fract 0 1,
           (fract 1 1,
             (d, (plus_rat,
                   (times_rat,
                     (max0_rat,
                       (ge_rat,
                         (rat_gt d,
                           (True,
                             (rat_name, ((\ _ -> True), rat_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (fract 0 1,
            (fract 1 1,
              (d, (plus_rat,
                    (times_rat,
                      (max0_rat,
                        (ge_rat,
                          (rat_gt d,
                            (True,
                              (rat_name,
                                ((\ _ -> True),
                                  rat_mono))))))))))))))))))))))))))
                  (get_comp_times
                    (mat n n,
                      (mat0I (fract 0 1) n n,
                        (mat1I (fract 0 1) (fract 1 1) n,
                          (mat1I (fract 0 1)
                             (get_comp_def
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                             n,
                            (mat_plusI plus_rat,
                              (mat_multI (fract 0 1) plus_rat times_rat n,
                                (mat_max0,
                                  (mat_ge,
                                    (mat_gtI
                                       (get_comp_gt
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                                       sd,
                                      (get_comp_mono
 ((\ _ -> True),
   (fract 0 1,
     (fract 1 1,
       (d, (plus_rat,
             (times_rat,
               (max0_rat,
                 (ge_rat,
                   (rat_gt d,
                     (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
(concat
   [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
     [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h', ' ',
       's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i',
       'o', 'n', ' '],
     shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
     get_comp_description
       ((\ _ -> True),
         (fract 0 1,
           (fract 1 1,
             (d, (plus_rat,
                   (times_rat,
                     (max0_rat,
                       (ge_rat,
                         (rat_gt d,
                           (True,
                             (rat_name, ((\ _ -> True), rat_mono))))))))))))],
  ((\ _ -> True),
    mat_mono sd
      (get_comp_check_mono
        ((\ _ -> True),
          (fract 0 1,
            (fract 1 1,
              (d, (plus_rat,
                    (times_rat,
                      (max0_rat,
                        (ge_rat,
                          (rat_gt d,
                            (True,
                              (rat_name,
                                ((\ _ -> True),
                                  rat_mono))))))))))))))))))))))))))))
            (\ f aa poly ->
              Alter5
                ((f, aa),
                  (fst poly,
                    fit_length
                      (get_comp_0
                        (mat n n,
                          (mat0I (fract 0 1) n n,
                            (mat1I (fract 0 1) (fract 1 1) n,
                              (mat1I (fract 0 1)
                                 (get_comp_def
                                   ((\ _ -> True),
                                     (fract 0 1,
                                       (fract 1 1,
 (d, (plus_rat,
       (times_rat,
         (max0_rat,
           (ge_rat,
             (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                                 n,
                                (mat_plusI plus_rat,
                                  (mat_multI (fract 0 1) plus_rat times_rat n,
                                    (mat_max0,
                                      (mat_ge,
(mat_gtI
   (get_comp_gt
     ((\ _ -> True),
       (fract 0 1,
         (fract 1 1,
           (d, (plus_rat,
                 (times_rat,
                   (max0_rat,
                     (ge_rat,
                       (rat_gt d,
                         (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
   sd,
  (get_comp_mono
     ((\ _ -> True),
       (fract 0 1,
         (fract 1 1,
           (d, (plus_rat,
                 (times_rat,
                   (max0_rat,
                     (ge_rat,
                       (rat_gt d,
                         (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
    (concat
       [shows_prec_nat Zero_nat n [], ['x'], shows_prec_nat Zero_nat n [],
         [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ', 'w', 'i', 't', 'h',
           ' ', 's', 't', 'r', 'i', 'c', 't', ' ', 'd', 'i', 'm', 'e', 'n', 's',
           'i', 'o', 'n', ' '],
         shows_prec_nat Zero_nat sd [], [' ', 'o', 'v', 'e', 'r', ' '],
         get_comp_description
           ((\ _ -> True),
             (fract 0 1,
               (fract 1 1,
                 (d, (plus_rat,
                       (times_rat,
                         (max0_rat,
                           (ge_rat,
                             (rat_gt d,
                               (True,
                                 (rat_name,
                                   ((\ _ -> True), rat_mono))))))))))))],
      ((\ _ -> True),
        mat_mono sd
          (get_comp_check_mono
            ((\ _ -> True),
              (fract 0 1,
                (fract 1 1,
                  (d, (plus_rat,
                        (times_rat,
                          (max0_rat,
                            (ge_rat,
                              (rat_gt d,
                                (True,
                                  (rat_name,
                                    ((\ _ -> True),
                                      rat_mono))))))))))))))))))))))))))
                      aa (snd poly))));
      }))
    (\ typea pia ->
      (case typea of {
        Naturala deg ->
          (if less_eq_nat deg one_nat then Int_carrier (map alter1a pia)
            else Int_nl_carrier (map alter9a pia));
        Arctic -> Arctic_carrier (map alter3a pia);
        Arctic_rat d -> Arctic_rat_carrier d (map alter7a pia);
        Int_mat n sd -> Int_mat_carrier n sd (map alter4a pia);
        Arc_mat n -> Arc_mat_carrier n (map alter6a pia);
        Arc_rat_mat d n -> Arc_rat_mat_carrier d n (map alter8a pia);
        Rationala d deg ->
          (if less_eq_nat deg one_nat then Rat_carrier d (map alter2a pia)
            else Rat_nl_carrier d (map alter10a pia));
        Rat_mat d n sd -> Rat_mat_carrier d n sd (map alter5a pia);
      }));

xml1to2elements ::
  forall a b c.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) -> (a -> Maybe b -> c) -> Xml -> Sum [Char] c;
xml1to2elements tag p1 p2 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             (null atts &&
               (less_eq_nat one_nat l &&
                 (less_eq_nat l (nat 2) && is_none text)))
         then binda (p1 (nth cs Zero_nat))
                (\ x1 ->
                  (if equal_nat l (nat 2)
                    then binda (p2 (nth cs one_nat))
                           (\ x2 -> Inr (f x1 (Just x2)))
                    else Inr (f x1 Nothing)))
         else no_transform tag (Xml name atts cs text));

precedence ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [((a, Nat), Nat)];
precedence xml2name =
  xml_many_elements
    ['s', 't', 'a', 't', 'u', 's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c',
      'e']
    (xml4elements
      ['s', 't', 'a', 't', 'u', 's', 'P', 'r', 'e', 'c', 'e', 'd', 'e', 'n',
        'c', 'e', 'E', 'n', 't', 'r', 'y']
      xml2name (nata ['a', 'r', 'i', 't', 'y'])
      (nata ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e'])
      (xml_options
        [(['l', 'e', 'x'], xml0elements ['l', 'e', 'x'] ()),
          (['m', 'u', 'l'],
            (\ _ ->
              Inl ['m', 'u', 'l', 't', 'i', 's', 'e', 't', '-', 'c', 'o', 'm',
                    'p', 'a', 'r', 'i', 's', 'o', 'n', 's', ' ', 'n', 'o', 't',
                    ' ', 's', 'u', 'p', 'p', 'o', 'r', 't', 'e', 'd']))])
      (\ f a p _ -> ((f, a), p)))
    id;

path_order :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Redpair a);
path_order xml2name =
  xml1to2elements ['p', 'a', 't', 'h', 'O', 'r', 'd', 'e', 'r']
    (precedence xml2name) (afs xml2name)
    (\ pr a -> (case a of {
                 Nothing -> Lpo pr [];
                 Just aa -> Lpo pr aa;
               }));

redpair :: forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Redpair a);
redpair xml2name =
  xml1element ['r', 'e', 'd', 'P', 'a', 'i', 'r']
    (xml_options
      [(['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o', 'n'],
         interpretation xml2name),
        (['p', 'a', 't', 'h', 'O', 'r', 'd', 'e', 'r'], path_order xml2name)])
    id;

height :: forall a. (a, Nat) -> Nat;
height (f, h) = h;

rm_lookup :: forall a b. (Linorder a) => a -> Rbt a b -> Maybe b;
rm_lookup k m = lookupa m k;

rs_memb :: forall a. (Linorder a) => a -> Rbt a () -> Bool;
rs_memb = s_memb rm_lookup;

shows_lpo_repr :: forall a. (Showa a) => [((a, Nat), Nat)] -> [Char] -> [Char];
shows_lpo_repr prs =
  shows_string
    ['L', 'P', 'O', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f', 'o',
      'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'p', 'r', 'e', 'c', 'e', 'd', 'e',
      'n', 'c', 'e'] .
    shows_nl .
      shows_map
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ pr ->
                 shows_string
                   ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '('] .
                   shows_prec Zero_nat f .
                     shows_string ['['] .
                       shows_prec_nat Zero_nat n .
                         shows_string [']', ')', ' ', '=', ' '] .
                           shows_prec_nat Zero_nat pr . shows_nl)
            b)
        prs .
        shows_nl .
          shows_string
            ['p', 'r', 'e', 'c', 'e', 'd', 'e', 'n', 'c', 'e', '(', '_', ')',
              ' ', '=', ' ', '0'] .
            shows_nl;

s_isEmpty :: forall a. a -> a;
s_isEmpty isEmpty = isEmpty;

wf_terms_list :: forall a b. (Eq a) => [(a, Nat)] -> Term a b -> Bool;
wf_terms_list uu (Var uv) = True;
wf_terms_list fa (Fun f ts) =
  membera fa (f, size_list ts) && all (wf_terms_list fa) ts;

full_af :: forall a. a -> Nat -> Nat -> Bool;
full_af f n i = True;

rp_desc ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      [Char] -> [Char];
rp_desc (v, (s, (n, (a, (m, d))))) = d;

rp_mono ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      Sum ([Char] -> [Char]) ();
rp_mono (v, (s, (n, (a, (m, d))))) = m;

defs_list :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
defs_list [] = [];
defs_list ((Fun f ss, r) : ts) =
  let {
    fs = defs_list ts;
    n = size_list ss;
  } in (if membera fs (f, n) then fs else (f, n) : fs);
defs_list ((Var x, r) : ts) = defs_list ts;

shows_pos :: Pos -> [Char] -> [Char];
shows_pos (PCons i Empty) = shows_prec_nat Zero_nat i;
shows_pos (PCons i (PCons v va)) =
  shows_prec_nat Zero_nat i . shows_pos (PCons v va);
shows_pos Empty = shows_string ['e'];

apply_args :: forall a b. a -> Term a b -> [Term a b] -> Term a b;
apply_args a t [] = t;
apply_args a t (s : ss) = apply_args a (Fun a [t, s]) ss;

get_symbol :: forall a. (a -> Nat -> [a]) -> a -> Nat -> Nat -> a;
get_symbol sm f n i = nth (sm f n) i;

sig_list_to_sig_map ::
  forall a.
    (Eq a) => a -> [((a, Nat), [a])] ->
                     ([((a, Nat), [a])] -> a -> Nat -> a) -> a -> Nat -> [a];
sig_list_to_sig_map a sml fmapa =
  let {
    fm = fmapa sml;
  } in (\ f n ->
         (case lookupb (f, n) sml of {
           Nothing -> [fm f n];
           Just xs -> (if null xs then [fm f n] else xs);
         }));

vars_rule_list :: forall a b. (Term a b, Term a b) -> [b];
vars_rule_list r = vars_term_lista (fst r) ++ vars_term_lista (snd r);

vars_rule_lista :: forall a b. (Eq b) => (Term a b, Term a b) -> [b];
vars_rule_lista r = remdups (vars_rule_list r);

aarity_term ::
  forall a b. (Eq a) => a -> (a -> Nat -> [a]) -> Term a b -> Maybe Nat;
aarity_term a sm t =
  (case unapp a t of {
    (Var _, ts) -> Nothing;
    (Fun f ss, ts) ->
      Just (minus_nat (aarity sm f (size_list ss)) (size_list ts));
  });

eta_closed_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> (a -> Nat -> [a]) ->
                         [(Term a b, Term a b)] ->
                           [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
eta_closed_rules a sm ra r =
  catcha
    (forallm
      (\ (l, raa) ->
        (case aarity_term a sm l of {
          Nothing -> Inr ();
          Just Zero_nat -> Inr ();
          Just (Suc _) ->
            check (any (\ (lll, rrr) ->
                         (case (lll, rrr) of {
                           (Var _, b) -> False;
                           (Fun f [], b) -> False;
                           (Fun f [ll], b) -> False;
                           (Fun f [ll, Var x], Var _) -> False;
                           (Fun f [ll, Var x], Fun g []) -> False;
                           (Fun f [ll, Var x], Fun g [rr]) -> False;
                           (Fun f [ll, Var x], Fun g [rr, Var y]) ->
                             f == a &&
                               (g == a &&
                                 (x == y &&
                                   (not (membera (vars_rule_lista (ll, rr))
  x) &&
                                     instance_rule (l, raa) (ll, rr))));
                           (Fun f [ll, Var x], Fun g (rr : Var y : _ : _)) ->
                             False;
                           (Fun f [ll, Var x], Fun g (rr : Fun _ _ : list)) ->
                             False;
                           (Fun f (ll : Var x : _ : _), b) -> False;
                           (Fun f (ll : Fun _ _ : list), b) -> False;
                         }))
                    r)
              (shows_prec_list Zero_nat
                 ['e', 't', 'a', ' ', 'e', 'x', 'p', 'a', 'n', 's', 'i', 'o',
                   'n', ' ', 'o', 'f', ' '] .
                shows_rule (l, raa) .
                  shows_prec_list Zero_nat
                    [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);
        }))
      ra)
    (\ x -> Inl (snd x));

check_subseteq :: forall a. (Eq a) => [a] -> [a] -> Sum a ();
check_subseteq xs ys =
  catcha (forallm (\ x -> (if membera ys x then Inr () else Inl x)) xs)
    (\ x -> Inl (snd x));

uncurry_term ::
  forall a b. (Eq a) => a -> (a -> Nat -> [a]) -> Term a b -> Term a b;
uncurry_term a sm t =
  (case unapp a t of {
    (Var x, ts) -> apply_args a (Var x) (map (uncurry_term a sm) ts);
    (Fun f ss, ts) ->
      let {
        n = size_list ss;
        uss = map (uncurry_term a sm) ss;
        uts = map (uncurry_term a sm) ts;
        aa = aarity sm f n;
        m = mina (size_list ts) aa;
        fm = get_symbol sm f n m;
      } in apply_args a (Fun fm (uss ++ takea m uts)) (dropa m uts);
  });

uncurry_rules ::
  forall a b.
    (Eq a) => a -> (a -> Nat -> [a]) ->
                     [(Term a b, Term a b)] -> [(Term a b, Term a b)];
uncurry_rules a sm =
  map (\ (l, r) -> (uncurry_term a sm l, uncurry_term a sm r));

check_CS_subseteq ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 [(Term a b, Term a b)] -> Sum (Term a b, Term a b) ();
check_CS_subseteq r s =
  catcha
    (forallm (\ (l, ra) -> check (any (instance_rule (l, ra)) s) (l, ra)) r)
    (\ x -> Inl (snd x));

generate_var :: Nat -> [Char];
generate_var i = 'x' : shows_prec_nat Zero_nat i [];

generate_f_xs :: forall a. a -> Nat -> Term a [Char];
generate_f_xs f n = Fun f (map (\ i -> Var (generate_var i)) (upt Zero_nat n));

uncurry_of_sig_list ::
  forall a.
    a -> [((a, Nat), [a])] ->
           (a -> Nat -> [a]) -> [(Term a [Char], Term a [Char])];
uncurry_of_sig_list a sml sm =
  concatMap
    (\ (b, c) ->
      let {
        (f, n) = b;
      } in (\ _ ->
             let {
               g = get_symbol sm f n;
             } in map (\ i ->
                        (Fun a [generate_f_xs (g i) (plus_nat n i),
                                 Var (generate_var (plus_nat n i))],
                          generate_f_xs (g (Suc i)) (plus_nat n (Suc i))))
                    (upt Zero_nat (aarity sm f n)))
        c)
    sml;

uncurry_tt ::
  forall a.
    (Eq a,
      Showa a) => (a, ([((a, Nat), [a])],
                        ([(Term a [Char], Term a [Char])],
                          [(Term a [Char], Term a [Char])]))) ->
                    [(Term a [Char], Term a [Char])] ->
                      [(Term a [Char], Term a [Char])] ->
                        Sum ([Char] -> [Char]) ();
uncurry_tt (a, (sml, (u, e))) ra r =
  let {
    sm = sig_list_to_sig_map a sml (\ _ f _ -> f);
    r_eta = e ++ ra;
  } in binda (catcha
               (forallm
                 (\ (l, _) ->
                   check (hvf_term a l)
                     (shows_prec_list Zero_nat
                        ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b',
                          'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                       shows_term l .
                         shows_prec_list Zero_nat
                           [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                             'e', 'd']))
                 r_eta)
               (\ x -> Inl (snd x)))
         (\ _ ->
           binda (eta_closed_rules a sm r_eta r_eta)
             (\ _ ->
               binda (catcha (check_subseteq (uncurry_rules a sm r_eta) r)
                       (\ x ->
                         Inl (shows_prec_list Zero_nat
                                ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd',
                                  ' ', 'r', 'u', 'l', 'e', ' '] .
                               shows_rule x .
                                 shows_prec_list Zero_nat
                                   [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                     'n', 'g'])))
                 (\ _ ->
                   catcha (check_CS_subseteq (uncurry_of_sig_list a sml sm) r)
                     (\ x ->
                       Inl (shows_prec_list Zero_nat
                              ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u',
                                'l', 'e', ' '] .
                             shows_rule x .
                               shows_prec_list Zero_nat
                                 [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                   'n', 'g'])))));

concat_lists :: forall a. [[a]] -> [[a]];
concat_lists [] = [[]];
concat_lists (asa : xs) =
  let {
    a = concat_lists xs;
  } in concatMap (\ vec -> map (\ aa -> aa : vec) asa) a;

enum_vectors :: forall a b. [a] -> [b] -> [[(b, a)]];
enum_vectors c [] = [[]];
enum_vectors c (x : xs) =
  let {
    a = enum_vectors c xs;
  } in concatMap (\ vec -> map (\ ca -> (x, ca) : vec) c) a;

letters :: [Char];
letters =
  ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
    'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
    'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '_', '0', '1', '2', '3', '4', '5', '6',
    '7', '8', '9', '&', ';', ':'];

xml1or2many_elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) ->
            (a -> Maybe b -> [c] -> d) -> Xml -> Sum [Char] d;
xml1or2many_elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (is_none text && less_eq_nat (nat 2) (size_list cs)))
    then binda (p1 (nth cs Zero_nat))
           (\ x ->
             catcha
               (binda (p2 (nth cs one_nat))
                 (\ y ->
                   binda (mapa p3 (tl (tl cs)))
                     (\ xs -> Inr (f x (Just y) xs))))
               (\ _ -> binda (mapa p3 (tl cs)) (\ xs -> Inr (f x Nothing xs))))
    else no_transform tag (Xml name atts cs text));

arith_fun :: Xml -> Sum [Char] ArithFun;
arith_fun xml =
  xml1element ['a', 'r', 'i', 't', 'h', 'F', 'u', 'n', 'c', 't', 'i', 'o', 'n']
    (xml_options
      [(['n', 'a', 't', 'u', 'r', 'a', 'l'],
         xml_change (nata ['n', 'a', 't', 'u', 'r', 'a', 'l']) Const),
        (['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'],
          xml_change (nata ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e'])
            (\ n -> Arg (minus_nat n one_nat))),
        (['s', 'u', 'm'], xml_many_elements ['s', 'u', 'm'] arith_fun Sum),
        (['p', 'r', 'o', 'd', 'u', 'c', 't'],
          xml_many_elements ['p', 'r', 'o', 'd', 'u', 'c', 't'] arith_fun Prod),
        (['m', 'i', 'n'], xml_many_elements ['m', 'i', 'n'] arith_fun Min),
        (['m', 'a', 'x'], xml_many_elements ['m', 'a', 'x'] arith_fun Max)])
    id xml;

xml0to1elements ::
  forall a b.
    [Char] -> (Xml -> Sum [Char] a) -> (Maybe a -> b) -> Xml -> Sum [Char] b;
xml0to1elements tag p1 f (Xml name atts cs text) =
  let {
    l = size_list cs;
  } in (if name == tag &&
             (null atts &&
               (less_eq_nat Zero_nat l &&
                 (less_eq_nat l one_nat && is_none text)))
         then (if equal_nat l one_nat
                then binda (p1 (nth cs Zero_nat)) (\ x1 -> Inr (f (Just x1)))
                else Inr (f Nothing))
         else no_transform tag (Xml name atts cs text));

sl_inter ::
  forall a b.
    (Xml -> Sum [Char] (Lab a b)) ->
      Xml -> Sum [Char] (Sum (Sl_inter a b, Bool) (Maybe (Lab a b)));
sl_inter xml2name =
  xml1element ['m', 'o', 'd', 'e', 'l']
    (xml_options
      [(['f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e', 'l'],
         xml1or2many_elements
           ['f', 'i', 'n', 'i', 't', 'e', 'M', 'o', 'd', 'e', 'l']
           (xml_change
             (nata ['c', 'a', 'r', 'r', 'i', 'e', 'r', 'S', 'i', 'z', 'e'])
             (\ n -> minus_nat n one_nat))
           (xml1element ['t', 'u', 'p', 'l', 'e', 'O', 'r', 'd', 'e', 'r']
             (xml0elements ['p', 'o', 'i', 'n', 't', 'W', 'i', 's', 'e'] ()) id)
           (xml3elements ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't'] xml2name
             (nata ['a', 'r', 'i', 't', 'y']) arith_fun
             (\ f a -> (\ b -> ((f, a), b))))
           (\ n model xs -> Inl (SL_Inter n xs, is_none model))),
        (['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g'],
          xml0to1elements
            ['r', 'o', 'o', 't', 'L', 'a', 'b', 'e', 'l', 'i', 'n', 'g']
            xml2name Inr)])
    id;

xml_term :: forall a b. (Showa a, Showa b) => Term a b -> Xml;
xml_term (Var x) = Xml ['v', 'a', 'r'] [] [] (Just (shows_prec Zero_nat x []));
xml_term (Fun f ts) =
  Xml ['f', 'u', 'n', 'a', 'p', 'p'] []
    (Xml ['n', 'a', 'm', 'e'] [] [] (Just (shows_prec Zero_nat f [])) :
      map (\ t -> Xml ['a', 'r', 'g'] [] [xml_term t] Nothing) ts)
    Nothing;

less_eq_list :: forall a. (Eq a, Order a) => [a] -> [a] -> Bool;
less_eq_list (x : xs) (y : ys) = less x y || x == y && less_eq_list xs ys;
less_eq_list [] xs = True;
less_eq_list (x : xs) [] = False;

less_list :: forall a. (Eq a, Order a) => [a] -> [a] -> Bool;
less_list (x : xs) (y : ys) = less x y || x == y && less_list xs ys;
less_list [] (x : xs) = True;
less_list xs [] = False;

instance (Eq a, Order a) => Orda [a] where {
  less_eq = less_eq_list;
  less = less_list;
};

data DPDisproof a b =
  DPLoop (Term a b) [(Pos, ((Term a b, Term a b), (Bool, Term a b)))]
    [(b, Term a b)] (Ctxt a b)
  | DPRuleRemoval [(Term a b, Term a b)] [(Term a b, Term a b)]
      (DPDisproof a b);

data TRSDisproof a b =
  Loop (Term a b) [(Pos, ((Term a b, Term a b), Term a b))] [(b, Term a b)]
    (Ctxt a b)
  | NotWellFormed | NTRuleRemoval [(Term a b, Term a b)] (TRSDisproof a b)
  | NTStringReversal [(Term a b, Term a b)] (TRSDisproof a b)
  | NTDPTrans [(Term a b, Term a b)] (DPDisproof a b);

poly_var :: forall a b. (Eq a) => [([(a, Nat)], b)] -> a -> Bool;
poly_var p v = membera (concatMap (map fst . fst) p) v;

data CertificationResult = Certified [Char] | Unsupported [Char] | Error [Char];

rm_alpha :: forall a b. (Linorder a) => Rbt a b -> a -> Maybe b;
rm_alpha = lookupa;

rm_empty :: forall a b. (Linorder a) => Rbt a b;
rm_empty = empty;

rs_alpha :: forall a. (Linorder a) => Rbt a () -> a -> Bool;
rs_alpha = s_alpha rm_alpha;

rs_empty :: forall a. (Linorder a) => Rbt a ();
rs_empty = s_empty rm_empty;

rs_union :: forall a. (Linorder a) => Rbt a () -> Rbt a () -> Rbt a ();
rs_union = s_union rm_add;

funs_term_list :: forall a b. Term a b -> [a];
funs_term_list (Var uu) = [];
funs_term_list (Fun f ts) = f : concatMap funs_term_list ts;

funs_rule_list :: forall a b. (Term a b, Term a b) -> [a];
funs_rule_list r = funs_term_list (fst r) ++ funs_term_list (snd r);

funs_trs_list :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [a];
funs_trs_list r = remdups (concatMap funs_rule_list r);

unlab :: forall a b. Lab a b -> Lab a b;
unlab (Lab f l) = f;
unlab (FunLab f l) = f;
unlab (UnLab v) = UnLab v;
unlab (Sharp v) = Sharp v;

is_labeled :: forall a b. Lab a b -> Bool;
is_labeled (UnLab uu) = False;
is_labeled (Sharp uv) = False;
is_labeled (Lab v va) = True;
is_labeled (FunLab v va) = True;

label_decomp :: forall a b. Lab a b -> (Lab a b, Sum b [Lab a b]);
label_decomp (Lab f l) = (f, Inl l);
label_decomp (FunLab f l) = (f, Inr l);

lge_to_lgr ::
  forall a b.
    (Eq b) => (a -> Nat -> b -> b -> Bool) ->
                (a -> Nat -> b -> Bool) -> a -> Nat -> b -> b -> Bool;
lge_to_lgr lge ls f n =
  let {
    lSfn = ls f n;
    lgefn = lge f n;
  } in (\ l la ->
         not (l == la) && (member l lSfn && (member la lSfn && lgefn l la)));

check_sl_only_lab_and_decr ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => (Lab a b ->
                    Nat -> Sum b [Lab a b] -> Sum b [Lab a b] -> Bool) ->
                    (Lab a b -> Nat -> Sum b [Lab a b] -> Bool) ->
                      [(Term (Lab a b) c, Term (Lab a b) c)] ->
                        [(Term (Lab a b) c, Term (Lab a b) c)] ->
                          Sum ([Char] -> [Char]) ();
check_sl_only_lab_and_decr lge ls r =
  (\ xs ->
    catcha
      (forallm
        (\ (ll, lr) ->
          let {
            l = map_term unlab ll;
            ra = map_term unlab lr;
            n = num_args l;
          } in check (if equal_term l ra
                       then not (is_var ll) &&
                              (not (is_var lr) &&
                                (args ll == args lr &&
                                  (is_labeled (the (root ll)) &&
                                    (is_labeled (the (root lr)) &&
                                      (ls (the (root l)) n
 (snd (label_decomp (the (root ll)))) &&
(ls (the (root l)) n (snd (label_decomp (the (root lr)))) &&
  lge_to_lgr lge ls (the (root l)) n (snd (label_decomp (the (root ll))))
    (snd (label_decomp (the (root lr)))))))))) ||
                              equal_term ll lr && membera r (l, ra)
                       else membera r (l, ra))
                 (shows_string
                    ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                      ' '] .
                   shows_rule (l, ra) .
                     shows_string
                       [' ', 'i', 's', ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r',
                         ' ', 'd', 'e', 'c', 'r', 'e', 'a', 's', 'i', 'n', 'g',
                         ' ', 'n', 'o', 'r', ' ', 'a', ' ', 'l', 'a', 'b', 'e',
                         'l', 'e', 'd', ' ', 'v', 'e', 'r', 's', 'i', 'o', 'n',
                         ' ', 'o', 'f', ' ', 'a', 'n', ' ', 'o', 'r', 'i', 'g',
                         'i', 'n', 'a', 'l', ' ', 'r', 'u', 'l', 'e']))
        xs)
      (\ x -> Inl (snd x)));

check_sem_lab_tt ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => Sum ([Char] -> [Char]) () ->
                    ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                      Sum ([Char] -> [Char]) ()) ->
                      ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                        [(Term (Lab a b) c, Term (Lab a b) c)] ->
                          Sum ([Char] -> [Char]) ()) ->
                        ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                          [(Term (Lab a b) c, Term (Lab a b) c)] ->
                            Sum ([Char] -> [Char]) ()) ->
                          (Lab a b ->
                            Nat ->
                              Sum b [Lab a b] -> Sum b [Lab a b] -> Bool) ->
                            (Lab a b -> Nat -> Sum b [Lab a b] -> Bool) ->
                              [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                  Sum ([Char] -> [Char]) ();
check_sem_lab_tt valid check_decr check_lab check_model_lab lge ls r lR =
  binda valid
    (\ _ ->
      catcha
        (binda (check_decr lR)
          (\ _ ->
            binda (check_model_lab lR r)
              (\ _ -> check_sl_only_lab_and_decr lge ls r lR)))
        (\ x ->
          Inl (shows_string
                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                   ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S',
                   ':'] .
                shows_nl . x)));

model_check_valid :: forall a. Sum a ();
model_check_valid = Inr ();

model_check_decr :: forall a b. a -> Sum b ();
model_check_decr = (\ _ -> Inr ());

eval_lab ::
  forall a b c d e.
    (a -> [b] -> b) ->
      (a -> [b] -> c) ->
        (a -> Nat -> c -> d) -> (e -> b) -> Term a e -> (b, Term d e);
eval_lab i l lc alpha (Var x) = (alpha x, Var x);
eval_lab i l lc alpha (Fun f ts) =
  let {
    clts = map (eval_lab i l lc alpha) ts;
    cs = map fst clts;
    c = i f cs;
    lts = map snd clts;
  } in (c, Fun (lc f (size_list ts) (l f cs)) lts);

check_sl_rule_ass ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => Bool ->
                    (a -> [b] -> b) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          (b -> b -> Bool) ->
                            [(Term d e, Term d e)] ->
                              (e -> b) ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_ass mc i la lc cge lR alpha (l, r) =
  let {
    cl_ll = eval_lab i la lc alpha l;
    cr_lr = eval_lab i la lc alpha r;
  } in binda (check (implies mc (cge (fst cl_ll) (fst cr_lr)))
               (shows_string ['r', 'u', 'l', 'e', ' '] .
                 shows_rule (l, r) .
                   shows_string
                     [' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 't',
                       'h', 'e', ' ', 'm', 'o', 'd', 'e', 'l', ' ', 'c', 'o',
                       'n', 'd', 'i', 't', 'i', 'o', 'n', ',', ' ', '[', 'l',
                       'h', 's', ']', ' ', '=', ' '] .
                     shows_prec Zero_nat (fst cl_ll :: b) .
                       shows_string
                         [',', ' ', '[', 'r', 'h', 's', ']', ' ', '=', ' '] .
                         shows_prec Zero_nat (fst cr_lr :: b)))
         (\ _ ->
           check (membera lR (snd cl_ll, snd cr_lr))
             (shows_string
                ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                  ' '] .
               shows_rule (snd cl_ll, snd cr_lr) .
                 shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']));

assign_from_vector :: forall a b. (Eq a) => [(a, b)] -> a -> b;
assign_from_vector vec x = the (lookupb x vec);

check_sl_rule ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> Nat -> c -> d) ->
                        [b] ->
                          (b -> b -> Bool) ->
                            Bool ->
                              [(Term d e, Term d e)] ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule i la lc c cge mc lR (l, r) =
  catcha
    (forallm (\ alpha -> check_sl_rule_ass mc i la lc cge lR alpha (l, r))
      (map assign_from_vector
        (enum_vectors c (list_union (vars_term_list l) (vars_term_list r)))))
    (\ x -> Inl (snd x));

labela :: forall a b. Lab a b -> Nat -> Sum b [Lab a b] -> Lab a b;
labela f n (Inl l) = Lab f l;
labela f n (Inr l) = FunLab f l;

check_sl_lab_trs ::
  forall a b c d.
    (Eq a, Showa a, Eq b, Showa b, Showa c, Eq d,
      Showa d) => (Lab a b -> [c] -> c) ->
                    (Lab a b -> [c] -> Sum b [Lab a b]) ->
                      [c] ->
                        (c -> c -> Bool) ->
                          [(Term (Lab a b) d, Term (Lab a b) d)] ->
                            [(Term (Lab a b) d, Term (Lab a b) d)] ->
                              Sum ([Char] -> [Char]) ();
check_sl_lab_trs i l c cge lP p =
  catcha (forallm (check_sl_rule i l labela c cge False lP) p)
    (\ x -> Inl (snd x));

rl_inter_list :: forall a b. (Eq a) => [a] -> a -> a -> b -> a;
rl_inter_list fs d g uu = (if membera fs g then g else d);

rl_fun :: forall a b c. a -> b -> Sum c b;
rl_fun uu cs = Inr cs;

model_cge :: forall a. (Eq a) => a -> a -> Bool;
model_cge = (\ a b -> a == b);

check_sl_model_lab_trs ::
  forall a b c d.
    (Eq a, Showa a, Eq b, Showa b, Showa c, Eq d,
      Showa d) => (Lab a b -> [c] -> c) ->
                    (Lab a b -> [c] -> Sum b [Lab a b]) ->
                      [c] ->
                        (c -> c -> Bool) ->
                          [(Term (Lab a b) d, Term (Lab a b) d)] ->
                            [(Term (Lab a b) d, Term (Lab a b) d)] ->
                              Sum ([Char] -> [Char]) ();
check_sl_model_lab_trs i l c cge lR r =
  catcha (forallm (check_sl_rule i l labela c cge True lR) r)
    (\ x -> Inl (snd x));

model_lge :: forall a b c. (Eq c) => a -> b -> c -> c -> Bool;
model_lge = (\ _ _ -> (\ a b -> a == b));

model_LS :: forall a b c. a -> b -> c -> Bool;
model_LS = (\ _ _ -> top_fun);

rl_tt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                    [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                      Sum ([Char] -> [Char]) ();
rl_tt r lR =
  let {
    fs = funs_trs_list r;
  } in binda (check (not (null fs))
               (shows_string
                  ['r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                    'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
                    't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ',
                    'f', 'u', 'n', 'c', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm',
                    'b', 'o', 'l', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 's',
                    'i', 'g', 'n', 'a', 't', 'u', 'r', 'e'] .
                 shows_nl))
         (\ _ ->
           let {
             d = head fs;
           } in check_sem_lab_tt model_check_valid model_check_decr
                  (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs model_cge)
                  (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs
                    model_cge)
                  model_lge model_LS r lR);

rp_valid ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      Sum ([Char] -> [Char]) ();
rp_valid (v, (s, (n, (a, (m, d))))) = v;

uncurry_top ::
  forall a b. (Eq a) => a -> Nat -> (a -> Nat -> [a]) -> Term a b -> Term a b;
uncurry_top a n sm (Fun f ts) =
  let {
    mt = map (map_term_wa (\ fa na -> get_symbol sm fa na Zero_nat));
  } in (if f == a &&
             (equal_nat (size_list ts) n &&
               (not (is_var (head ts)) &&
                 not (equal_nat
                       (aarity sm (the (root (head ts))) (num_args (head ts)))
                       Zero_nat)))
         then let {
                (Fun g ss) = head ts;
              } in Fun (get_symbol sm g (size_list ss) one_nat)
                     (mt (ss ++ tl ts))
         else Fun (get_symbol sm f (size_list ts) Zero_nat) (mt ts));
uncurry_top a n sm (Var x) = Var x;

xml_rules ::
  forall a b. (Showa a, Showa b) => [Char] -> [(Term a b, Term a b)] -> Xml;
xml_rules tag rls =
  Xml tag []
    [Xml ['r', 'u', 'l', 'e', 's'] []
       (map (\ (l, r) ->
              Xml ['r', 'u', 'l', 'e'] []
                [Xml ['l', 'h', 's'] [] [xml_term l] Nothing,
                  Xml ['r', 'h', 's'] [] [xml_term r] Nothing]
                Nothing)
         rls)
       Nothing]
    Nothing;

lexord_code :: forall a. (Eq a) => (a -> a -> Bool) -> [a] -> [a] -> Bool;
lexord_code lt uu [] = False;
lexord_code lt [] (v : va) = True;
lexord_code lt (x : xs) (y : ys) = lt x y || x == y && lexord_code lt xs ys;

less_lab ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Lab a b -> Lab a b -> Bool;
less_lab (UnLab f) (UnLab g) = less f g;
less_lab (UnLab f) (Lab v va) = True;
less_lab (UnLab f) (FunLab v va) = True;
less_lab (UnLab f) (Sharp v) = True;
less_lab (Lab f la) (Lab g l) = less la l || la == l && less_lab f g;
less_lab (Lab f l) (UnLab uv) = False;
less_lab (Lab f l) (FunLab v va) = True;
less_lab (Lab f l) (Sharp v) = True;
less_lab (Sharp f) (Sharp g) = less_lab f g;
less_lab (Sharp f) (UnLab ux) = False;
less_lab (Sharp f) (Lab uy uz) = False;
less_lab (Sharp f) (FunLab v vb) = True;
less_lab (FunLab f fs) (FunLab g gs) =
  less_lab f g || equal_lab f g && lexord_code less_lab fs gs;
less_lab (FunLab f fs) (Lab v va) = False;
less_lab (FunLab f fs) (UnLab v) = False;
less_lab (FunLab f fs) (Sharp v) = False;

less_eq_lab ::
  forall a b. (Eq a, Orda a, Eq b, Orda b) => Lab a b -> Lab a b -> Bool;
less_eq_lab xs ys = less_lab xs ys || equal_lab xs ys;

instance (Eq a, Orda a, Eq b, Orda b) => Orda (Lab a b) where {
  less_eq = less_eq_lab;
  less = less_lab;
};

class (Semiring_0 a, Ordered_ab_semigroup a) => Ordered_semiring_0 a where {
};

class (Semiring_1 a, Ordered_semiring_0 a) => Ordered_semiring_1 a where {
};

class (Comm_semiring_1 a, Ordered_semiring_1 a) => Poly_carrier a where {
};

class (Max_ordered_monoid_add a,
        Ordered_semiring_1 a) => Max_ordered_semiring_1 a where {
};

class (Poly_carrier a, Max_ordered_semiring_1 a) => Poly_carriera a where {
};

default_I :: forall a. (Poly_carriera a) => a -> Nat -> [([(Nat, Nat)], a)];
default_I def n =
  ([], def) : map (\ i -> ([(i, one_nat)], one)) (upt Zero_nat n);

poly_power ::
  forall a b.
    (Eq a, Eq b,
      Comm_semiring_1 b) => [([(a, Nat)], b)] -> Nat -> [([(a, Nat)], b)];
poly_power uu Zero_nat = one_poly;
poly_power p (Suc n) = poly_mult p (poly_power p n);

monom_subst ::
  forall a b c.
    (Eq b, Eq c,
      Comm_semiring_1 c) => (a -> [([(b, Nat)], c)]) ->
                              [(a, Nat)] -> [([(b, Nat)], c)];
monom_subst sigma [] = one_poly;
monom_subst sigma ((x, p) : m) =
  poly_mult (poly_power (sigma x) p) (monom_subst sigma m);

poly_subst ::
  forall a b c.
    (Eq b, Eq c,
      Comm_semiring_1 c) => (a -> [([(b, Nat)], c)]) ->
                              [([(a, Nat)], c)] -> [([(b, Nat)], c)];
poly_subst sigma [] = zero_poly;
poly_subst sigma ((m, c) : p) =
  poly_add (poly_mult [([], c)] (monom_subst sigma m)) (poly_subst sigma p);

eval_term ::
  forall a b c.
    (Eq b, Poly_carriera b,
      Eq c) => (a -> Nat -> [([(Nat, Nat)], b)]) ->
                 Term a c -> [([(c, Nat)], b)];
eval_term uu (Var x) = [([(x, one_nat)], one)];
eval_term i (Fun f ts) =
  let {
    ps = map (eval_term i) ts;
    n = size_list ts;
  } in poly_subst (\ ia -> (if less_nat ia n then nth ps ia else zero_poly))
         (i f n);

eval_monom :: forall a b. (Comm_semiring_1 b) => (a -> b) -> [(a, Nat)] -> b;
eval_monom alpha [] = one;
eval_monom alpha ((x, p) : m) = times (eval_monom alpha m) (power (alpha x) p);

eval_poly ::
  forall a b. (Comm_semiring_1 b) => (a -> b) -> [([(a, Nat)], b)] -> b;
eval_poly alpha [] = zero;
eval_poly alpha (mc : p) =
  plus (times (eval_monom alpha (fst mc)) (snd mc)) (eval_poly alpha p);

indent :: ([Char] -> [Char]) -> [Char] -> [Char];
indent p s =
  concatMap (\ c -> (if c == newline then [newline, ' '] else [c])) (p s);

vars_trs_list :: forall a b. (Eq b) => [(Term a b, Term a b)] -> [b];
vars_trs_list trs = remdups (concatMap vars_rule_list trs);

funas_term_lista :: forall a b. Term a b -> [(a, Nat)];
funas_term_lista (Var uu) = [];
funas_term_lista (Fun f ts) = (f, size_list ts) : concatMap funas_term_lista ts;

funas_rule_list :: forall a b. (Term a b, Term a b) -> [(a, Nat)];
funas_rule_list r = funas_term_lista (fst r) ++ funas_term_lista (snd r);

funas_trs_list :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_trs_list r = remdups (concatMap funas_rule_list r);

shows_prec_ctxt ::
  forall a b. (Showa a, Showa b) => Nat -> Ctxt a b -> [Char] -> [Char];
shows_prec_ctxt d c = shows_ctxt c;

check_flat_ctxt ::
  forall a.
    (Eq a, Showa a) => [[Char]] -> Ctxt a [Char] -> Sum ([Char] -> [Char]) ();
check_flat_ctxt vs c =
  (case c of {
    Hole ->
      Inl (shows_prec_ctxt Zero_nat c .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a',
                't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
              shows_nl);
    More _ ss1 Hole ss2 ->
      let {
        ss = ss1 ++ ss2;
      } in binda (check (distinct ss)
                   (shows_prec_ctxt Zero_nat c .
                     shows_string
                       [' ', 'c', 'o', 'n', 't', 'a', 'i', 'n', 's', ' ', 'd',
                         'u', 'p', 'l', 'i', 'c', 'a', 't', 'e', ' ', 'v', 'a',
                         'r', 'i', 'a', 'b', 'l', 'e', 's'] .
                       shows_nl))
             (\ _ ->
               binda (check (all is_var ss)
                       (shows_prec_ctxt Zero_nat c .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'l',
                             'a', 't', ',', ' ', 'i', '.', 'e', '.', ',', ' ',
                             'h', 'a', 's', ' ', 'd', 'e', 'p', 't', 'h', ' ',
                             'g', 'r', 'e', 'a', 't', 'e', 'r', ' ', 't', 'h',
                             'a', 'n', ' ', 'o', 'n', 'e'] .
                           shows_nl))
                 (\ _ ->
                   check (all (\ t -> not (membera vs (the_var t)))
                           (ss1 ++ ss2))
                     (shows_prec_ctxt Zero_nat c .
                       shows_string
                         [' ', 'h', 'a', 's', ' ', 't', 'o', ' ', 'c', 'o', 'n',
                           't', 'a', 'i', 'n', ' ', 'o', 'n', 'l', 'y', ' ',
                           'f', 'r', 'e', 's', 'h', ' ', 'v', 'a', 'r', 'i',
                           'a', 'b', 'l', 'e', 's'] .
                         shows_nl)));
    More _ ss1 (More _ _ _ _) ss2 ->
      Inl (shows_prec_ctxt Zero_nat c .
            shows_string
              [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a',
                't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
              shows_nl);
  });

is_flat_ctxt_list ::
  forall a b. (Eq a, Eq b) => [a] -> [(b, Nat)] -> Ctxt b a -> Bool;
is_flat_ctxt_list vs fas (More f ss1 Hole ss2) =
  let {
    ss = ss1 ++ ss2;
  } in membera fas (f, Suc (size_list ss)) &&
         (all is_var ss &&
           (distinct ss && null (list_inter (map the_var ss) vs)));
is_flat_ctxt_list vs fas Hole = False;
is_flat_ctxt_list vs fas (More v va (More vd ve vf vg) vc) = False;

check_is_flat_ctxt ::
  forall a.
    (Eq a,
      Showa a) => [[Char]] ->
                    [(a, Nat)] -> Ctxt a [Char] -> Sum ([Char] -> [Char]) ();
check_is_flat_ctxt vs fas c =
  check (is_flat_ctxt_list vs fas c)
    (shows_prec_ctxt Zero_nat c .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'f', 'l', 'a', 't',
          ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
        shows_nl);

all_interval_nat :: (Nat -> Bool) -> Nat -> Nat -> Bool;
all_interval_nat p i j = less_eq_nat j i || p i && all_interval_nat p (Suc i) j;

hole_at :: forall a b. (Eq a) => Nat -> Nat -> a -> Ctxt a b -> Bool;
hole_at n i f (More g ss1 Hole ss2) =
  g == f &&
    (equal_nat (size_list ss1) i &&
      equal_nat (size_list ss2) (minus_nat (minus_nat n i) one_nat));
hole_at n i f Hole = False;
hole_at n i f (More v va (More vd ve vf vg) vc) = False;

check_flat_ctxt_complete ::
  forall a.
    (Eq a, Showa a) => [Ctxt a [Char]] -> (a, Nat) -> Sum ([Char] -> [Char]) ();
check_flat_ctxt_complete fcs fa =
  check (all_interval_nat (\ i -> any (hole_at (snd fa) i (fst fa)) fcs)
          Zero_nat (snd fa))
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'i', 's', 't', ' ', 'o', 'f', ' ', 'f', 'l',
         'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's', ' ', 'i', 's',
         ' ', 'i', 'n', 'c', 'o', 'm', 'p', 'l', 'e', 't', 'e'] .
      shows_nl);

check_rule_preserving ::
  forall a.
    (Eq a,
      Showa a) => [Ctxt a [Char]] ->
                    [(Term a [Char], Term a [Char])] ->
                      (Term a [Char], Term a [Char]) ->
                        Sum ([Char] -> [Char]) ();
check_rule_preserving fcs rs rule =
  check (membera rs rule ||
          all (\ c ->
                membera rs (ctxt_apply c (fst rule), ctxt_apply c (snd rule)))
            fcs)
    (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
      shows_rule rule .
        shows_string
          [' ', 'i', 's', ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
            'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 't', 'h',
            'e', ' ', 'r', 'e', 's', 'u', 'l', 't', 'i', 'n', 'g', ' ', 's',
            'e', 't', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'n',
            'o', 'r', ' ', 'c', 'l', 'o', 's', 'e', 'd', ' ', 'u', 'n', 'd',
            'e', 'r', ' ', 'a', 'l', 'l', ' ', 'f', 'l', 'a', 't', ' ', 'c',
            'o', 'n', 't', 'e', 'x', 't', 's'] .
          shows_nl);

check_rule_reflecting ::
  forall a.
    (Eq a,
      Showa a) => [Ctxt a [Char]] ->
                    [(Term a [Char], Term a [Char])] ->
                      (Term a [Char], Term a [Char]) ->
                        Sum ([Char] -> [Char]) ();
check_rule_reflecting fcs rs rule =
  check (any (\ (l, r) ->
               any (\ c ->
                     equal_term (fst rule) (ctxt_apply c l) &&
                       equal_term (snd rule) (ctxt_apply c r))
                 (Hole : fcs))
          rs)
    (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
      shows_rule rule .
        shows_string
          [' ', 'i', 's', ' ', 'n', 'e', 'i', 't', 'h', 'e', 'r', ' ', 'c', 'o',
            'n', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'i', 'n', ' ', 't', 'h',
            'e', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 's', 'e',
            't', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', 's', ' ', 'n', 'o',
            'r', ' ', 'o', 'b', 't', 'a', 'i', 'n', 'e', 'd', ' ', 'b', 'y',
            ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ', 'a', ' ', 'f',
            'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x', 't'] .
          shows_nl);

fcc_tt ::
  forall a.
    (Eq a,
      Showa a) => [(Term a [Char], Term a [Char])] ->
                    [Ctxt a [Char]] ->
                      [(Term a [Char], Term a [Char])] ->
                        Sum ([Char] -> [Char]) ();
fcc_tt r fcs lR =
  let {
    vs = vars_trs_list r;
    fas = funas_trs_list r;
  } in binda (check (not (null fcs))
               (shows_string
                  [' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n',
                    'e', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e',
                    'x', 't', ' ', 'i', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                    'e', 'd', ' ', 'f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'c',
                    'l', 'o', 's', 'u', 'r', 'e', ' ', 'u', 'n', 'd', 'e', 'r',
                    ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
                    't', 's'] .
                 shows_nl))
         (\ _ ->
           binda (catcha (forallm (check_flat_ctxt vs) fcs)
                   (\ x -> Inl (snd x)))
             (\ _ ->
               binda (catcha (forallm (check_is_flat_ctxt vs fas) fcs)
                       (\ x -> Inl (snd x)))
                 (\ _ ->
                   binda (catcha (forallm (check_flat_ctxt_complete fcs) fas)
                           (\ x -> Inl (snd x)))
                     (\ _ ->
                       binda (catcha (forallm (check_rule_preserving fcs lR) r)
                               (\ x -> Inl (snd x)))
                         (\ _ ->
                           catcha (forallm (check_rule_reflecting fcs r) lR)
                             (\ x -> Inl (snd x)))))));

check_supt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_supt s t =
  check (member (s, t) supt)
    (shows_prec_term Zero_nat t .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p', 'r', 'o', 'p',
          'e', 'r', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f',
          ' '] .
        shows_prec_term Zero_nat s);

rp_strict ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
rp_strict (v, (s, (n, (a, (m, d))))) = s;

eq_rule_mod_vars ::
  forall a b.
    (Eq a, Eq b) => (Term a b, Term a b) -> (Term a b, Term a b) -> Bool;
eq_rule_mod_vars lr st = instance_rule lr st && instance_rule st lr;

shows_prec_pos :: Nat -> Pos -> [Char] -> [Char];
shows_prec_pos d p = shows_pos p;

ctxt_of_pos_term :: forall a b. Pos -> Term a b -> Ctxt a b;
ctxt_of_pos_term Empty t = Hole;
ctxt_of_pos_term (PCons i ps) (Fun f ts) =
  More f (takea i ts) (ctxt_of_pos_term ps (nth ts i)) (dropa (Suc i) ts);

check_rstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Pos ->
                      (Term a b, Term a b) ->
                        Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rstep ra p r s t =
  let {
    sps = poss_list s;
    tps = poss_list t;
  } in binda (check (any (eq_rule_mod_vars r) ra)
               (shows_rule r .
                 shows_string
                   [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'r', 'u',
                     'l', 'e', ' ', 'o', 'f'] .
                   shows_nl . shows_trs ra . shows_nl))
         (\ _ ->
           binda (check (membera sps p)
                   (shows_prec_pos Zero_nat p .
                     shows_string
                       [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 'p',
                         'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                         ' '] .
                       shows_prec_term Zero_nat s . shows_nl))
             (\ _ ->
               binda (check (membera tps p)
                       (shows_prec_pos Zero_nat p .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                             'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' ', 'o',
                             'f', ' '] .
                           shows_prec_term Zero_nat t . shows_nl))
                 (\ _ ->
                   let {
                     c = ctxt_of_pos_term p s;
                     d = ctxt_of_pos_term p t;
                     u = subt_at s p;
                     v = subt_at t p;
                   } in (case match_list [(u, fst r), (v, snd r)] of {
                          Nothing ->
                            Inl (shows_string
                                   ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm',
                                     ' '] .
                                  shows_prec_term Zero_nat t .
                                    shows_string
                                      [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
't', ' ', 'r', 'e', 's', 'u', 'l', 't', ' ', 'f', 'r', 'o', 'm', ' ', 'a', ' ',
'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
'o', 'n', ' ', 'o', 'f', ' ', 'r', 'u', 'l', 'e', ' '] .
                                      shows_nl .
shows_rule r .
  shows_string
    [' ', 'a', 't', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' '] .
    shows_prec_pos Zero_nat p . shows_nl);
                          Just _ ->
                            check (equal_ctxt c d)
                              (shows_string
                                 ['t', 'h', 'e', ' ', 't', 'e', 'r', 'm', ' '] .
                                shows_prec_term Zero_nat t .
                                  shows_string
                                    [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
                                      't', ' ', 'r', 'e', 's', 'u', 'l', 't',
                                      ' ', 'f', 'r', 'o', 'm', ' ', 'a', ' ',
                                      'p', 'r', 'o', 'p', 'e', 'r', ' ', 'a',
                                      'p', 'p', 'l', 'i', 'c', 'a', 't', 'i',
                                      'o', 'n', ' ', 'o', 'f', ' ', 'r', 'u',
                                      'l', 'e', ' '] .
                                    shows_nl .
                                      shows_rule r .
shows_string [' ', 'a', 't', ' ', 'p', 'o', 's', 'i', 't', 'i', 'o', 'n', ' '] .
  shows_prec_pos Zero_nat p . shows_nl);
                        }))));

check_rrstep ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    (Term a b, Term a b) ->
                      Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rrstep ra r s t = check_rstep ra Empty r s t;

check_steps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                        Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_steps p r [] s u =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
         ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e'] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_string
            (shows_nl
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm']) .
            shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_steps p ra ((uu, (r, (True, t))) : prts) s u =
  binda (check_rrstep p r s t) (\ _ -> check_steps p ra prts t u);
check_steps pa ra ((p, (r, (False, t))) : prts) s u =
  binda (check_rstep ra p r s t) (\ _ -> check_steps pa ra prts t u);

only_eta_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
only_eta_rules e r_eta =
  catcha
    (forallm
      (\ (l, r) ->
        check (case (l, r) of {
                (Var _, b) -> False;
                (Fun f ls, Var _) -> False;
                (Fun f ls, Fun g rs) ->
                  f == g &&
                    (equal_nat (size_list ls) (nat 2) &&
                      (equal_nat (size_list rs) (nat 2) &&
                        (equal_term (nth ls one_nat) (nth rs one_nat) &&
                          any (\ (la, ra) ->
                                instance_rule (head ls, head rs) (la, ra))
                            r_eta)));
              })
          (shows_prec_list Zero_nat ['r', 'u', 'l', 'e', ' '] .
            shows_rule (l, r) .
              shows_prec_list Zero_nat
                [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n', ' ', '(',
                  'e', 't', 'a', '-', 'e', 'x', 'p', 'a', 'n', 'd', 'e', 'd',
                  ')', ' ', 'o', 'r', 'i', 'g', 'i', 'n', 'a', 'l', ' ', 'r',
                  'u', 'l', 'e']))
      e)
    (\ x -> Inl (snd x));

uncurry_proc ::
  forall a.
    (Eq a,
      Showa a) => (a, ([((a, Nat), [a])],
                        ([(Term a [Char], Term a [Char])],
                          [(Term a [Char], Term a [Char])]))) ->
                    ([((a, Nat), [a])] -> a -> Nat -> a) ->
                      (a -> Nat ->
                              [((a, Nat), [a])] -> Sum ([Char] -> [Char]) ()) ->
                        [(Term a [Char], Term a [Char])] ->
                          [(Term a [Char], Term a [Char])] ->
                            [(Term a [Char], Term a [Char])] ->
                              [(Term a [Char], Term a [Char])] ->
                                Maybe ((a, Nat) ->
[(Term a [Char], Term a [Char])]) ->
                                  Sum ([Char] -> [Char]) ();
uncurry_proc (a, (sml, (uu, e))) fmapa check_inj pa ra p r rm =
  let {
    sm = sig_list_to_sig_map a sml fmapa;
    r_eta = e ++ ra;
    u = uncurry_rules a sm r_eta;
    s = uncurry_of_sig_list a sml sm;
  } in binda (only_eta_rules e r_eta)
         (\ _ ->
           binda (check_inj a (nat 2) sml)
             (\ _ ->
               binda (catcha
                       (forallm
                         (\ (l, _) ->
                           check (hvf_term a l)
                             (shows_prec_list Zero_nat
                                ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i',
                                  'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l',
                                  'h', 's', ' '] .
                               shows_term l .
                                 shows_prec_list Zero_nat
                                   [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                     'w', 'e', 'd']))
                         pa)
                       (\ x -> Inl (snd x)))
                 (\ _ ->
                   binda (catcha
                           (forallm
                             (\ (l, _) ->
                               check (hvf_term a l)
                                 (shows_prec_list Zero_nat
                                    ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r',
                                      'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n',
                                      ' ', 'l', 'h', 's', ' '] .
                                   shows_term l .
                                     shows_prec_list Zero_nat
                                       [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l',
 'o', 'w', 'e', 'd']))
                             r_eta)
                           (\ x -> Inl (snd x)))
                     (\ _ ->
                       binda (eta_closed_rules a sm r_eta r_eta)
                         (\ _ ->
                           binda (catcha
                                   (check_subseteq (uncurry_rules a sm pa) p)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a', 'i', 'r',
      ' '] .
   shows_rule x .
     shows_prec_list Zero_nat
       [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                             (\ _ ->
                               binda (catcha (check_subseteq u r)
                                       (\ x ->
 Inl (shows_prec_list Zero_nat
        ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
          ' '] .
       shows_rule x .
         shows_prec_list Zero_nat
           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                 (\ _ ->
                                   binda (catcha (check_CS_subseteq s r)
   (\ x ->
     Inl (shows_prec_list Zero_nat
            ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e', ' '] .
           shows_rule x .
             shows_prec_list Zero_nat
               [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                     (\ _ ->
                                       catcha (check_CS_subseteq r (u ++ s))
 (\ x ->
   Inl (shows_prec_list Zero_nat
          ['a', 'd', 'd', 'i', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'r', 'u', 'l',
            'e', ' '] .
         shows_rule x .
           shows_prec_list Zero_nat
             [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
               'e', 'd']))))))))));

pi_full :: forall a. a -> Nat -> Nat -> Bool;
pi_full uu uv uw = True;

update_tokens :: forall a. ([a] -> [a]) -> [a] -> Sum [Char] ([a], [a]);
update_tokens f ts = Inr (ts, f ts);

remove_comments_aux_acc :: [Char] -> Bool -> [Char] -> [Char];
remove_comments_aux_acc a False (c : cs) =
  (if c == '<' && takea (nat 3) cs == ['!', '-', '-']
    then remove_comments_aux_acc a True (tl cs)
    else remove_comments_aux_acc (c : a) False cs);
remove_comments_aux_acc a True (c : cs) =
  (if c == '-' && takea (nat 2) cs == ['-', '>']
    then remove_comments_aux_acc a False (dropa (nat 2) cs)
    else remove_comments_aux_acc a True cs);
remove_comments_aux_acc a uu [] = a;

remove_comments_aux :: Bool -> [Char] -> [Char];
remove_comments_aux False (c : cs) =
  (if c == '<' && takea (nat 3) cs == ['!', '-', '-']
    then remove_comments_aux True (tl cs)
    else c : remove_comments_aux False cs);
remove_comments_aux True (c : cs) =
  (if c == '-' && takea (nat 2) cs == ['-', '>']
    then remove_comments_aux False (dropa (nat 2) cs)
    else remove_comments_aux True cs);
remove_comments_aux uu [] = [];

remove_comments :: [Char] -> [Char];
remove_comments =
  (if False then reverse . remove_comments_aux_acc [] False
    else remove_comments_aux False);

scan_upto :: [Char] -> [Char] -> Sum [Char] ([Char], [Char]);
scan_upto end [] =
  Inl (['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'e',
         'n', 'd', '-', 'm', 'a', 'r', 'k', 'e', 'r', ' '] ++
        quote (shows_prec_list Zero_nat end) []);
scan_upto end (t : ts) =
  (if map snd (zip end (t : ts)) == end
    then Inr (end, dropa (size_list end) (t : ts))
    else binda (scan_upto end ts) (\ (res, tsa) -> Inr (t : res, tsa)));

parse_header :: [Char] -> Sum [Char] ([Char], [Char]);
parse_header ts =
  (if takea (nat 2) (trim ts) == ['<', '?']
    then bind (scan_upto ['?', '>'])
           (\ h -> bind parse_header (\ hs tsa -> Inr (h ++ hs, tsa))) ts
    else bind spaces (\ _ tsa -> Inr ([], tsa)) ts);

parse_name :: [Char] -> Sum [Char] ([Char], [Char]);
parse_name =
  bind (manyof letters)
    (\ n ->
      bind spaces
        (\ _ ->
          (if null n then (\ _ -> Inl letters) else (\ ts -> Inr (n, ts)))));

parse_attribute_value :: [Char] -> Sum [Char] ([Char], [Char]);
parse_attribute_value =
  bind (oneof [dquote])
    (\ _ ->
      bind (many (\ c -> not ([c] == dquote)))
        (\ v -> bind (oneof [dquote]) (\ _ ts -> Inr (v, ts))));

parse_attributes :: [Char] -> Sum [Char] ([([Char], [Char])], [Char]);
parse_attributes [] = Inr ([], []);
parse_attributes (c : s) =
  (if membera ['/', '>'] c then Inr ([], c : s)
    else bind parse_name
           (\ k ->
             bind (oneof [['=']])
               (\ _ ->
                 bind parse_attribute_value
                   (\ v ->
                     bind parse_attributes
                       (\ atts ts -> Inr ((k, v) : atts, ts)))))
           (c : s));

parse_text :: [Char] -> Sum [Char] (Maybe [Char], [Char]);
parse_text =
  bind (many (\ c -> not ([c] == ['<'])))
    (\ ts ->
      let {
        text = trim ts;
      } in (if null text then (\ tsa -> Inr (Nothing, tsa))
             else (\ tsa -> Inr (Just (reverse (trim (reverse text))), tsa))));

parse_nodes :: [Char] -> Sum [Char] ([Xml], [Char]);
parse_nodes ts =
  (if null ts || takea (nat 2) ts == ['<', '/'] then Inr ([], ts)
    else bind (oneof [['<']])
           (\ _ ->
             bind parse_name
               (\ n ->
                 bind parse_attributes
                   (\ atts ->
                     bind (oneof [['/', '>'], ['>']])
                       (\ e tsa ->
                         (if e == ['/', '>']
                           then bind parse_nodes
                                  (\ cs tsb ->
                                    Inr (Xml n atts [] Nothing : cs, tsb))
                                  tsa
                           else bind parse_text
                                  (\ t ->
                                    bind parse_nodes
                                      (\ cs ->
bind (oneof [['<', '/'] ++ n ++ ['>']])
  (\ _ -> bind parse_nodes (\ ns tsb -> Inr (Xml n atts cs t : ns, tsb)))))
                                  tsa)))))
           ts);

parse_node :: [Char] -> Sum [Char] (Xml, [Char]);
parse_node =
  bind (oneof [['<']])
    (\ _ ->
      bind parse_name
        (\ n ->
          bind parse_attributes
            (\ atts ->
              bind (oneof [['/', '>'], ['>']])
                (\ e ->
                  (if e == ['/', '>']
                    then (\ ts -> Inr (Xml n atts [] Nothing, ts))
                    else bind parse_text
                           (\ t ->
                             bind parse_nodes
                               (\ cs ->
                                 bind (oneof [['<', '/'] ++ n ++ ['>']])
                                   (\ _ ts -> Inr (Xml n atts cs t, ts)))))))));

parse_doc :: [Char] -> Sum [Char] (Xmldoc, [Char]);
parse_doc =
  debug ['0']
    ['p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 't', 'x', 't', ' ', 't', 'o', ' ',
      'x', 'm', 'l']
    (bind (update_tokens remove_comments)
      (\ _ ->
        bind parse_header
          (\ h ->
            bind parse_node
              (\ xml -> bind eoi (\ _ ts -> Inr (Xmldoc h xml, ts))))));

data Filtered a = FPair a Nat;

plain_name :: Xml -> Sum [Char] (Lab [Char] [Nat]);
plain_name x =
  xml_options
    [(['n', 'a', 'm', 'e'], xml_change (text ['n', 'a', 'm', 'e']) UnLab),
      (['s', 'h', 'a', 'r', 'p'],
        xml1element ['s', 'h', 'a', 'r', 'p'] plain_name Sharp),
      (['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l'],
        xml2elements
          ['l', 'a', 'b', 'e', 'l', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l']
          plain_name (label plain_name)
          (\ f a ->
            (case a of {
              Inl aa -> Lab f aa;
              Inr aa -> FunLab f aa;
            })))]
    x;

xml2many_elements ::
  forall a b c d.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) -> (a -> b -> [c] -> d) -> Xml -> Sum [Char] d;
xml2many_elements tag p1 p2 p3 f (Xml name atts cs text) =
  (if name == tag &&
        (null atts && (is_none text && less_eq_nat (nat 2) (size_list cs)))
    then binda (p1 (nth cs Zero_nat))
           (\ x ->
             binda (p2 (nth cs one_nat))
               (\ y -> binda (mapa p3 (tl (tl cs))) (\ xs -> Inr (f x y xs))))
    else no_transform tag (Xml name atts cs text));

transition_lhs ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml -> Sum [Char] (Sum ((a, Nat), [[Char]]) [Char]);
transition_lhs xml2name x =
  (if let {
        (Xml _ _ cs _) = x;
      } in equal_nat (size_list cs) one_nat
    then xml1element ['l', 'h', 's'] state Inr x
    else xml2many_elements ['l', 'h', 's'] xml2name
           (nata ['h', 'e', 'i', 'g', 'h', 't']) state
           (\ f h qs -> Inl ((f, h), qs)) x);

transition ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml -> Sum [Char] (Sum (Ta_rule [Char] (a, Nat)) ([Char], [Char]));
transition xml2name =
  xml2elements ['t', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n']
    (transition_lhs xml2name) (xml1element ['r', 'h', 's'] state id)
    (\ lhs q ->
      (case lhs of {
        Inl (f, qs) -> Inl (TA_rule f qs q);
        Inr qa -> Inr (qa, q);
      }));

instance Orda Char where {
  less_eq = (\ a b -> a <= b);
  less = (\ a b -> a < b);
};

instance (Eq a, Order a) => Preorder [a] where {
};

instance (Eq a, Order a) => Order [a] where {
};

gen_list_to_map_aux :: forall a b c. (a -> b -> c -> c) -> c -> [(a, b)] -> c;
gen_list_to_map_aux upd accs [] = accs;
gen_list_to_map_aux upd accs ((k, v) : l) =
  gen_list_to_map_aux upd (upd k v accs) l;

gen_list_to_map :: forall a b c. a -> (b -> c -> a -> a) -> [(b, c)] -> a;
gen_list_to_map empt upd l = gen_list_to_map_aux upd empt (reverse l);

data RelTRSDisproof a b =
  RelLoop (Term a b) [(Pos, ((Term a b, Term a b), (Bool, Term a b)))]
    [(b, Term a b)] (Ctxt a b)
  | RelNotWellFormed
  | NTRelRuleRemoval [(Term a b, Term a b)] [(Term a b, Term a b)]
      (RelTRSDisproof a b)
  | RNotSN (TRSDisproof a b);

shows_monom :: forall a. (Showa a) => [(a, Nat)] -> [Char] -> [Char];
shows_monom [(x, p)] =
  (if equal_nat p one_nat then shows_prec Zero_nat x
    else shows_prec Zero_nat x .
           shows_string ['^'] . shows_prec_nat Zero_nat p);
shows_monom ((x, p) : v : va) =
  (if equal_nat p one_nat then shows_prec Zero_nat x
    else shows_prec Zero_nat x .
           shows_string ['^'] . shows_prec_nat Zero_nat p) .
    shows_string ['*'] . shows_monom (v : va);
shows_monom [] = shows_string ['1'];

shows_poly ::
  forall a b.
    (Showa a, Eq b, Poly_carriera b,
      Showa b) => [([(a, Nat)], b)] -> [Char] -> [Char];
shows_poly [] = shows_string ['0'];
shows_poly ((m, c) : p) =
  (if c == one then shows_monom m
    else (if null m then shows_prec Zero_nat c
           else shows_prec Zero_nat c . shows_string ['*'] . shows_monom m)) .
    (if null p then shows_string []
      else shows_string [' ', '+', ' '] . shows_poly p);

poly_split ::
  forall a b.
    (Eq a, Zero b) => [(a, Nat)] -> [([(a, Nat)], b)] -> (b, [([(a, Nat)], b)]);
poly_split m p =
  (case extract (\ (n, _) -> eq_monom m n) p of {
    Nothing -> (zero, p);
    Just (p1, ((_, c), p2)) -> (c, p1 ++ p2);
  });

list_to_rm :: forall a b. (Linorder a) => [(a, b)] -> Rbt a b;
list_to_rm = gen_list_to_map rm_empty rm_update;

rm_isEmpty :: forall a b. (Eq a, Linorder a, Eq b) => Rbt a b -> Bool;
rm_isEmpty m = m == empty;

rito_map_to_sorted_list ::
  forall a b c.
    ((a -> b -> [(a, b)] -> [(a, b)]) -> c -> [(a, b)] -> [(a, b)]) ->
      c -> [(a, b)];
rito_map_to_sorted_list iteratea m =
  iteratea (\ u v -> (\ a -> (u, v) : a)) m [];

itoi_reverse_iterateo ::
  forall a b c d.
    ((a -> Bool) -> (b -> c -> a -> a) -> d -> a -> a) ->
      (b -> c -> a -> a) -> d -> a -> a;
itoi_reverse_iterateo it = it (\ _ -> True);

rm_reverse_iterateoi ::
  forall a b c. (a -> Bool) -> (b -> c -> a -> a) -> Rbta b c -> a -> a;
rm_reverse_iterateoi c f Emptya sigma = sigma;
rm_reverse_iterateoi c f (Branch col l k v r) sigma =
  (if c sigma
    then let {
           sigmaa = rm_reverse_iterateoi c f r sigma;
         } in (if c sigmaa then rm_reverse_iterateoi c f l (f k v sigmaa)
                else sigmaa)
    else sigma);

rm_reverse_iterateoia ::
  forall a b c.
    (Linorder b) => (a -> Bool) -> (b -> c -> a -> a) -> Rbt b c -> a -> a;
rm_reverse_iterateoia c f r = rm_reverse_iterateoi c f (impl_of r);

rm_reverse_iterateo ::
  forall a b c. (Linorder a) => (a -> b -> c -> c) -> Rbt a b -> c -> c;
rm_reverse_iterateo = itoi_reverse_iterateo rm_reverse_iterateoia;

rm_to_list :: forall a b. (Linorder a) => Rbt a b -> [(a, b)];
rm_to_list = rito_map_to_sorted_list rm_reverse_iterateo;

gen_list_to_set_aux :: forall a b. (a -> b -> b) -> b -> [a] -> b;
gen_list_to_set_aux ins accs [] = accs;
gen_list_to_set_aux ins accs (x : l) = gen_list_to_set_aux ins (ins x accs) l;

gen_list_to_set :: forall a b. a -> (b -> a -> a) -> [b] -> a;
gen_list_to_set empt ins = gen_list_to_set_aux ins empt;

list_to_rs :: forall a. (Linorder a) => [a] -> Rbt a ();
list_to_rs = gen_list_to_set rs_empty rs_ins;

rs_isEmpty :: forall a. (Eq a, Linorder a) => Rbt a () -> Bool;
rs_isEmpty = s_isEmpty rm_isEmpty;

rito_set_to_sorted_list ::
  forall a b. ((a -> [a] -> [a]) -> b -> [a] -> [a]) -> b -> [a];
rito_set_to_sorted_list iteratea m = iteratea (\ a b -> a : b) m [];

s_reverse_iterateo ::
  forall a b c.
    (Linorder a) => ((a -> () -> b -> b) -> c -> b -> b) ->
                      (a -> b -> b) -> c -> b -> b;
s_reverse_iterateo iteratea f s sigma_0 = iteratea (\ k _ -> f k) s sigma_0;

rs_reverse_iterateo ::
  forall a b. (Linorder a) => (a -> b -> b) -> Rbt a () -> b -> b;
rs_reverse_iterateo = s_reverse_iterateo rm_reverse_iterateo;

rs_to_list :: forall a. (Linorder a) => Rbt a () -> [a];
rs_to_list = rito_set_to_sorted_list rs_reverse_iterateo;

funas_below_term_list :: forall a b. Term a b -> [(a, Nat)];
funas_below_term_list t = concatMap funas_term_lista (args t);

funas_below_rule_list :: forall a b. (Term a b, Term a b) -> [(a, Nat)];
funas_below_rule_list r =
  funas_below_term_list (fst r) ++ funas_below_term_list (snd r);

funas_below_trs_list ::
  forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_below_trs_list trs = remdups (concatMap funas_below_rule_list trs);

check_sem_lab_proc ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => Sum ([Char] -> [Char]) () ->
                    ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                      Sum ([Char] -> [Char]) ()) ->
                      ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                        [(Term (Lab a b) c, Term (Lab a b) c)] ->
                          Sum ([Char] -> [Char]) ()) ->
                        ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                          [(Term (Lab a b) c, Term (Lab a b) c)] ->
                            Sum ([Char] -> [Char]) ()) ->
                          (Lab a b ->
                            Nat ->
                              Sum b [Lab a b] -> Sum b [Lab a b] -> Bool) ->
                            (Lab a b -> Nat -> Sum b [Lab a b] -> Bool) ->
                              [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                  [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                    [(Term (Lab a b) c, Term (Lab a b) c)] ->
                                      Sum ([Char] -> [Char]) ();
check_sem_lab_proc valid check_decr check_lab check_model_lab lge ls p r lP lR =
  binda valid
    (\ _ ->
      binda (check_decr lR)
        (\ _ ->
          binda (catcha (check_lab lP p)
                  (\ x ->
                    Inl (shows_string
                           ['n', 'o', 't', ' ', 'a', 'l', 'l', ' ', 'l', 'a',
                             'b', 'e', 'l', 'e', 'd', ' ', 'D', 'P', 's', ' ',
                             'p', 'r', 'e', 's', 'e', 'n', 't', ':'] .
                          shows_nl . x)))
            (\ _ ->
              binda (catcha (check_model_lab lR r)
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'l', 'a', 'b', 'e', 'l',
                                 'e', 'd', ' ', 'T', 'R', 'S', ':'] .
                              shows_nl . x)))
                (\ _ ->
                  catcha (check_sl_only_lab_and_decr lge ls r lR)
                    (\ x -> Inl (x . shows_nl))))));

rl_fun_delt :: forall a b. (Eq a) => Maybe a -> [a] -> a -> [a] -> Sum b [a];
rl_fun_delt Nothing uu uv cs = Inr cs;
rl_fun_delt (Just delt) fs f cs =
  Inr (if membera (delt : fs) f then cs else map (\ _ -> delt) cs);

rl_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Maybe (Lab a b) ->
                    [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                      [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                        [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                          [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                            Sum ([Char] -> [Char]) ();
rl_proc delt_opt p r lP lR =
  let {
    fs = filter (\ f -> not (Just f == delt_opt))
           (list_union (funs_trs_list r) (map fst (funas_below_trs_list p)));
  } in binda (check (not (null fs))
               (shows_string
                  ['r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                    'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
                    't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ',
                    'f', 'u', 'n', 'c', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm',
                    'b', 'o', 'l', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 's',
                    'i', 'g', 'n', 'a', 't', 'u', 'r', 'e'] .
                 shows_nl))
         (\ _ ->
           let {
             d = head fs;
           } in check_sem_lab_proc model_check_valid model_check_decr
                  (check_sl_lab_trs (rl_inter_list fs d)
                    (rl_fun_delt delt_opt fs) fs model_cge)
                  (check_sl_model_lab_trs (rl_inter_list fs d)
                    (rl_fun_delt delt_opt fs) fs model_cge)
                  model_lge model_LS p r lP lR);

subst_apply_term :: forall a b c. Term a b -> (b -> Term a c) -> Term a c;
subst_apply_term (Var x) sigma = sigma x;
subst_apply_term (Fun f ss) sigma =
  Fun f (map (\ t -> subst_apply_term t sigma) ss);

data Ta_rule_impl a b = TA_rule_impl b [a] a (Rbt a ());

rq_impl :: forall a b. Ta_rule_impl a b -> Rbt a ();
rq_impl (TA_rule_impl f qsa q qs) = qs;

check_rsteps ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Pos, ((Term a b, Term a b), Term a b))] ->
                      Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_rsteps r prts s u =
  check_steps [] r (map (\ (p, (ra, t)) -> (p, (ra, (False, t)))) prts) s u;

check_wf_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_wf_trs r =
  catcha
    (catcha
      (forallm
        (\ rule ->
          catcha
            (binda
              (check (not (is_var (fst rule)))
                (shows_string
                  ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'l', 'e', 'f',
                    't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', ' ',
                    'i', 'n']))
              (\ _ ->
                catcha
                  (check_subseteq (vars_term_list (snd rule))
                    (vars_term_list (fst rule)))
                  (\ x ->
                    Inl (shows_string
                           ['f', 'r', 'e', 'e', ' ', 'v', 'a', 'r', 'i', 'a',
                             'b', 'l', 'e', ' '] .
                          shows_prec Zero_nat x .
                            shows_string
                              [' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't', '-',
                                'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e',
                                ' ', 'o', 'f']))))
            (\ x ->
              Inl (x . shows_string [' ', 'r', 'u', 'l', 'e', ' '] .
                         shows_rule rule . shows_nl)))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (shows_string
             ['t', 'h', 'e', ' ', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'n', 'o',
               't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e',
               'd'] .
            shows_nl . x));

bounds_type :: Xml -> Sum [Char] Boundstype;
bounds_type =
  xml1element ['t', 'y', 'p', 'e']
    (xml_options
      [(['r', 'o', 'o', 'f'], xml0elements ['r', 'o', 'o', 'f'] Roof),
        (['m', 'a', 't', 'c', 'h'],
          xml0elements ['m', 'a', 't', 'c', 'h'] Match)])
    id;

bounds_bound :: Xml -> Sum [Char] Nat;
bounds_bound = nata ['b', 'o', 'u', 'n', 'd'];

final_states :: Xml -> Sum [Char] [[Char]];
final_states =
  xml_many_elements ['f', 'i', 'n', 'a', 'l', 'S', 't', 'a', 't', 'e', 's']
    state id;

transitions ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml -> Sum [Char] ([Ta_rule [Char] (a, Nat)], [([Char], [Char])]);
transitions xml2name =
  xml_many_elements ['t', 'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n', 's']
    (transition xml2name)
    (\ rls ->
      let {
        (rules, eps) =
          partition (\ a -> (case a of {
                              Inl _ -> True;
                              Inr _ -> False;
                            }))
            rls;
        ruls = map (\ (Inl r) -> r) rules;
        a = map (\ (Inr e) -> e) eps;
      } in (ruls, a));

bounds_tree_automaton ::
  forall a.
    (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Tree_automaton [Char] (a, Nat));
bounds_tree_automaton xml2name =
  xml2elements ['t', 'r', 'e', 'e', 'A', 'u', 't', 'o', 'm', 'a', 't', 'o', 'n']
    final_states (transitions xml2name) (\ f (a, b) -> Tree_Automaton f a b);

bounds_info ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Bounds_info a [Char]);
bounds_info xml2name =
  xml4elements ['b', 'o', 'u', 'n', 'd', 's'] bounds_type bounds_bound
    final_states (bounds_tree_automaton xml2name) Bounds_Info;

id_labeling :: forall a b. Lab a [b] -> [b] -> Sum [b] [Lab a [b]];
id_labeling uu cs = Inl cs;

ordering_constraint_proof ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (Redpair a);
ordering_constraint_proof xml2name =
  xml1element
    ['o', 'r', 'd', 'e', 'r', 'i', 'n', 'g', 'C', 'o', 'n', 's', 't', 'r', 'a',
      'i', 'n', 't', 'P', 'r', 'o', 'o', 'f']
    (redpair xml2name) id;

xml1many2elements_gen ::
  forall a b c d e.
    [Char] ->
      (Xml -> Sum [Char] a) ->
        (a -> Xml -> Sum [Char] b) ->
          (Xml -> Sum [Char] c) ->
            (Xml -> Sum [Char] d) ->
              (a -> [b] -> c -> d -> e) -> Xml -> Sum [Char] e;
xml1many2elements_gen tag p1 p2 p3 p4 f (Xml name atts cs text) =
  let {
    ds = reverse cs;
    l = size_list cs;
  } in (if name == tag && (null atts && (is_none text && less_eq_nat (nat 3) l))
         then binda (p1 (nth cs Zero_nat))
                (\ x ->
                  binda (mapa (p2 x) (tl (takea (minus_nat l (nat 2)) cs)))
                    (\ xs ->
                      binda (p3 (nth ds one_nat))
                        (\ y ->
                          binda (p4 (nth ds Zero_nat))
                            (\ z -> Inr (f x xs y z)))))
         else no_transform tag (Xml name atts cs text));

projected_rseq ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] a) ->
                (a -> Nat) ->
                  Xml ->
                    Sum [Char]
                      ((Term a [Char], Term a [Char]),
                        [(Pos, ((Term a [Char], Term a [Char]),
                                 Term a [Char]))]);
projected_rseq xml2name pia (Xml name atts cs text) =
  let {
    tag = ['p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', 'R', 'e', 'w', 'r', 'i',
            't', 'e', 'S', 'e', 'q', 'u', 'e', 'n', 'c', 'e'];
  } in (if name == tag &&
             (null atts && (is_none text && equal_nat (size_list cs) (nat 2)))
         then binda (rule xml2name (nth cs Zero_nat))
                (\ r -> rseq xml2name pia r (nth cs one_nat))
         else no_transform tag (Xml name atts cs text));

create_proj :: forall a. (Eq a) => ProjL a -> a -> Nat;
create_proj (Projection p) f =
  (case lookupb f p of {
    Nothing -> Zero_nat;
    Just n -> n;
  });

flat_contexts ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] [Ctxt a [Char]];
flat_contexts xml2name =
  xml_many_elements ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 's']
    (ctxt xml2name) id;

uncurry_info ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          (a, ([((a, Nat), [a])],
                ([(Term a [Char], Term a [Char])],
                  [(Term a [Char], Term a [Char])])));
uncurry_info xml2name =
  xml4elements
    ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'I', 'n', 'f', 'o', 'r', 'm', 'a', 't',
      'i', 'o', 'n']
    xml2name
    (xml_many_elements
      ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm', 'b', 'o',
        'l', 's']
      (xml2many_elements
        ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', 'S', 'y', 'm', 'b', 'o',
          'l', 'E', 'n', 't', 'r', 'y']
        xml2name (nata ['a', 'r', 'i', 't', 'y']) xml2name
        (\ f n -> (\ a -> ((f, n), a))))
      id)
    (xml1element ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'R', 'u', 'l', 'e', 's']
      (rules xml2name) id)
    (xml1element ['e', 't', 'a', 'R', 'u', 'l', 'e', 's'] (rules xml2name) id)
    (\ a sml u e -> (a, (sml, (u, e))));

xml2dpProof ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] (Lab a [Nat])) ->
                Xml -> Sum [Char] (DPProof a [Nat] [Char]);
xml2dpProof xml2name x =
  xml1element ['d', 'p', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['p', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['p', 'I', 's', 'E', 'm', 'p', 't', 'y'] PisEmpty),
        (['d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c'],
          xml_many_elements
            ['d', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c']
            (xml2nd_choice ['c', 'o', 'm', 'p', 'o', 'n', 'e', 'n', 't']
              (xml1element ['d', 'p', 's'] (rules xml2name) id)
              ['r', 'e', 'a', 'l', 'S', 'c', 'c'] (xml2dpProof xml2name)
              (\ dps prfOpt -> (prfOpt, dps)))
            DepGraphProc),
        (['r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c'],
          xml3elements ['r', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml2dpProof xml2name) RedPairProc),
        (['r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c'],
          xml4elements
            ['r', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P', 'r', 'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dpProof xml2name) RedPairUrProc),
        (['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o',
           'c'],
          xml4elements
            ['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r',
              'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dpProof xml2name) MonoRedPairProc),
        (['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r', 'P',
           'r', 'o', 'c'],
          xml5elements
            ['m', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'U', 'r',
              'P', 'r', 'o', 'c']
            (ordering_constraint_proof xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
              (rules xml2name) id)
            (xml2dpProof xml2name) MonoRedPairUrProc),
        (['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c'],
          xml4elements ['s', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
            (sl_inter xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dpProof xml2name)
            (\ a ->
              (case a of {
                Inl (i, True) -> SemlabProc i (Inl id_labeling);
                Inl (i, False) -> SemlabProc i (Inr ([], True));
                Inr aa -> RootlabProc aa;
              }))),
        (['s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c'],
          xml1many2elements_gen
            ['s', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c']
            (proj xml2name) (\ pia -> projected_rseq xml2name (create_proj pia))
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml2dpProof xml2name) SubtermCriterionProc),
        (['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
           's', 'u', 'r', 'e', 'P', 'r', 'o', 'c'],
          xml5elements
            ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l',
              'o', 's', 'u', 'r', 'e', 'P', 'r', 'o', 'c']
            (xml1element ['f', 'r', 'e', 's', 'h', 'S', 'y', 'm', 'b', 'o', 'l']
              xml2name id)
            (flat_contexts xml2name)
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dpProof xml2name) FccProc),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c'],
          xml_choice ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
            [xml4elements
               ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
               (uncurry_info xml2name)
               (xml1element ['d', 'p', 's'] (rules xml2name) id)
               (xml1element ['t', 'r', 's'] (rules xml2name) id)
               (xml2dpProof xml2name) (UncurryProc Nothing),
              xml5elements
                ['u', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
                (nata ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'v', 'e',
                        'T', 'o', 'p'])
                (uncurry_info xml2name)
                (xml1element ['d', 'p', 's'] (rules xml2name) id)
                (xml1element ['t', 'r', 's'] (rules xml2name) id)
                (xml2dpProof xml2name) (\ n -> UncurryProc (Just n))]),
        (['u', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c'],
          xml3elements ['u', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2dpProof xml2name) UnlabProc),
        (['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o', 'c'],
          xml1many_elements
            ['s', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'P', 'r', 'o',
              'c']
            (xml_options
              [(['s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i', 't', 'e',
                  'r', 'i', 'o', 'n'],
                 xml0elements
                   ['s', 'u', 'b', 't', 'e', 'r', 'm', 'C', 'r', 'i', 't', 'e',
                     'r', 'i', 'o', 'n']
                   Nothing),
                (['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P', 'a', 'i',
                   'r'],
                  xml2elements
                    ['r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n', 'P', 'a', 'i',
                      'r']
                    (ordering_constraint_proof xml2name)
                    (xml1element
                      ['u', 's', 'a', 'b', 'l', 'e', 'R', 'u', 'l', 'e', 's']
                      (rules xml2name) id)
                    (\ redp ur -> Just (redp, ur)))])
            (scg xml2name)
            (\ a ->
              (case a of {
                Nothing -> SizeChangeSubTerm;
                Just redp_ur -> SizeChangeRedPair (fst redp_ur) (snd redp_ur);
              })))])
    id x;

instance (Eq a, Order a, Eq b, Order b) => Preorder (Lab a b) where {
};

instance (Eq a, Order a, Eq b, Order b) => Order (Lab a b) where {
};

unsharp :: forall a b. Lab a b -> Lab a b;
unsharp (Sharp f) = f;
unsharp (Lab v va) = Lab v va;
unsharp (FunLab v va) = FunLab v va;
unsharp (UnLab v) = UnLab v;

data RelTRSProof a b c =
  RelRuleRemoval (Redpair (Lab a b)) [(Term (Lab a b) c, Term (Lab a b) c)]
    [(Term (Lab a b) c, Term (Lab a b) c)] (RelTRSProof a b c)
  | RelSisEmpty (TRSProof a b c)
  | RelUnlab [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (RelTRSProof a b c)
  | RelSemlab (Sl_inter a b) (Sum (Lab a b -> [Nat] -> Sum b [Lab a b]) c)
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (RelTRSProof a b c)
  | RelRootlab [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (RelTRSProof a b c)
  | RelUncurry
      (Lab a b,
        ([((Lab a b, Nat), [Lab a b])],
          ([(Term (Lab a b) c, Term (Lab a b) c)],
            [(Term (Lab a b) c, Term (Lab a b) c)])))
      [(Term (Lab a b) c, Term (Lab a b) c)]
      [(Term (Lab a b) c, Term (Lab a b) c)] (RelTRSProof a b c)
  | RelRisEmpty;

data CertificationProblem a b c =
  TRSProof [(Term a c, Term a c)] (TRSProof a b c)
  | TRSDisproof [(Term a c, Term a c)] (TRSDisproof (Lab a b) c)
  | RelTRSProof ([(Term a c, Term a c)], [(Term a c, Term a c)])
      (RelTRSProof a b c)
  | RelTRSDisproof ([(Term a c, Term a c)], [(Term a c, Term a c)])
      (RelTRSDisproof (Lab a b) c);

lpo_strict_unbounded ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat -> Nat) ->
                    (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
lpo_strict_unbounded pr =
  (\ (s, t) ->
    check (fst (lpo_unbounded pr s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', 'L', 'P', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

lpo_nstrict_unbounded ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat -> Nat) ->
                    (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
lpo_nstrict_unbounded pr =
  (\ (s, t) ->
    check (snd (lpo_unbounded pr s t))
      (shows_string
         ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i', 'e',
           'n', 't', ' '] .
        shows_prec_term Zero_nat s .
          shows_string [' ', '>', '=', 'L', 'P', 'O', ' '] .
            shows_prec_term Zero_nat t . shows_nl));

create_LPO_redpair ::
  forall a b c.
    (Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([((a, Nat), Nat)] -> b -> Nat -> Nat) ->
                    [((a, Nat), Nat)] ->
                      (Sum ([Char] -> [Char]) (),
                        ((Term b c, Term b c) -> Sum ([Char] -> [Char]) (),
                          ((Term b c, Term b c) -> Sum ([Char] -> [Char]) (),
                            (b -> Nat -> Nat -> Bool,
                              (Sum ([Char] -> [Char]) (), [Char] -> [Char])))));
create_LPO_redpair lpo_repr_to_pr pr =
  let {
    p = lpo_repr_to_pr pr;
  } in (Inr (),
         (lpo_strict_unbounded p,
           (lpo_nstrict_unbounded p, (full_af, (Inr (), shows_lpo_repr pr)))));

check_left_linear_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_left_linear_trs trs =
  catcha
    (catcha
      (forallm (\ x -> (if linear_term (fst x) then Inr () else Inl x)) trs)
      (\ x -> Inl (snd x)))
    (\ _ ->
      Inl (shows_prec_list Zero_nat trs .
            shows_nl .
              shows_string
                ['i', 's', ' ', 'n', 'o', 't', ' ', 'l', 'e', 'f', 't', '-',
                  'l', 'i', 'n', 'e', 'a', 'r'] .
                shows_nl));

check_no_var ::
  forall a b. (Showa a, Showa b) => Term a b -> Sum ([Char] -> [Char]) ();
check_no_var t =
  check (not (is_var t))
    (shows_string
       ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'f', 'o', 'u', 'n', 'd'] .
      shows_nl);

check_no_defined_root_defs ::
  forall a b.
    (Eq a, Showa a,
      Showa b) => [(a, Nat)] -> Term a b -> Sum ([Char] -> [Char]) ();
check_no_defined_root_defs f t =
  check (not (membera f (the (root t), num_args t)))
    (shows_string ['t', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' '] .
      shows_prec_term Zero_nat t .
        shows_string [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd']);

block_term :: forall a b. a -> Term a b -> Term a b;
block_term f (Var x) = Var x;
block_term f (Fun g ts) = Fun g (map (\ t -> Fun f [t]) ts);

block_rule :: forall a b. a -> (Term a b, Term a b) -> (Term a b, Term a b);
block_rule f r = (block_term f (fst r), block_term f (snd r));

check_superset_of_blocked ::
  forall a.
    (Eq a,
      Showa a) => a -> [(Term a [Char], Term a [Char])] ->
                         [(Term a [Char], Term a [Char])] ->
                           Sum ([Char] -> [Char]) ();
check_superset_of_blocked f ra r =
  catcha
    (catcha
      (catcha
        (forallm
          (\ x -> (if membera ra (block_rule f x) then Inr () else Inl x)) r)
        (\ x -> Inl (snd x)))
      (\ x ->
        Inl (shows_string ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
              shows_rule (block_rule f x) .
                shows_string
                  [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'] .
                  shows_nl)))
    (\ x ->
      Inl (shows_trs r .
            shows_string
              ['i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 's',
                'e', 't', ' ', 'o', 'f'] .
              shows_trs ra . x . shows_nl));

fcc_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Nat ->
                    [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                      [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                        Lab a b ->
                          [Ctxt (Lab a b) [Char]] ->
                            [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                              [(Term (Lab a b) [Char],
                                 Term (Lab a b) [Char])] ->
                                Sum ([Char] -> [Char]) ();
fcc_proc n pa ra f fcs p r =
  let {
    fa = (f, Suc Zero_nat);
    cf = More f [] Hole [];
    fcsa = cf : fcs;
    vs = vars_trs_list ra;
    fs = list_union (funas_below_trs_list pa) (funas_trs_list ra);
    fas = fa : fs;
    unlabn = funpow n unlab;
    ds = defs_list (map_trs_list unlabn ra);
  } in binda (check (not (membera ds (unlabn f, one_nat)))
               (shows_prec_lab Zero_nat (unlabn f) .
                 shows_string
                   [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'r', 'e', 's',
                     'h'] .
                   shows_nl))
         (\ _ ->
           binda (check_left_linear_trs ra)
             (\ _ ->
               binda (check_wf_trs ra)
                 (\ _ ->
                   binda (catcha
                           (forallm
                             (\ rb ->
                               binda (check_no_var (fst rb))
                                 (\ _ ->
                                   binda (check_no_var (snd rb))
                                     (\ _ ->
                                       check_no_defined_root_defs ds
 (map_term unlabn (snd rb)))))
                             pa)
                           (\ x -> Inl (snd x)))
                     (\ _ ->
                       binda (catcha (forallm (check_flat_ctxt vs) fcsa)
                               (\ x -> Inl (snd x)))
                         (\ _ ->
                           binda (catcha
                                   (forallm (check_is_flat_ctxt vs fas) fcsa)
                                   (\ x -> Inl (snd x)))
                             (\ _ ->
                               binda (catcha
                                       (forallm (check_flat_ctxt_complete fcsa)
 fas)
                                       (\ x -> Inl (snd x)))
                                 (\ _ ->
                                   binda (catcha
   (forallm (check_rule_preserving fcsa r) ra) (\ x -> Inl (snd x)))
                                     (\ _ ->
                                       binda
 (catcha (forallm (check_rule_reflecting fcsa ra) r) (\ x -> Inl (snd x)))
 (\ _ -> check_superset_of_blocked f p pa)))))))));

check_supteq ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b -> Term a b -> Sum ([Char] -> [Char]) ();
check_supteq s t =
  check (member (s, t) supteq)
    (shows_prec_term Zero_nat t .
      shows_string
        [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ', 's', 'u', 'b', 't',
          'e', 'r', 'm', ' ', 'o', 'f', ' '] .
        shows_prec_term Zero_nat s);

rqs_impl :: forall a b. Ta_rule_impl a b -> [a];
rqs_impl (TA_rule_impl f qsa q qs) = qsa;

uncurry_eta_split ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 [(Term a b, Term a b)] ->
                   ([(Term a b, Term a b)], [(Term a b, Term a b)]);
uncurry_eta_split eboth rtest =
  let {
    test =
      (\ (l, r) ->
        less_nat Zero_nat (num_args l) &&
          (less_nat Zero_nat (num_args r) &&
            any (eq_rule_mod_vars (head (args l), head (args r))) rtest));
  } in partition test eboth;

uncurry_rel_tt ::
  forall a.
    (Eq a,
      Showa a) => (a, ([((a, Nat), [a])],
                        ([(Term a [Char], Term a [Char])],
                          [(Term a [Char], Term a [Char])]))) ->
                    [(Term a [Char], Term a [Char])] ->
                      [(Term a [Char], Term a [Char])] ->
                        [(Term a [Char], Term a [Char])] ->
                          [(Term a [Char], Term a [Char])] ->
                            Sum ([Char] -> [Char]) ();
uncurry_rel_tt (a, (sml, (u, eb))) ra rwa r rw =
  let {
    (e, ew) = uncurry_eta_split eb ra;
    sm = sig_list_to_sig_map a sml (\ _ f _ -> f);
    r_eta = e ++ ra;
    rw_eta = ew ++ rwa;
    rb_eta = r_eta ++ rw_eta;
  } in binda (catcha
               (forallm
                 (\ (l, _) ->
                   check (hvf_term a l)
                     (shows_prec_list Zero_nat
                        ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a', 'b',
                          'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                       shows_term l .
                         shows_prec_list Zero_nat
                           [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                             'e', 'd']))
                 r_eta)
               (\ x -> Inl (snd x)))
         (\ _ ->
           binda (catcha
                   (forallm
                     (\ (l, _) ->
                       check (hvf_term a l)
                         (shows_prec_list Zero_nat
                            ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r', 'i', 'a',
                              'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's',
                              ' '] .
                           shows_term l .
                             shows_prec_list Zero_nat
                               [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o',
                                 'w', 'e', 'd']))
                     rw_eta)
                   (\ x -> Inl (snd x)))
             (\ _ ->
               binda (eta_closed_rules a sm r_eta r_eta)
                 (\ _ ->
                   binda (eta_closed_rules a sm rb_eta rb_eta)
                     (\ _ ->
                       binda (catcha
                               (forallm
                                 (\ (l, rb) ->
                                   check (not (is_var l))
                                     (shows_prec_list Zero_nat
['l', 'h', 's', ' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't', ' ', 'b', 'e', ' ',
  'a', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'r',
  'u', 'l', 'e', ' '] .
                                       shows_rule (l, rb)))
                                 rw_eta)
                               (\ x -> Inl (snd x)))
                         (\ _ ->
                           binda (catcha
                                   (check_subseteq (uncurry_rules a sm r_eta) r)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
      ' '] .
   shows_rule x .
     shows_prec_list Zero_nat
       [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                             (\ _ ->
                               binda (catcha
                                       (check_subseteq
 (uncurry_rules a sm rw_eta) rw)
                                       (\ x ->
 Inl (shows_prec_list Zero_nat
        ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
          ' '] .
       shows_rule x .
         shows_prec_list Zero_nat
           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                 (\ _ ->
                                   catcha
                                     (check_CS_subseteq
                                       (uncurry_of_sig_list a sml sm) rw)
                                     (\ x ->
                                       Inl
 (shows_prec_list Zero_nat
    ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e', ' '] .
   shows_rule x .
     shows_prec_list Zero_nat
       [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))))))));

xml2trsInput ::
  forall a.
    (Xml -> Sum [Char] a) ->
      Xml ->
        Sum [Char]
          ([(Term a [Char], Term a [Char])],
            Maybe [(Term a [Char], Term a [Char])]);
xml2trsInput xml2name =
  xml1element ['i', 'n', 'p', 'u', 't']
    (xml1to2elements ['t', 'r', 's', 'I', 'n', 'p', 'u', 't']
      (xml1element ['t', 'r', 's'] (rules xml2name) id)
      (xml1element
        ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's']
        (rules xml2name) id)
      (\ a b -> (a, b)))
    id;

xml2trsProof ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] (Lab a [Nat])) ->
                Xml -> Sum [Char] (TRSProof a [Nat] [Char]);
xml2trsProof xml2name x =
  xml1element
    ['t', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P',
      'r', 'o', 'o', 'f']
    (xml_options
      [(['r', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['r', 'I', 's', 'E', 'm', 'p', 't', 'y'] RisEmpty),
        (['d', 'p', 'T', 'r', 'a', 'n', 's'],
          xml3elements ['d', 'p', 'T', 'r', 'a', 'n', 's']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (bool ['m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                    's'])
            (xml2dpProof xml2name) (\ dps _ -> DPTrans dps)),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml3elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (ordering_constraint_proof xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name) RuleRemoval),
        (['u', 'n', 'l', 'a', 'b'],
          xml2elements ['u', 'n', 'l', 'a', 'b']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name) Unlab),
        (['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'],
          xml2elements
            ['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a',
              'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name) StringReversal),
        (['s', 'e', 'm', 'l', 'a', 'b'],
          xml3elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_inter xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name)
            (\ a ->
              (case a of {
                Inl (i, True) -> Semlab i (Inl id_labeling);
                Inl (i, False) -> Semlab i (Inr []);
                Inr _ -> Rootlab;
              }))),
        (['b', 'o', 'u', 'n', 'd', 's'],
          (\ xml -> binda (bounds_info xml2name xml) (Inr . Bounds))),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y'],
          xml3elements ['u', 'n', 'c', 'u', 'r', 'r', 'y']
            (uncurry_info xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name) Uncurry),
        (['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l', 'o',
           's', 'u', 'r', 'e'],
          xml3elements
            ['f', 'l', 'a', 't', 'C', 'o', 'n', 't', 'e', 'x', 't', 'C', 'l',
              'o', 's', 'u', 'r', 'e']
            (flat_contexts xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsProof xml2name) Fcc)])
    id x;

instance Preorder Char where {
};

instance Order Char where {
};

data L_poly a b = LPoly b [(a, b)];

sym_collect :: forall a b. (Term a b -> Bool) -> Term a b -> [a];
sym_collect p (Var x) = [];
sym_collect p (Fun f ts) =
  (if p (Fun f ts) then [f] else []) ++ concatMap (sym_collect p) ts;

fresh_strings_list :: [Char] -> Nat -> [[Char]] -> Nat -> [[Char]];
fresh_strings_list name offset used n =
  takea n
    (filter (\ s -> not (membera (remdups used) s))
      (map (\ i -> name ++ shows_prec_nat Zero_nat (plus_nat i offset) [])
        (upt Zero_nat (plus_nat n (size_list (remdups used))))));

subst_of :: forall a b. (Eq a) => [(a, Term b a)] -> a -> Term b a;
subst_of xts =
  (\ x -> (case lookupb x xts of {
            Nothing -> Var x;
            Just t -> t;
          }));

rm_update_dj :: forall a b. (Linorder a) => a -> b -> Rbt a b -> Rbt a b;
rm_update_dj = rm_update;

check_sem_lab_rel_tt ::
  forall a b c.
    (Showa a, Showa b,
      Showa c) => Sum ([Char] -> [Char]) () ->
                    ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                      Sum ([Char] -> [Char]) ()) ->
                      ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                        [(Term (Lab a b) c, Term (Lab a b) c)] ->
                          Sum ([Char] -> [Char]) ()) ->
                        ([(Term (Lab a b) c, Term (Lab a b) c)] ->
                          [(Term (Lab a b) c, Term (Lab a b) c)] ->
                            Sum ([Char] -> [Char]) ()) ->
                          (Lab a b -> Nat -> Sum b [Lab a b] -> Bool) ->
                            ([(Term (Lab a b) c, Term (Lab a b) c)],
                              [(Term (Lab a b) c, Term (Lab a b) c)]) ->
                              ([(Term (Lab a b) c, Term (Lab a b) c)],
                                [(Term (Lab a b) c, Term (Lab a b) c)]) ->
                                Sum ([Char] -> [Char]) ();
check_sem_lab_rel_tt valid check_decr check_lab check_model_lab ls rs lRS =
  binda valid
    (\ _ ->
      catcha
        (binda (check_decr (snd lRS))
          (\ _ ->
            binda (check_model_lab (fst lRS) (fst rs))
              (\ _ -> check_model_lab (snd lRS) (snd rs))))
        (\ x ->
          Inl (shows_string
                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                   ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S',
                   ':'] .
                shows_nl . x)));

rl_rel_tt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ([(Term (Lab a b) [Char], Term (Lab a b) [Char])],
                    [(Term (Lab a b) [Char], Term (Lab a b) [Char])]) ->
                    ([(Term (Lab a b) [Char], Term (Lab a b) [Char])],
                      [(Term (Lab a b) [Char], Term (Lab a b) [Char])]) ->
                      Sum ([Char] -> [Char]) ();
rl_rel_tt rs lRS =
  let {
    fs = funs_trs_list (fst rs ++ snd rs);
  } in binda (check (not (null fs))
               (shows_string
                  ['r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n',
                    'g', ' ', 'r', 'e', 'q', 'u', 'i', 'r', 'e', 's', ' ', 'a',
                    't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ',
                    'f', 'u', 'n', 'c', 't', 'i', 'o', 'n', ' ', 's', 'y', 'm',
                    'b', 'o', 'l', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ', 's',
                    'i', 'g', 'n', 'a', 't', 'u', 'r', 'e'] .
                 shows_nl))
         (\ _ ->
           let {
             d = head fs;
           } in check_sem_lab_rel_tt model_check_valid model_check_decr
                  (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs model_cge)
                  (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs
                    model_cge)
                  model_LS rs lRS);

rsym_impl :: forall a b. Ta_rule_impl a b -> (b, Nat);
rsym_impl (TA_rule_impl f qsa q qs) = (f, size_list qsa);

computeRuleMapH ::
  forall a b.
    (Eq a) => [(Term a b, Term a b)] ->
                Maybe [((a, Nat), [(Term a b, Term a b)])];
computeRuleMapH [] = Just [];
computeRuleMapH ((Fun f ts, r) : rules) =
  let {
    n = size_list ts;
  } in (case computeRuleMapH rules of {
         Nothing -> Nothing;
         Just rm ->
           (case extract (\ (fa, _) -> fa == (f, n)) rm of {
             Nothing -> Just (((f, n), [(Fun f ts, r)]) : rm);
             Just (bef, ((fa, rls), aft)) ->
               Just ((fa, (Fun f ts, r) : rls) : bef ++ aft);
           });
       });
computeRuleMapH ((Var uu, uv) : rules) = Nothing;

computeRuleMap ::
  forall a b.
    (Eq a) => [(Term a b, Term a b)] ->
                Maybe ((a, Nat) -> [(Term a b, Term a b)]);
computeRuleMap rls =
  (case computeRuleMapH rls of {
    Nothing -> Nothing;
    Just rm ->
      Just (\ f ->
             (case lookupb f rm of {
               Nothing -> [];
               Just rlsa -> rlsa;
             }));
  });

apply_af_entry :: forall a b. a -> Af_entry -> [Term a b] -> Term a b;
apply_af_entry uu (Collapse i) ts = nth ts i;
apply_af_entry f (AFList is) ts = Fun f (map (nth ts) is);

afs_term ::
  forall a b. (a -> Nat -> Af_entry) -> Term a b -> Term (Filtered a) b;
afs_term pia (Fun f ts) =
  let {
    l = size_list ts;
  } in apply_af_entry (FPair f l) (pia f l) (map (afs_term pia) ts);
afs_term pia (Var x) = Var x;

search_len :: [Char] -> [Char] -> Nat -> Maybe ([Char], [Char]);
search_len t s n =
  (if takea n t == s then Just ([], dropa n t)
    else (case t of {
           [] -> Nothing;
           h : tt ->
             (case search_len tt s n of {
               Nothing -> Nothing;
               Just (bef, aft) -> Just (h : bef, aft);
             });
         }));

reverse_trs :: forall a b. [(Term a b, Term a b)] -> [(Term a b, Term a b)];
reverse_trs r = map (\ lr -> (snd lr, fst lr)) r;

check_graph_decomp ::
  forall a.
    (a -> [Char] -> [Char]) ->
      ((a, a) -> Bool) -> [(Bool, [a])] -> Sum ([Char] -> [Char]) ();
check_graph_decomp ss g rcs =
  binda (check_pairwise (check_edges ss g) (map snd rcs))
    (\ _ ->
      catcha
        (forallm (\ c -> check_edges ss g c c)
          (map_filter (\ x -> (if not (fst x) then Just (snd x) else Nothing))
            rcs))
        (\ x -> Inl (snd x)));

add_varI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> b -> a -> [(b, a)] -> [(b, a)];
add_varI uu uv x a [] = [(x, a)];
add_varI ze pl x a ((y, b) : vas) =
  (if x == y then let {
                    s = pl a b;
                  } in (if s == ze then vas else (x, s) : vas)
    else (y, b) : add_varI ze pl x a vas);

sum_pvarsI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> [(b, a)] -> [(b, a)] -> [(b, a)];
sum_pvarsI uu uv [] vbs = vbs;
sum_pvarsI ze pl ((x, a) : vas) vbs =
  (if a == ze then sum_pvarsI ze pl vas vbs
    else sum_pvarsI ze pl vas (add_varI ze pl x a vbs));

sum_lpolyI ::
  forall a b.
    (Eq a,
      Eq b) => a -> (a -> a -> a) -> L_poly b a -> L_poly b a -> L_poly b a;
sum_lpolyI ze pl (LPoly a vas) (LPoly b vbs) =
  LPoly (pl a b) (sum_pvarsI ze pl vas vbs);

list_sumI :: forall a. a -> (a -> a -> a) -> [a] -> a;
list_sumI ze pl [] = ze;
list_sumI ze pl (x : xs) = pl x (list_sumI ze pl xs);

mul_pvarsI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> [(b, a)] -> [(b, a)];
mul_pvarsI uu uv a [] = [];
mul_pvarsI ze ti a ((x, b) : vas) =
  let {
    p = ti a b;
    res = mul_pvarsI ze ti a vas;
  } in (if p == ze then res else (x, p) : res);

mul_lpolyI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> L_poly b a -> L_poly b a;
mul_lpolyI ze ti a (LPoly b vas) = LPoly (ti a b) (mul_pvarsI ze ti a vas);

pleftI ::
  forall a b c.
    (Eq a,
      Eq c) => a -> a -> (a -> a -> a) ->
                           (a -> a -> a) ->
                             (a -> a) ->
                               (b -> Nat -> (a, [a])) -> Term b c -> L_poly c a;
pleftI ze on pl ti mz pia (Var x) = LPoly ze [(x, on)];
pleftI ze on pl ti mz pia (Fun f ts) =
  let {
    (c, asa) = pia f (size_list ts);
  } in (case sum_lpolyI ze pl (LPoly c [])
               (list_sumI (LPoly ze []) (sum_lpolyI ze pl)
                 (map (\ at ->
                        mul_lpolyI ze ti (fst at)
                          (pleftI ze on pl ti mz pia (snd at)))
                   (zip asa ts)))
         of {
         LPoly d [] -> LPoly (mz d) [];
         LPoly d (ab : lista) -> LPoly d (ab : lista);
       });

instance (Eq a, Linorder a) => Linorder [a] where {
};

check_DPs ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> a) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] -> Sum (Term a b, Term a b) ();
check_DPs unshp r p =
  let {
    d = defs_list r;
  } in catcha
         (forallm
           (\ x ->
             (if let {
                   (ll, rr) = x;
                 } in not (is_var rr) &&
                        (not (membera d (the (root rr), num_args rr)) &&
                          any (\ (l, ra) ->
                                equal_term l (sharp_term unshp ll) &&
                                  member (ra, sharp_term unshp rr) supteq)
                            r)
               then Inr () else Inl x))
           p)
         (\ x -> Inl (snd x));

check_poly_ge ::
  forall a b.
    (Eq a,
      Ordered_semiring_0 b) => [([(a, Nat)], b)] -> [([(a, Nat)], b)] -> Bool;
check_poly_ge [] q = all (\ (_, a) -> ge zero a) q;
check_poly_ge ((m, c) : p) q =
  (case extract (\ nd -> eq_monom (fst nd) m) q of {
    Nothing -> ge c zero && check_poly_ge p q;
    Just (q1, ((_, d), q2)) -> ge c d && check_poly_ge p (q1 ++ q2);
  });

check_poly_gt ::
  forall a b.
    (Ordered_semiring_0 a,
      Eq b) => (a -> a -> Bool) ->
                 [([(b, Nat)], a)] -> [([(b, Nat)], a)] -> Bool;
check_poly_gt gt p q =
  let {
    (a1, p1) = poly_split [] p;
    (b1, q1) = poly_split [] q;
  } in gt a1 b1 && check_poly_ge p1 q1;

rev_term :: forall a b. Ctxt a b -> Term a b -> Term a b;
rev_term c (Var x) = ctxt_apply c (Var x);
rev_term c (Fun f [t]) = rev_term (More f [] c []) t;

rev_rule :: forall a b. (Term a b, Term a b) -> (Term a b, Term a b);
rev_rule (l, r) = (rev_term Hole l, rev_term Hole r);

rp_non_strict ::
  forall a b.
    (Sum ([Char] -> [Char]) (),
      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
          (a -> Nat -> Nat -> Bool,
            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
      (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
rp_non_strict (v, (s, (n, (a, (m, d))))) = n;

data Ta_impl_ext a b c =
  Ta_impl_ext (Rbt a ()) (Rbt (b, Nat) [Ta_rule_impl a b]) [a] (Rbt a ())
    (a -> Rbt a ()) c;

ta_rqs_set :: forall a b c. Ta_impl_ext a b c -> Rbt a ();
ta_rqs_set
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    more)
  = ta_rqs_set;

check_no_defined_root ::
  forall a b.
    (Showa a,
      Showa b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                    Term a b -> Sum ([Char] -> [Char]) ();
check_no_defined_root rm t =
  check (null (rm (the (root t), num_args t)))
    (shows_string ['t', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'o', 'f', ' '] .
      shows_prec_term Zero_nat t .
        shows_string [' ', 'i', 's', ' ', 'd', 'e', 'f', 'i', 'n', 'e', 'd']);

uncurry_of_top_sig_list ::
  forall a.
    a -> Nat ->
           [((a, Nat), [a])] ->
             (a -> Nat -> [a]) -> [(Term a [Char], Term a [Char])];
uncurry_of_top_sig_list a m sml sm =
  concatMap
    (\ (b, c) ->
      let {
        (f, n) = b;
      } in (\ _ ->
             let {
               g = get_symbol sm f n;
             } in map (\ i ->
                        (Fun a (generate_f_xs (g i) (plus_nat n i) :
                                 map (\ ia -> Var (generate_var ia))
                                   (upt (plus_nat n i)
                                     (plus_nat (plus_nat n i)
                                       (minus_nat m one_nat)))),
                          generate_f_xs (g (Suc i))
                            (plus_nat (plus_nat n i) (minus_nat m one_nat))))
                    (upt Zero_nat (aarity sm f n)))
        c)
    sml;

eta_closed_top_rules ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> Nat ->
                         (a -> Nat -> [a]) ->
                           [(Term a b, Term a b)] ->
                             [(Term a b, Term a b)] ->
                               Sum ([Char] -> [Char]) ();
eta_closed_top_rules a n sm r p =
  catcha
    (forallm
      (\ (l, ra) ->
        (case l of {
          Var _ -> Inr ();
          Fun ff ls ->
            check (equal_nat (aarity sm ff (size_list ls)) Zero_nat ||
                    any (\ (lll, rrr) ->
                          (case (lll, rrr) of {
                            (Var _, b) -> False;
                            (Fun f [], b) -> False;
                            (Fun f (ll : yy), Var _) -> False;
                            (Fun f (ll : yy), Fun g []) -> False;
                            (Fun f (ll : yy), Fun g (rr : zz)) ->
                              f == a &&
                                (g == a &&
                                  (zz == yy &&
                                    (equal_nat (size_list yy)
                                       (minus_nat n one_nat) &&
                                      (distinct yy &&
(all is_var yy &&
  (null (list_inter (map the_var yy) (vars_rule_lista (ll, rr))) &&
    instance_rule (l, ra) (ll, rr)))))));
                          }))
                      p)
              (shows_prec_list Zero_nat
                 ['e', 't', 'a', ' ', 'e', 'x', 'p', 'a', 'n', 's', 'i', 'o',
                   'n', ' ', 'o', 'f', ' '] .
                shows_rule (l, ra) .
                  shows_prec_list Zero_nat
                    [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);
        }))
      r)
    (\ x -> Inl (snd x));

uncurry_top_rules ::
  forall a b.
    (Eq a) => a -> Nat ->
                     (a -> Nat -> [a]) ->
                       [(Term a b, Term a b)] -> [(Term a b, Term a b)];
uncurry_top_rules a n sm =
  map (\ (l, r) -> (uncurry_top a n sm l, uncurry_top a n sm r));

uncurry_top_proc ::
  forall a.
    (Eq a,
      Showa a) => (a, ([((a, Nat), [a])],
                        ([(Term a [Char], Term a [Char])],
                          [(Term a [Char], Term a [Char])]))) ->
                    Nat ->
                      ([((a, Nat), [a])] -> a -> Nat -> a) ->
                        (a -> Nat ->
                                [((a, Nat), [a])] ->
                                  Sum ([Char] -> [Char]) ()) ->
                          [(Term a [Char], Term a [Char])] ->
                            [(Term a [Char], Term a [Char])] ->
                              [(Term a [Char], Term a [Char])] ->
                                [(Term a [Char], Term a [Char])] ->
                                  Maybe ((a, Nat) ->
  [(Term a [Char], Term a [Char])]) ->
                                    Sum ([Char] -> [Char]) ();
uncurry_top_proc (a, (sml, (uu, e))) n fmapa check_inj pa ra p r (Just rm) =
  binda (check (not (equal_nat n Zero_nat))
          (shows_prec_list Zero_nat
            ['t', 'h', 'e', ' ', 'a', 'r', 'i', 't', 'y', ' ', 'o', 'f', ' ',
              't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd',
              ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'm', 'u', 's', 't', ' ',
              'b', 'e', ' ', 'a', 't', ' ', 'l', 'e', 'a', 's', 't', ' ', '1']))
    (\ _ ->
      let {
        sm = sig_list_to_sig_map a sml fmapa;
        p_eta = e ++ pa;
        u = map_rules_wa (\ f na -> get_symbol sm f na Zero_nat) ra;
      } in binda (check_inj a n sml)
             (\ _ ->
               binda (catcha
                       (forallm
                         (\ (l, rb) ->
                           binda (check (hvf_top a n l)
                                   (shows_prec_list Zero_nat
                                      ['h', 'e', 'a', 'd', ' ', 'v', 'a', 'r',
'i', 'a', 'b', 'l', 'e', ' ', 'i', 'n', ' ', 'l', 'h', 's', ' '] .
                                     shows_term l .
                                       shows_prec_list Zero_nat
 [' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd']))
                             (\ _ ->
                               binda (check_no_var rb)
                                 (\ _ -> check_no_defined_root rm rb)))
                         p_eta)
                       (\ x -> Inl (snd x)))
                 (\ _ ->
                   binda (check (null (rm (a, n)))
                           (shows_prec_list Zero_nat
                              ['a', 'p', 'p', 'l', 'i', 'c', 'a', 't', 'i', 'o',
                                'n', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' '] .
                             shows_prec Zero_nat a .
                               shows_prec_list Zero_nat
                                 [' ', 'm', 'u', 's', 't', ' ', 'n', 'o', 't',
                                   ' ', 'b', 'e', ' ', 'd', 'e', 'f', 'i', 'n',
                                   'e', 'd', ' ', 'i', 'n', ' ', 'R']))
                     (\ _ ->
                       binda (if any (\ (_, rb) ->
                                       the (root rb) == a &&
 (equal_nat (num_args rb) n && equal_hterm (capRM2 rm (head (args rb))) HHole))
                                   pa
                               then binda (catcha
    (check_CS_subseteq (uncurry_of_top_sig_list a n sml sm) p)
    (\ x ->
      Inl (shows_prec_list Zero_nat
             ['u', 'n', 'c', 'u', 'r', 'r', 'y', ' ', 'r', 'u', 'l', 'e', ' '] .
            shows_rule x .
              shows_prec_list Zero_nat
                [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g', ' ',
                  'i', 'n', ' ', 'P'])))
                                      (\ _ ->
eta_closed_top_rules a n sm ra p_eta)
                               else Inr ())
                         (\ _ ->
                           binda (catcha
                                   (check_subseteq
                                     (uncurry_top_rules a n sm p_eta) p)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'p', 'a', 'i', 'r',
      ' '] .
   shows_rule x .
     shows_prec_list Zero_nat
       [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                             (\ _ ->
                               binda (catcha (check_subseteq u r)
                                       (\ x ->
 Inl (shows_prec_list Zero_nat
        ['u', 'n', 'c', 'u', 'r', 'r', 'i', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
          ' '] .
       shows_rule x .
         shows_prec_list Zero_nat
           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])))
                                 (\ _ ->
                                   catcha (check_subseteq r u)
                                     (\ x ->
                                       Inl
 (shows_prec_list Zero_nat
    ['a', 'd', 'd', 'i', 't', 'i', 'o', 'n', 'a', 'l', ' ', 'r', 'u', 'l', 'e',
      ' '] .
   shows_rule x .
     shows_prec_list Zero_nat
       [' ', 'i', 'n', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a',
         'l', 'l', 'o', 'w', 'e', 'd'])))))))));
uncurry_top_proc uv uw ux uy uz va vb vc Nothing = Inr ();

doc_of_string :: [Char] -> Sum [Char] Xmldoc;
doc_of_string s = binda (parse_doc s) (\ (doc, _) -> Inr doc);

filtered_fun :: forall a. Filtered a -> a;
filtered_fun (FPair f n) = f;

shows_prec_filtered ::
  forall a. (Showa a) => Nat -> Filtered a -> [Char] -> [Char];
shows_prec_filtered d ff = shows_prec Zero_nat (filtered_fun ff :: a);

shows_list_filtered :: forall a. (Showa a) => [Filtered a] -> [Char] -> [Char];
shows_list_filtered ffs = shows_list_aux (shows_prec_filtered Zero_nat) ffs;

instance (Showa a) => Showa (Filtered a) where {
  shows_prec = shows_prec_filtered;
  shows_list = shows_list_filtered;
};

afs_check ::
  forall a b.
    (Showa a,
      Showa b) => ([Char] -> [Char]) ->
                    (a -> Nat -> Af_entry) ->
                      ((Term (Filtered a) b, Term (Filtered a) b) ->
                        Sum ([Char] -> [Char]) ()) ->
                        (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
afs_check rel pia g lr =
  let {
    pl = afs_term pia (fst lr);
    pr = afs_term pia (snd lr);
  } in catcha (g (pl, pr))
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r',
                    'i', 'e', 'n', 't', ' '] .
                 shows_prec_term Zero_nat (fst lr) .
                   shows_string [' '] .
                     rel . shows_string [' '] .
                             shows_prec_term Zero_nat (snd lr) .
                               shows_nl .
                                 shows_string ['p', 'i', '(', ' '] .
                                   shows_prec_term Zero_nat (fst lr) .
                                     shows_string [' ', ')', ' ', '=', ' '] .
                                       shows_prec_term Zero_nat pl .
 shows_nl .
   shows_string ['p', 'i', '(', ' '] .
     shows_prec_term Zero_nat (snd lr) .
       shows_string [' ', ')', ' ', '=', ' '] .
         shows_prec_term Zero_nat pr . shows_nl . x));

shows_afs :: forall a. (Showa a) => [((a, Nat), Af_entry)] -> [Char] -> [Char];
shows_afs af =
  shows_map
    (\ (a, b) ->
      let {
        (f, n) = a;
      } in (\ e ->
             shows_string ['p', 'i', '('] .
               shows_prec Zero_nat f .
                 shows_string ['/'] .
                   shows_prec_nat Zero_nat n .
                     shows_string [')', ' ', '=', ' '] .
                       (case e of {
                         Collapse i -> shows_prec_nat Zero_nat (Suc i);
                         AFList ids -> shows_prec_list Zero_nat (map Suc ids);
                       }) .
                         shows_nl)
        b)
    af;

xml2dpDisproof ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (DPDisproof a [Char]);
xml2dpDisproof xml2name x =
  xml1element
    ['d', 'p', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
      'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
         xml3elements
           ['d', 'p', 'R', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
           (xml1element ['d', 'p', 's'] (rules xml2name) id)
           (xml1element ['t', 'r', 's'] (rules xml2name) id)
           (xml2dpDisproof xml2name) DPRuleRemoval),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) -> DPLoop s rseq a b))])
    id x;

starts_with :: [Char] -> [Char] -> Bool;
starts_with t s =
  (if not (null t) && membera wspace (head t) then starts_with (tl t) s
    else (case search_len t s (size_list s) of {
           Nothing -> False;
           Just _ -> True;
         }));

forallm_index_aux ::
  forall a b. (a -> Nat -> Sum b ()) -> Nat -> [a] -> Sum ((a, Nat), b) ();
forallm_index_aux p i [] = Inr ();
forallm_index_aux p i (x : xs) =
  binda (catcha (p x i) (\ xa -> Inl ((x, i), xa)))
    (\ _ -> forallm_index_aux p (Suc i) xs);

forallm_index ::
  forall a b. (a -> Nat -> Sum b ()) -> [a] -> Sum ((a, Nat), b) ();
forallm_index p xs = forallm_index_aux p Zero_nat xs;

instance (Eq a, Linorder a, Eq b, Linorder b) => Linorder (Lab a b) where {
};

prightI ::
  forall a b c.
    (Eq a,
      Eq c) => a -> a -> (a -> a -> a) ->
                           (a -> a -> a) ->
                             (a -> a) ->
                               (b -> Nat -> (a, [a])) -> Term b c -> L_poly c a;
prightI ze on pl ti mz pia (Var x) = LPoly ze [(x, on)];
prightI ze on pl ti mz pia (Fun f ts) =
  let {
    (c, asa) = pia f (size_list ts);
    (LPoly d a) =
      sum_lpolyI ze pl (LPoly c [])
        (list_sumI (LPoly ze []) (sum_lpolyI ze pl)
          (map (\ at ->
                 mul_lpolyI ze ti (fst at)
                   (prightI ze on pl ti mz pia (snd at)))
            (zip asa ts)));
  } in LPoly (mz d) a;

check_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b ->
                    [(Pos, ((Term a b, Term a b), Term a b))] ->
                      [(b, Term a b)] ->
                        Ctxt a b ->
                          [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_loop s rseq sigma c r =
  binda (check (not (null rseq))
          (shows_prec_list Zero_nat
            ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u', 'e',
              'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n',
              'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
    (\ _ ->
      check_rsteps r rseq s
        (ctxt_apply c (subst_apply_term s (subst_of sigma))));

poly_vars_list :: forall a b. (Eq a) => [([(a, Nat)], b)] -> [a];
poly_vars_list p = remdups (concatMap (map fst . fst) p);

funas_term_list :: forall a b. (Eq a) => Term a b -> [(a, Nat)];
funas_term_list = remdups . funas_term_lista;

rel_pow_impl ::
  forall a b.
    ([a] -> [a]) -> ([a] -> b -> b) -> (a -> b -> Bool) -> [a] -> b -> Nat -> b;
rel_pow_impl succa un memb new have Zero_nat = un new have;
rel_pow_impl succa un memb new have (Suc m) =
  (if null new then have
    else let {
           maybea = succa new;
           havea = un new have;
           newa = filter (\ n -> not (memb n havea)) maybea;
         } in rel_pow_impl succa un memb newa havea m);

rtrancl_impl ::
  forall a b.
    ([(a, a)] -> [a] -> [a]) ->
      ([a] -> b -> b) -> (a -> b -> Bool) -> b -> [(a, a)] -> [a] -> b;
rtrancl_impl gen_succ un memb emp rel =
  let {
    succa = gen_succ rel;
    n = size_list rel;
  } in (\ asa -> rel_pow_impl succa un memb asa emp n);

elem_list_to_rm :: forall a b. (Linorder b) => (a -> b) -> [a] -> Rbt b [a];
elem_list_to_rm key [] = rm_empty;
elem_list_to_rm key (d : ds) =
  let {
    rm = elem_list_to_rm key ds;
    k = key d;
  } in (case rm_alpha rm k of {
         Nothing -> rm_update_dj k [d] rm;
         Just dataa -> rm_update k (d : dataa) rm;
       });

rm_set_lookup :: forall a b. (Linorder a) => Rbt a [b] -> a -> [b];
rm_set_lookup rm =
  (\ a -> (case rm_alpha rm a of {
            Nothing -> [];
            Just rules -> rules;
          }));

rtrancl_rbt_impl :: forall a. (Linorder a) => [(a, a)] -> [a] -> Rbt a ();
rtrancl_rbt_impl =
  rtrancl_impl
    (\ rel ->
      let {
        rm = elem_list_to_rm fst rel;
      } in (\ asa ->
             rs_to_list
               (foldla
                 (\ a x ->
                   rs_union a (list_to_rs (map snd (rm_set_lookup rm x))))
                 rs_empty asa)))
    (\ asa bs -> rs_union bs (list_to_rs asa)) rs_memb rs_empty;

memo_rbt_rtrancl :: forall a. (Linorder a) => [(a, a)] -> a -> Rbt a ();
memo_rbt_rtrancl rel =
  let {
    tr = rtrancl_rbt_impl rel;
    rm = list_to_rm
           (map (\ a -> (a, tr [a]))
             (((rs_to_list . list_to_rs) . map fst) rel));
  } in (\ a ->
         (case rm_lookup a rm of {
           Nothing -> list_to_rs [a];
           Just asa -> asa;
         }));

conv_ta_rule :: forall a b. (a -> Rbt a ()) -> Ta_rule a b -> Ta_rule_impl a b;
conv_ta_rule eps (TA_rule f qs q) = TA_rule_impl f qs q (eps q);

generate_ta ::
  forall a b.
    (Linorder a, Eq b, Linorder b) => Tree_automaton a b -> Ta_impl_ext a b ();
generate_ta (Tree_Automaton fin rules eps) =
  let {
    ep = memo_rbt_rtrancl eps;
    rqs_rs = foldla (\ a x -> rs_union a (ep (rq x))) rs_empty rules;
    rrules = map (conv_ta_rule ep) rules;
  } in Ta_impl_ext (list_to_rs fin) (elem_list_to_rm rsym_impl rrules)
         (rs_to_list rqs_rs) rqs_rs ep ();

ta_eps_impl :: forall a b c. Ta_impl_ext a b c -> a -> Rbt a ();
ta_eps_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    more)
  = ta_eps_impl;

ta_res_list ::
  forall a b.
    (Eq a, Linorder a,
      Linorder b) => Rbt (a, Nat) [Ta_rule_impl b a] ->
                       (b -> Rbt b ()) -> Term a b -> Rbt b ();
ta_res_list ta eps (Var q) = eps q;
ta_res_list ta eps (Fun f ts) =
  let {
    rec = map (ta_res_list ta eps) ts;
    n = size_list ts;
    rules = rm_set_lookup ta (f, n);
    arules =
      filter
        (\ rule ->
          let {
            qs = rqs_impl rule;
          } in all_interval_nat (\ i -> rs_memb (nth qs i) (nth rec i)) Zero_nat
                 n)
        rules;
    a = map rq_impl arules;
  } in foldla rs_union rs_empty a;

ta_rqs_impl :: forall a b c. Ta_impl_ext a b c -> [a];
ta_rqs_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    more)
  = ta_rqs_impl;

check_not_wf_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_not_wf_trs r =
  check (not (isOK (check_wf_trs r)))
    (shows_string
       ['T', 'h', 'e', ' ', 'T', 'R', 'S', ' ', 'i', 's', ' ', 'w', 'e', 'l',
         'l', ' ', 'f', 'o', 'r', 'm', 'e', 'd'] .
      shows_nl);

uncurry_proc_both ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Maybe Nat ->
                    (Lab a b,
                      ([((Lab a b, Nat), [Lab a b])],
                        ([(Term (Lab a b) [Char], Term (Lab a b) [Char])],
                          [(Term (Lab a b) [Char], Term (Lab a b) [Char])]))) ->
                      [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                        [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                          [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                            [(Term (Lab a b) [Char], Term (Lab a b) [Char])] ->
                              Maybe ((Lab a b, Nat) ->
                                      [(Term (Lab a b) [Char],
 Term (Lab a b) [Char])]) ->
                                Sum ([Char] -> [Char]) ();
uncurry_proc_both Nothing (a, (sml, (u, e))) =
  uncurry_proc (a, (sml, (u, e))) (fmapa a (nat 2)) check_inj;
uncurry_proc_both (Just n) (a, (sml, (u, e))) =
  uncurry_top_proc (a, (sml, (u, e))) n (fmapa a n) check_inj;

afs_list_to_af ::
  forall a. (Eq a) => [((a, Nat), Af_entry)] -> a -> Nat -> Nat -> Bool;
afs_list_to_af pia f n =
  (\ i ->
    (case lookupb (f, n) pia of {
      Nothing -> True;
      Just (Collapse a) -> equal_nat i a;
      Just (AFList ids) -> membera ids i;
    }));

default_af_entry :: Nat -> Af_entry;
default_af_entry n = AFList (upt Zero_nat n);

afs_list_to_afs ::
  forall a. (Eq a) => [((a, Nat), Af_entry)] -> a -> Nat -> Af_entry;
afs_list_to_afs pia f n =
  (case lookupb (f, n) pia of {
    Nothing -> default_af_entry n;
    Just entry -> entry;
  });

check_wf_afs ::
  forall a. (Showa a) => [((a, Nat), Af_entry)] -> Sum ([Char] -> [Char]) ();
check_wf_afs pia =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ e ->
               catcha
                 (case e of {
                   Collapse i -> check (less_nat i n) i;
                   AFList ids ->
                     catcha
                       (forallm
                         (\ x -> (if less_nat x n then Inr () else Inl x)) ids)
                       (\ x -> Inl (snd x));
                 })
                 (\ x ->
                   Inl (shows_string ['i', 'n', 'd', 'e', 'x', ' '] .
                         shows_prec_nat Zero_nat x .
                           shows_string
                             [' ', 't', 'o', 'o', ' ', 'h', 'i', 'g', 'h', ' ',
                               'f', 'o', 'r', ' ', 'A', 'F', '-', 'e', 'n', 't',
                               'r', 'y', ' ', 'o', 'f'] .
                             shows_prec_prod Zero_nat (f, n) . shows_nl)))
          b)
      pia)
    (\ x -> Inl (snd x));

mono_af_entry :: Nat -> Af_entry -> Bool;
mono_af_entry n (Collapse i) = less_eq_nat n one_nat;
mono_af_entry n (AFList ids) = all_interval_nat (membera ids) Zero_nat n;

check_mono_afs_help ::
  forall a b c. (Showa a) => ((a, b), c) -> [Char] -> [Char];
check_mono_afs_help ((f, a), e) =
  shows_string ['e', 'n', 't', 'r', 'y', ' ', 'f', 'o', 'r', ' '] .
    shows_prec Zero_nat f .
      shows_string
        [' ', 'n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n', 'i', 'c'] .
        shows_nl;

check_mono_afs ::
  forall a. (Showa a) => [((a, Nat), Af_entry)] -> Sum ([Char] -> [Char]) ();
check_mono_afs pia =
  catcha
    (catcha
      (forallm
        (\ x -> (if mono_af_entry (snd (fst x)) (snd x) then Inr () else Inl x))
        pia)
      (\ x -> Inl (snd x)))
    (\ x -> Inl (check_mono_afs_help x));

af_redpair ::
  forall a b.
    (Eq a, Showa a,
      Showa b) => [((a, Nat), Af_entry)] ->
                    (Sum ([Char] -> [Char]) (),
                      ((Term (Filtered a) b, Term (Filtered a) b) ->
                         Sum ([Char] -> [Char]) (),
                        ((Term (Filtered a) b, Term (Filtered a) b) ->
                           Sum ([Char] -> [Char]) (),
                          (Filtered a -> Nat -> Nat -> Bool,
                            (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                      (Sum ([Char] -> [Char]) (),
                        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                          ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                            (a -> Nat -> Nat -> Bool,
                              (Sum ([Char] -> [Char]) (), [Char] -> [Char])))));
af_redpair pia rp =
  let {
    af = afs_list_to_af pia;
    afs = afs_list_to_afs pia;
  } in (binda (check_wf_afs pia) (\ _ -> rp_valid rp),
         (afs_check (shows_string ['>']) afs (rp_strict rp),
           (afs_check (shows_string ['>', '=']) afs (rp_non_strict rp),
             (af, (binda (check_mono_afs pia) (\ _ -> rp_mono rp),
                    shows_string
                      ['A', 'r', 'g', 'u', 'm', 'e', 'n', 't', ' ', 'F', 'i',
                        'l', 't', 'e', 'r', ':', ' '] .
                      shows_nl . shows_afs pia . shows_nl . rp_desc rp)))));

xml2trsDisproof ::
  forall a. (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (TRSDisproof a [Char]);
xml2trsDisproof xml2name x =
  xml1element
    ['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
      'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
          'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'],
         xml0elements
           ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i',
             't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd']
           NotWellFormed),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) ->
              Loop s (map (\ (xa, (y, (_, z))) -> (xa, (y, z))) rseq) a b)),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml2elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsDisproof xml2name) NTRuleRemoval),
        (['d', 'p', 'T', 'r', 'a', 'n', 's'],
          xml3elements ['d', 'p', 'T', 'r', 'a', 'n', 's']
            (xml1element ['d', 'p', 's'] (rules xml2name) id)
            (bool ['m', 'a', 'r', 'k', 'e', 'd', 'S', 'y', 'm', 'b', 'o', 'l',
                    's'])
            (xml2dpDisproof xml2name) (\ p _ -> NTDPTrans p)),
        (['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l'],
          xml2elements
            ['s', 't', 'r', 'i', 'n', 'g', 'R', 'e', 'v', 'e', 'r', 's', 'a',
              'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2trsDisproof xml2name) NTStringReversal)])
    id x;

xml2relTrsProof ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] (Lab a [Nat])) ->
                Xml -> Sum [Char] (RelTRSProof a [Nat] [Char]);
xml2relTrsProof xml2name x =
  xml1element
    ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i', 'n', 'a',
      't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['r', 'I', 's', 'E', 'm', 'p', 't', 'y'],
         xml0elements ['r', 'I', 's', 'E', 'm', 'p', 't', 'y'] RelRisEmpty),
        (['s', 'I', 's', 'E', 'm', 'p', 't', 'y'],
          xml1element ['s', 'I', 's', 'E', 'm', 'p', 't', 'y']
            (xml2trsProof xml2name) RelSisEmpty),
        (['u', 'n', 'l', 'a', 'b'],
          xml3elements ['u', 'n', 'l', 'a', 'b']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2relTrsProof xml2name) RelUnlab),
        (['s', 'e', 'm', 'l', 'a', 'b'],
          xml4elements ['s', 'e', 'm', 'l', 'a', 'b'] (sl_inter xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2relTrsProof xml2name)
            (\ a ->
              (case a of {
                Inl (i, True) -> RelSemlab i (Inl id_labeling);
                Inl (i, False) -> RelSemlab i (Inr []);
                Inr _ -> RelRootlab;
              }))),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml4elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (ordering_constraint_proof xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2relTrsProof xml2name) RelRuleRemoval),
        (['u', 'n', 'c', 'u', 'r', 'r', 'y'],
          xml4elements ['u', 'n', 'c', 'u', 'r', 'r', 'y']
            (uncurry_info xml2name)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2relTrsProof xml2name) RelUncurry)])
    id x;

xml2relTrsDisproof ::
  forall a.
    (Xml -> Sum [Char] a) -> Xml -> Sum [Char] (RelTRSDisproof a [Char]);
xml2relTrsDisproof xml2name x =
  xml1element
    ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e', 'r', 'm',
      'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f']
    (xml_options
      [(['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i', 't',
          'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd'],
         xml0elements
           ['v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', 'C', 'o', 'n', 'd', 'i',
             't', 'i', 'o', 'n', 'V', 'i', 'o', 'l', 'a', 't', 'e', 'd']
           RelNotWellFormed),
        (['l', 'o', 'o', 'p'],
          xml_change (loopa xml2name)
            (\ (s, (rseq, (a, b))) -> RelLoop s rseq a b)),
        (['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
           'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'],
          xml_change (xml2trsDisproof xml2name) RNotSN),
        (['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l'],
          xml3elements ['r', 'u', 'l', 'e', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml1element ['t', 'r', 's'] (rules xml2name) id)
            (xml2relTrsDisproof xml2name) NTRelRuleRemoval)])
    id x;

xml2certProblem ::
  forall a.
    (Eq a) => (Xml -> Sum [Char] a) ->
                (Xml -> Sum [Char] (Lab a [Nat])) ->
                  Xml -> Sum [Char] (CertificationProblem a [Nat] [Char]);
xml2certProblem xml2uname xml2name (Xml name egal cs texta) =
  (if name ==
        ['c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o', 'n', 'P',
          'r', 'o', 'b', 'l', 'e', 'm'] &&
        (is_none texta && less_eq_nat (nat 3) (size_list cs))
    then binda (xml2trsInput xml2uname (nth cs Zero_nat))
           (\ (r, s_opt) ->
             binda (text ['c', 'p', 'f', 'V', 'e', 'r', 's', 'i', 'o', 'n']
                     (nth cs one_nat))
               (\ _ ->
                 (case s_opt of {
                   Nothing ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['t', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a',
                             't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f'],
                            xml_change (xml2trsProof xml2name) (TRSProof r)),
                           (['t', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm',
                              'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o',
                              'o', 'f'],
                             xml_change (xml2trsDisproof xml2name)
                               (TRSDisproof r))])
                       id (nth cs (nat 2));
                   Just s ->
                     xml1element ['p', 'r', 'o', 'o', 'f']
                       (xml_options
                         [(['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e',
                             'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P',
                             'r', 'o', 'o', 'f'],
                            xml_change (xml2relTrsProof xml2name)
                              (RelTRSProof (r, s))),
                           (['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o',
                              'n', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i',
                              'o', 'n', 'P', 'r', 'o', 'o', 'f'],
                             xml_change (xml2relTrsDisproof xml2name)
                               (RelTRSDisproof (r, s)))])
                       id (nth cs (nat 2));
                 })))
    else Inl ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'p', 'a', 'r',
               's', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'e', 'l', 'e', 'm', 'e',
               'n', 't']);

eqWhiteSpace :: [Char] -> [Char] -> Bool;
eqWhiteSpace s1 s2 =
  debug ['0']
    ['c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'w', 'h', 'e', 't', 'h',
      'e', 'r', ' ', 'p', 'a', 'r', 's', 'e', 'd', ' ', 'i', 'n', 'p', 'u', 't',
      ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', 's', ' ', 't', 'o',
      ' ', 'r', 'e', 'a', 'l', ' ', 'i', 'n', 'p', 'u', 't']
    (filter (\ c -> not (membera wspace c)) s1 ==
      filter (\ c -> not (membera wspace c)) s2);

xml_certificationProblem ::
  forall a b c. (Showa a, Showa c) => CertificationProblem a b c -> Xml;
xml_certificationProblem (TRSProof trs uu) =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    [xml_rules ['t', 'r', 's'] trs] Nothing;
xml_certificationProblem (TRSDisproof trs uv) =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    [xml_rules ['t', 'r', 's'] trs] Nothing;
xml_certificationProblem (RelTRSProof (r, s) uw) =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    [xml_rules ['t', 'r', 's'] r,
      xml_rules
        ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'] s]
    Nothing;
xml_certificationProblem (RelTRSDisproof (r, s) ux) =
  Xml ['t', 'r', 's', 'I', 'n', 'p', 'u', 't'] []
    [xml_rules ['t', 'r', 's'] r,
      xml_rules
        ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'R', 'u', 'l', 'e', 's'] s]
    Nothing;

check_DP ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> a) ->
                    [(Term a b, Term a b)] ->
                      Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                        [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_DP shp r rm p =
  catcha
    (binda (check_wf_trs r)
      (\ _ ->
        let {
          d = defs_list r;
        } in binda (catcha
                     (forallm
                       (\ (f, n) ->
                         check (not (membera d (shp f, n)))
                           (shows_string
                              ['s', 'h', 'a', 'r', 'p', 'i', 'n', 'g', ' '] .
                             shows_prec Zero_nat f .
                               shows_string
                                 [' ', 'y', 'i', 'e', 'l', 'd', 's', ' ', 't',
                                   'h', 'e', ' ', 'd', 'e', 'f', 'i', 'n', 'e',
                                   'd', ' ', 's', 'y', 'm', 'b', 'o', 'l',
                                   ' '] .
                                 shows_prec Zero_nat (shp f)))
                       d)
                     (\ x -> Inl (snd x)))
               (\ _ ->
                 catcha (check_subseteq (dP_list shp r rm) p)
                   (\ x ->
                     Inl (shows_string ['t', 'h', 'e', ' ', 'D', 'P', ' '] .
                           shows_rule x .
                             shows_string
                               [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't',
                                 ' ', 'a', 'p', 'p', 'e', 'a', 'r', ' ', 'i',
                                 'n', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ',
                                 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                               shows_nl)))))
    (\ x ->
      Inl (shows_string
             ['t', 'h', 'e', ' ', 'D', 'P', '-', 't', 'r', 'a', 'n', 's', 'f',
               'o', 'r', 'm', 'a', 't', 'i', 'o', 'n', ' ', 'i', 's', ' ', 'n',
               'o', 't', ' ', 'a', 'p', 'p', 'l', 'i', 'e', 'd', ' ', 'c', 'o',
               'r', 'r', 'e', 'c', 't', 'l', 'y', '.'] .
            shows_nl . x));

internal_error :: [Char] -> [Char];
internal_error =
  shows_string
    ['i', 'n', 't', 'e', 'r', 'n', 'a', 'l', ' ', 'e', 'r', 'r', 'o', 'r'] .
    shows_nl .
      shows_string
        ['p', 'l', 'e', 'a', 's', 'e', ' ', 'c', 'o', 'n', 't', 'a', 'c', 't',
          ' ', 'o', 'n', 'e', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'C', 'e',
          'T', 'A', ' ', 'm', 'e', 'm', 'b', 'e', 'r', 's'] .
        shows_nl .
          shows_string
            ['h', 't', 't', 'p', ':', '/', '/', 'c', 'l', '-', 'i', 'n', 'f',
              'o', 'r', 'm', 'a', 't', 'i', 'k', '.', 'u', 'i', 'b', 'k', '.',
              'a', 'c', '.', 'a', 't', '/', 's', 'o', 'f', 't', 'w', 'a', 'r',
              'e', '/', 'c', 'e', 't', 'a'] .
            shows_nl .
              shows_string
                ['c', 'e', 't', 'a', '@', 'i', 'n', 'f', 'o', 'r', 'm', 'a',
                  't', 'i', 'k', '.', 'u', 'i', 'b', 'k', '.', 'a', 'c', '.',
                  'a', 't'] .
                shows_nl;

check_redpair_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_redpair_proc rp p rP r =
  catcha
    (let {
       (_, nSp) = partition (\ foo -> isOK (rp_strict rp foo)) p;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha
                      (catcha (forallm (rp_non_strict rp) r)
                        (\ x -> Inl (snd x)))
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n',
                                 't', 'i', 'n', 'g', ' ', 'T', 'R', 'S'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha
                          (catcha (forallm (rp_non_strict rp) nSp)
                            (\ x -> Inl (snd x)))
                          (\ x ->
                            Inl (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e',
                                     'n', 't', 'i', 'n', 'g', ' ', 'D', 'P',
                                     's'] .
                                  shows_nl . x)))
                    (\ _ ->
                      catcha (check_subseteq nSp rP)
                        (\ x ->
                          Inl (case rp_strict rp x of {
                                Inl err -> err;
                                Inr _ -> internal_error;
                              }))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't',
               'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
               'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h',
               'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl . rp_desc rp . shows_nl . x));

to_lpoly_inter ::
  forall a b.
    (Eq b) => a -> a -> [((b, Nat), (a, [a]))] -> b -> Nat -> (a, [a]);
to_lpoly_inter d on i f n =
  (case lookupb (f, n) i of {
    Nothing -> (d, replicatea n on);
    Just cas -> cas;
  });

check_lpoly_coeffs ::
  forall a b.
    (Showa a,
      Showa b) => (a -> Bool) ->
                    (a -> Bool) ->
                      (a -> Bool) ->
                        [((b, Nat), (a, [a]))] -> Sum ([Char] -> [Char]) ();
check_lpoly_coeffs wfe apos ch i =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ (c, cs) ->
               catcha
                 (binda
                   (check (wfe c)
                     (shows_string
                        ['c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a',
                          'r', 't', ' '] .
                       shows_prec Zero_nat c .
                         shows_string
                           [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'w', 'e',
                             'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd'] .
                           shows_nl))
                   (\ _ ->
                     binda (check (apos c || any apos (takea n cs))
                             (shows_string
                                ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't',
                                  ' ', 'f', 'i', 'n', 'd', ' ', 'p', 'o', 's',
                                  'i', 't', 'i', 'v', 'e', ' ', 'e', 'n', 't',
                                  'r', 'y', ' ', 'w', 'h', 'i', 'c', 'h', ' ',
                                  'i', 's', ' ', 'r', 'e', 'q', 'u', 'i', 'r',
                                  'e', 'd', ' ', 'f', 'o', 'r', ' ', 'a', 'r',
                                  'c', 't', 'i', 'c', ' ', 'i', 'n', 't', 'e',
                                  'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o',
                                  'n', 's'] .
                               shows_nl))
                       (\ _ ->
                         catcha
                           (forallm
                             (\ aa ->
                               check (ch aa && wfe aa)
                                 (shows_string
                                    ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i',
                                      'e', 'n', 't', ' '] .
                                   shows_prec Zero_nat aa .
                                     shows_string
                                       [' ', 'i', 's', ' ', 'n', 'o', 't', ' ',
 'a', 'l', 'l', 'o', 'w', 'e', 'd'] .
                                       shows_nl))
                             cs)
                           (\ x -> Inl (snd x)))))
                 (\ x ->
                   Inl (shows_string
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                            't', 'h', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r',
                            'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                            ' '] .
                         shows_prec Zero_nat f .
                           shows_string ['/'] .
                             shows_prec_nat Zero_nat n . shows_nl . x)))
          b)
      i)
    (\ x -> Inl (snd x));

get_comp_wf_elem ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> Bool;
get_comp_wf_elem
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  wfe;

get_comp_ge ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> a -> Bool;
get_comp_ge
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  ge;

lookup_rest :: forall a b. (Eq a) => a -> [(a, b)] -> Maybe (b, [(a, b)]);
lookup_rest x [] = Nothing;
lookup_rest x ((y, c) : ycs) =
  (if x == y then Just (c, ycs)
    else (case lookup_rest x ycs of {
           Nothing -> Nothing;
           Just (d, yccs) -> Just (d, (y, c) : yccs);
         }));

check_pvarsI ::
  forall a b.
    (Eq b) => a -> (a -> a -> Bool) -> [(b, a)] -> [(b, a)] -> Sum b ();
check_pvarsI ze ge vas [] =
  catcha (forallm (\ va -> check (ge (snd va) ze) (fst va)) vas)
    (\ x -> Inl (snd x));
check_pvarsI ze ge vas ((x, b) : vbs) =
  let {
    (a, vasa) =
      (case lookup_rest x vas of {
        Nothing -> (ze, vas);
        Just (a, ba) -> (a, ba);
      });
  } in binda (check (ge a b) x) (\ _ -> check_pvarsI ze ge vasa vbs);

show_pvars :: forall a b. (Eq a, Showa a, Showa b) => a -> [(b, a)] -> [[Char]];
show_pvars on [] = [];
show_pvars on ((x, c) : vas) =
  ((if on == c then [] else shows_prec Zero_nat c []) ++
    shows_prec Zero_nat x []) :
    show_pvars on vas;

shows_lpoly ::
  forall a b.
    (Eq a, Showa a, Showa b) => a -> a -> L_poly b a -> [Char] -> [Char];
shows_lpoly ze on (LPoly c cs) =
  (case show_pvars on cs of {
    [] -> shows_prec Zero_nat c;
    a : list ->
      (if c == ze then id
        else shows_prec Zero_nat c . shows_string [' ', '+', ' ']) .
        shows_list_gen (shows_prec_list Zero_nat) [] [] [' ', '+', ' '] []
          (a : list);
  });

check_lpoly_sI ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Bool ->
                    a -> a -> (a -> a -> Bool) ->
                                (a -> a -> Bool) ->
                                  L_poly b a ->
                                    L_poly b a -> Sum ([Char] -> [Char]) ();
check_lpoly_sI mode ze on ge gt (LPoly a vas) (LPoly b vbs) =
  catcha
    (binda
      (check (gt a b)
        (shows_string
          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
            'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c', 'o', 'n', 's',
            't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't']))
      (\ _ ->
        catcha (check_pvarsI ze (if mode then ge else gt) vas vbs)
          (\ x ->
            Inl (shows_string
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                     ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                     'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                     'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                     ' '] .
                  shows_prec Zero_nat x))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
               'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' '] .
            shows_lpoly ze on (LPoly a vas) .
              shows_string [' ', '>', ' '] .
                shows_lpoly ze on (LPoly b vbs) . shows_nl . x));

check_polo_sI ::
  forall a b c.
    (Eq a, Showa a, Showa b, Eq c,
      Showa c) => Bool ->
                    a -> a -> (a -> a -> a) ->
                                (a -> a -> a) ->
                                  (a -> a) ->
                                    (a -> a -> Bool) ->
                                      (a -> a -> Bool) ->
(b -> Nat -> (a, [a])) -> (Term b c, Term b c) -> Sum ([Char] -> [Char]) ();
check_polo_sI mode ze on pl ti mz ge gt pia (s, t) =
  let {
    left = pleftI ze on pl ti mz pia s;
    right = prightI ze on pl ti mz pia t;
  } in catcha (check_lpoly_sI mode ze on ge gt left right)
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'n',
                    's', 'u', 'r', 'e', ' '] .
                 shows_prec_term Zero_nat s .
                   shows_string [' ', '>', ' '] .
                     shows_prec_term Zero_nat t . shows_nl . x));

get_comp_max0 ::
  forall a.
    (a -> Bool,
      (a, (a, (a, (a -> a -> a,
                    (a -> a -> a,
                      (a -> a,
                        (a -> a -> Bool,
                          (a -> a -> Bool,
                            (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
      a -> a;
get_comp_max0
  (wfe, (ze, (on, (def, (pl, (ti, (mz, (ge, (gt, (mo, (de, (ap, cm)))))))))))) =
  mz;

check_lpoly_nsI ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => a -> a -> (a -> a -> Bool) ->
                              L_poly b a ->
                                L_poly b a -> Sum ([Char] -> [Char]) ();
check_lpoly_nsI ze on ge (LPoly a vas) (LPoly b vbs) =
  catcha
    (binda
      (check (ge a b)
        (shows_string
          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
            'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c', 'o', 'n', 's',
            't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't', 's']))
      (\ _ ->
        catcha (check_pvarsI ze ge vas vbs)
          (\ x ->
            Inl (shows_string
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                     ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                     'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                     'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e',
                     ' '] .
                  shows_prec Zero_nat x))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
               'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' '] .
            shows_lpoly ze on (LPoly a vas) .
              shows_string [' ', '>', '=', ' '] .
                shows_lpoly ze on (LPoly b vbs) . shows_nl . x));

check_polo_nsI ::
  forall a b c.
    (Eq a, Showa a, Showa b, Eq c,
      Showa c) => a -> a -> (a -> a -> a) ->
                              (a -> a -> a) ->
                                (a -> a) ->
                                  (a -> a -> Bool) ->
                                    (b -> Nat -> (a, [a])) ->
                                      (Term b c, Term b c) ->
Sum ([Char] -> [Char]) ();
check_polo_nsI ze on pl ti mz ge pia (s, t) =
  let {
    left = pleftI ze on pl ti mz pia s;
    right = prightI ze on pl ti mz pia t;
  } in catcha (check_lpoly_nsI ze on ge left right)
         (\ x ->
           Inl (shows_string
                  ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e', 'n',
                    's', 'u', 'r', 'e', ' '] .
                 shows_prec_term Zero_nat s .
                   shows_string [' ', '>', '=', ' '] .
                     shows_prec_term Zero_nat t . shows_nl . x));

create_afI ::
  forall a b.
    (Eq a, Eq b) => a -> [((b, Nat), (a, [a]))] -> b -> Nat -> Nat -> Bool;
create_afI ze [] f n i = True;
create_afI ze (((g, m), (c, coeffs)) : rest) f n i =
  (if f == g && equal_nat m n
    then less_nat i (size_list coeffs) && not (nth coeffs i == ze)
    else create_afI ze rest f n i);

check_poly_monoI ::
  forall a b.
    (Showa a,
      Showa b) => a -> (a -> a -> Bool) ->
                         (a -> Bool) ->
                           [((b, Nat), (a, [a]))] -> Sum ([Char] -> [Char]) ();
check_poly_monoI ze ge cm =
  (\ xs ->
    catcha
      (forallm
        (\ (a, b) ->
          let {
            (f, n) = a;
          } in (\ (c, cs) ->
                 catcha
                   (binda
                     (check (ge c ze)
                       (shows_string
                          ['c', 'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p',
                            'a', 'r', 't', ' '] .
                         shows_prec Zero_nat c .
                           shows_string
                             [' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'a',
                               't', ' ', 'l', 'e', 'a', 's', 't', ' '] .
                             shows_prec Zero_nat ze . shows_nl))
                     (\ _ ->
                       binda (check (less_eq_nat n (size_list cs))
                               (shows_string
                                  ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ',
                                    'a', 'r', 'g', 'u', 'm', 'e', 'n', 't', ' ',
                                    'i', 's', ' ', 'i', 'g', 'n', 'o', 'r', 'e',
                                    'd'] .
                                 shows_nl))
                         (\ _ ->
                           catcha
                             (forallm
                               (\ d ->
                                 check (cm d)
                                   (shows_string
                                      ['c', 'o', 'e', 'f', 'f', 'i', 'c', 'i',
'e', 'n', 't', ' '] .
                                     shows_prec Zero_nat d .
                                       shows_string
 [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd'] .
 shows_nl))
                               cs)
                             (\ x -> Inl (snd x)))))
                   (\ x ->
                     Inl (shows_string
                            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                              't', 'h', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n',
                              'i', 'c', 'i', 't', 'y', ' ', 'd', 'u', 'e', ' ',
                              't', 'o', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r',
                              'e', 't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f',
                              ' '] .
                           shows_prec Zero_nat f .
                             shows_string ['/'] .
                               shows_prec_nat Zero_nat n . shows_nl . x)))
            b)
        xs)
      (\ x -> Inl (snd x)));

create_lpoly_repr ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Bool,
                    (a, (a, (a, (a -> a -> a,
                                  (a -> a -> a,
                                    (a -> a,
                                      (a -> a -> Bool,
(a -> a -> Bool, (Bool, ([Char], (a -> Bool, a -> Bool)))))))))))) ->
                    [((b, Nat), (a, [a]))] -> [Char] -> [Char];
create_lpoly_repr c i =
  let {
    pia = to_lpoly_inter (get_comp_def c) (get_comp_1 c) i;
  } in shows_string
         ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i', 'n', 't',
           'e', 'r', 'p', 'r', 'e', 't', 'r', 'a', 't', 'i', 'o', 'n', ' ', 'o',
           'v', 'e', 'r', ' '] .
         shows_string (get_comp_description c) .
           shows_nl .
             shows_sep
               (\ (f, n) ->
                 let {
                   t = Fun f (map Var
                               (fresh_strings_list ['x', '_'] one_nat [] n));
                 } in shows_string ['P', 'o', 'l', '('] .
                        shows_prec_term Zero_nat t .
                          shows_string [')', ' ', '=', ' '] .
                            shows_lpoly (get_comp_0 c) (get_comp_1 c)
                              (pleftI (get_comp_0 c) (get_comp_1 c)
                                (get_comp_plus c) (get_comp_times c)
                                (get_comp_max0 c) pia t))
               shows_nl (remdups (map fst i));

create_poly_redpair ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([((a, Nat), (b, [b]))] -> Sum ([Char] -> [Char]) ()) ->
                    (b -> Bool,
                      (b, (b, (b, (b -> b -> b,
                                    (b -> b -> b,
                                      (b -> b,
(b -> b -> Bool,
  (b -> b -> Bool, (Bool, ([Char], (b -> Bool, b -> Bool)))))))))))) ->
                      [((a, Nat), (b, [b]))] ->
                        (Sum ([Char] -> [Char]) (),
                          ((Term a c, Term a c) -> Sum ([Char] -> [Char]) (),
                            ((Term a c, Term a c) -> Sum ([Char] -> [Char]) (),
                              (a -> Nat -> Nat -> Bool,
                                (Sum ([Char] -> [Char]) (),
                                  [Char] -> [Char])))));
create_poly_redpair cI c i =
  let {
    pia = to_lpoly_inter (get_comp_def c) (get_comp_1 c) i;
  } in (binda (cI i)
          (\ _ ->
            check_lpoly_coeffs (get_comp_wf_elem c) (get_comp_apos c)
              (\ x -> get_comp_ge c x (get_comp_0 c)) i),
         (check_polo_sI (get_comp_mono c) (get_comp_0 c) (get_comp_1 c)
            (get_comp_plus c) (get_comp_times c) (get_comp_max0 c)
            (get_comp_ge c) (get_comp_gt c) pia,
           (check_polo_nsI (get_comp_0 c) (get_comp_1 c) (get_comp_plus c)
              (get_comp_times c) (get_comp_max0 c) (get_comp_ge c) pia,
             (create_afI (get_comp_0 c) i,
               (binda (check (get_comp_mono c)
                        (shows_string
                          ['m', 'o', 'n', 'o', 't', 'o', 'n', 'e', ' ', 'p',
                            'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', 's',
                            ' ', 'n', 'o', 't', ' ', 's', 'u', 'p', 'p', 'o',
                            'r', 't', ' ', 'f', 'o', 'r', ' ', 't', 'h', 'i',
                            's', ' ', 'd', 'o', 'm', 'a', 'i', 'n']))
                  (\ _ ->
                    check_poly_monoI (get_comp_0 c) (get_comp_ge c)
                      (get_comp_check_mono c) i),
                 create_lpoly_repr c i)))));

poly_inter_list_to_inter ::
  forall a b.
    (Poly_carriera a,
      Eq b) => a -> [((b, Nat), [([(Nat, Nat)], a)])] ->
                      b -> Nat -> [([(Nat, Nat)], a)];
poly_inter_list_to_inter def i =
  (\ f n ->
    (case lookupb (f, n) i of {
      Nothing -> default_I def n;
      Just p -> p;
    }));

check_poly_weak_mono_discrete ::
  forall a b. (Eq a, Eq b, Poly_carrier b) => [([(a, Nat)], b)] -> a -> Bool;
check_poly_weak_mono_discrete p v =
  check_poly_ge
    (poly_subst
      (\ w -> poly_of (if w == v then PSum [PNum one, PVar v] else PVar w)) p)
    p;

check_poly_weak_mono_all ::
  forall a b. (Ordered_semiring_0 b) => [([(a, Nat)], b)] -> Bool;
check_poly_weak_mono_all p = all (\ (_, c) -> ge c zero) p;

check_poly_weak_mono_and_pos ::
  forall a b. (Eq a, Eq b, Poly_carrier b) => Bool -> [([(a, Nat)], b)] -> Bool;
check_poly_weak_mono_and_pos discrete p =
  (if discrete
    then all (check_poly_weak_mono_discrete p) (poly_vars_list p) &&
           ge (eval_poly (\ _ -> zero) p :: b) (zero :: b)
    else check_poly_weak_mono_all p);

check_poly_inter_list ::
  forall a b.
    (Eq b,
      Poly_carriera b) => Bool ->
                            [((a, Nat), [([(Nat, Nat)], b)])] ->
                              Sum (a, [([(Nat, Nat)], b)]) ();
check_poly_inter_list discrete i =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (_, a) = x;
              } in check_poly_weak_mono_and_pos discrete a
            then Inr () else Inl x))
        i)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (let {
             (a, b) = x;
           } in let {
                  (f, _) = a;
                } in (\ aa -> (f, aa))
                  b));

poly_inter_to_af ::
  forall a b. (a -> Nat -> [([(Nat, Nat)], b)]) -> a -> Nat -> Nat -> Bool;
poly_inter_to_af i f n = poly_var (i f n);

check_poly_strict_mono_discrete ::
  forall a b.
    (Eq a, Poly_carrier a,
      Eq b) => (a -> a -> Bool) -> [([(b, Nat)], a)] -> b -> Bool;
check_poly_strict_mono_discrete gt p v =
  check_poly_gt gt
    (poly_subst
      (\ w -> poly_of (if w == v then PSum [PNum one, PVar v] else PVar w)) p)
    p;

check_monom_strict_mono :: forall a. (Eq a) => Bool -> [(a, Nat)] -> a -> Bool;
check_monom_strict_mono pm m v =
  not (null m) &&
    (null (tl m) &&
      (fst (head m) == v &&
        (if pm then less_eq_nat one_nat (snd (head m))
          else equal_nat (snd (head m)) one_nat)));

check_poly_strict_mono ::
  forall a b. (Eq a, Poly_carrier b) => Bool -> [([(a, Nat)], b)] -> a -> Bool;
check_poly_strict_mono pm p v =
  any (\ (m, c) -> ge c one && check_monom_strict_mono pm m v) p;

check_poly_strict_mono_smart ::
  forall a b.
    (Eq a, Poly_carrier a,
      Eq b) => Bool ->
                 Bool -> (a -> a -> Bool) -> [([(b, Nat)], a)] -> b -> Bool;
check_poly_strict_mono_smart discrete pm gt p v =
  (if discrete then check_poly_strict_mono_discrete gt p v
    else check_poly_strict_mono pm p v);

create_nlpoly_redpair ::
  forall a b c.
    (Eq a, Poly_carriera a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => a -> (a -> a -> Bool) ->
                         Bool ->
                           Bool ->
                             [((b, Nat), [([(Nat, Nat)], a)])] ->
                               (Sum ([Char] -> [Char]) (),
                                 ((Term b c, Term b c) ->
                                    Sum ([Char] -> [Char]) (),
                                   ((Term b c, Term b c) ->
                                      Sum ([Char] -> [Char]) (),
                                     (b -> Nat -> Nat -> Bool,
                                       (Sum ([Char] -> [Char]) (),
 [Char] -> [Char])))));
create_nlpoly_redpair def gt power_mono discrete i =
  let {
    j = poly_inter_list_to_inter def i;
    x = poly_subst
          (\ n -> poly_of (PVar (['x', '_'] ++ shows_prec_nat Zero_nat n [])));
  } in (catcha (check_poly_inter_list discrete i)
          (\ xa ->
            Inl (let {
                   (f, p) = xa;
                 } in shows_string
                        ['i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
                          'i', 'o', 'n', ' '] .
                        shows_poly (x p) .
                          shows_string [' ', 'o', 'f', ' '] .
                            shows_prec Zero_nat f .
                              shows_string
                                [' ', 'i', 'n', 'v', 'a', 'l', 'i', 'd', ' '])),
         ((\ (s, t) ->
            let {
              p = eval_term j s;
              q = eval_term j t;
            } in check (check_poly_gt gt p q)
                   (shows_string
                      ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e',
                        'n', 's', 'u', 'r', 'e', ' '] .
                     shows_poly p .
                       shows_string [' ', '>', ' '] . shows_poly q)),
           ((\ (s, t) ->
              let {
                p = eval_term j s;
                q = eval_term j t;
              } in check (check_poly_ge p q)
                     (shows_string
                        ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'e',
                          'n', 's', 'u', 'r', 'e', ' '] .
                       shows_poly p .
                         shows_string [' ', '>', '=', ' '] . shows_poly q)),
             (poly_inter_to_af j,
               (catcha
                  (catcha
                    (forallm
                      (\ xa ->
                        (if let {
                              (a, b) = xa;
                            } in let {
                                   (_, n) = a;
                                 } in (\ p ->
all_interval_nat (check_poly_strict_mono_smart discrete power_mono gt p)
  Zero_nat n)
                                   b
                          then Inr () else Inl xa))
                      i)
                    (\ xa -> Inl (snd xa)))
                  (\ xa ->
                    Inl (let {
                           (a, b) = xa;
                         } in let {
                                (f, _) = a;
                              } in (\ p ->
                                     shows_string
                                       ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o',
 't', ' ', 'e', 'n', 's', 'u', 'r', 'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n',
 'i', 'c', 't', 'y', ' ', 'o', 'f', ' '] .
                                       shows_poly (x p) .
 shows_string
   [' ', 'a', 's', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't',
     'i', 'o', 'n', ' ', 'o', 'f', ' '] .
   shows_prec Zero_nat f)
                                b)),
                 shows_string
                   ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'i',
                     'n', 't', 'e', 'r', 'p', 'r', 'e', 't', 'a', 't', 'i', 'o',
                     'n'] .
                   shows_nl .
                     shows_sep
                       (\ (a, b) ->
                         let {
                           (f, n) = a;
                         } in (\ p ->
                                shows_string ['P', 'o', 'l', '('] .
                                  shows_prec Zero_nat f .
                                    shows_string ['/'] .
                                      shows_prec_nat Zero_nat n .
shows_string [')', ' ', '=', ' '] . shows_poly (x p))
                           b)
                       shows_nl i)))));

instance Ordered_semiring_0 Integer where {
};

instance Ordered_semiring_1 Integer where {
};

instance Poly_carrier Integer where {
};

instance Max_ordered_semiring_1 Integer where {
};

instance Poly_carriera Integer where {
};

instance Ordered_semiring_0 Rat where {
};

instance Ordered_semiring_1 Rat where {
};

instance Poly_carrier Rat where {
};

instance Max_ordered_semiring_1 Rat where {
};

instance Poly_carriera Rat where {
};

faulty_redpair_impl ::
  forall a b.
    (Showa a,
      Showa b) => [Char] ->
                    (Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                          (a -> Nat -> Nat -> Bool,
                            (Sum ([Char] -> [Char]) (), [Char] -> [Char])))));
faulty_redpair_impl s =
  (Inl (shows_string s),
    ((\ _ -> Inr ()), ((\ _ -> Inr ()), (full_af, (Inr (), id)))));

shows_arctic :: Arctic -> [Char] -> [Char];
shows_arctic (Num_arc i) = shows_prec_int Zero_nat i;
shows_arctic MinInfty = shows_string ['-', 'i', 'n', 'f'];

shows_prec_arctic :: Nat -> Arctic -> [Char] -> [Char];
shows_prec_arctic d ai = shows_arctic ai;

shows_list_arctic :: [Arctic] -> [Char] -> [Char];
shows_list_arctic asa = shows_list_aux (shows_prec_arctic Zero_nat) asa;

instance Showa Arctic where {
  shows_prec = shows_prec_arctic;
  shows_list = shows_list_arctic;
};

shows_arctic_rat :: Arctic_rat -> [Char] -> [Char];
shows_arctic_rat (Num_arc_rat i) = shows_prec_rat Zero_nat i;
shows_arctic_rat MinInfty_rat = shows_string ['-', 'i', 'n', 'f'];

shows_prec_arctic_rat :: Nat -> Arctic_rat -> [Char] -> [Char];
shows_prec_arctic_rat d ari = shows_arctic_rat ari;

shows_list_arctic_rat :: [Arctic_rat] -> [Char] -> [Char];
shows_list_arctic_rat asa = shows_list_aux (shows_prec_arctic_rat Zero_nat) asa;

instance Showa Arctic_rat where {
  shows_prec = shows_prec_arctic_rat;
  shows_list = shows_list_arctic_rat;
};

equal_filtered :: forall a. (Eq a) => Filtered a -> Filtered a -> Bool;
equal_filtered (FPair fa nata) (FPair f nat) = fa == f && equal_nat nata nat;

instance (Eq a) => Eq (Filtered a) where {
  a == b = equal_filtered a b;
};

lpo_repr_to_pr ::
  forall a. (Eq a) => [((a, Nat), Nat)] -> Filtered a -> Nat -> Nat;
lpo_repr_to_pr prs (FPair f a) uu =
  (case lookupb (f, a) prs of {
    Nothing -> Zero_nat;
    Just x -> x;
  });

get_redpair ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Redpair a ->
                    (Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                          (a -> Nat -> Nat -> Bool,
                            (Sum ([Char] -> [Char]) (), [Char] -> [Char])))));
get_redpair (Int_carrier i) =
  create_poly_redpair (\ _ -> Inr ())
    ((\ _ -> True),
      (0, (1, (1, ((\ a b -> a + b),
                    ((\ a b -> a * b),
                      (max0_int,
                        (ge_int,
                          ((\ x y -> y < x),
                            (True,
                              (int_name, ((\ _ -> True), int_mono))))))))))))
    i;
get_redpair (Int_nl_carrier i) =
  create_nlpoly_redpair 1 (\ x y -> y < x) True True i;
get_redpair (Rat_carrier d i) =
  create_poly_redpair
    (\ _ ->
      check (less_rat (fract 0 1) d)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'e', 'l', 't', 'a']))
    ((\ _ -> True),
      (fract 0 1,
        (fract 1 1,
          (d, (plus_rat,
                (times_rat,
                  (max0_rat,
                    (ge_rat,
                      (rat_gt d,
                        (True, (rat_name, ((\ _ -> True), rat_mono))))))))))))
    i;
get_redpair (Rat_nl_carrier d i) =
  (if less_rat (fract 0 1) d
    then create_nlpoly_redpair d (rat_gt d) (less_eq_rat (fract 1 1) d) False i
    else faulty_redpair_impl
           ['d', 'e', 'l', 't', 'a', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
             ' ', 'l', 'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ',
             'z', 'e', 'r', 'o']);
get_redpair (Arctic_carrier i) =
  create_poly_redpair (\ _ -> Inr ())
    ((\ _ -> True),
      (zero_arctic,
        (one_arctic,
          (one_arctic,
            (plus_arctic,
              (times_arctic,
                (id, (ge_arctic,
                       (gt_arctic,
                         (False,
                           (arctic_name, (pos_arctic, (\ _ -> False)))))))))))))
    i;
get_redpair (Arctic_rat_carrier d i) =
  create_poly_redpair
    (\ _ ->
      check (less_rat (fract 0 1) d)
        (shows_string
          ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'e', 'l', 't', 'a']))
    ((\ _ -> True),
      (zero_arctic_rat,
        (one_arctic_rat,
          (one_arctic_rat,
            (plus_arctic_rat,
              (times_arctic_rat,
                (id, (ge_arctic_rat,
                       (gt_arctic_rat d,
                         (False,
                           (arctic_rat_name,
                             (pos_arctic_rat, (\ _ -> False)))))))))))))
    i;
get_redpair (Int_mat_carrier n sd i) =
  create_poly_redpair
    (\ _ ->
      binda (Inr ())
        (\ _ ->
          check (less_nat Zero_nat sd && less_eq_nat sd n)
            (shows_string
              ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n',
                's', 'i', 'o', 'n'])))
    (mat n n,
      (mat0I 0 n n,
        (mat1I 0 1 n,
          (mat1I 0
             (get_comp_def
               ((\ _ -> True),
                 (0, (1, (1, ((\ a b -> a + b),
                               ((\ a b -> a * b),
                                 (max0_int,
                                   (ge_int,
                                     ((\ x y -> y < x),
                                       (True,
 (int_name, ((\ _ -> True), int_mono)))))))))))))
             n,
            (mat_plusI (\ a b -> a + b),
              (mat_multI 0 (\ a b -> a + b) (\ a b -> a * b) n,
                (mat_max0,
                  (mat_ge,
                    (mat_gtI
                       (get_comp_gt
                         ((\ _ -> True),
                           (0, (1, (1, ((\ a b -> a + b),
 ((\ a b -> a * b),
   (max0_int,
     (ge_int,
       ((\ x y -> y < x),
         (True, (int_name, ((\ _ -> True), int_mono)))))))))))))
                       sd,
                      (get_comp_mono
                         ((\ _ -> True),
                           (0, (1, (1, ((\ a b -> a + b),
 ((\ a b -> a * b),
   (max0_int,
     (ge_int,
       ((\ x y -> y < x),
         (True, (int_name, ((\ _ -> True), int_mono)))))))))))),
                        (concat
                           [shows_prec_nat Zero_nat n [], ['x'],
                             shows_prec_nat Zero_nat n [],
                             [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c',
                               't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o',
                               'n', ' '],
                             shows_prec_nat Zero_nat sd [],
                             [' ', 'o', 'v', 'e', 'r', ' '],
                             get_comp_description
                               ((\ _ -> True),
                                 (0, (1, (1,
   ((\ a b -> a + b),
     ((\ a b -> a * b),
       (max0_int,
         (ge_int,
           ((\ x y -> y < x),
             (True, (int_name, ((\ _ -> True), int_mono))))))))))))],
                          ((\ _ -> True),
                            mat_mono sd
                              (get_comp_check_mono
                                ((\ _ -> True),
                                  (0, (1,
(1, ((\ a b -> a + b),
      ((\ a b -> a * b),
        (max0_int,
          (ge_int,
            ((\ x y -> y < x),
              (True,
                (int_name, ((\ _ -> True), int_mono)))))))))))))))))))))))))
    i;
get_redpair (Rat_mat_carrier d n sd i) =
  create_poly_redpair
    (\ _ ->
      binda (check (less_rat (fract 0 1) d)
              (shows_string
                ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'e', 'l', 't',
                  'a']))
        (\ _ ->
          check (less_nat Zero_nat sd && less_eq_nat sd n)
            (shows_string
              ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n',
                's', 'i', 'o', 'n'])))
    (mat n n,
      (mat0I (fract 0 1) n n,
        (mat1I (fract 0 1) (fract 1 1) n,
          (mat1I (fract 0 1)
             (get_comp_def
               ((\ _ -> True),
                 (fract 0 1,
                   (fract 1 1,
                     (d, (plus_rat,
                           (times_rat,
                             (max0_rat,
                               (ge_rat,
                                 (rat_gt d,
                                   (True,
                                     (rat_name,
                                       ((\ _ -> True), rat_mono)))))))))))))
             n,
            (mat_plusI plus_rat,
              (mat_multI (fract 0 1) plus_rat times_rat n,
                (mat_max0,
                  (mat_ge,
                    (mat_gtI
                       (get_comp_gt
                         ((\ _ -> True),
                           (fract 0 1,
                             (fract 1 1,
                               (d, (plus_rat,
                                     (times_rat,
                                       (max0_rat,
 (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))
                       sd,
                      (get_comp_mono
                         ((\ _ -> True),
                           (fract 0 1,
                             (fract 1 1,
                               (d, (plus_rat,
                                     (times_rat,
                                       (max0_rat,
 (ge_rat, (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))),
                        (concat
                           [shows_prec_nat Zero_nat n [], ['x'],
                             shows_prec_nat Zero_nat n [],
                             [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's', ' ',
                               'w', 'i', 't', 'h', ' ', 's', 't', 'r', 'i', 'c',
                               't', ' ', 'd', 'i', 'm', 'e', 'n', 's', 'i', 'o',
                               'n', ' '],
                             shows_prec_nat Zero_nat sd [],
                             [' ', 'o', 'v', 'e', 'r', ' '],
                             get_comp_description
                               ((\ _ -> True),
                                 (fract 0 1,
                                   (fract 1 1,
                                     (d, (plus_rat,
   (times_rat,
     (max0_rat,
       (ge_rat,
         (rat_gt d, (True, (rat_name, ((\ _ -> True), rat_mono))))))))))))],
                          ((\ _ -> True),
                            mat_mono sd
                              (get_comp_check_mono
                                ((\ _ -> True),
                                  (fract 0 1,
                                    (fract 1 1,
                                      (d,
(plus_rat,
  (times_rat,
    (max0_rat,
      (ge_rat,
        (rat_gt d,
          (True, (rat_name, ((\ _ -> True), rat_mono)))))))))))))))))))))))))
    i;
get_redpair (Arc_mat_carrier n i) =
  create_poly_redpair
    (\ _ ->
      binda (Inr ())
        (\ _ ->
          check (less_nat Zero_nat n)
            (shows_string
              ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n',
                's', 'i', 'o', 'n'])))
    (mat n n,
      (mat0I zero_arctic n n,
        (mat1I zero_arctic one_arctic n,
          (mat1I zero_arctic
             (get_comp_def
               ((\ _ -> True),
                 (zero_arctic,
                   (one_arctic,
                     (one_arctic,
                       (plus_arctic,
                         (times_arctic,
                           (id, (ge_arctic,
                                  (gt_arctic,
                                    (False,
                                      (arctic_name,
(pos_arctic, (\ _ -> False))))))))))))))
             n,
            (mat_plusI plus_arctic,
              (mat_multI zero_arctic plus_arctic times_arctic n,
                (id, (mat_ge,
                       (mat_comp_all
                          (get_comp_gt
                            ((\ _ -> True),
                              (zero_arctic,
                                (one_arctic,
                                  (one_arctic,
                                    (plus_arctic,
                                      (times_arctic,
(id, (ge_arctic,
       (gt_arctic,
         (False, (arctic_name, (pos_arctic, (\ _ -> False)))))))))))))),
                         (False,
                           (concat
                              [shows_prec_nat Zero_nat n [], ['x'],
                                shows_prec_nat Zero_nat n [],
                                [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                  ' ', 'o', 'v', 'e', 'r', ' '],
                                get_comp_description
                                  ((\ _ -> True),
                                    (zero_arctic,
                                      (one_arctic,
(one_arctic,
  (plus_arctic,
    (times_arctic,
      (id, (ge_arctic,
             (gt_arctic,
               (False, (arctic_name, (pos_arctic, (\ _ -> False)))))))))))))],
                             (mat_arc_posI
                                (get_comp_apos
                                  ((\ _ -> True),
                                    (zero_arctic,
                                      (one_arctic,
(one_arctic,
  (plus_arctic,
    (times_arctic,
      (id, (ge_arctic,
             (gt_arctic,
               (False, (arctic_name, (pos_arctic, (\ _ -> False)))))))))))))),
                               (\ _ -> False)))))))))))))
    i;
get_redpair (Arc_rat_mat_carrier d n i) =
  create_poly_redpair
    (\ _ ->
      binda (check (less_rat (fract 0 1) d)
              (shows_string
                ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'e', 'l', 't',
                  'a']))
        (\ _ ->
          check (less_nat Zero_nat n)
            (shows_string
              ['i', 'n', 'v', 'a', 'l', 'i', 'd', ' ', 'd', 'i', 'm', 'e', 'n',
                's', 'i', 'o', 'n'])))
    (mat n n,
      (mat0I zero_arctic_rat n n,
        (mat1I zero_arctic_rat one_arctic_rat n,
          (mat1I zero_arctic_rat
             (get_comp_def
               ((\ _ -> True),
                 (zero_arctic_rat,
                   (one_arctic_rat,
                     (one_arctic_rat,
                       (plus_arctic_rat,
                         (times_arctic_rat,
                           (id, (ge_arctic_rat,
                                  (gt_arctic_rat d,
                                    (False,
                                      (arctic_rat_name,
(pos_arctic_rat, (\ _ -> False))))))))))))))
             n,
            (mat_plusI plus_arctic_rat,
              (mat_multI zero_arctic_rat plus_arctic_rat times_arctic_rat n,
                (id, (mat_ge,
                       (mat_comp_all
                          (get_comp_gt
                            ((\ _ -> True),
                              (zero_arctic_rat,
                                (one_arctic_rat,
                                  (one_arctic_rat,
                                    (plus_arctic_rat,
                                      (times_arctic_rat,
(id, (ge_arctic_rat,
       (gt_arctic_rat d,
         (False, (arctic_rat_name, (pos_arctic_rat, (\ _ -> False)))))))))))))),
                         (False,
                           (concat
                              [shows_prec_nat Zero_nat n [], ['x'],
                                shows_prec_nat Zero_nat n [],
                                [' ', 'm', 'a', 't', 'r', 'i', 'c', 'e', 's',
                                  ' ', 'o', 'v', 'e', 'r', ' '],
                                get_comp_description
                                  ((\ _ -> True),
                                    (zero_arctic_rat,
                                      (one_arctic_rat,
(one_arctic_rat,
  (plus_arctic_rat,
    (times_arctic_rat,
      (id, (ge_arctic_rat,
             (gt_arctic_rat d,
               (False,
                 (arctic_rat_name, (pos_arctic_rat, (\ _ -> False)))))))))))))],
                             (mat_arc_posI
                                (get_comp_apos
                                  ((\ _ -> True),
                                    (zero_arctic_rat,
                                      (one_arctic_rat,
(one_arctic_rat,
  (plus_arctic_rat,
    (times_arctic_rat,
      (id, (ge_arctic_rat,
             (gt_arctic_rat d,
               (False,
                 (arctic_rat_name, (pos_arctic_rat, (\ _ -> False)))))))))))))),
                               (\ _ -> False)))))))))))))
    i;
get_redpair (Lpo prec pia) =
  af_redpair pia (create_LPO_redpair lpo_repr_to_pr prec);

matchCapBelowI ::
  forall a b.
    (Eq a, Eq b) => Term a b -> [(Term a b, Term a b)] -> Term a b -> Bool;
matchCapBelowI l r (Fun f ts) = gcMatchI (HFun f (map (capI r) ts)) l;

check_ur_closed_term_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      (a -> Nat -> Nat -> Bool) ->
                        Term a b -> Sum ([Char] -> [Char]) ();
check_ur_closed_term_af uu uv uw (Var x) = Inr ();
check_ur_closed_term_af r ur pia (Fun f ts) =
  let {
    n = size_list ts;
  } in binda (catcha
               (forallm_index
                 (\ t i ->
                   (if pia f n i then check_ur_closed_term_af r ur pia t
                     else Inr ()))
                 ts)
               (\ x -> Inl (snd x)))
         (\ _ ->
           catcha
             (forallm
               (\ lr ->
                 check (membera ur lr ||
                         not (matchCapBelowI (fst lr) r (Fun f ts)))
                   (shows_string
                      ['d', 'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ',
                        's', 'u', 'b', 't', 'e', 'r', 'm', ' '] .
                     shows_prec_term Zero_nat (Fun f ts) .
                       shows_string
                         [' ', 'o', 'f', ' ', 's', 'o', 'm', 'e', ' ', 'u', 's',
                           'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',', ' ',
                           'r', 'u', 'l', 'e', ' '] .
                         shows_rule lr .
                           shows_string
                             [' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.']))
               r)
             (\ x -> Inl (snd x)));

check_ur_P_closed_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        (a -> Nat -> Nat -> Bool) -> Sum ([Char] -> [Char]) ();
check_ur_P_closed_af r ur p pia =
  binda (catcha
          (catcha
            (forallm (\ lr -> check_ur_closed_term_af r ur pia (snd lr)) ur)
            (\ x -> Inl (snd x)))
          (\ x ->
            Inl (shows_string
                   ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
                     'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l', 'o', 's',
                     'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r', 't', 'i',
                     'e', 's', ' ', 'o', 'f', ' ', 'r', 'h', 's', ' ', 'o', 'f',
                     ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                     's'] .
                  shows_nl . x)))
    (\ _ ->
      catcha
        (catcha (forallm (\ st -> check_ur_closed_term_af r ur pia (snd st)) p)
          (\ x -> Inl (snd x)))
        (\ x ->
          Inl (shows_string
                 ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ', 'c',
                   'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l', 'o', 's',
                   'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r', 't', 'i',
                   'e', 's', ' ', 'o', 'f', ' ', 'r', 'h', 's', ' ', 'o', 'f',
                   ' ', 'D', 'P', 's'] .
                shows_nl . x)));

matchCapRMBelow ::
  forall a b.
    (Eq a,
      Eq b) => Term a b ->
                 Maybe ((a, Nat) -> [(Term a b, Term a b)]) -> Term a b -> Bool;
matchCapRMBelow l rm (Fun f ts) = gcMatchI (HFun f (map (capRM rm) ts)) l;

check_ur_closed_term_rm_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                    [(Term a b, Term a b)] ->
                      (a -> Nat -> Nat -> Bool) ->
                        Term a b -> Sum ([Char] -> [Char]) ();
check_ur_closed_term_rm_af uu uv uw (Var x) = Inr ();
check_ur_closed_term_rm_af rm ur pia (Fun f ts) =
  let {
    n = size_list ts;
  } in binda (catcha
               (forallm_index
                 (\ t i ->
                   (if pia f n i then check_ur_closed_term_rm_af rm ur pia t
                     else Inr ()))
                 ts)
               (\ x -> Inl (snd x)))
         (\ _ ->
           catcha
             (forallm
               (\ lr ->
                 check (membera ur lr ||
                         not (matchCapRMBelow (fst lr) (Just rm) (Fun f ts)))
                   (shows_string
                      ['d', 'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ',
                        's', 'u', 'b', 't', 'e', 'r', 'm', ' '] .
                     shows_prec_term Zero_nat (Fun f ts) .
                       shows_string
                         [' ', 'o', 'f', ' ', 's', 'o', 'm', 'e', ' ', 'u', 's',
                           'a', 'b', 'l', 'e', ' ', 'r', 'h', 's', ',', ' ',
                           'r', 'u', 'l', 'e', ' '] .
                         shows_rule lr .
                           shows_string
                             [' ', 's', 'h', 'o', 'u', 'l', 'd', ' ', 'b', 'e',
                               ' ', 'u', 's', 'a', 'b', 'l', 'e', '.']))
               (rm (f, n)))
             (\ x -> Inl (snd x)));

check_ur_P_closed_rm_af ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          (a -> Nat -> Nat -> Bool) ->
                            Sum ([Char] -> [Char]) ();
check_ur_P_closed_rm_af r rmo ur p pia =
  (case rmo of {
    Nothing -> check_ur_P_closed_af r ur p pia;
    Just rm ->
      binda (catcha
              (catcha
                (forallm (\ lr -> check_ur_closed_term_rm_af rm ur pia (snd lr))
                  ur)
                (\ x -> Inl (snd x)))
              (\ x ->
                Inl (shows_string
                       ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                         'c', 'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l',
                         'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e',
                         'r', 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'r', 'h',
                         's', ' ', 'o', 'f', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                         ' ', 'r', 'u', 'l', 'e', 's'] .
                      shows_nl . x)))
        (\ _ ->
          catcha
            (catcha
              (forallm (\ st -> check_ur_closed_term_rm_af rm ur pia (snd st))
                p)
              (\ x -> Inl (snd x)))
            (\ x ->
              Inl (shows_string
                     ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
                       'c', 'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ', 'c', 'l',
                       'o', 's', 'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e',
                       'r', 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'r', 'h',
                       's', ' ', 'o', 'f', ' ', 'D', 'P', 's'] .
                    shows_nl . x)));
  });

check_redpair_ur_af_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                            [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_redpair_ur_af_proc rp p rP r rm ur =
  catcha
    (let {
       (_, nSp) = partition (\ foo -> isOK (rp_strict rp foo)) p;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha (check_ur_P_closed_rm_af r rm ur p (rp_af rp))
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'c', 'l', 'o', 's', 'u',
                                 'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e', 'r',
                                 't', 'i', 'e', 's', ' ', 'o', 'f', ' ', 'u',
                                 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l',
                                 'e', 's', ' ', 'w', '.', 'r', '.', 't', '.',
                                 ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n', 't',
                                 ' ', 'f', 'i', 'l', 't', 'e', 'r'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha
                          (catcha (forallm (rp_non_strict rp) ur)
                            (\ x -> Inl (snd x)))
                          (\ x ->
                            Inl (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e',
                                     'n', 't', 'i', 'n', 'g', ' ', 'u', 's',
                                     'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l',
                                     'e', 's'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha
                              (catcha (forallm (rp_non_strict rp) nSp)
                                (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D',
 'P', 's'] .
                                      shows_nl . x)))
                        (\ _ ->
                          catcha (check_subseteq nSp rP)
                            (\ x ->
                              Inl (case rp_strict rp x of {
                                    Inl err -> err;
                                    Inr _ -> internal_error;
                                  })))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't',
               'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
               'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h',
               'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl .
              rp_desc rp .
                shows_nl .
                  shows_string
                    ['a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                      'o', 'w', 'i', 'n', 'g', ' ', 'u', 's', 'a', 'b', 'l',
                      'e', ' ', 'r', 'u', 'l', 'e', 's'] .
                    shows_nl . shows_rules ur . shows_nl . x));

check_dep_graph_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Bool, [(Term a b, Term a b)])] ->
                      Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                        Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                          Sum ([Char] -> [Char]) ();
check_dep_graph_proc p dps rm rrm =
  binda (catcha (check_subseteq p (concatMap snd dps))
          (\ x ->
            Inl (shows_string
                   ['D', 'e', 'p', 'e', 'n', 'd', 'e', 'n', 'c', 'y', ' ', 'P',
                     'a', 'i', 'r', ' '] .
                  shows_rule x .
                    shows_string
                      [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n', 'g',
                        ' ', 'i', 'n', ' ', 'd', 'e', 'c', 'o', 'm', 'p', 'o',
                        's', 'i', 't', 'i', 'o', 'n'] .
                      shows_nl)))
    (\ _ ->
      catcha
        (check_graph_decomp (shows_prec_prod Zero_nat . fst)
          (\ (a, b) ->
            let {
              (aa, ba) = a;
            } in let {
                   (_, t) = aa;
                 } in (\ (_, ct) (ab, bb) ->
                        let {
                          (u, _) = ab;
                        } in (\ (cu, _) -> gcMatchI ct u && gcMatchI cu t)
                          bb)
                   ba
              b)
          (map (\ (real, cs) ->
                 (real,
                   map (\ (s, t) -> ((s, t), (capRM rrm s, capRM rm t))) cs))
            dps))
        (\ x ->
          Inl (shows_string
                 ['o', 'u', 'r', ' ', 'e', 's', 't', 'i', 'm', 'a', 't', 'i',
                   'o', 'n', ' ', '(', 'E', 'D', 'G', '*', '*', '*', ')', ' ',
                   'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 's', 'h',
                   'o', 'w', ' ', 't', 'h', 'a', 't', ' ', 'y', 'o', 'u', ' ',
                   'h', 'a', 'v', 'e', ' ', 'a', ' ', 'v', 'a', 'l', 'i', 'd',
                   ' ', 'd', 'e', 'c', 'o', 'm', 'p', 'o', 's', 'i', 't', 'i',
                   'o', 'n', ' '] .
                shows_string
                  ['d', 'u', 'e', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'f',
                    'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a',
                    's', 'o', 'n'] .
                  shows_nl . x)));

check_allm_gen_index ::
  forall a b.
    (a -> Nat -> Nat) -> (a -> Nat -> Sum b ()) -> Nat -> [a] -> Sum b ();
check_allm_gen_index g f n xs =
  snd (foldla (\ (i, m) x -> (g x i, binda m (\ _ -> f x i))) (n, Inr ()) xs);

check_mono_redpair_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_mono_redpair_proc rp p rP r rR =
  catcha
    (let {
       (_, nSp) = partition (\ foo -> isOK (rp_strict rp foo)) p;
       (_, nSr) = partition (\ foo -> isOK (rp_strict rp foo)) r;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha (rp_mono rp)
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                                 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                                 'f', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                                 'o', 'r', 'd', 'e', 'r'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha
                          (catcha (forallm (rp_non_strict rp) nSr)
                            (\ x -> Inl (snd x)))
                          (\ x ->
                            Inl (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e',
                                     'n', 't', 'i', 'n', 'g', ' ', 'T', 'R',
                                     'S'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha
                              (catcha (forallm (rp_non_strict rp) nSp)
                                (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D',
 'P', 's'] .
                                      shows_nl . x)))
                        (\ _ ->
                          binda (catcha (check_subseteq rR r)
                                  (\ _ ->
                                    Inl (shows_string
  ['p', 'r', 'o', 'b', 'l', 'e', 'm', ':', ' ', 'a', 'd', 'd', 'i', 't', 'i',
    'o', 'n', 'a', 'l', ' '])))
                            (\ _ ->
                              binda (catcha (check_subseteq nSr rR)
                                      (\ x ->
Inl (case rp_strict rp x of {
      Inl err -> err;
      Inr _ -> internal_error;
    })))
                                (\ _ ->
                                  catcha (check_subseteq nSp rP)
                                    (\ x ->
                                      Inl
(case rp_strict rp x of {
  Inl err -> err;
  Inr _ -> internal_error;
})))))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't',
               'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
               'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h',
               'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl . rp_desc rp . shows_nl . x));

check_mono_redpair_ur_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                            [(Term a b, Term a b)] ->
                              [(Term a b, Term a b)] ->
                                Sum ([Char] -> [Char]) ();
check_mono_redpair_ur_proc rp p rP r rm rR ur =
  catcha
    (let {
       us = concatMap (funas_term_list . snd) (p ++ ur);
       filt = filter (\ lr -> wf_terms_list us (fst lr));
       (_, nSp) = partition (\ foo -> isOK (rp_strict rp foo)) p;
       (_, nSr) = partition (\ foo -> isOK (rp_strict rp foo)) ur;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha (rp_mono rp)
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                                 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                                 'f', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                                 'o', 'r', 'd', 'e', 'r'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha (check_ur_P_closed_rm_af r rm ur p pi_full)
                          (\ x ->
                            Inl (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'i', 't', 'h', ' ', 'c', 'l', 'o', 's',
                                     'u', 'r', 'e', ' ', 'p', 'r', 'o', 'p',
                                     'e', 'r', 't', 'i', 'e', 's', ' ', 'o',
                                     'f', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                                     ' ', 'r', 'u', 'l', 'e', 's'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha
                              (catcha (forallm (rp_non_strict rp) nSr)
                                (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u',
 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's'] .
                                      shows_nl . x)))
                        (\ _ ->
                          binda (catcha
                                  (catcha (forallm (rp_non_strict rp) nSp)
                                    (\ x -> Inl (snd x)))
                                  (\ x ->
                                    Inl (shows_string
   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o', 'r',
     'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'D', 'P', 's'] .
  shows_nl . x)))
                            (\ _ ->
                              binda (catcha (check_subseteq rR r)
                                      (\ _ ->
Inl (shows_string
       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ':', ' ', 'a', 'd', 'd', 'i', 't',
         'i', 'o', 'n', 'a', 'l', ' ', 'r', 'u', 'l', 'e', '(', 's', ')', ' ',
         't', 'o', ' ', 'T', 'R', 'S', ' ', 'a', 'd', 'd', 'e', 'd'] .
      shows_nl)))
                                (\ _ ->
                                  binda (catcha (check_subseteq (filt nSr) rR)
  (\ x ->
    Inl (case rp_strict rp x of {
          Inl err -> err;
          Inr _ -> internal_error;
        })))
                                    (\ _ ->
                                      catcha (check_subseteq (filt nSp) rP)
(\ x ->
  Inl (case rp_strict rp x of {
        Inl err -> err;
        Inr _ -> internal_error;
      }))))))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', '(', 'm', 'o', 'n', 'o', 't',
               'o', 'n', 'i', 'c', ')', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i',
               'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c', 'e',
               's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e',
               ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl .
              rp_desc rp .
                shows_nl .
                  shows_string
                    ['a', 'n', 'd', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                      'o', 'w', 'i', 'n', 'g', ' ', 'u', 's', 'a', 'b', 'l',
                      'e', ' ', 'r', 'u', 'l', 'e', 's'] .
                    shows_rules ur . shows_nl . x));

unlab_of_trs_list ::
  forall a b c.
    (Eq a, Eq b,
      Eq c) => [(Term (Lab a b) c, Term (Lab a b) c)] ->
                 [(Term (Lab a b) c, Term (Lab a b) c)];
unlab_of_trs_list r =
  remdups
    (map_trs_list unlab
      (filter
        (\ (l, ra) ->
          equal_term l ra ||
            not (equal_term (map_term unlab l) (map_term unlab ra)))
        r));

check_unlab_proc ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => [(Term (Lab a b) c, Term (Lab a b) c)] ->
                    [(Term (Lab a b) c, Term (Lab a b) c)] ->
                      [(Term (Lab a b) c, Term (Lab a b) c)] ->
                        [(Term (Lab a b) c, Term (Lab a b) c)] ->
                          Sum ([Char] -> [Char]) ();
check_unlab_proc p r ulP ulR =
  let {
    ulr = unlab_of_trs_list r;
  } in binda (catcha (check_subseteq (map_trs_list unlab p) ulP)
               (\ x ->
                 Inl (shows_string
                        ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'p',
                          'a', 'i', 'r', ' '] .
                       shows_rule x .
                         shows_prec_list Zero_nat
                           [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i', 'n',
                             'g'] .
                           shows_nl)))
         (\ _ ->
           binda (catcha (check_subseteq ulr ulR)
                   (\ x ->
                     Inl (shows_string
                            ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ',
                              'r', 'u', 'l', 'e', ' '] .
                           shows_rule x .
                             shows_prec_list Zero_nat
                               [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                 'n', 'g'] .
                               shows_nl)))
             (\ _ ->
               catcha (check_subseteq ulR ulr)
                 (\ x ->
                   Inl (shows_string ['r', 'u', 'l', 'e', ' '] .
                         shows_rule x .
                           shows_prec_list Zero_nat
                             [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', 'n',
                               ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd',
                               ' ', 'r', 'u', 'l', 'e'] .
                             shows_nl))));

get_largest_element :: forall a b. Sl_inter a b -> Nat;
get_largest_element (SL_Inter n uu) = n;

take_default :: forall a. a -> [a] -> Nat -> a;
take_default def [] uu = def;
take_default uv (x : xs) Zero_nat = x;
take_default def (uw : xs) (Suc i) = take_default def xs i;

eval_arithFun_unbound :: [Nat] -> ArithFun -> Nat;
eval_arithFun_unbound nats (Arg i) = take_default Zero_nat nats i;
eval_arithFun_unbound nats (Const n) = n;
eval_arithFun_unbound nats (Sum []) = Zero_nat;
eval_arithFun_unbound nats (Sum (f : fs)) =
  plus_nat (eval_arithFun_unbound nats f) (eval_arithFun_unbound nats (Sum fs));
eval_arithFun_unbound nats (Prod []) = one_nat;
eval_arithFun_unbound nats (Prod (f : fs)) =
  times_nat (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Prod fs));
eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (Max (f : v : va)) =
  maxa (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Max (v : va)));
eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f;
eval_arithFun_unbound nats (Min (f : v : va)) =
  mina (eval_arithFun_unbound nats f)
    (eval_arithFun_unbound nats (Min (v : va)));

eval_arithFun :: Nat -> ArithFun -> [Nat] -> Nat;
eval_arithFun c f nats = mod_nat (eval_arithFun_unbound nats f) c;

sl_inter_to_inter ::
  forall a b. (Eq a, Eq b) => Sl_inter a b -> Lab a b -> [Nat] -> Nat;
sl_inter_to_inter (SL_Inter c ls) fl cs =
  (case lookupb (fl, size_list cs) ls of {
    Nothing -> Zero_nat;
    Just f -> eval_arithFun (Suc c) f cs;
  });

shows_arithFun :: ArithFun -> [Char] -> [Char];
shows_arithFun (Arg i) =
  shows_string ['x', '_'] . shows_prec_nat Zero_nat (Suc i);
shows_arithFun (Const n) = shows_prec_nat Zero_nat n;
shows_arithFun (Sum fs) =
  shows_list_gen shows_arithFun [] [] [' ', '+', ' '] [] fs;
shows_arithFun (Prod fs) =
  shows_list_gen shows_arithFun [] [] [' ', '*', ' '] [] fs;
shows_arithFun (Min fs) =
  shows_list_gen shows_arithFun [] ['m', 'i', 'n', '('] [','] [')'] fs;
shows_arithFun (Max fs) =
  shows_list_gen shows_arithFun [] ['m', 'a', 'x', '('] [','] [')'] fs;

shows_prec_arithFun :: Nat -> ArithFun -> [Char] -> [Char];
shows_prec_arithFun d f = shows_arithFun f;

shows_sl_inter ::
  forall a b. (Showa a, Showa b) => Sl_inter a b -> [Char] -> [Char];
shows_sl_inter (SL_Inter n fnas) =
  shows_string
    ['c', 'a', 'r', 'r', 'i', 'e', 'r', ' ', '{', '0', ',', '.', '.', '.',
      ','] .
    shows_prec_nat Zero_nat n .
      shows_string ['}'] .
        shows_nl .
          shows_map
            (\ (a, b) ->
              let {
                (f, na) = a;
              } in (\ aa ->
                     shows_string ['['] .
                       shows_prec_lab Zero_nat f .
                         shows_string ['/'] .
                           shows_prec_nat Zero_nat na .
                             shows_string [']', ' ', '=', ' '] .
                               shows_prec_arithFun Zero_nat aa . shows_nl)
                b)
            fnas;

shows_prec_sl_inter ::
  forall a b. (Showa a, Showa b) => Nat -> Sl_inter a b -> [Char] -> [Char];
shows_prec_sl_inter d sli = shows_sl_inter sli;

qmodel_L ::
  forall a b c d.
    (Eq a, Eq b) => [(Lab a b, Nat)] -> Lab a b -> [c] -> Sum [c] d;
qmodel_L sig =
  (\ f cs -> (if membera sig (f, size_list cs) then Inl cs else Inl []));

enum_vectors_nat :: forall a. [a] -> Nat -> [[a]];
enum_vectors_nat c Zero_nat = [[]];
enum_vectors_nat c (Suc n) =
  let {
    a = enum_vectors_nat c n;
  } in concatMap (\ vec -> map (\ ca -> ca : vec) c) a;

qmodel_check_interpretation ::
  ArithFun -> Nat -> Nat -> Sum ([Char] -> [Char]) ();
qmodel_check_interpretation f n c =
  let {
    ca = upt Zero_nat (Suc c);
    css = enum_vectors_nat ca n;
  } in catcha
         (forallm
           (\ cs ->
             catcha
               (forallm
                 (\ i ->
                   catcha
                     (forallm
                       (\ l ->
                         check (less_eq_nat (eval_arithFun (Suc c) f cs)
                                 (eval_arithFun (Suc c) f (list_update cs i l)))
                           (shows_prec_list Zero_nat
                              ['n', 'o', 't', ' ', 'm', 'o', 'n', 'o', 't', 'o',
                                'n', 'e', ' ', 'i', 'n', ' '] .
                             shows_prec_nat Zero_nat (Suc i) .
                               shows_prec_list Zero_nat
                                 ['.', ' ', 'a', 'r', 'g', 'u', 'm', 'e', 'n',
                                   't']))
                       (upt (nth cs i) (Suc c)))
                     (\ x -> Inl (snd x)))
                 (upt Zero_nat n))
               (\ x -> Inl (snd x)))
           css)
         (\ x -> Inl (snd x));

qmodel_check_valid ::
  forall a b. (Showa a, Showa b) => Sl_inter a b -> Sum ([Char] -> [Char]) ();
qmodel_check_valid (SL_Inter c ls) =
  catcha
    (forallm
      (\ (a, b) ->
        let {
          (f, n) = a;
        } in (\ g ->
               catcha (qmodel_check_interpretation g n c)
                 (\ x ->
                   Inl (shows_prec_list Zero_nat
                          ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'i', 'n',
                            ' ', 'w', 'e', 'a', 'k', '-', 'm', 'o', 'n', 'o',
                            't', 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                            'f', ' ', 'i', 'n', 't', 'e', 'r', 'p', 'r', 'e',
                            't', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' '] .
                         shows_prec_lab Zero_nat f . shows_nl . x)))
          b)
      ls)
    (\ x -> Inl (snd x));

check_decr_present_aux_1 ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a (b, Nat), Term a (b, Nat))] ->
                 b -> a -> a -> Nat ->
                                  Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present_aux_1 r v f1 f2 n =
  let {
    vs = map (\ na -> Var (v, na)) (upt Zero_nat n);
    rule = (Fun f1 vs, Fun f2 vs);
  } in check (not (is_none (find (instance_rule rule) r))) rule;

check_decr_present_aux_2 ::
  forall a b.
    (Eq a,
      Eq b) => [(Term a b, Term a b)] ->
                 b -> [(a, (a, Nat))] ->
                        Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present_aux_2 r v req =
  let {
    add_nats = map_vars (\ va -> (va, Zero_nat));
    ra = map (\ (l, ra) -> (add_nats l, add_nats ra)) r;
  } in catcha
         (forallm (\ (f1, (a, b)) -> check_decr_present_aux_1 ra v f1 a b) req)
         (\ x -> Inl (snd x));

check_decr_present ::
  forall a b.
    (Eq a,
      Eq b) => [(a, Nat)] ->
                 (a -> [Nat] -> a) ->
                   b -> Nat ->
                          [(Term a b, Term a b)] ->
                            Sum (Term a (b, Nat), Term a (b, Nat)) ();
check_decr_present sig l v c r =
  let {
    ca = upt Zero_nat (Suc c);
    ls = (\ (f, n) ->
           concatMap
             (\ cs ->
               concatMap
                 (\ i ->
                   let {
                     ci = nth cs i;
                   } in (if less_nat ci c
                          then [(l f (list_update cs i (Suc ci)), (l f cs, n))]
                          else []))
                 (upt Zero_nat n))
             (enum_vectors_nat ca n));
  } in check_decr_present_aux_2 r v (concatMap ls sig);

qmodel_check_decr ::
  forall a b.
    (Eq a, Showa a,
      Eq b) => [(Lab a [Nat], Nat)] ->
                 b -> Nat ->
                        [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                          Sum ([Char] -> [Char]) ();
qmodel_check_decr sig v c =
  (\ lR ->
    catcha (check_decr_present sig Lab v c lR)
      (\ x ->
        Inl (let {
               display =
                 map_vars
                   (\ (_, n) ->
                     (shows_prec_list Zero_nat ['x'] .
                       shows_prec_nat Zero_nat n)
                       []);
             } in shows_prec_list Zero_nat
                    ['d', 'e', 'c', 'r', 'e', 'a', 's', 'i', 'n', 'g', ' ', 'r',
                      'u', 'l', 'e', ' '] .
                    shows_rule (display (fst x), display (snd x)) .
                      shows_prec_list Zero_nat
                        [' ', 'm', 'i', 's', 's', 'i', 'n', 'g'])));

qmodel_cge :: Nat -> Nat -> Bool;
qmodel_cge = (\ x y -> less_eq_nat y x);

pointwise_ext ::
  forall a. (a -> a -> (Bool, Bool)) -> [a] -> [a] -> (Bool, Bool);
pointwise_ext f [] [] = (False, True);
pointwise_ext f (uu : uv) [] = (False, False);
pointwise_ext f [] (uw : ux) = (False, False);
pointwise_ext f (a : asa) (b : bs) =
  (case f a b of {
    (stri, True) ->
      (case pointwise_ext f asa bs of {
        (strir, True) -> (stri || strir, True);
        (strir, False) -> (False, False);
      });
    (stri, False) -> (False, False);
  });

qmodel_lge :: forall a b c d. a -> b -> Sum [Nat] c -> Sum [Nat] d -> Bool;
qmodel_lge f n =
  (\ l r ->
    (case (l, r) of {
      (Inl cs1, Inl cs2) ->
        snd (pointwise_ext (\ x y -> (less_nat y x, less_eq_nat y x)) cs1 cs2);
      (Inl cs1, Inr _) -> False;
      (Inr _, b) -> False;
    }));

qmodel_LS ::
  forall a b c.
    (Eq a, Eq b, Eq c) => [(a, Nat)] -> [b] -> a -> Nat -> Sum [b] c -> Bool;
qmodel_LS sig ls =
  (\ f n ->
    (if membera sig (f, n) then set (map Inl (enum_vectors_nat ls n))
      else inserta (Inl []) bot_fun));

funas_root_term :: forall a b. Term a b -> [(a, Nat)];
funas_root_term (Fun f ts) = [(f, size_list ts)];
funas_root_term (Var uu) = [];

funas_root_rule :: forall a b. (Eq a) => (Term a b, Term a b) -> [(a, Nat)];
funas_root_rule (l, r) = remdups (funas_root_term l ++ funas_root_term r);

funas_root_rules :: forall a b. (Eq a) => [(Term a b, Term a b)] -> [(a, Nat)];
funas_root_rules r = remdups (concatMap funas_root_rule r);

check_sem_lab_root_proc ::
  forall a b c d e f g h i j k l.
    (Eq b, Showa b, Eq c, Showa c, Eq d, Showa d, Showa g, Showa h, Showa i,
      Showa j) => Sum ([Char] -> [Char]) a ->
                    ([(Term (Lab b c) d, Term (Lab b c) d)] ->
                      Sum ([Char] -> [Char]) e) ->
                      (f -> [(Term g h, Term i j)] ->
                              Sum ([Char] -> [Char]) k) ->
                        ([(Term (Lab b c) d, Term (Lab b c) d)] ->
                          [(Term (Lab b c) d, Term (Lab b c) d)] ->
                            Sum ([Char] -> [Char]) l) ->
                          (Lab b c ->
                            Nat ->
                              Sum c [Lab b c] -> Sum c [Lab b c] -> Bool) ->
                            (Lab b c -> Nat -> Sum c [Lab b c] -> Bool) ->
                              Maybe ((i, Nat) -> [(Term i j, Term i j)]) ->
                                [(Term g h, Term i j)] ->
                                  [(Term (Lab b c) d, Term (Lab b c) d)] ->
                                    f -> [(Term (Lab b c) d,
    Term (Lab b c) d)] ->
   Sum ([Char] -> [Char]) ();
check_sem_lab_root_proc valid check_decr check_lab check_model_lab lge ls rm p r
  lP lR =
  (case rm of {
    Nothing -> Inr ();
    Just irm ->
      binda valid
        (\ _ ->
          binda (catcha
                  (forallm
                    (\ (s, t) ->
                      binda (check_no_var s)
                        (\ _ ->
                          binda (check_no_var t)
                            (\ _ -> check_no_defined_root irm t)))
                    p)
                  (\ x -> Inl (snd x)))
            (\ _ ->
              binda (check_decr lR)
                (\ _ ->
                  binda (catcha (check_lab lP p)
                          (\ x ->
                            Inl (shows_string
                                   ['n', 'o', 't', ' ', 'a', 'l', 'l', ' ', 'l',
                                     'a', 'b', 'e', 'l', 'e', 'd', ' ', 'D',
                                     'P', 's', ' ', 'p', 'r', 'e', 's', 'e',
                                     'n', 't', ':'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha (check_model_lab lR r)
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'i', 't', 'h', ' ', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'T', 'R', 'S',
 ':'] .
                                      shows_nl . x)))
                        (\ _ ->
                          catcha (check_sl_only_lab_and_decr lge ls r lR)
                            (\ x -> Inl (x . shows_nl)))))));
  });

lab_root ::
  forall a b c d e.
    (a -> [b] -> b) ->
      (a -> [b] -> c) ->
        (a -> [b] -> c) ->
          (a -> Nat -> c -> d) -> (e -> b) -> Term a e -> Term d e;
lab_root i la l lc alpha (Fun f ts) =
  let {
    clts = map (eval_lab i la lc alpha) ts;
    cs = map fst clts;
    a = map snd clts;
  } in Fun (lc f (size_list ts) (l f cs)) a;
lab_root uu uv uw ux uy (Var x) = Var x;

check_sl_rule_all_ass ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          (c -> [c]) ->
                            [(Term d e, Term d e)] ->
                              (e -> b) ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_all_ass i lb la lc gen_smaller lR alpha (l, Fun f ts) =
  let {
    ll = lab_root i lb la lc alpha l;
    clts = map (eval_lab i lb lc alpha) ts;
    lts = map snd clts;
    ld = la f (map fst clts);
    n = size_list ts;
    small = gen_smaller ld;
  } in catcha
         (forallm
           (\ le ->
             check (membera lR (ll, Fun (lc f n le) lts))
               (shows_string
                  ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
                    ' '] .
                 shows_rule (ll, Fun (lc f n le) lts) .
                   shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']))
           small)
         (\ x -> Inl (snd x));
check_sl_rule_all_ass i lb la lc gen_smaller lR alpha (l, Var x) =
  let {
    ll = lab_root i lb la lc alpha l;
    lr = lab_root i lb la lc alpha (Var x);
  } in check (membera lR (ll, lr))
         (shows_string
            ['l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e', ' '] .
           shows_rule (ll, lr) .
             shows_string [' ', 'm', 'i', 's', 's', 'i', 'n', 'g']);

check_sl_rule_all ::
  forall a b c d e.
    (Showa a, Showa b, Eq d, Showa d, Eq e,
      Showa e) => (a -> [b] -> b) ->
                    (a -> [b] -> c) ->
                      (a -> [b] -> c) ->
                        (a -> Nat -> c -> d) ->
                          [b] ->
                            (c -> [c]) ->
                              [(Term d e, Term d e)] ->
                                (Term a e, Term a e) ->
                                  Sum ([Char] -> [Char]) ();
check_sl_rule_all i lb la lc c gen_smaller lR (l, r) =
  catcha
    (forallm
      (\ alpha -> check_sl_rule_all_ass i lb la lc gen_smaller lR alpha (l, r))
      (map assign_from_vector
        (enum_vectors c (list_union (vars_term_list l) (vars_term_list r)))))
    (\ x -> Inl (snd x));

check_sl_lab_all_trs ::
  forall a b c d.
    (Eq a, Showa a, Eq b, Showa b, Showa c, Eq d,
      Showa d) => (Lab a b -> [c] -> c) ->
                    (Lab a b -> [c] -> Sum b [Lab a b]) ->
                      (Lab a b -> [c] -> Sum b [Lab a b]) ->
                        [c] ->
                          (Sum b [Lab a b] -> [Sum b [Lab a b]]) ->
                            [(Term (Lab a b) d, Term (Lab a b) d)] ->
                              [(Term (Lab a b) d, Term (Lab a b) d)] ->
                                Sum ([Char] -> [Char]) ();
check_sl_lab_all_trs i la l c gen lP p =
  catcha (forallm (check_sl_rule_all i la l labela c gen lP) p)
    (\ x -> Inl (snd x));

qmodel_gen :: [Nat] -> [[Nat]];
qmodel_gen [] = [[]];
qmodel_gen (x : xs) =
  let {
    ys = upt Zero_nat (Suc x);
    zs = qmodel_gen xs;
  } in concatMap (\ y -> map (\ a -> y : a) zs) ys;

qmodel_gen2 :: forall a b. Sum [Nat] a -> [Sum [Nat] b];
qmodel_gen2 (Inl l) = map Inl (qmodel_gen l);

create_sem_lab_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Sl_inter a [Nat] ->
                    Sum (Lab a [Nat] -> [Nat] -> Sum [Nat] [Lab a [Nat]])
                      (b, Bool) ->
                      Maybe ((Lab a [Nat], Nat) ->
                              [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)]) ->
                        [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                          [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                            [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                              [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                                Sum ([Char] -> [Char]) ();
create_sem_lab_proc sli (Inl l) rm r1 r2 r3 r4 =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
  } in catcha
         (check_sem_lab_proc model_check_valid model_check_decr
           (check_sl_lab_trs i l ca model_cge)
           (check_sl_model_lab_trs i l ca model_cge) model_lge model_LS r1 r2 r3
           r4)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));
create_sem_lab_proc sli (Inr (v, False)) rm r1 r2 r3 r4 =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
    sig = funas_trs_list (r1 ++ r2);
    l = qmodel_L sig;
  } in catcha
         (check_sem_lab_proc (qmodel_check_valid sli)
           (qmodel_check_decr sig v c) (check_sl_lab_trs i l ca qmodel_cge)
           (check_sl_model_lab_trs i l ca qmodel_cge) qmodel_lge
           (qmodel_LS sig ca) r1 r2 r3 r4)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));
create_sem_lab_proc sli (Inr (v, True)) rm r1 r2 r3 r4 =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
    sig1 = list_union (funas_below_trs_list r1) (funas_trs_list r2);
    sig2 = funas_root_rules r1;
    l = qmodel_L sig1;
    la = qmodel_L sig2;
  } in catcha
         (check_sem_lab_root_proc (qmodel_check_valid sli)
           (qmodel_check_decr sig1 v c)
           (check_sl_lab_all_trs i l la ca qmodel_gen2)
           (check_sl_model_lab_trs i l ca qmodel_cge) qmodel_lge
           (qmodel_LS sig1 ca) rm r1 r2 r3 r4)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));

create_rseq_map ::
  forall a b.
    (Eq a,
      Eq b) => [((Term a b, Term a b),
                  [(Pos, ((Term a b, Term a b), Term a b))])] ->
                 (Term a b, Term a b) ->
                   Maybe [(Pos, ((Term a b, Term a b), Term a b))];
create_rseq_map rseqs r = lookupb r rseqs;

check_strict ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    ((Term a b, Term a b) ->
                      Maybe [(Pos, ((Term a b, Term a b), Term a b))]) ->
                      (a -> Nat) ->
                        (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_strict ra rseqm p r =
  let {
    s = proj_term p (fst r);
    t = proj_term p (snd r);
  } in (case rseqm r of {
         Nothing -> check_supt s t;
         Just rseq ->
           (if equal_nat (size_list rseq) Zero_nat then check_supt s t
             else let {
                    u = snd (snd (last rseq));
                  } in binda (check_rsteps ra rseq s u)
                         (\ _ -> check_supteq u t));
       });

check_weak ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat) ->
                    (Term a b, Term a b) -> Sum ([Char] -> [Char]) ();
check_weak p r =
  catcha
    (check (equal_term (proj_term p (fst r)) (proj_term p (snd r)))
      (shows_string
         ['t', 'h', 'e', ' ', 'p', 'r', 'o', 'j', 'e', 'c', 't', 'e', 'd', ' ',
           'l', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'q', 'u',
           'a', 'l', ' ', 't', 'o', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'o', 'j',
           'e', 'c', 't', 'e', 'd', ' ', 'r', 'h', 's'] .
        shows_nl))
    (\ x ->
      Inl (shows_string
             ['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
               'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '] .
            shows_rule r .
              shows_string [',', ' ', 's', 'i', 'n', 'c', 'e'] .
                shows_nl .
                  shows_prec_term Zero_nat (proj_term p (fst r)) .
                    shows_string [' ', '!', '=', ' '] .
                      shows_prec_term Zero_nat (proj_term p (snd r)) .
                        shows_nl . x));

check_subterm_criterion_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (a -> Nat) ->
                    ((Term a b, Term a b) ->
                      Maybe [(Pos, ((Term a b, Term a b), Term a b))]) ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Term a b, Term a b)] ->
                            Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                              Sum ([Char] -> [Char]) ();
check_subterm_criterion_proc pb rseqm pa p r Nothing = Inr ();
check_subterm_criterion_proc pb rseqm pa p r (Just rm) =
  binda (catcha
          (forallm
            (\ (l, ra) ->
              binda (check_no_var l)
                (\ _ ->
                  binda (check_no_var ra) (\ _ -> check_no_defined_root rm ra)))
            pa)
          (\ x -> Inl (snd x)))
    (\ _ ->
      binda (catcha (forallm (check_strict r rseqm pb) (list_diff pa p))
              (\ x -> Inl (snd x)))
        (\ _ -> catcha (forallm (check_weak pb) p) (\ x -> Inl (snd x))));

create_subterm_criterion_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ProjL a ->
                    [((Term a b, Term a b),
                       [(Pos, ((Term a b, Term a b), Term a b))])] ->
                      [(Term a b, Term a b)] ->
                        [(Term a b, Term a b)] ->
                          [(Term a b, Term a b)] ->
                            Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                              Sum ([Char] -> [Char]) ();
create_subterm_criterion_proc pa rseqs p rP r rm =
  let {
    pia = create_proj pa;
    rseqm = create_rseq_map rseqs;
  } in check_subterm_criterion_proc pia rseqm p rP r rm;

get_arg :: forall a b. Term a b -> Nat -> Term a b;
get_arg t Zero_nat = t;
get_arg t (Suc n) = nth (args t) n;

data Scg a b = Null | Scg a a [(b, b)] [(b, b)];

comp :: forall a. (Eq a) => [(a, a)] -> [(a, a)] -> [(a, a)];
comp esa es =
  remdups
    (concatMap
      (\ (x, y) ->
        concatMap (\ (ya, z) -> (if y == ya then [(x, z)] else [])) es)
      esa);

subtract_list :: forall a. (Eq a) => [a] -> [a] -> [a];
subtract_list xs ys = filter (\ x -> not (membera ys x)) xs;

scg_comp ::
  forall a b. (Eq b) => (a -> a -> Bool) -> Scg a b -> Scg a b -> Scg a b;
scg_comp conn (Scg pa qa stra wka) (Scg p q str wk) =
  (if not (conn qa p) then Null
    else let {
           strs = remdups (comp stra str ++ comp stra wk ++ comp wka str);
           a = subtract_list (remdups (comp wka wk)) strs;
         } in Scg pa q strs a);
scg_comp conn Null g = Null;
scg_comp conn (Scg v va vb vc) Null = Null;

equal_scg :: forall a b. (Eq a, Eq b) => Scg a b -> Scg a b -> Bool;
equal_scg (Scg pp1 pp2 list1 list2) Null = False;
equal_scg Null (Scg pp1 pp2 list1 list2) = False;
equal_scg (Scg pp1a pp2a list1a list2a) (Scg pp1 pp2 list1 list2) =
  pp1a == pp1 && (pp2a == pp2 && (list1a == list1 && list2a == list2));
equal_scg Null Null = True;

subsumes :: forall a b. (Eq a, Eq b) => Scg a b -> Scg a b -> Bool;
subsumes (Scg pa qa stra wka) (Scg p q str wk) =
  pa == p &&
    (qa == q && (all (membera str) stra && all (membera (str ++ wk)) wka));
subsumes g Null = True;
subsumes Null (Scg v va vb vc) = False;

insert_scg ::
  forall a b. (Eq a, Eq b) => Scg a b -> [Scg a b] -> Maybe [Scg a b];
insert_scg g gs =
  (if equal_scg g Null then Nothing
    else (if any (\ ga -> subsumes ga g) gs then Nothing
           else Just (g : filter (\ ga -> not (subsumes g ga)) gs)));

mk_trancl ::
  forall a b.
    (Eq a, Eq b) => (a -> a -> Bool) -> [Scg a b] -> [Scg a b] -> [Scg a b];
mk_trancl conn [] s = s;
mk_trancl conn (g : gs) s =
  (case insert_scg g s of {
    Nothing -> mk_trancl conn gs s;
    Just sa ->
      mk_trancl conn
        (map (scg_comp conn g) sa ++ map (\ f -> scg_comp conn f g) sa ++ gs)
        sa;
  });

in_situ :: forall a b. (Eq b) => Scg a b -> Bool;
in_situ Null = True;
in_situ (Scg p q str wk) = any (\ (a, b) -> a == b) str;

combine :: forall a b. (Eq b) => Scg a b -> Scg a b -> Scg a b;
combine (Scg pa qa stra wka) (Scg p q str wk) =
  Scg pa qa (remdups (stra ++ str)) (remdups (wka ++ wk));
combine Null s = Null;
combine (Scg v va vb vc) Null = Null;

sagiv :: forall a b. (Eq a, Eq b) => (a -> a -> Bool) -> Scg a b -> Bool;
sagiv conn g =
  (if in_situ g then True
    else let {
           gg = scg_comp conn g g;
         } in (if subsumes gg g then False else sagiv conn (combine g gg)));

check_SCT :: forall a b. (Eq a, Eq b) => (a -> a -> Bool) -> [Scg a b] -> Bool;
check_SCT conn gs = let {
                      a = mk_trancl conn gs [];
                    } in all (sagiv conn) a;

check_sct_subterm_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [((Term a b, Term a b), ([(Nat, Nat)], [(Nat, Nat)]))] ->
                      Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                        Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                          Sum ([Char] -> [Char]) ();
check_sct_subterm_proc p gs rm rrm =
  (case rm of {
    Nothing -> Inr ();
    Just irm ->
      catcha
        (binda
          (catcha (check_subseteq p (map fst gs))
            (\ x ->
              Inl (shows_string
                     ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o',
                       ' ', 's', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g',
                       'e', ' ', 'g', 'r', 'a', 'p', 'h', ' ', 'f', 'o', 'r',
                       ' ', 'D', 'P', ' '] .
                    shows_rule x . shows_nl)))
          (\ _ ->
            binda (catcha
                    (forallm
                      (\ (a, b) ->
                        let {
                          (s, t) = a;
                        } in (\ (stri, nstri) ->
                               catcha
                                 (binda (check_no_var s)
                                   (\ _ ->
                                     binda (check_no_var t)
                                       (\ _ ->
 binda (check_no_defined_root irm t)
   (\ _ ->
     let {
       m = num_args t;
       n = num_args s;
     } in binda (catcha
                  (forallm
                    (\ (i, j) ->
                      check (less_eq_nat i n &&
                              (less_eq_nat j m &&
                                isOK (check_supt (get_arg s i) (get_arg t j))))
                        (shows_string
                           ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i',
                             't', 'h', ' ', 'e', 'd', 'g', 'e', ' '] .
                          shows_prec_nat Zero_nat i .
                            shows_string [' ', '|', '>', ' '] .
                              shows_prec_nat Zero_nat j . shows_nl))
                    stri)
                  (\ x -> Inl (snd x)))
            (\ _ ->
              catcha
                (forallm
                  (\ (i, j) ->
                    check (less_eq_nat i n &&
                            (less_eq_nat j m &&
                              isOK (check_supteq (get_arg s i) (get_arg t j))))
                      (shows_string
                         ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't',
                           'h', ' ', 'e', 'd', 'g', 'e', ' '] .
                        shows_prec_nat Zero_nat i .
                          shows_string [' ', '|', '>', '=', ' '] .
                            shows_prec_nat Zero_nat j . shows_nl))
                  nstri)
                (\ x -> Inl (snd x)))))))
                                 (\ x ->
                                   Inl (shows_string
  ['p', 'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w', 'i', 't', 'h', ' ', 'D',
    'P', ' '] .
 shows_rule (s, t) . shows_nl . x)))
                          b)
                      gs)
                    (\ x -> Inl (snd x)))
              (\ _ ->
                check (check_SCT
                        (\ (_, (t, (_, ct))) (u, (_, (cu, _))) ->
                          gcMatchI ct u && gcMatchI cu t)
                        (map (\ (a, b) ->
                               let {
                                 (s, t) = a;
                               } in (\ (stri, nstri) ->
                                      let {
e = (s, (t, (capRM rrm s, capRM rm t)));
                                      } in Scg e e stri nstri)
                                 b)
                          gs))
                  (shows_string
                     ['s', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e',
                       ' ', 'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f',
                       'a', 'i', 'l', 'e', 'd'] .
                    shows_nl))))
        (\ x ->
          Inl (shows_string
                 ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                   'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e',
                   '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c',
                   'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 's',
                   'u', 'b', 't', 'e', 'r', 'm', '-', 'r', 'e', 'l', 'a', 't',
                   'i', 'o', 'n', ' ', 'f', 'o', 'r', ' ', 't', 'h', 'e', ' ',
                   'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e',
                   'a', 's', 'o', 'n'] .
                shows_nl . x));
  });

check_sct_ur_af_entry ::
  forall a b c d e f.
    (Eq a, Showa a, Eq b,
      Showa b) => ((a, Nat) -> [(Term a b, Term a b)]) ->
                    [(Term a b, Term a b)] ->
                      (a -> Nat -> Nat -> Bool) ->
                        ((Term a b, Term a b) -> Sum c d) ->
                          ((Term a b, Term a b) -> Sum e f) ->
                            Term a b ->
                              Term a b ->
                                [(Nat, Nat)] ->
                                  [(Nat, Nat)] -> Sum ([Char] -> [Char]) ();
check_sct_ur_af_entry rm ur pia sa ns s t stri nstri =
  catcha
    (binda (check_no_var s)
      (\ _ ->
        binda (check_no_var t)
          (\ _ ->
            binda (check_no_defined_root rm t)
              (\ _ ->
                let {
                  m = num_args t;
                  n = num_args s;
                } in binda (catcha
                             (forallm
                               (\ i ->
                                 check (less_eq_nat i n)
                                   (shows_string
                                      ['l', 'e', 'f', 't', '-', 'i', 'n', 'd',
'e', 'x', ' ', 't', 'o', ' ', 'l', 'a', 'r', 'g', 'e'] .
                                     shows_prec_nat Zero_nat i . shows_nl))
                               (remdups (map fst (stri ++ nstri))))
                             (\ x -> Inl (snd x)))
                       (\ _ ->
                         binda (catcha
                                 (forallm
                                   (\ j ->
                                     check (less_eq_nat j m &&
     isOK (check_ur_closed_term_rm_af rm ur pia (get_arg t j)))
                                       (shows_string
  ['r', 'i', 'g', 'h', 't', '-', 'i', 'n', 'd', 'e', 'x', ' ', 't', 'o', ' ',
    'l', 'a', 'r', 'g', 'e', ' ', 'o', 'r', ' ', 'a', 'r', 'g', 'u', 'm', 'e',
    'n', 't', ' ', 'v', 'i', 'o', 'l', 'a', 't', 'e', 's', ' ', 'u', 's', 'a',
    'b', 'l', 'e', '-', 'r', 'u', 'l', 'e', 's', ' ', 'c', 'o', 'n', 'd', 'i',
    't', 'i', 'o', 'n'] .
 shows_prec_nat Zero_nat j))
                                   (remdups (map snd (stri ++ nstri))))
                                 (\ x -> Inl (snd x)))
                           (\ _ ->
                             let {
                               _ = args s;
                               _ = args t;
                             } in binda (catcha
  (forallm
    (\ (i, j) ->
      check (isOK (sa (get_arg s i, get_arg t j)))
        (shows_string
           ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ',
             'e', 'd', 'g', 'e', ' '] .
          shows_prec_nat Zero_nat i .
            shows_string [' ', '-', 'S', '-', '>', ' '] .
              shows_prec_nat Zero_nat j))
    stri)
  (\ x -> Inl (snd x)))
                                    (\ _ ->
                                      catcha
(forallm
  (\ (i, j) ->
    check (isOK (ns (get_arg s i, get_arg t j)))
      (shows_string
         ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h', ' ', 'e',
           'd', 'g', 'e', ' '] .
        shows_prec_nat Zero_nat i .
          shows_string [' ', '-', 'N', 'S', '-', '>', ' '] .
            shows_prec_nat Zero_nat j))
  nstri)
(\ x -> Inl (snd x)))))))))
    (\ x ->
      Inl (shows_string
             ['p', 'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w', 'i', 't', 'h',
               ' ', 'D', 'P', ' '] .
            shows_rule (s, t) . shows_nl . x));

check_sct_ur_af_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [((Term a b, Term a b), ([(Nat, Nat)], [(Nat, Nat)]))] ->
                        Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                          Maybe ((a, Nat) -> [(Term a b, Term a b)]) ->
                            [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_sct_ur_af_proc rp p gs rm rrm ur =
  (case rm of {
    Nothing -> Inr ();
    Just irm ->
      catcha
        (let {
           ggs = filter (\ g -> membera p (fst g)) gs;
           pia = rp_af rp;
           s = rp_strict rp;
           ns = rp_non_strict rp;
         } in binda (rp_valid rp)
                (\ _ ->
                  binda (catcha
                          (catcha
                            (forallm
                              (\ lr ->
                                check_ur_closed_term_rm_af irm ur pia (snd lr))
                              ur)
                            (\ x -> Inl (snd x)))
                          (\ x ->
                            Inl (shows_string
                                   ['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                                     'n', ' ', 'c', 'h', 'e', 'c', 'k', 'i',
                                     'n', 'g', ' ', 'c', 'l', 'o', 's', 'u',
                                     'r', 'e', ' ', 'p', 'r', 'o', 'p', 'e',
                                     'r', 't', 'i', 'e', 's', ' ', 'o', 'f',
                                     ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ',
                                     'u', 's', 'a', 'b', 'l', 'e', ' ', 'r',
                                     'u', 'l', 'e', 's'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha (check_subseteq p (map fst ggs))
                              (\ x ->
                                Inl (shows_string
                                       ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's',
 ' ', 'n', 'o', ' ', 's', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ',
 'g', 'r', 'a', 'p', 'h', ' ', 'f', 'o', 'r', ' ', 'D', 'P', ' '] .
                                      shows_rule x)))
                        (\ _ ->
                          binda (catcha
                                  (forallm
                                    (\ (a, b) ->
                                      let {
(sa, t) = a;
                                      } in
(\ (aa, ba) -> check_sct_ur_af_entry irm ur pia s ns sa t aa ba)
b)
                                    ggs)
                                  (\ x -> Inl (snd x)))
                            (\ _ ->
                              binda (catcha
                                      (catcha (forallm (rp_non_strict rp) ur)
(\ x -> Inl (snd x)))
                                      (\ x ->
Inl (shows_string
       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ', 'o',
         'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'u', 's', 'a', 'b', 'l',
         'e', ' ', 'r', 'u', 'l', 'e', 's'] .
      shows_nl . x)))
                                (\ _ ->
                                  check (check_SCT
  (\ (_, (t, (_, ct))) (u, (_, (cu, _))) -> gcMatchI ct u && gcMatchI cu t)
  (map (\ (a, b) ->
         let {
           (sa, t) = a;
         } in (\ (stri, nstri) ->
                let {
                  e = (sa, (t, (capRM rrm sa, capRM rm t)));
                } in Scg e e stri nstri)
           b)
    ggs))
                                    (shows_string
                                       ['s', 'i', 'z', 'e', '-', 'c', 'h', 'a',
 'n', 'g', 'e', ' ', 'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f', 'a', 'i',
 'l', 'e', 'd'] .
                                      shows_nl)))))))
        (\ x ->
          Inl (shows_string
                 ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p',
                   'p', 'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e',
                   '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c',
                   'e', 's', 's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't',
                   'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g',
                   ' '] .
                shows_nl .
                  rp_desc rp .
                    shows_nl .
                      shows_string
                        ['f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l',
                          'l', 'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a', 's',
                          'o', 'n'] .
                        shows_nl . x));
  });

checkDPProof ::
  forall a.
    (Eq a,
      Showa a) => ([Char] -> [Char]) ->
                    [(Term (Lab a [Nat]) [Char], Term (Lab a [Nat]) [Char])] ->
                      [(Term (Lab a [Nat]) [Char],
                         Term (Lab a [Nat]) [Char])] ->
                        Nat ->
                          Maybe ((Lab a [Nat], Nat) ->
                                  [(Term (Lab a [Nat]) [Char],
                                     Term (Lab a [Nat]) [Char])]) ->
                            Maybe ((Lab a [Nat], Nat) ->
                                    [(Term (Lab a [Nat]) [Char],
                                       Term (Lab a [Nat]) [Char])]) ->
                              DPProof a [Nat] [Char] ->
                                Sum ([Char] -> [Char]) ();
checkDPProof i p r m rm rrm PisEmpty =
  debug (i []) ['P', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (case p of {
      [] -> Inr ();
      _ : _ ->
        Inl (i . shows_string
                   [':', ' ', 'P', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e',
                     'm', 'p', 't', 'y', ' ', 'i', 'n', ' ', 't', 'h', 'e', ' ',
                     'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ', 'D', 'P',
                     '-', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                   shows_nl . shows_dpp p r);
    });
checkDPProof i p r m rm rrm (RedPairProc redp rP prf) =
  debug (i []) ['R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
    (binda
      (catcha (check_redpair_proc (get_redpair redp) p rP r)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i',
                       'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o',
                       'c', 'e', 's', 's', 'o', 'r', ' ', 't', 'o', ' ', 's',
                       'w', 'i', 't', 'c', 'h', ' ', 'f', 'r', 'o', 'm'] .
                     shows_nl .
                       shows_dpp p r .
                         shows_nl .
                           shows_string
                             ['t', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P', ' ',
                               'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                             shows_nl . shows_dpp rP r . shows_nl . x)))
      (\ _ ->
        catcha (checkDPProof (i . shows_string ['.', '1']) rP r m rm rrm prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
                         'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ',
                         'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (RedPairUrProc redp rP ur prf) =
  debug (i [])
    ['R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c', ' ', '(', 'u', 's',
      'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ' ', '+', ' ', 'a', 'f',
      ')']
    (binda
      (catcha (check_redpair_ur_af_proc (get_redpair redp) p rP r rm ur)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i',
                       'o', 'n', ' ', 'p', 'a', 'i', 'r', ' '] .
                     shows_string
                       ['p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', '(',
                         'w', 'i', 't', 'h', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                         ' ', 'r', 'u', 'l', 'e', 's', ' ', 'a', 'n', 'd', ' ',
                         'a', 'f', ')', ' ', 't', 'o', ' ', 's', 'w', 'i', 't',
                         'c', 'h', ' ', 'f', 'r', 'o', 'm', ' '] .
                       shows_nl .
                         shows_dpp p r .
                           shows_nl .
                             shows_string
                               ['t', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                 ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                               shows_nl . shows_dpp rP r . shows_nl . x)))
      (\ _ ->
        catcha (checkDPProof (i . shows_string ['.', '1']) rP r m rm rrm prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'd', 'u',
                         'c', 't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ',
                         'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (DepGraphProc edpts) =
  debug (i []) ['D', 'e', 'p', 'G', 'r', 'a', 'p', 'h', 'P', 'r', 'o', 'c']
    (binda
      (catcha
        (check_dep_graph_proc p
          (map (\ x ->
                 ((case fst x of {
                    Nothing -> False;
                    Just _ -> True;
                  }),
                   snd x))
            edpts)
          rm rrm)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i',
                       'l', 'e', ' ', 't', 'r', 'y', 'i', 'n', 'g', ' ', 't',
                       'o', ' ', 'p', 'e', 'r', 'f', 'o', 'r', 'm', ' ', 'S',
                       'C', 'C', '-', 'd', 'e', 'c', 'o', 'm', 'p', 'o', 's',
                       'i', 't', 'i', 'o', 'n', ' ', ' ', 'o', 'n', ' '] .
                     shows_nl . shows_dpp p r . shows_nl . x)))
      (\ _ ->
        catcha
          (check_allm_gen_index
            (\ rp j -> (case fst rp of {
                         Nothing -> j;
                         Just _ -> Suc j;
                       }))
            (\ rp j ->
              (case fst rp of {
                Nothing -> Inr ();
                Just a ->
                  checkDPProof
                    (i . shows_string ['.'] . shows_prec_nat Zero_nat j)
                    (snd rp) r m rm rrm a;
              }))
            one_nat edpts)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'd', 'e', 'p', 'e',
                         'n', 'd', 'e', 'n', 'c', 'y', ' ', 'g', 'r', 'a', 'p',
                         'h', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                         'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (MonoRedPairProc redp rP rR prf) =
  debug (i [])
    ['M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c']
    (binda
      (catcha (check_mono_redpair_proc (get_redpair redp) p rP r rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n',
                       'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                       'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p', 'r', 'o', 'c',
                       'e', 's', 's', 'o', 'r', ' '] .
                     shows_string
                       ['t', 'o', ' ', 's', 'w', 'i', 't', 'c', 'h', ' ', 'f',
                         'r', 'o', 'm', ' '] .
                       shows_nl .
                         shows_dpp p r .
                           shows_nl .
                             shows_string
                               ['t', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                 ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm', ' '] .
                               shows_nl . shows_dpp rP rR . shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) rP rR m
            (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o',
                         't', 'o', 'n', 'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c',
                         't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' ', 'p',
                         'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (MonoRedPairUrProc redp rP rR ur prf) =
  debug (i [])
    ['M', 'o', 'n', 'o', 'R', 'e', 'd', 'P', 'a', 'i', 'r', 'P', 'r', 'o', 'c',
      ' ', '(', 'u', 's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e', 's', ')']
    (binda
      (catcha (check_mono_redpair_ur_proc (get_redpair redp) p rP r rm rR ur)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'm', 'o', 'n', 'o', 't', 'o', 'n',
                       'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o',
                       'n', ' ', 'p', 'a', 'i', 'r', ' '] .
                     shows_string
                       ['p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r', ' ', '(',
                         'w', 'i', 't', 'h', ' ', 'u', 's', 'a', 'b', 'l', 'e',
                         ' ', 'r', 'u', 'l', 'e', 's', ')', ' ', 't', 'o', ' ',
                         's', 'w', 'i', 't', 'c', 'h', ' ', 'f', 'r', 'o', 'm',
                         ' '] .
                       shows_nl .
                         shows_dpp p r .
                           shows_nl .
                             shows_string
                               ['t', 'o', ' ', 't', 'h', 'e', ' ', 'D', 'P',
                                 ' ', 'p', 'r', 'o', 'b', 'l', 'e', 'm'] .
                               shows_nl . shows_dpp rP rR . shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) rP rR m
            (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'm', 'o', 'n', 'o',
                         't', 'o', 'n', 'i', 'c', ' ', 'r', 'e', 'd', 'u', 'c',
                         't', 'i', 'o', 'n', ' ', 'p', 'a', 'i', 'r', ' '] .
                       shows_string
                         ['(', 'w', 'i', 't', 'h', ' ', 'u', 's', 'a', 'b', 'l',
                           'e', ' ', 'r', 'u', 'l', 'e', 's', ')', ' ', 'p',
                           'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                         shows_nl . indent x))));
checkDPProof i p r (Suc m) rm rrm (UnlabProc ulP ulR prf) =
  debug (i []) ['U', 'n', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
    (binda
      (catcha (check_unlab_proc p r ulP ulR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l',
                       'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                       'o', 'r'] .
                     shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) ulP ulR m
            (computeRuleMap ulR) (computeRuleMap (reverse_trs ulR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'l', 'a',
                         'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
                         'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkDPProof i p r Zero_nat rm rrm (UnlabProc ulP ulR prf) =
  Inl (i . shows_string
             [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n', ' ',
               'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ', 't', 'h', 'e', ' ',
               'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r',
               'o', 'c', 'e', 's', 's', 'o', 'r'] .
             shows_nl .
               shows_string
                 ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'n',
                   'o', 't', ' ', 'a', 'l', 'l', 'o', 'w', 'e', 'd', ' ', 'd',
                   'u', 'e', ' ', 't', 'o', ' ', 'l', 'a', 'b', 'e', 'l', '-',
                   'd', 'e', 'p', 't', 'h', ' ', 'o', 'f', ' ', '0'] .
                 shows_nl);
checkDPProof i p r m rm rrm (SemlabProc sli l lP lR prf) =
  debug (i []) ['S', 'e', 'm', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
    (binda
      (catcha (create_sem_lab_proc sli l rm p r lP lR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i',
                       'c', ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ',
                       'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                     shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) lP lR (Suc m)
            (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'e', 'm', 'a',
                         'n', 't', 'i', 'c', ' ', 'l', 'a', 'b', 'e', 'l', 'i',
                         'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                         'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (RootlabProc delt_opt lP lR prf) =
  debug (i []) ['R', 'o', 'o', 't', 'l', 'a', 'b', 'P', 'r', 'o', 'c']
    (binda
      (catcha (rl_proc delt_opt p r lP lR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'r', 'o', 'o', 't', ' ', 'l', 'a',
                       'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
                       'e', 's', 's', 'o', 'r'] .
                     shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) lP lR (Suc m)
            (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'o', 'o', 't',
                         '-', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ', 'p',
                         'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkDPProof i pa r m rm rrm (SubtermCriterionProc p rseq rP prf) =
  debug (i []) ['S', 'u', 'b', 't', 'e', 'r', 'm', 'P', 'r', 'o', 'c']
    (binda
      (catcha (create_subterm_criterion_proc p rseq pa rP r rm)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm',
                       ' ', 'c', 'r', 'i', 't', 'e', 'r', 'i', 'o', 'n', ' ',
                       'p', 'r', 'o', 'c', 'e', 's', 's', 'o', 'r'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkDPProof (i . shows_string ['.', '1']) rP r m rm rrm prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 's', 'u', 'b', 't',
                         'e', 'r', 'm', ' ', 'c', 'r', 'i', 't', 'e', 'r', 'i',
                         'o', 'n', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                         'r'] .
                       shows_nl . indent x))));
checkDPProof i p r m rm rrm (SizeChangeSubTerm graphs) =
  debug (i [])
    ['S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'S', 'u', 'b', 'T', 'e',
      'r', 'm']
    (check_sct_subterm_proc p graphs rm rrm);
checkDPProof i p r m rm rrm (SizeChangeRedPair redp ur graphs) =
  debug (i [])
    ['S', 'i', 'z', 'e', 'C', 'h', 'a', 'n', 'g', 'e', 'R', 'e', 'd', 'P', 'a',
      'i', 'r']
    (check_sct_ur_af_proc (get_redpair redp) p graphs rm rrm ur);
checkDPProof i p r m rm rrm (FccProc f fcs rP rR prf) =
  debug (i []) ['F', 'c', 'c', 'P', 'r', 'o', 'c']
    (binda
      (catcha (fcc_proc m p r f fcs rP rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e',
                       ' ', 'u', 'n', 'd', 'e', 'r', ' ', 'f', 'l', 'a', 't',
                       ' ', 'c', 'o', 'n', 't', 'e', 'x', 't', 's'] .
                     shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) rP rR m
            (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n',
                         't', 'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r',
                         'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's', 'o',
                         'r'] .
                       shows_nl . indent x))));
checkDPProof i pa ra m rm rrm (UncurryProc mode u_info p r prf) =
  debug (i []) ['U', 'n', 'c', 'u', 'r', 'r', 'y', 'P', 'r', 'o', 'c']
    (binda
      (catcha (uncurry_proc_both mode u_info pa ra p r rm)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       't', 'h', 'e', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y',
                       'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                       'o', 'r'] .
                     shows_nl . x)))
      (\ _ ->
        catcha
          (checkDPProof (i . shows_string ['.', '1']) p r Zero_nat
            (computeRuleMap r) (computeRuleMap (reverse_trs r)) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'c', 'u',
                         'r', 'r', 'y', 'i', 'n', 'g', ' ', 'p', 'r', 'o', 'c',
                         'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));

bounds_condition ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => Boundstype ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
bounds_condition Roof uu = Inr ();
bounds_condition Match r =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (_, a) = x;
              } in linear_term a
            then Inr () else Inl x))
        r)
      (\ x -> Inl (snd x)))
    (\ x ->
      Inl (let {
             (l, ra) = x;
           } in shows_prec_list Zero_nat ['r', 'u', 'l', 'e', ' '] .
                  shows_rule (l, ra) .
                    shows_prec_list Zero_nat
                      [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'r', 'i', 'g',
                        'h', 't', '-', 'l', 'i', 'n', 'e', 'a', 'r', ' ']));

ta_rules_impl ::
  forall a b c. Ta_impl_ext a b c -> Rbt (b, Nat) [Ta_rule_impl a b];
ta_rules_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    more)
  = ta_rules_impl;

ta_final_impl :: forall a b c. Ta_impl_ext a b c -> Rbt a ();
ta_final_impl
  (Ta_impl_ext ta_final_impl ta_rules_impl ta_rqs_impl ta_rqs_set ta_eps_impl
    more)
  = ta_final_impl;

ta_contains_impl ::
  forall a b.
    (Eq a, Linorder a, Eq b,
      Linorder b) => [(a, Nat)] ->
                       Ta_impl_ext b a () -> [b] -> Sum (Sum b (Term a b)) ();
ta_contains_impl f ta qs =
  let {
    fin = ta_final_impl ta;
    look = rm_set_lookup (ta_rules_impl ta);
    _ = ta_eps_impl ta;
  } in binda (catcha
               (catcha (forallm (\ q -> check (rs_memb q fin) q) qs)
                 (\ x -> Inl (snd x)))
               (\ x -> Inl (Inl x)))
         (\ _ ->
           catcha
             (catcha
               (forallm
                 (\ (fa, n) ->
                   let {
                     rules = look (fa, n);
                   } in catcha
                          (catcha
                            (forallm
                              (\ x ->
                                (if any (\ rule ->
  rqs_impl rule == x &&
    let {
      qq = rq_impl rule;
    } in any (\ q -> rs_memb q qq) qs)
                                      rules
                                  then Inr () else Inl x))
                              (concat_lists (replicatea n qs)))
                            (\ x -> Inl (snd x)))
                          (\ x -> Inl (Fun fa (map Var x))))
                 f)
               (\ x -> Inl (snd x)))
             (\ x -> Inl (Inr x)));

ta_rule_compatible_list ::
  forall a b c.
    (Linorder a, Eq b, Linorder b,
      Eq c) => Ta_impl_ext a b () ->
                 ((Term b c, Term b c), [c]) ->
                   Sum ((Term b a, Term b a), a) ();
ta_rule_compatible_list ta =
  let {
    rm = ta_rules_impl ta;
    eps = ta_eps_impl ta;
  } in (\ (a, b) ->
         let {
           (l, r) = a;
         } in (\ xs ->
                catcha
                  (forallm
                    (\ vec ->
                      let {
                        tau = assign_from_vector vec;
                        l_tau = map_vars tau l;
                        r_tau = map_vars tau r;
                        rhs = rs_alpha (ta_res_list rm eps r_tau);
                      } in catcha
                             (forallm
                               (\ q -> check (member q rhs) ((l_tau, r_tau), q))
                               (rs_to_list (ta_res_list rm eps l_tau)))
                             (\ x -> Inl (snd x)))
                    (enum_vectors (ta_rqs_impl ta) xs))
                  (\ x -> Inl (snd x)))
           b);

ta_trs_compatible_list ::
  forall a b c.
    (Linorder a, Eq b, Linorder b,
      Eq c) => Ta_impl_ext a b () ->
                 [((Term b c, Term b c), [c])] ->
                   Sum ((Term b c, Term b c), ((Term b a, Term b a), a)) ();
ta_trs_compatible_list ta r =
  let {
    check = ta_rule_compatible_list ta;
  } in catcha (forallm (\ lr -> catcha (check lr) (\ x -> Inl (fst lr, x))) r)
         (\ x -> Inl (snd x));

flatten_term_enum_filter ::
  forall a b. (Term a b -> Bool) -> Term [a] b -> [Term a b];
flatten_term_enum_filter f (Var x) =
  let {
    tx = Var x;
  } in (if f tx then [tx] else []);
flatten_term_enum_filter f (Fun fs ts) =
  let {
    lts = map (flatten_term_enum_filter f) ts;
  } in (if any null lts then []
         else let {
                ss = concat_lists lts;
              } in filter f (concatMap (\ fa -> map (Fun fa) ss) fs));

inverse_base_term_filter ::
  forall a b. (Term (a, Nat) b -> Bool) -> Term a b -> Nat -> [Term (a, Nat) b];
inverse_base_term_filter filt l c =
  let {
    hs = upt Zero_nat (Suc c);
  } in flatten_term_enum_filter filt
         (map_term (\ f -> map (\ h -> lift h f) hs) l);

cover_bound_list_filter ::
  forall a b.
    (Eq b) => (Term (a, Nat) b -> Bool) ->
                ((Term a b, Term a b) -> Term a b -> Bool) ->
                  Nat ->
                    [(Term a b, Term a b)] ->
                      [((Term (a, Nat) b, Term (a, Nat) b), [b])];
cover_bound_list_filter filt ff c r =
  concatMap
    (\ (l, ra) ->
      let {
        ee = ff (l, ra);
        xs = vars_term_list l;
      } in map (\ la ->
                 ((la, map_term
                         (lift (Suc (min_list
                                      (map height
(sym_collect (\ t -> ee (map_term base t)) la)))))
                         ra),
                   xs))
             (inverse_base_term_filter filt l c))
    r;

ta_res_list_all ::
  forall a b c.
    (Linorder a, Eq b,
      Linorder b) => Rbt a () ->
                       Rbt (b, Nat) [Ta_rule_impl a b] -> Term b c -> Rbt a ();
ta_res_list_all q ta (Var uu) = q;
ta_res_list_all q ta (Fun f ts) =
  let {
    rec = map (ta_res_list_all q ta) ts;
    n = size_list ts;
    rules = rm_set_lookup ta (f, n);
    arules =
      filter
        (\ rule ->
          let {
            qs = rqs_impl rule;
          } in all_interval_nat (\ i -> rs_memb (nth qs i) (nth rec i)) Zero_nat
                 n)
        rules;
    a = map rq_impl arules;
  } in foldla rs_union rs_empty a;

ta_rule_compatible_heuristic ::
  forall a b c.
    (Eq a, Linorder a, Eq b,
      Linorder b) => Ta_impl_ext a b () -> Term b c -> Bool;
ta_rule_compatible_heuristic ta l =
  rs_isEmpty (ta_res_list_all (ta_rqs_set ta) (ta_rules_impl ta) l);

boundstype_fun ::
  forall a b. (Eq b) => Boundstype -> (Term a b, Term a b) -> Term a b -> Bool;
boundstype_fun Roof = roof;
boundstype_fun Match = match;

check_bounds ::
  forall a b.
    (Eq a, Linorder a, Showa a, Eq b, Linorder b,
      Showa b) => Bounds_info a b ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_bounds (Bounds_Info typea c qfin preTA) r =
  let {
    ta = generate_ta preTA;
  } in binda (check_wf_trs r)
         (\ _ ->
           binda (check_left_linear_trs r)
             (\ _ ->
               binda (bounds_condition typea r)
                 (\ _ ->
                   binda (check (any (\ q -> rs_memb q (ta_rqs_set ta)) qfin)
                           (shows_prec_list Zero_nat
                             ['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                               'n', 'd', ' ', 'm', 'e', 'n', 't', 'i', 'o', 'n',
                               'e', 'd', ' ', 'f', 'i', 'n', 'a', 'l', ' ', 's',
                               't', 'a', 't', 'e', ' ', 'i', 'n', ' ', 'T',
                               'A']))
                     (\ _ ->
                       binda (catcha
                               (catcha
                                 (forallm
                                   (\ x ->
                                     (if let {
   (f, _) = x;
 } in less_eq_nat (height f) c
                                       then Inr () else Inl x))
                                   (map fst (rm_to_list (ta_rules_impl ta))))
                                 (\ x -> Inl (snd x)))
                               (\ x ->
                                 Inl (let {
(f, _) = x;
                                      } in
shows_prec_prod Zero_nat f .
  shows_prec_list Zero_nat
    [' ', 'i', 's', ' ', 's', 'y', 'm', 'b', 'o', 'l', ' ', 'i', 'n', ' ', 'T',
      'A', ' ', 'w', 'i', 't', 'h', ' ', 'h', 'e', 'i', 'g', 'h', 't', ' ', 'l',
      'a', 'r', 'g', 'e', 'r', ' ', 't', 'h', 'a', 'n', ' ', 'c', ' ', '=',
      ' '] .
    shows_prec_nat Zero_nat c)))
                         (\ _ ->
                           binda (catcha
                                   (ta_contains_impl
                                     (map (\ (f, a) -> (lift Zero_nat f, a))
                                       (funas_trs_list r))
                                     ta qfin)
                                   (\ x ->
                                     Inl (shows_prec_list Zero_nat
    ['i', 't', ' ', 'c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'b', 'e',
      ' ', 'g', 'u', 'a', 'r', 'a', 'n', 't', 'e', 'e', 'd', ' ', 't', 'h', 'a',
      't', ' ', 'l', 'i', 'f', 't', '0', '(', 'T', '(', 'S', 'i', 'g', 'm', 'a',
      ')', ')', ' ', 'i', 's', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd', ' ',
      'b', 'y', ' ', 'T', 'A'] .
   shows_nl .
     (case x of {
       Inl q ->
         shows_prec_list Zero_nat ['s', 't', 'a', 't', 'e', ' '] .
           shows_prec Zero_nat q .
             shows_prec_list Zero_nat
               [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'a',
                 'l'];
       Inr fqs ->
         shows_prec_list Zero_nat
           ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n', 'o', ' ', 't',
             'r', 'a', 'n', 's', 'i', 't', 'i', 'o', 'n', ' ', 'f', 'r', 'o',
             'm', ' '] .
           shows_prec_term Zero_nat fqs .
             shows_prec_list Zero_nat
               [' ', 't', 'o', ' ', 'a', ' ', 'f', 'i', 'n', 'a', 'l', ' ', 's',
                 't', 'a', 't', 'e'];
     }))))
                             (\ _ ->
                               catcha
                                 (ta_trs_compatible_list ta
                                   (cover_bound_list_filter
                                     (\ l ->
                                       not (ta_rule_compatible_heuristic ta l))
                                     (boundstype_fun typea) c r))
                                 (\ x ->
                                   Inl (let {
  (lr, (lrq, q)) = x;
} in shows_prec_list Zero_nat
       ['T', 'A', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'm', 'p',
         'a', 't', 'i', 'b', 'l', 'e', ' ', 'w', 'i', 't', 'h', ' ', 'R'] .
       shows_nl .
         shows_prec_list Zero_nat
           ['f', 'o', 'r', ' ', 'r', 'u', 'l', 'e', ' '] .
           shows_rule lr .
             shows_nl .
               shows_prec_list Zero_nat
                 ['w', 'h', 'i', 'c', 'h', ' ', 'i', 's', ' ', 'i', 'n', 's',
                   't', 'a', 'n', 't', 'i', 'a', 't', 'e', 'd', ' ', 'b', 'y',
                   ' ', 's', 't', 'a', 't', 'e', 's', ' ', 't', 'o', ' '] .
                 shows_rule lrq .
                   shows_nl .
                     shows_prec_list Zero_nat
                       ['t', 'h', 'e', ' ', 's', 't', 'a', 't', 'e', ' '] .
                       shows_prec Zero_nat q .
                         shows_prec_list Zero_nat
                           [' ', 'i', 's', ' ', 'o', 'n', 'l', 'y', ' ', 'r',
                             'e', 'a', 'c', 'h', 'a', 'b', 'l', 'e', ' ', 'f',
                             'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'l', 'h',
                             's'] .
                           shows_nl))))))));

instance Linorder Char where {
};

check_rule_removal_proc ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_rule_removal_proc rp r rR =
  catcha
    (let {
       (_, nSr) = partition (\ foo -> isOK (rp_strict rp foo)) r;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha (rp_mono rp)
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                                 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                                 'f', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                                 'o', 'r', 'd', 'e', 'r'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha (check_subseteq rR r)
                          (\ x ->
                            Inl (shows_string
                                   ['n', 'o', 't', ' ', 'a', 'l', 'l', 'o', 'w',
                                     'e', 'd', ' ', 't', 'o', ' ', 'i', 'n',
                                     't', 'r', 'o', 'd', 'u', 'c', 'e', ' ',
                                     'n', 'e', 'w', ' ', 'r', 'u', 'l', 'e',
                                     ' '] .
                                  shows_nl . shows_rule x)))
                    (\ _ ->
                      binda (catcha
                              (catcha (forallm (rp_non_strict rp) nSr)
                                (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'r',
 'u', 'l', 'e', 's'] .
                                      shows_nl . x)))
                        (\ _ ->
                          catcha (check_subseteq nSr rR)
                            (\ x ->
                              Inl (case rp_strict rp x of {
                                    Inl err -> err;
                                    Inr _ -> internal_error;
                                  })))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r',
               'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's',
               's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ',
               'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl . rp_desc rp . shows_nl . x));

unlab_trs_tt ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => [(Term (Lab a b) c, Term (Lab a b) c)] ->
                    [(Term (Lab a b) c, Term (Lab a b) c)] ->
                      Sum ([Char] -> [Char]) ();
unlab_trs_tt r rR =
  let {
    u = unlab_of_trs_list r;
  } in catcha (binda (check_subseteq rR u) (\ _ -> check_subseteq u rR))
         (\ _ ->
           Inl (shows_rules rR .
                 shows_nl .
                   shows_string
                     [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c',
                       'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ', 't',
                       'o', ' ', 'u', 'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd',
                       ' ', 'T', 'R', 'S', ' '] .
                     shows_nl . shows_rules u . shows_nl));

create_sem_lab_tt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Sl_inter a [Nat] ->
                    Sum (Lab a [Nat] -> [Nat] -> Sum [Nat] [Lab a [Nat]]) b ->
                      [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                        [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)] ->
                          Sum ([Char] -> [Char]) ();
create_sem_lab_tt sli (Inl l) r lr =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
  } in catcha
         (check_sem_lab_tt model_check_valid model_check_decr
           (check_sl_lab_trs i l ca model_cge)
           (check_sl_model_lab_trs i l ca model_cge) model_lge model_LS r lr)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));
create_sem_lab_tt sli (Inr v) r lr =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    sig = funas_trs_list r;
    i = sl_inter_to_inter sli;
    l = qmodel_L sig;
  } in catcha
         (check_sem_lab_tt (qmodel_check_valid sli) (qmodel_check_decr sig v c)
           (check_sl_lab_trs i l ca qmodel_cge)
           (check_sl_model_lab_trs i l ca qmodel_cge) qmodel_lge
           (qmodel_LS sig ca) r lr)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));

unary_term :: forall a b. Term a b -> Bool;
unary_term (Var x) = True;
unary_term (Fun f [t]) = unary_term t;
unary_term (Fun v []) = False;
unary_term (Fun v (vb : vd : ve)) = False;

check_unary_signature ::
  forall a b. [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_unary_signature r =
  catcha
    (catcha
      (forallm
        (\ x ->
          (if let {
                (l, ra) = x;
              } in unary_term l && unary_term ra
            then Inr () else Inl x))
        r)
      (\ x -> Inl (snd x)))
    (\ _ ->
      Inl (shows_string
            ['t', 'h', 'e', ' ', 't', 'r', 's', ' ', 'd', 'o', 'e', 's', ' ',
              'n', 'o', 't', ' ', 'h', 'a', 'v', 'e', ' ', 'a', ' ', 'u', 'n',
              'a', 'r', 'y', ' ', 's', 'i', 'g', 'n', 'a', 't', 'u', 'r',
              'e']));

check_string_rev ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_string_rev r rR =
  let {
    rev = map rev_rule r;
  } in binda (check_unary_signature r)
         (\ _ ->
           binda (catcha (check_subseteq rev rR)
                   (\ x ->
                     Inl (shows_string
                            ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
                           shows_rule x .
                             shows_string
                               [' ', 'i', 's', ' ', 'm', 'i', 's', 's', 'i',
                                 'n', 'g'] .
                               shows_nl)))
             (\ _ ->
               catcha (check_subseteq rR rev)
                 (\ x ->
                   Inl (shows_string
                          ['t', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' '] .
                         shows_rule x .
                           shows_string
                             [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'a', ' ',
                               'r', 'e', 'v', 'e', 'r', 's', 'e', 'd', ' ', 'r',
                               'u', 'l', 'e'] .
                             shows_nl))));

checkTRSProof ::
  forall a.
    (Eq a, Linorder a,
      Showa a) => ([Char] -> [Char]) ->
                    [(Term (Lab a [Nat]) [Char], Term (Lab a [Nat]) [Char])] ->
                      Nat ->
                        TRSProof a [Nat] [Char] -> Sum ([Char] -> [Char]) ();
checkTRSProof i r m (DPTrans p prf) =
  debug (i []) ['D', 'P', ' ', 't', 'r', 'a', 'n', 's']
    (let {
       rm = computeRuleMap r;
     } in binda (catcha (check_DP Sharp r rm p)
                  (\ x ->
                    Inl (i . shows_string
                               [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w',
                                 'h', 'e', 'n', ' ', 's', 'w', 'i', 't', 'c',
                                 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm',
                                 ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S'] .
                               shows_nl .
                                 shows_trs r .
                                   shows_nl .
                                     shows_string
                                       ['t', 'o', ' ', 't', 'h', 'e', ' ', 'i',
 'n', 'i', 't', 'i', 'a', 'l', ' ', 'D', 'P', ' ', 'p', 'r', 'o', 'b', 'l', 'e',
 'm'] .
                                       shows_nl .
 shows_dpp p r . shows_nl . x)))
            (\ _ ->
              catcha
                (checkDPProof (i . shows_string ['.', '1']) p r Zero_nat rm
                  (computeRuleMap (reverse_trs r)) prf)
                (\ x ->
                  Inl (i . shows_string
                             [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e',
                               'l', 'o', 'w', ' ', 's', 'w', 'i', 't', 'c', 'h',
                               ' ', 't', 'o', ' ', 'd', 'e', 'p', 'e', 'n', 'd',
                               'e', 'n', 'c', 'y', ' ', 'p', 'a', 'i', 'r',
                               's'] .
                             shows_nl . indent x))));
checkTRSProof i r m (Bounds info) =
  debug (i []) ['B', 'o', 'u', 'n', 'd', 's']
    (catcha (check_bounds info r)
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n',
                     ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ', 'b', 'o',
                     'u', 'n', 'd', 's', ' ', 'o', 'n', ' ', 't', 'h', 'e', ' ',
                     'T', 'R', 'S'] .
                   shows_nl . shows_trs r . shows_nl . x)));
checkTRSProof i r m (RuleRemoval rp rR prf) =
  debug (i []) ['r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l']
    (binda
      (catcha (check_rule_removal_proc (get_redpair rp) r rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a',
                       'l', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i', 't',
                       'c', 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ',
                       't', 'h', 'e', ' ', 'T', 'R', 'S'] .
                     shows_nl .
                       shows_trs r .
                         shows_nl .
                           shows_string
                             ['t', 'o', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                               'S'] .
                             shows_nl . shows_trs rR . shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) rR m prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
                         ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r',
                         'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkTRSProof i r (Suc m) (Unlab uR prf) =
  debug (i []) ['u', 'n', 'l', 'a', 'b', 'e', 'l']
    (binda
      (catcha (unlab_trs_tt r uR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ',
                       'o', 'f', ' ', 'T', 'R', 'S', ' ', 'w', 'h', 'e', 'n',
                       ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ',
                       'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                       'S'] .
                     shows_nl .
                       shows_trs r .
                         shows_nl .
                           shows_string
                             ['t', 'o', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                               'S'] .
                             shows_nl . shows_trs uR . shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) uR m prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'l', 'a',
                         'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkTRSProof i r Zero_nat (Unlab uR prf) =
  debug (i []) ['u', 'n', 'l', 'a', 'b', 'e', 'l']
    (binda
      (catcha (check_subseteq (map_trs_list unlab r) uR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ',
                       'o', 'f', ' ', 'T', 'R', 'S', ' ', 'w', 'h', 'e', 'n',
                       ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ',
                       'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                       'S'] .
                     shows_nl .
                       shows_trs r .
                         shows_nl .
                           shows_string
                             ['t', 'o', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                               'S'] .
                             shows_nl .
                               shows_trs uR .
                                 shows_nl .
                                   shows_string
                                     ['u', 'n', 'l', 'a', 'b', 'e', 'l', 'e',
                                       'd', ' ', 'r', 'u', 'l', 'e', ' '] .
                                     shows_rule x .
                                       shows_string
 [' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'p', 'r', 'e', 's', 'e', 'n', 't'])))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) uR Zero_nat prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'l', 'a',
                         'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkTRSProof i r m RisEmpty =
  debug (i []) ['R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (check (null r)
      (i . shows_string
             [':', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm',
               'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ', 'i', 't', ' ', 'i',
               's', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S'] .
             shows_trs r . shows_nl));
checkTRSProof i r m (Semlab sli l lR prf) =
  debug (i []) ['S', 'e', 'm', 'l', 'a', 'b']
    (binda
      (catcha (create_sem_lab_tt sli l r lR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a',
                       'b', 'e', 'l', 'i', 'n', 'g'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) lR (Suc m) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c',
                         ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkTRSProof i r m (Rootlab lR prf) =
  debug (i []) ['R', 'o', 'o', 't', 'l', 'a', 'b']
    (binda
      (catcha (rl_tt r lR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       'r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i',
                       'n', 'g'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) lR (Suc m) prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 'r', 'o', 'o', 't', '-', 'l', 'a', 'b',
                         'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkTRSProof i r m (Fcc fcs rR prf) =
  debug (i []) ['F', 'c', 'c']
    (binda
      (catcha (fcc_tt r fcs rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       'f', 'l', 'a', 't', ' ', 'c', 'o', 'n', 't', 'e', 'x',
                       't', ' ', 'c', 'l', 'o', 's', 'u', 'r', 'e'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) rR m prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 'f', 'l', 'a', 't', ' ', 'c', 'o', 'n',
                         't', 'e', 'x', 't', ' ', 'c', 'l', 'o', 's', 'u', 'r',
                         'e'] .
                       shows_nl . indent x))));
checkTRSProof i r m (StringReversal rR prf) =
  debug (i [])
    ['S', 't', 'r', 'i', 'n', 'g', ' ', 'R', 'e', 'v', 'e', 'r', 's', 'a', 'l']
    (binda
      (catcha (check_string_rev r rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e',
                       'r', 's', 'a', 'l'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) rR m prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 's', 't', 'r', 'i', 'n', 'g', ' ', 'r',
                         'e', 'v', 'e', 'r', 's', 'a', 'l'] .
                       shows_nl . indent x))));
checkTRSProof i r m (Uncurry u_info rR prf) =
  debug (i []) ['U', 'n', 'c', 'u', 'r', 'r', 'y']
    (binda
      (catcha (uncurry_tt u_info r rR)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g', ' ', 't', 'r',
                       'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i', 'o',
                       'n'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) rR Zero_nat prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i',
                         'n', 'g'] .
                       shows_nl . indent x))));

init_lab :: forall a b. a -> Lab a b;
init_lab f = UnLab f;

check_dp_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b ->
                    [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                      [(b, Term a b)] ->
                        Ctxt a b ->
                          [(Term a b, Term a b)] ->
                            [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_dp_loop s prseq sigma c p r =
  (if equal_ctxt c Hole
    then binda (check (not (null prseq))
                 (shows_prec_list Zero_nat
                   ['r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u',
                     'e', 'n', 'c', 'e', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e',
                     ' ', 'n', 'o', 'n', '-', 'e', 'm', 'p', 't', 'y']))
           (\ _ ->
             check_steps p r prseq s (subst_apply_term s (subst_of sigma)))
    else check_loop s (map (\ (x, (y, (_, z))) -> (x, (y, z))) prseq) sigma c
           r);

check_dp_disproof ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => ([Char] -> [Char]) ->
                    [(Term a b, Term a b)] ->
                      [(Term a b, Term a b)] ->
                        DPDisproof a b -> Sum ([Char] -> [Char]) ();
check_dp_disproof i p r (DPLoop s rseq sigma c) =
  debug (i []) ['L', 'o', 'o', 'p']
    (catcha (check_dp_loop s rseq sigma c p r)
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'l',
                     'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f'] .
                   shows_nl . x)));
check_dp_disproof i pa ra (DPRuleRemoval p r prf) =
  debug (i []) ['R', 'u', 'l', 'e', ' ', 'R', 'e', 'm', 'o', 'v', 'a', 'l']
    (binda
      (catcha (check_subseteq p pa)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'r', 'u', 'l', 'e', '-', 'r', 'e', 'm', 'o', 'v', 'a',
                       'l', ' ', 'p', 'r', 'o', 'o', 'f'] .
                     shows_nl .
                       shows_rule x .
                         shows_string
                           [' ', 'i', 's', ' ', 'a', ' ', 'n', 'e', 'w', ' ',
                             'p', 'a', 'i', 'r', ' ', 'n', 'o', 't', ' ', 'o',
                             'c', 'c', 'u', 'r', 'r', 'i', 'n', 'g', ' ', 'i',
                             'n', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'e', 'v',
                             'i', 'o', 'u', 's', ' ', 's', 'e', 't', ' ', 'o',
                             'f', ' ', 'p', 'a', 'i', 'r', 's'] .
                           shows_nl)))
      (\ _ ->
        binda (catcha (check_subseteq r ra)
                (\ x ->
                  Inl (i . shows_string
                             [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n',
                               ' ', 'r', 'u', 'l', 'e', '-', 'r', 'e', 'm', 'o',
                               'v', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f'] .
                             shows_nl .
                               shows_rule x .
                                 shows_string
                                   [' ', 'i', 's', ' ', 'a', ' ', 'n', 'e', 'w',
                                     ' ', 'r', 'u', 'l', 'e', ' ', 'n', 'o',
                                     't', ' ', 'o', 'c', 'c', 'u', 'r', 'r',
                                     'i', 'n', 'g', ' ', 'i', 'n', ' ', 't',
                                     'h', 'e', ' ', 'p', 'r', 'e', 'v', 'i',
                                     'o', 'u', 's', ' ', 'T', 'R', 'S'] .
                                   shows_nl)))
          (\ _ -> check_dp_disproof (i . shows_string ['.', '1']) p r prf)));

checkTRSDisproof ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    [(Term (Lab a b) c, Term (Lab a b) c)] ->
                      TRSDisproof (Lab a b) c -> Sum ([Char] -> [Char]) ();
checkTRSDisproof i r (Loop s rseq sigma c) =
  debug (i []) ['L', 'o', 'o', 'p']
    (catcha (check_loop s rseq sigma c r)
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'l',
                     'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f'] .
                   shows_nl . x)));
checkTRSDisproof i r NotWellFormed =
  debug (i [])
    ['n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd']
    (catcha (check_not_wf_trs r)
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'n',
                     'o', 't', '-', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm',
                     'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f'] .
                   shows_nl . x)));
checkTRSDisproof i ra (NTRuleRemoval r prf) =
  debug (i []) ['r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l']
    (binda
      (catcha (check_subseteq r ra)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'r', 'u', 'l', 'e', '-', 'r', 'e', 'm', 'o', 'v', 'a',
                       'l', ' ', 'p', 'r', 'o', 'o', 'f'] .
                     shows_nl .
                       shows_rule x .
                         shows_string
                           [' ', 'i', 's', ' ', 'a', ' ', 'n', 'e', 'w', ' ',
                             'r', 'u', 'l', 'e', ' ', 'n', 'o', 't', ' ', 'o',
                             'c', 'c', 'u', 'r', 'r', 'i', 'n', 'g', ' ', 'i',
                             'n', ' ', 't', 'h', 'e', ' ', 'p', 'r', 'e', 'v',
                             'i', 'o', 'u', 's', ' ', 'T', 'R', 'S'] .
                           shows_nl)))
      (\ _ -> checkTRSDisproof (i . shows_string ['.', '1']) r prf));
checkTRSDisproof i ra (NTStringReversal r prf) =
  debug (i [])
    ['s', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e', 'r', 's', 'a', 'l']
    (binda
      (catcha (check_string_rev ra r)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       's', 't', 'r', 'i', 'n', 'g', ' ', 'r', 'e', 'v', 'e',
                       'r', 's', 'a', 'l', ' ', 'p', 'r', 'o', 'o', 'f'] .
                     shows_nl . x)))
      (\ _ -> checkTRSDisproof (i . shows_string ['.', '1']) r prf));
checkTRSDisproof i r (NTDPTrans p prf) =
  debug (i [])
    ['D', 'P', ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r', 'm', 'a', 't', 'i',
      'o', 'n']
    (binda
      (catcha (check_DPs unsharp r p)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'D', 'P', ' ', 't', 'r', 'a', 'n', 's', 'f', 'o', 'r',
                       'm', 'a', 't', 'i', 'o', 'n'] .
                     shows_nl .
                       shows_prec_list Zero_nat
                         ['p', 'r', 'o', 'b', 'l', 'e', 'm', 'a', 't', 'i', 'c',
                           ' ', 'r', 'u', 'l', 'e', ' '] .
                         shows_rule x)))
      (\ _ -> check_dp_disproof (i . shows_string ['.', '1']) p r prf));

check_relative_rule_removal ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => (Sum ([Char] -> [Char]) (),
                    ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                      ((Term a b, Term a b) -> Sum ([Char] -> [Char]) (),
                        (a -> Nat -> Nat -> Bool,
                          (Sum ([Char] -> [Char]) (), [Char] -> [Char]))))) ->
                    ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                      ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                        Sum ([Char] -> [Char]) ();
check_relative_rule_removal rp (r, s) (rR, rS) =
  catcha
    (let {
       (_, nSr) = partition (\ foo -> isOK (rp_strict rp foo)) r;
       (_, nSs) = partition (\ foo -> isOK (rp_strict rp foo)) s;
     } in binda (rp_valid rp)
            (\ _ ->
              binda (catcha (rp_mono rp)
                      (\ x ->
                        Inl (shows_string
                               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                 'i', 't', 'h', ' ', 'm', 'o', 'n', 'o', 't',
                                 'o', 'n', 'i', 'c', 'i', 't', 'y', ' ', 'o',
                                 'f', ' ', 's', 't', 'r', 'i', 'c', 't', ' ',
                                 'o', 'r', 'd', 'e', 'r'] .
                              shows_nl . x)))
                (\ _ ->
                  binda (catcha
                          (catcha (forallm (rp_non_strict rp) nSr)
                            (\ x -> Inl (snd x)))
                          (\ x ->
                            Inl (shows_string
                                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w',
                                     'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e',
                                     'n', 't', 'i', 'n', 'g', ' ', 'r', 'u',
                                     'l', 'e', 's'] .
                                  shows_nl . x)))
                    (\ _ ->
                      binda (catcha
                              (catcha (forallm (rp_non_strict rp) nSs)
                                (\ x -> Inl (snd x)))
                              (\ x ->
                                Inl (shows_string
                                       ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ',
 'w', 'h', 'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i', 'n', 'g', ' ', 'r',
 'u', 'l', 'e', 's'] .
                                      shows_nl . x)))
                        (\ _ ->
                          binda (catcha (check_subseteq nSs rS)
                                  (\ x ->
                                    Inl (case rp_strict rp x of {
  Inl err -> err;
  Inr _ -> internal_error;
})))
                            (\ _ ->
                              catcha (check_subseteq nSr rR)
                                (\ x ->
                                  Inl (case rp_strict rp x of {
Inl err -> err;
Inr _ -> internal_error;
                                      }))))))))
    (\ x ->
      Inl (shows_string
             ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
               'l', 'y', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e', ' ', 'r',
               'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r', 'o', 'c', 'e', 's',
               's', 'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ',
               'f', 'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g'] .
            shows_nl . rp_desc rp . shows_nl . x));

unlab_reltrs_tt ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([(Term (Lab a b) c, Term (Lab a b) c)],
                    [(Term (Lab a b) c, Term (Lab a b) c)]) ->
                    ([(Term (Lab a b) c, Term (Lab a b) c)],
                      [(Term (Lab a b) c, Term (Lab a b) c)]) ->
                      Sum ([Char] -> [Char]) ();
unlab_reltrs_tt rs rRS =
  catcha
    (binda (check_subseteq (map_trs_list unlab (fst rs)) (fst rRS))
      (\ _ ->
        catcha
          (forallm
            (\ x ->
              (if equal_term (fst x) (snd x) || membera (snd rRS) x then Inr ()
                else Inl x))
            (map_trs_list unlab (snd rs)))
          (\ x -> Inl (snd x))))
    (\ x ->
      Inl (shows_string
             ['c', 'a', 'n', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 'u',
               'n', 'l', 'a', 'b', 'e', 'l', 'e', 'd', ' ', 'r', 'u', 'l', 'e',
               ' '] .
            shows_nl . shows_rule x));

create_sem_lab_rel_tt ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Sl_inter a [Nat] ->
                    Sum (Lab a [Nat] -> [Nat] -> Sum [Nat] [Lab a [Nat]]) b ->
                      ([(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)],
                        [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)]) ->
                        ([(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)],
                          [(Term (Lab a [Nat]) b, Term (Lab a [Nat]) b)]) ->
                          Sum ([Char] -> [Char]) ();
create_sem_lab_rel_tt sli (Inl l) rs lrs =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
  } in catcha
         (check_sem_lab_rel_tt model_check_valid model_check_decr
           (check_sl_lab_trs i l ca model_cge)
           (check_sl_model_lab_trs i l ca model_cge) model_LS rs lrs)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));
create_sem_lab_rel_tt sli (Inr v) rs lrs =
  let {
    c = get_largest_element sli;
    ca = upt Zero_nat (Suc c);
    i = sl_inter_to_inter sli;
    sig = funas_trs_list (fst rs ++ snd rs);
    l = qmodel_L sig;
  } in catcha
         (check_sem_lab_rel_tt (qmodel_check_valid sli)
           (qmodel_check_decr sig v c) (check_sl_lab_trs i l ca qmodel_cge)
           (check_sl_model_lab_trs i l ca qmodel_cge) (qmodel_LS sig ca) rs lrs)
         (\ x -> Inl (shows_prec_sl_inter Zero_nat sli . shows_nl . x));

checkRelTRSProof ::
  forall a.
    (Eq a, Linorder a,
      Showa a) => ([Char] -> [Char]) ->
                    [(Term (Lab a [Nat]) [Char], Term (Lab a [Nat]) [Char])] ->
                      [(Term (Lab a [Nat]) [Char],
                         Term (Lab a [Nat]) [Char])] ->
                        RelTRSProof a [Nat] [Char] -> Sum ([Char] -> [Char]) ();
checkRelTRSProof i r s RelRisEmpty =
  debug (i []) ['R', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (check (null r)
      (i . shows_string
             [':', ' ', 'R', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm',
               'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ', 'i', 't', ' ', 'i',
               's', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S'] .
             shows_trs r . shows_nl));
checkRelTRSProof i r s (RelSisEmpty prf) =
  debug (i []) ['S', ' ', 'i', 's', ' ', 'e', 'm', 'p', 't', 'y']
    (binda
      (check (null s)
        (i . shows_string
               [':', ' ', 'S', ' ', 'i', 's', ' ', 'n', 'o', 't', ' ', 'e', 'm',
                 'p', 't', 'y', ',', ' ', 'b', 'u', 't', ' ', 'i', 't', ' ',
                 'i', 's', ' ', 't', 'h', 'e', ' ', 'T', 'R', 'S'] .
               shows_trs s . shows_nl))
      (\ _ ->
        catcha (checkTRSProof (i . shows_string ['.', '1']) r Zero_nat prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
                         ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r',
                         'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkRelTRSProof i r s (RelRuleRemoval rp rR rS prf) =
  debug (i []) ['r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l']
    (binda
      (catcha (check_relative_rule_removal (get_redpair rp) (r, s) (rR, rS))
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a',
                       'l', ' ', 'w', 'h', 'e', 'n', ' ', 's', 'w', 'i', 't',
                       'c', 'h', 'i', 'n', 'g', ' ', 'f', 'r', 'o', 'm', ' ',
                       't', 'h', 'e', ' ', 'T', 'R', 'S', 's'] .
                     shows_nl .
                       shows_trs r .
                         shows_nl .
                           shows_string [' ', 'a', 'n', 'd', ' '] .
                             shows_nl .
                               shows_trs s .
                                 shows_nl .
                                   shows_string
                                     ['t', 'o', ' ', 't', 'h', 'e', ' ', 'T',
                                       'R', 'S', 's'] .
                                     shows_nl .
                                       shows_trs rR .
 shows_nl .
   shows_string [' ', 'a', 'n', 'd', ' '] . shows_nl . shows_trs rS . x)))
      (\ _ ->
        catcha (checkRelTRSProof (i . shows_string ['.', '1']) rR rS prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'r', 'u', 'l', 'e',
                         ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l', ' ', 'p', 'r',
                         'o', 'c', 'e', 's', 's', 'o', 'r'] .
                       shows_nl . indent x))));
checkRelTRSProof i r s (RelUnlab uR uS prf) =
  debug (i []) ['u', 'n', 'l', 'a', 'b', 'e', 'l']
    (binda
      (catcha (unlab_reltrs_tt (r, s) (uR, uS))
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'u', 'n', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g', ' ',
                       'o', 'f', ' ', 'T', 'R', 'S', ' ', 'w', 'h', 'e', 'n',
                       ' ', 's', 'w', 'i', 't', 'c', 'h', 'i', 'n', 'g', ' ',
                       'f', 'r', 'o', 'm', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                       'S', ' ', 'R'] .
                     shows_nl .
                       shows_trs r .
                         shows_nl .
                           shows_string ['a', 'n', 'd', ' ', 'S'] .
                             shows_nl .
                               shows_trs s .
                                 shows_string
                                   ['t', 'o', ' ', 't', 'h', 'e', ' ', 'T', 'R',
                                     'S', 's', ' '] .
                                   shows_nl .
                                     shows_trs uR .
                                       shows_nl . shows_trs uS . x)))
      (\ _ ->
        catcha (checkRelTRSProof (i . shows_string ['.', '1']) uR uS prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 't', 'h', 'e', ' ', 'u', 'n', 'l', 'a',
                         'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkRelTRSProof i r s (RelSemlab sli l lR lS prf) =
  debug (i []) ['S', 'e', 'm', 'l', 'a', 'b']
    (binda
      (catcha (create_sem_lab_rel_tt sli l (r, s) (lR, lS))
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       's', 'e', 'm', 'a', 'n', 't', 'i', 'c', ' ', 'l', 'a',
                       'b', 'e', 'l', 'i', 'n', 'g'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkRelTRSProof (i . shows_string ['.', '1']) lR lS prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c',
                         ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkRelTRSProof i r s (RelRootlab lR lS prf) =
  debug (i []) ['R', 'o', 'o', 't', 'l', 'a', 'b']
    (binda
      (catcha (rl_rel_tt (r, s) (lR, lS))
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       'r', 'o', 'o', 't', '-', 'l', 'a', 'b', 'e', 'l', 'i',
                       'n', 'g'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkRelTRSProof (i . shows_string ['.', '1']) lR lS prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 's', 'e', 'm', 'a', 'n', 't', 'i', 'c',
                         ' ', 'l', 'a', 'b', 'e', 'l', 'i', 'n', 'g'] .
                       shows_nl . indent x))));
checkRelTRSProof i r s (RelUncurry info uR uS prf) =
  debug (i []) ['U', 'n', 'c', 'u', 'r', 'r', 'y']
    (binda
      (catcha (uncurry_rel_tt info r s uR uS)
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e',
                       'n', ' ', 'a', 'p', 'p', 'l', 'y', 'i', 'n', 'g', ' ',
                       'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i', 'n', 'g'] .
                     shows_nl . x)))
      (\ _ ->
        catcha (checkRelTRSProof (i . shows_string ['.', '1']) uR uS prf)
          (\ x ->
            Inl (i . shows_string
                       [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'b', 'e', 'l',
                         'o', 'w', ' ', 'u', 'n', 'c', 'u', 'r', 'r', 'y', 'i',
                         'n', 'g'] .
                       shows_nl . indent x))));

check_rel_seq ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => [(Term a b, Term a b)] ->
                    [(Term a b, Term a b)] ->
                      [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                        Term a b ->
                          Term a b -> Bool -> Sum ([Char] -> [Char]) ();
check_rel_seq r sa [] s u True =
  check (equal_term s u)
    (shows_string
       ['t', 'h', 'e', ' ', 'l', 'a', 's', 't', ' ', 't', 'e', 'r', 'm', ' ',
         'o', 'f', ' ', 't', 'h', 'e', ' ', 'r', 'e', 'w', 'r', 'i', 't', 'e',
         ' ', 's', 'e', 'q', 'u', 'e', 'n', 'c', 'e'] .
      shows_nl .
        shows_prec_term Zero_nat s .
          shows_string
            (shows_nl
              ['d', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'c', 'o', 'r', 'r',
                'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 't', 'h', 'e',
                ' ', 'g', 'o', 'a', 'l', ' ', 't', 'e', 'r', 'm']) .
            shows_nl . shows_prec_term Zero_nat u . shows_nl);
check_rel_seq r sa [] s u False =
  Inl (shows_prec_list Zero_nat
        ['d', 'i', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i', 'n', 'd', ' ', 's',
          't', 'r', 'i', 'c', 't', ' ', 's', 't', 'e', 'p', ' ', 'i', 'n', ' ',
          'r', 'e', 'w', 'r', 'i', 't', 'e', ' ', 's', 'e', 'q', 'u', 'e', 'n',
          'c', 'e']);
check_rel_seq ra sa ((p, (r, (True, t))) : prts) s u b =
  binda (check_rstep ra p r s t) (\ _ -> check_rel_seq ra sa prts t u True);
check_rel_seq ra sa ((p, (r, (False, t))) : prts) s u b =
  binda (check_rstep sa p r s t) (\ _ -> check_rel_seq ra sa prts t u b);

check_rel_loop ::
  forall a b.
    (Eq a, Showa a, Eq b,
      Showa b) => Term a b ->
                    [(Pos, ((Term a b, Term a b), (Bool, Term a b)))] ->
                      [(b, Term a b)] ->
                        Ctxt a b ->
                          [(Term a b, Term a b)] ->
                            [(Term a b, Term a b)] -> Sum ([Char] -> [Char]) ();
check_rel_loop sa rseq sigma c r s =
  check_rel_seq r s rseq sa
    (ctxt_apply c (subst_apply_term sa (subst_of sigma))) False;

check_wf_rel_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                    Sum ([Char] -> [Char]) ();
check_wf_rel_trs (r, s) =
  binda (check_wf_trs r)
    (\ _ ->
      (if null r then Inr ()
        else catcha
               (forallm
                 (\ rule ->
                   catcha
                     (catcha
                       (check_subseteq (vars_term_list (snd rule))
                         (vars_term_list (fst rule)))
                       (\ x ->
                         Inl (shows_string
                                ['f', 'r', 'e', 'e', ' ', 'v', 'a', 'r', 'i',
                                  'a', 'b', 'l', 'e', ' '] .
                               shows_prec Zero_nat x .
                                 shows_string
                                   [' ', 'i', 'n', ' ', 'r', 'i', 'g', 'h', 't',
                                     '-', 'h', 'a', 'n', 'd', ' ', 's', 'i',
                                     'd', 'e', ' ', 'o', 'f'])))
                     (\ x ->
                       Inl (x . shows_string [' ', 'r', 'u', 'l', 'e', ' '] .
                                  shows_rule rule . shows_nl)))
                 s)
               (\ x -> Inl (snd x))));

check_not_wf_rel_trs ::
  forall a b.
    (Showa a, Eq b,
      Showa b) => ([(Term a b, Term a b)], [(Term a b, Term a b)]) ->
                    Sum ([Char] -> [Char]) ();
check_not_wf_rel_trs rs =
  check (not (isOK (check_wf_rel_trs rs)))
    (shows_string
       ['T', 'h', 'e', ' ', 'T', 'R', 'S', 's', ' ', 'R', ' ', 'a', 'n', 'd',
         ' ', 'S', ' ', 'a', 'r', 'e', ' ', 'w', 'e', 'l', 'l', ' ', 'f', 'o',
         'r', 'm', 'e', 'd'] .
      shows_nl);

checkRelTRSDisproof ::
  forall a b c.
    (Eq a, Showa a, Eq b, Showa b, Eq c,
      Showa c) => ([Char] -> [Char]) ->
                    ([(Term (Lab a b) c, Term (Lab a b) c)],
                      [(Term (Lab a b) c, Term (Lab a b) c)]) ->
                      RelTRSDisproof (Lab a b) c -> Sum ([Char] -> [Char]) ();
checkRelTRSDisproof i (r, sa) (RelLoop s rseq sigma c) =
  debug (i []) ['L', 'o', 'o', 'p']
    (catcha (check_rel_loop s rseq sigma c r sa)
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'l',
                     'o', 'o', 'p', ' ', 'p', 'r', 'o', 'o', 'f'] .
                   shows_nl . x)));
checkRelTRSDisproof i (r, s) (RNotSN prf) =
  checkTRSDisproof (i . shows_string ['.', '1']) r prf;
checkRelTRSDisproof i (r, s) RelNotWellFormed =
  debug (i [])
    ['n', 'o', 't', ' ', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm', 'e', 'd']
    (catcha (check_not_wf_rel_trs (r, s))
      (\ x ->
        Inl (i . shows_string
                   [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'n',
                     'o', 't', '-', 'w', 'e', 'l', 'l', '-', 'f', 'o', 'r', 'm',
                     'e', 'd', ' ', 'p', 'r', 'o', 'o', 'f'] .
                   shows_nl . x)));
checkRelTRSDisproof i (ra, sa) (NTRelRuleRemoval r s prf) =
  debug (i []) ['r', 'u', 'l', 'e', ' ', 'r', 'e', 'm', 'o', 'v', 'a', 'l']
    (binda
      (catcha (binda (check_subseteq r ra) (\ _ -> check_subseteq s sa))
        (\ x ->
          Inl (i . shows_string
                     [':', ' ', 'e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ',
                       'r', 'u', 'l', 'e', '-', 'r', 'e', 'm', 'o', 'v', 'a',
                       'l', ' ', 'p', 'r', 'o', 'o', 'f'] .
                     shows_nl .
                       shows_rule x .
                         shows_string
                           [' ', 'i', 's', ' ', 'a', ' ', 'n', 'e', 'w', ' ',
                             'r', 'u', 'l', 'e', ' '] .
                           shows_nl)))
      (\ _ -> checkRelTRSDisproof (i . shows_string ['.', '1']) (r, s) prf));

certifyCertProblem ::
  forall a.
    (Eq a, Linorder a,
      Showa a) => CertificationProblem a [Nat] [Char] -> CertificationResult;
certifyCertProblem (TRSProof trs prf) =
  (case checkTRSProof (shows_string ['1']) (map (map_rule init_lab) trs)
          Zero_nat prf
    of {
    Inl err ->
      Error ((shows_string
                ['t', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ',
                  'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c',
                  'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ ->
      Certified
        ['<', 't', 'r', 's', 'T', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o',
          'n', 'P', 'r', 'o', 'o', 'f', '>'];
  });
certifyCertProblem (TRSDisproof trs prf) =
  (case checkTRSDisproof (shows_string ['1']) (map (map_rule init_lab) trs) prf
    of {
    Inl err ->
      Error ((shows_string
                ['n', 'o', 'n', '-', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't',
                  'i', 'o', 'n', ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o',
                  't', ' ', 'a', 'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ ->
      Certified
        ['<', 't', 'r', 's', 'N', 'o', 'n', 't', 'e', 'r', 'm', 'i', 'n', 'a',
          't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
  });
certifyCertProblem (RelTRSProof (r, s) prf) =
  (case checkRelTRSProof (shows_string ['1']) (map (map_rule init_lab) r)
          (map (map_rule init_lab) s) prf
    of {
    Inl err ->
      Error ((shows_string
                ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 't', 'e', 'r',
                  'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', ' ', 'p', 'r', 'o',
                  'o', 'f', ' ', 'n', 'o', 't', ' ', 'a', 'c', 'c', 'e', 'p',
                  't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ ->
      Certified
        ['<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'T', 'e', 'r', 'm', 'i',
          'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f', '>'];
  });
certifyCertProblem (RelTRSDisproof (r, s) prf) =
  (case checkRelTRSDisproof (shows_string ['1'])
          (map (map_rule init_lab) r, map (map_rule init_lab) s) prf
    of {
    Inl err ->
      Error ((shows_string
                ['r', 'e', 'l', 'a', 't', 'i', 'v', 'e', ' ', 'n', 'o', 'n',
                  '-', 't', 'e', 'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n',
                  ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'n', 'o', 't', ' ', 'a',
                  'c', 'c', 'e', 'p', 't', 'e', 'd'] .
               shows_nl . err)
              []);
    Inr _ ->
      Certified
        ['<', 'r', 'e', 'l', 'a', 't', 'i', 'v', 'e', 'N', 'o', 'n', 't', 'e',
          'r', 'm', 'i', 'n', 'a', 't', 'i', 'o', 'n', 'P', 'r', 'o', 'o', 'f',
          '>'];
  });

extract_input_proof :: [Char] -> Maybe ([Char], ([Char], [Char]));
extract_input_proof s =
  (case search_len s ['<', 'i', 'n', 'p', 'u', 't', '>']
          (size_list ['<', 'i', 'n', 'p', 'u', 't', '>'])
    of {
    Nothing -> Nothing;
    Just (_, minput) ->
      (case search_len minput ['<', '/', 'i', 'n', 'p', 'u', 't', '>']
              (size_list ['<', '/', 'i', 'n', 'p', 'u', 't', '>'])
        of {
        Nothing -> Nothing;
        Just (the_input, sa) ->
          (case search_len sa ['<', 'p', 'r', 'o', 'o', 'f', '>']
                  (size_list ['<', 'p', 'r', 'o', 'o', 'f', '>'])
            of {
            Nothing -> Nothing;
            Just (version, prf) -> Just (the_input, (version, prf));
          });
      });
  });

xmldoc2certProblem ::
  Xmldoc -> Sum [Char] (CertificationProblem [Char] [Nat] [Char]);
xmldoc2certProblem (Xmldoc header xml) =
  debug ['0']
    ['p', 'a', 'r', 's', 'i', 'n', 'g', ' ', 'x', 'm', 'l', ' ', 't', 'o', ' ',
      'c', 'e', 'r', 't', 'i', 'f', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 'p',
      'r', 'o', 'b', 'l', 'e', 'm']
    (xml2certProblem (text ['n', 'a', 'm', 'e']) plain_name xml);

parseCertProblem ::
  [Char] -> Sum [Char] (CertificationProblem [Char] [Nat] [Char]);
parseCertProblem s = binda (doc_of_string s) xmldoc2certProblem;

certifyProof :: [Char] -> CertificationResult;
certifyProof problemString =
  (case parseCertProblem problemString of {
    Inl err ->
      Unsupported
        (['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'i', 'l', 'e', ' ', 'p', 'a',
           'r', 's', 'i', 'n', 'g'] ++
          [newline] ++ err);
    Inr cp ->
      (case extract_input_proof problemString of {
        Nothing ->
          Error ['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                  'n', 'd', ' ', 'i', 'n', 'p', 'u', 't', ' ', 'a', 'n', 'd',
                  ' ', 'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ', 's', 't',
                  'r', 'i', 'n', 'g', ' ', 'f', 'o', 'r', ' ', 'c', 'o', 'm',
                  'p', 'a', 'r', 'i', 's', 'o', 'n', ' ', 'w', 'i', 't', 'h',
                  ' ', 'p', 'a', 'r', 's', 'e', 'd', ' ', 'r', 'e', 's', 'u',
                  'l', 't'];
        Just (the_input, (_, the_proof)) ->
          (if eqWhiteSpace the_input
                (shows_prec_xml Zero_nat (xml_certificationProblem cp) [])
            then (case certifyCertProblem cp of {
                   Certified tag ->
                     (if starts_with the_proof tag then Certified tag
                       else Error (['p', 'r', 'o', 'v', 'e', 'n', ' ', 'p', 'r',
                                     'o', 'p', 'e', 'r', 't', 'y', ' '] ++
                                    tag ++
                                      [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o',
't', ' ', 'c', 'o', 'r', 'r', 'e', 's', 'p', 'o', 'n', 'd', ' ', 't', 'o', ' ',
'p', 'r', 'o', 'o', 'f', ' ', 'i', 'n', ' ', 'i', 'n', 'p', 'u', 't', ' ', 's',
't', 'r', 'i', 'n', 'g', ':', ' '] ++
takea (nat 20) the_proof));
                   Unsupported a -> Unsupported a;
                   Error a -> Error a;
                 })
            else Error (concat
                         [['p', 'a', 'r', 's', 'e', 'd', ' ', 'p', 'r', 'o',
                            'b', 'l', 'e', 'm', ' ', 'd', 'o', 'e', 's', ' ',
                            'n', 'o', 't', ' ', 'c', 'o', 'r', 'r', 'e', 's',
                            'p', 'o', 'n', 'd', ' ', 't', 'o', ' ', 'i', 'n',
                            'p', 'u', 't'],
                           [newline], ['i', 'n', 'p', 'u', 't', ':'], [newline],
                           filter (\ c -> not (membera wspace c)) the_input,
                           [newline], [newline],
                           ['p', 'a', 'r', 's', 'e', 'd', ':'], [newline],
                           filter (\ c -> not (membera wspace c))
                             (shows_prec_xml Zero_nat
                               (xml_certificationProblem cp) []),
                           [newline]]));
      });
  });

}
